/*
 ************************************************************************
 *                                                                      *
 * simpleCSM.c -- simple OpenCSM application                            *
 *                                                                      *
 *              Written by John Dannenhoffer @ Syracuse University      *
 *                                                                      *
 ************************************************************************
 */

/*
 * Copyright (C) 2018  John F. Dannenhoffer, III (Syracuse University)
 *
 * This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *     MA  02110-1301  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "egads.h"
#include "OpenCSM.h"


/***********************************************************************/
/*                                                                     */
/*   main program                                                      */
/*                                                                     */
/***********************************************************************/

int
main(int       argc,                    /* (in)  number of arguments */
     char      *argv[])                 /* (in)  array  of arguments */
{
    int          status, imajor, iminor, nbrch, npmtr, nbody;
    int          ipmtr, builtTo, type, ichld, ileft, irite;
    int          nnode, nedge, nface, iface, npnt, ntri, ipnt;
    const int    *ptype, *pindx, *tris, *tric;
    double       vals[10], *vel=NULL, vbest, vtest;
    const double *xyz, *uv;
    void         *modl;
    modl_T       *MODL;

    /* --------------------------------------------------------------- */

    /* report version */
    status = ocsmVersion(&imajor, &iminor);
    printf("ocsmVersion -> status=%d, imajor=%d, iminor=%d\n",
           status, imajor, iminor);

    /* load a .csm file */
    status = ocsmLoad("bolt.csm", &modl);
    printf("ocsmLoad -> status=%d\n",
           status);
    if (status != EGADS_SUCCESS) return(EXIT_FAILURE);

    /* get info about the current MODL */
    status = ocsmInfo(modl, &nbrch, &npmtr, &nbody);
    printf("ocsmInfo -> status=%d, nbrch=%d, npmtr=%d, nbody=%d\n",
           status, nbrch, npmtr, nbody);

    /* change the shaft's diameter (Dhsaft) to 0.50 */
    status = ocsmFindPmtr(modl, "Dshaft", OCSM_EXTERNAL, 1, 1, &ipmtr);
    printf("ocsmFindPmtr -> status=%d, ipmtr=%d\n",
           status, ipmtr);

    status = ocsmSetValuD(modl, ipmtr, 1, 1, 0.5);
    printf("ocsmSetValuD -> status=%d\n", status);

    /* build the MODL (and do not return Bodys directly) */
    nbody = 0;
    status = ocsmBuild(modl, 0, &builtTo, &nbody, NULL);
    printf("ocsmBuild -> status=%d, builtTo=%d\n",
           status, builtTo);

    /* find the sensitivity with respect to the shaft's length (Lshaft) */
    status = ocsmFindPmtr(modl, "Lshaft", OCSM_EXTERNAL, 1, 1, &ipmtr);
    printf("ocsmFindPmtr -> status=%d, ipmtr=%d\n",
           status, ipmtr);

    status = ocsmSetVelD(modl, 0,     0, 0, 0.0);
    printf("ocsmSetVelD(0) -> status=%d\n",
           status);

    status = ocsmSetVelD(modl, ipmtr, 1, 1, 1.0);
    printf("ocsmSetVelD(1) -> status=%d\n",
           status);

    status = ocsmPrintPmtrs(modl, stdout);
    printf("ocsmPrintPmtrs -> status=%d\n", status);

    /* rebuild the MODL (needed to propagate velocities to the arguments) */
    nbody = 0;
    status = ocsmBuild(modl, 0, &builtTo, &nbody, NULL);
    printf("ocsmBuild -> status=%d, builtTo=%d\n",
           status, builtTo);

    /* get info about the current MODL (which should now contain Bodys) */
    status = ocsmInfo(modl, &nbrch, &npmtr, &nbody);
    printf("ocsmInfo -> status=%d, nbrch=%d, npmtr=%d, nbody=%d\n",
           status, nbrch, npmtr, nbody);

    /* get info about last Body (which is on the stack) */
    status = ocsmGetBody(modl, nbody, &type, &ichld, &ileft, &irite, vals, &nnode, &nedge, &nface);
    printf("ocsmGetBody -> status=%d, type=%s, ichld=%d, ileft=%d, irite=%d, nnode=%d, nedge=%d, nface=%d\n",
           status, ocsmGetText(type), ichld, ileft, irite, nnode, nedge, nface);

    /* find the maximum velocity for each Face */
    for (iface = 1; iface <= nface; iface++) {
        printf("iface=%d\n", iface);

        /* get tessellation info (poke into MODL structure to get etess) */
        MODL = (modl_T *)modl;
        status = EG_getTessFace(MODL->body[nbody].etess, iface,
                                &npnt,  &xyz, &uv, &ptype, &pindx,
                                &ntri, &tris, &tric);
        printf("    EG_getTessFace -> status=%d, npnt=%d, ntri=%d\n",
               status, npnt, ntri);

        /* storage for velocities */
        vel = (double*) EG_alloc(3*npnt*sizeof(double));
        if (vel == NULL) {
            printf("malloc failed\n");
            return(EXIT_FAILURE);
        }

        /* getting velocity automatically computes sensitivities */
        status = ocsmGetVel(modl, nbody, OCSM_FACE, iface, npnt, NULL, vel);
        printf("    ocsmGetVel -> status=%d\n",
               status);

        vbest = 0;
        for (ipnt = 0; ipnt < npnt; ipnt++) {
            vtest = sqrt(vel[3*ipnt]*vel[3*ipnt]+vel[3*ipnt+1]*vel[3*ipnt+1]+vel[3*ipnt+2]*vel[3*ipnt+2]);
            if (vtest > vbest) vbest = vtest;
        }
        
        printf("        vbest=%12.5e\n", vbest);

        EG_free(vel);
    }

    /* free up the MODL */
    status = ocsmFree(modl);
    printf("ocsmFree(modl) -> status=%d\n",
           status);

    /* remove tmp files (if they exist) and cleanup udp storage */
    status = ocsmFree(NULL);
    printf("ocsmFree(NULL) -> status=%d\n",
           status);

    return(EXIT_SUCCESS);
}
