// AFLR4 interface functions - Modified from functions provided with
//	AFLR4_LIB source (aflr4_cad_geom_setup.c) written by David L. Marcum

//  Modified by Dr. Ryan Durscher AFRL/RQVC

#include "AFLR4_LIB.h" // Bring in AFLR4 API library

#include "egads.h"		   // Bring in EGADS

#include "meshUtils.h" // Collection of helper functions for meshing
#include "cfdUtils.h"  // Collection of helper functions for cfd analysis
#include "miscUtils.h"

static INT_ egads_eval_bedge (INT_ idef, void *cadData, INT_ *nbedge, INT_2D ** inibe,
                              DOUBLE_2D ** u, DOUBLE_3D ** x)
{

  /*
   * Generate a boundary edge grid for a given surface (idef) with a CAD geometry
   * definition. The boundary edge connectivity and coordinate arrays do
   * not need to be allocated prior to calling this routine. They should be set to
   * NULL if they are not allocated. In either case they will be reallocated for
   * the generated boundary edge grid size.
   *
   * Modifications are required for alternative CAD systems.
   *
   * Input Parameters
   *
   *  idef  : identification label for surface definition
   *  inibe : boundary edge node connectivity pointer
   *          must at least be set to NULL it will be reallocated
   *  u     : u,v-coordinates pointer
   *          must at least be set to NULL it will be reallocated
   *  x     : x,y,z-coordinates pointer
   *          must at least be set to NULL it will be reallocated
   *
   * Output Parameters
   *
   *  nbedge : number of boundary edges in boundary edge grid
   *  inibe  : boundary edge node connectivity for each boundary edge of
   *              boundary edge grid
   *  u      : u,v-coordinates of boundary edge grid
   *  x      : x,y,z-coordinates of boundary edge grid
   *
   * Return Value
   *
   *  0     : no errors
   *  > 0   : there were errors
   *
   */

    INT_         ierr;
    int          cnt, i, iloop, iedge, index, last, mtype, oclass,
                 n, nloop, nedge, *sen, *senses, status;

    double       range[2], uvbox[4], t, uv[2];

    const double *xyzs = NULL, *ts = NULL;

    ego          tess, geom, *loops = NULL, *edges = NULL, *nodes = NULL;

    int tessStatus, numPoints, numFace;
    ego body, *faces = NULL;


    tess = (ego) cadData;

    // Get body from tessellation
    status = EG_statusTessBody(tess, &body, &tessStatus, &numPoints);
    if (status != EGADS_SUCCESS) {
        status =  -status;
        goto cleanup;
    }

    // Get faces
    status = EG_getBodyTopos(body, NULL, FACE, &numFace, &faces);
    if (status != EGADS_SUCCESS) {
        status =  -status;
        goto cleanup;
    }

    status = EG_getTopology(faces[idef-1], &geom, &oclass, &mtype, uvbox,
                            &nloop, &loops, &senses);
    if (status != EGADS_SUCCESS) {
        status =  -status;
        goto cleanup;
    }

    for (cnt = iloop = 0; iloop < nloop; iloop++) {
        status = EG_getTopology(loops[iloop], &geom, &oclass, &mtype,
                                NULL, &nedge, &edges, &senses);
        if (status != EGADS_SUCCESS) {
            status =  -status;
            goto cleanup;
        }

        for (iedge = 0; iedge < nedge; iedge++) {
            status = EG_getTopology(edges[iedge], &geom, &oclass,
                                    &mtype, range, &n, &nodes, &sen);
            if (status != EGADS_SUCCESS) {
                status =  -status;
                goto cleanup;
            }

            if (mtype == DEGENERATE) continue;

            index = EG_indexBodyTopo(body, edges[iedge]);
            if (index < EGADS_SUCCESS) {
                status = -index;
                goto cleanup;
            }

            status = EG_getTessEdge(tess, index, &n, &xyzs, &ts);
            if (status != EGADS_SUCCESS) {
                status =  -status;
                goto cleanup;
            }

            cnt += n-1;
        }
    }

    ierr = 0;

    *nbedge = cnt;

    *inibe = (INT_2D *) ug_malloc (&ierr, ((*nbedge)+1) * sizeof (INT_2D));
    *u = (DOUBLE_2D *)  ug_malloc (&ierr, ((*nbedge)+1) * sizeof (DOUBLE_2D));
    *x = (DOUBLE_3D *)  ug_malloc (&ierr, ((*nbedge)+1) * sizeof (DOUBLE_3D));

    if (ierr) {
        ug_error_message ("*** ERROR 104111 : unable to allocate required memory ***");
        return 104111;
    }

    for (cnt = iloop = 0; iloop < nloop; iloop++) {
        status = EG_getTopology(loops[iloop], &geom, &oclass, &mtype,
                                NULL, &nedge, &edges, &senses);
        if (status != EGADS_SUCCESS) {
            status =  -status;
            goto cleanup;
        }


        last = cnt;
        for (iedge = 0; iedge < nedge; iedge++) {
            status = EG_getTopology(edges[iedge], &geom, &oclass,
                                    &mtype, range, &n, &nodes, &sen);
            if (status != EGADS_SUCCESS) {
                status =  -status;
                goto cleanup;
            }

            if (mtype == DEGENERATE) continue;

            index = EG_indexBodyTopo(body, edges[iedge]);
            if (index < EGADS_SUCCESS) {
                status = -index;
                goto cleanup;
            }

            status = EG_getTessEdge(tess, index, &n, &xyzs, &ts);
            if (status != EGADS_SUCCESS) {
                status =  -status;
                goto cleanup;
            }


            for (i = 0; i < n-1; i++, cnt++) {
                if (senses[iedge] == 1) {
                    t = ts[i];
                    (*x)[cnt+1][0] = xyzs[3*i  ];
                    (*x)[cnt+1][1] = xyzs[3*i+1];
                    (*x)[cnt+1][2] = xyzs[3*i+2];
                } else {
                    t = ts[n-i-1];
                    (*x)[cnt+1][0] = xyzs[3*(n-i-1)  ];
                    (*x)[cnt+1][1] = xyzs[3*(n-i-1)+1];
                    (*x)[cnt+1][2] = xyzs[3*(n-i-1)+2];
                }

                status = EG_getEdgeUV(faces[idef-1], edges[iedge],
                                      senses[iedge], t, uv);
                if (status != EGADS_SUCCESS) {
                    status =  -status;
                    goto cleanup;
                }

                #ifdef NORMALIZE
                (*u)[cnt+1][0] = (uv[0]-uvbox[0])/(uvbox[1]-uvbox[0]);
                (*u)[cnt+1][1] = (uv[1]-uvbox[2])/(uvbox[3]-uvbox[2]);
                #else
                (*u)[cnt+1][0] = uv[0];
                (*u)[cnt+1][1] = uv[1];
                #endif
                (*inibe)[cnt+1][0] = cnt+1;
                (*inibe)[cnt+1][1] = cnt+2;
            }
        }
        (*inibe)[cnt][1] = last+1;
    }

    status = EGADS_SUCCESS;

    cleanup:

        //if (ts != NULL) EG_free(ts);
        ts = NULL;
        //if (xyzs != NULL) EG_free(xyzs);
        xyzs = NULL;
        // if (nodes != NULL) EG_free(nodes);
        nodes = NULL;
        // if (edges != NULL) EG_free(edges);
        edges = NULL;
        // if (loops != NULL) EG_free(loops);
        loops = NULL;

        if (faces != NULL) EG_free(faces);

        //printf("egads_eval_bedge- cleanup complete\n");
        return status;
}

static INT_ egads_eval_uv_bounds (INT_ idef, void *cadData, double *umax, double *umin,
                                  double *vmax, double *vmin)
{

  /*
   * Set uv-bounding box for a given CAD geometry definition. This is the bounding
   * box for the full definition (not that for a trimming curve). It is used to
   * determine if the mapping has singular bounding curves in physical space or
   * requires scaling.
   *
   * Modifications are required for alternative CAD systems.
   *
   * Input Parameters
   *
   *  idef  : identification label for surface definition
   *
   * Output Parameters
   *
   *  umax  : u-coordinate maximum value
   *  umin  : u-coordinate minimum value
   *  vmax  : v-coordinate maximum value
   *  vmin  : v-coordinate minimum value
   *
   * Return Value
   *
   *  0     : no errors
   *  > 0   : there were errors
   *
   */

    double uvbox[4];

    #ifdef NORMALIZE
        uvbox[0] = uvbox[2] = 0.0;
        uvbox[1] = uvbox[3] = 1.0;
    #else

        int    i, stat, per;

        int tessStatus, numPoints, numFace;
        ego tess, body, *faces = NULL;

        tess = (ego ) cadData;

        // Get body from tessellation
        stat = EG_statusTessBody(tess, &body, &tessStatus, &numPoints);
        if (stat != EGADS_SUCCESS) return -stat;

        // Get faces
        stat = EG_getBodyTopos(body, NULL, FACE, &numFace, &faces);
        if (stat != EGADS_SUCCESS) {
            if (faces != NULL) EG_free(faces);
            return -stat;
        }

        i = idef;

        stat = EG_getRange(faces[i-1], uvbox, &per);
        if (stat != EGADS_SUCCESS) {
            if (faces != NULL) EG_free(faces);
            return -stat;
        }

        if (faces != NULL) EG_free(faces);

    #endif

    *umin = uvbox[0];
    *umax = uvbox[1];
    *vmin = uvbox[2];
    *vmax = uvbox[3];

    return 0;
}

static INT_ egads_eval_xyz_at_uv (INT_ idef, void *cadData, double u1, double u2, double *x1,
                                  double *x2, double *x3)
{

  /*
   *
   * Evaluate xyz coordinates (x1,x2,x3) on a given surface (idef) at a given
   * uv location (u1,u2).
   *
   * Modifications are required for alternative CAD systems.
   *
   */

    int    stat;
    double uv[2], result[18];

    int tessStatus, numPoints, numFace;
    ego tess, body, *faces = NULL;

    tess = (ego ) cadData;

    // Get body from tessellation
    stat = EG_statusTessBody(tess, &body, &tessStatus, &numPoints);
    if (stat != EGADS_SUCCESS) {
            stat =  -stat;
            goto cleanup;
    }

    // Get faces
    stat = EG_getBodyTopos(body, NULL, FACE, &numFace, &faces);
    if (stat != EGADS_SUCCESS) {
            stat =  -stat;
            goto cleanup;
    }

    {
    #ifdef NORMALIZE
        int    i, per;
        double uvbox[4];
        char   Text[133];

        i = idef;

        stat = EG_getRange(faces[i-1], uvbox, &per);
        if (stat != EGADS_SUCCESS) {
            stat =  -stat;
            goto cleanup;
        }

        uv[0] = uvbox[0] + u1*(uvbox[1]-uvbox[0]);
        uv[1] = uvbox[2] + u2*(uvbox[3]-uvbox[2]);
    #else
        int i;

        i = idef;

        uv[0] = u1;
        uv[1] = u2;
    #endif

        stat = EG_evaluate(faces[i-1], uv, result);
        if (stat != EGADS_SUCCESS) {
            stat =  -stat;
            goto cleanup;
        }
    }


    *x1 = result[0];
    *x2 = result[1];
    *x3 = result[2];

    stat = 0;

    goto cleanup;

    cleanup:

        if (faces != NULL) EG_free(faces);
        //printf("egads_eval_xyz_at_uv complete\n");

        return stat;
}


/*
static int aflr4_CAD_Cleanup (void) {

    // Cleanup EGADS objects.
     EG_deleteObject(tess);
     //EG_free(faces);
     return EGADS_SUCCESS;
}
*/


// Modified from aflr4_cad_geom_setup.c written by David L. Marcum
static int body_to_AFLR4_Geom_Setup (ego body, ego tess, mapAttrToIndexStruct attrMap)
{

  /* Output Parameters
   *
   *  EGADS surface data is saved internally
   *
   * Return Value
   *
   *  0         : no errors
   */

    INT_1D *bcID = NULL;   //BC id array for aflr4
    INT_1D *faceID = NULL; //Face id array for aflr4

    int status; //Return code
    int iFace; // Indexing of faces
    int numFace; // Number of faces on a body

    int bndID; // bndID being into faceId2BndMap

    const char *groupName = NULL; // Attribute name
    ego *faces = NULL;

    // Get Faces - save to global variable
    status = EG_getBodyTopos(body, NULL, FACE, &numFace, &faces);
    if (status != EGADS_SUCCESS) goto cleanup;

    // Allocate list of unique CAD surface ID values
    bcID    = (INT_1D *) ug_malloc (&status, (numFace+1) * sizeof (INT_1D));
    if (status != 0) {
        bcID = NULL;
        ug_error_message ("*** ERROR: unable to allocate required memory in body_to_AFLR4_Geom_Setup ***");
        status = EGADS_MALLOC;
        goto cleanup;
    }

    faceID  = (INT_1D *) ug_malloc (&status, (numFace+1) * sizeof (INT_1D));
    if (status != 0) {
        faceID = NULL;
        ug_error_message ("*** ERROR: unable to allocate required memory in body_to_AFLR4_Geom_Setup ***");
        status = EGADS_MALLOC;
        goto cleanup;
    }

    // Set list of unique CAD surface geometry ID values and
    //  set AFLR3 Grid BC values for each ID
    for (iFace = 1; iFace <= numFace; ++iFace) {

        // Set faceID array
        faceID[iFace] = iFace;

        // Set mesh bcID array -predefined AFLR3 Grid BC values:
        //  STD_UG3_GBC    : standard surface
        //  -STD_UG3_GBC   : standard BL generating surface
        //  BL_INT_UG3_GBC : surface that intersects BL region
        //  TRANSP_SRC_UG3_GBC  : embedded/transparent surface that is converted to source nodes by AFLR3
        //  TRANSP_BL_INT_UG3_GBC : embedded/transparent surface that intersects BL region
        //  TRANSP_UG3_GBC       : embedded/transparent surface
        //  -TRANSP_UG3_GBC      : embedded/transparent BL generating surface
        //  TRANSP_INTRNL_UG3_GBC : embedded/transparent surface that is converted to internal faces by AFLR3
        //  -TRANSP_INTRNL_UG3_GBC : embedded/transparent BL generating surface
        //                              that is converted to internal faces by AFLR3
        //  FIXED_BL_INT_UG3_GBC : fixed surface with a BL region that intersects volume BL region

        // Loop through attributes on face and get a bndID for that face if it is meshBC's
        // default gets set to 1 - standard surface
        bndID = STD_UG3_GBC;

/*
        if ((bndID !=  STD_UG3_GBC)           &&
            (bndID != -STD_UG3_GBC)           &&
            (bndID !=  BL_INT_UG3_GBC)        &&
            (bndID !=  TRANSP_SRC_UG3_GBC)    &&
            (bndID !=  TRANSP_BL_INT_UG3_GBC) &&
            (bndID !=  TRANSP_UG3_GBC)        &&
            (bndID != -TRANSP_UG3_GBC)        &&
            (bndID !=  TRANSP_INTRNL_UG3_GBC) &&
            (bndID != -TRANSP_INTRNL_UG3_GBC) &&
            (bndID !=  FIXED_BL_INT_UG3_GBC)) {

            printf("Unrecognized meshBC index/ID = %d for AFLR4 using the default of \"Standard Surface\"\n",bndID);
            bndID = STD_UG3_GBC;
        }
*/

        bcID[iFace] = bndID;

        // Loop through attributes on face and get the capsGroup name on the face
        status = retrieve_CAPSGroupAttr(faces[iFace-1], &groupName);
        if (status != CAPS_SUCCESS) {
            printf("Error occurred in retrieve_CAPSGroupAttr\n");
            goto cleanup;
        }

        // Pull out the index from the attribute map for the groupName
        status = get_mapAttrToIndexIndex(&attrMap, groupName, &bndID);
        if (status != CAPS_SUCCESS) {
            printf("Error occurred in get_mapAttrToIndexIndex\n");
            goto cleanup;
        }

        //printf("i = %d, ID = %d\n",iFace-1,(*faceId2BndIdMap)[iFace-1]);

    }

    // Save number of unique CAD surface ID values and list of unique CAD surface
    //  geometry ID values
    aflr4_set_cad_id_list (numFace, bcID, faceID, (void *) tess);

    // Register EGADS based CAD evaluation routines
    dgeom_register_cad_eval_xyz_at_uv (egads_eval_xyz_at_uv);
    dgeom_register_cad_eval_uv_bounds (egads_eval_uv_bounds);

    // Register EGADS based evaluation routine for boundary edge grids
    dgeom_register_cad_eval_bedge (egads_eval_bedge);

    status =  EGADS_SUCCESS;
    goto cleanup;

    cleanup:

        if (status != EGADS_SUCCESS) {
            if (faceID != NULL) ug_free (faceID);
            if (bcID   != NULL) ug_free (bcID);
        }

        if (faces  != NULL) EG_free(faces);

        //printf("body_to_AFLR4_Geom_Setup - cleanup comlete\n");
        return status;
}


static int buildTessellationObj(int surfaceIndex,
                                int numNodes,
                                int numTriFace,
                                int numQuadFace,
                                DOUBLE_2D *uv,
                                DOUBLE_3D *xyz,
                                INT_3D *triCon,
                                INT_4D *quadCon,
                                ego *egadsTess)
{

    int status; // Function return status

    int i; // Indexing

    int offSetIndex = 0;

    double *xyzSurf = NULL, *uvSurf = NULL; // Freeable - will get freed when tessellation object is deleted
    int *triSurf = NULL; // Freeable - will get freed when tessellation object is deleted

    xyzSurf = (double *) EG_alloc(3*numNodes*sizeof(double));
    if (xyzSurf == NULL) {
        status = EGADS_MALLOC;
        goto cleanup;
    }

    uvSurf = (double *) EG_alloc(2*numNodes*sizeof(double));
    if (uvSurf == NULL) {
        status = EGADS_MALLOC;
        goto cleanup;
    }

    triSurf = (int *) EG_alloc(3*(numTriFace + 2*numQuadFace)*sizeof(int));
    if (triSurf == NULL) {
        status = EGADS_MALLOC;
        goto cleanup;
    }

    for (i = 1; i <= numNodes; i++) {

        xyzSurf[3*(i-1) + 0] = xyz[i][0];
        xyzSurf[3*(i-1) + 1] = xyz[i][1];
        xyzSurf[3*(i-1) + 2] = xyz[i][2];

        uvSurf[2*(i-1) + 0] = uv[i][0];
        uvSurf[2*(i-1) + 1] = uv[i][1];
    }

    for (i = 0; i < numTriFace; i++) {
        triSurf[3*i + 0] = triCon[i+1][0];
        triSurf[3*i + 1] = triCon[i+1][1];
        triSurf[3*i + 2] = triCon[i+1][2];
    }

    // Split the quads
    offSetIndex = 3*numTriFace;
    for (i = 0; i < numQuadFace; i++) {
        /*
        printf("OffSet %d - %d\n", i, offSetIndex);
        printf("Index %d - %d\n", i, 3*i + offSetIndex + 0);
        printf("Index %d - %d\n", i, 3*i + offSetIndex + 1);
        printf("Index %d - %d\n", i, 3*i + offSetIndex + 2);
         */

        triSurf[3*i + 0 + offSetIndex] = quadCon[i+1][0];
        triSurf[3*i + 1 + offSetIndex] = quadCon[i+1][1];
        triSurf[3*i + 2 + offSetIndex] = quadCon[i+1][3];

        offSetIndex += 3;
        triSurf[3*i + 0 + offSetIndex] = quadCon[i+1][1];
        triSurf[3*i + 1 + offSetIndex] = quadCon[i+1][2];
        triSurf[3*i + 2 + offSetIndex] = quadCon[i+1][3];

        /*
        printf("OffSet %d - %d\n", i, offSetIndex);
        printf("Index %d - %d\n", i, 3*i + offSetIndex + 0);
        printf("Index %d - %d\n", i, 3*i + offSetIndex + 1);
        printf("Index %d - %d\n", i, 3*i + offSetIndex + 2);
         */
    }

    status = EG_setTessFace((*egadsTess),
                            surfaceIndex,
                            numNodes,
                            xyzSurf,
                            uvSurf,
                            (numTriFace + 2*numQuadFace),
                            triSurf);
    goto cleanup;

    cleanup:
        if (xyzSurf != NULL) EG_free(xyzSurf);
        if (uvSurf != NULL) EG_free(uvSurf);
        if (triSurf != NULL) EG_free(triSurf);

        return status;

}

int aflr4_Surface_Mesh(ego bodyIn,
                       meshInputStruct meshInput,
                       mapAttrToIndexStruct attrMap,
                       meshStruct *surfaceMesh)
{

    int status; // Function return status

    int i, surf; // Indexing

    int numSurface = 0;
    int numTriFace = 0;
    int numNodes = 0;
    int numQuadFace = 0;

    double box[6], size, params[3]; // Bounding box variables

    ego tess = NULL;

//  int idef, idefc, found, ndef, itype; // DGEOM parameters

    int Message_Flag; // AFRL4 message return flag

    UG_Param_Struct *AFLR4_Param_Struct_Ptr = NULL; // AFRL4 input structure

//  int geomType = 3; // AFRL4 geometry type - test Mode

    // AFRL4 output arrays
    INT_1D *bcFlag = NULL;
    INT_1D *blFlag = NULL;
    INT_3D *triCon = NULL;
    INT_4D *quadCon = NULL;
    DOUBLE_2D *uv = NULL;
    DOUBLE_3D *xyz = NULL;

    int glueId; // Id of glued composite surface

     // Commandline inputs
    int  prog_argc   = 0;    // Number of arguments
    char **prog_argv = NULL; // String arrays

    // Get bounding box for scaling
    status = EG_getBoundingBox(bodyIn, box);
    if (status != EGADS_SUCCESS) goto cleanup;

    size = sqrt((box[0]-box[3])*(box[0]-box[3]) +
                (box[1]-box[4])*(box[1]-box[4]) +
                (box[2]-box[5])*(box[2]-box[5]));

    params[0] = meshInput.paramTess[0]*size;
    params[1] = meshInput.paramTess[1]*size;
    params[2] = meshInput.paramTess[2];

    // Get Tessellation  - save to global variable
    status = EG_makeTessBody(bodyIn, params, &tess);
    if (status != EGADS_SUCCESS) goto cleanup;

    /*if (surfaceMesh->bodyTessMap.egadsTess != NULL) {
        status = EG_deleteObject(surfaceMesh->bodyTessMap.egadsTess);
        if (status != EGADS_SUCCESS) return status;
    }*/

    status = EG_makeTessBody(bodyIn, params, &surfaceMesh->bodyTessMap.egadsTess);
    if (status != EGADS_SUCCESS) goto cleanup;

    // Parse input string
    if (meshInput.aflr4Input.meshInputString != NULL) {
        status = string_toProgArgs(meshInput.aflr4Input.meshInputString, &prog_argc, &prog_argv);

        /*printf("Number of args = %d\n",prog_argc);
        for (i = 0; i <prog_argc ; i++) printf("Arg %d = %s\n", i, prog_argv[i]);
        */

    } else {
        // Get other command options
    }

    // Initialize and setup AFLR4 input parameter structure, write version info if
    // specified in the command line arguments, write documentation if specified
    // in the command line arguments, start memory monitor, file status monitor,
    // and create TMP file directory if specified.
    // Note that prog_argv is not used if prog_argc = 0.

    // Setup AFRL4_Param_Struct
    //status = aflr4_setup(&geomType, prog_argc, prog_argv, "", "", &AFLR4_Param_Struct_Ptr);
    //if (status != CAPS_SUCCESS) return -1; // Add error code


    status = ug_malloc_param (&AFLR4_Param_Struct_Ptr);
    if ((status != CAPS_SUCCESS) || (AFLR4_Param_Struct_Ptr == NULL)) {
        printf("ug_malloc_param status = %d\n", status);
        goto cleanup;
    }

    status = ug_initialize_param (AFLR4_Param_Struct_Ptr);
    if (status != CAPS_SUCCESS) {
            printf("ug_initialize_param status = %d\n", status);
            goto cleanup;
        }

    status = ug_initialize_aflr_param (4, AFLR4_Param_Struct_Ptr);
    if (status != CAPS_SUCCESS) {
            printf("ug_initialize_aflr_param status = %d\n", status);
            goto cleanup;
        }

    status = aflr2_initialize_param (AFLR4_Param_Struct_Ptr);
    if (status != CAPS_SUCCESS) {
        printf("aflr2_initialize_param status = %d\n", status);
        goto cleanup;
    }

    status = aflr4_initialize_param (AFLR4_Param_Struct_Ptr);
    if (status != CAPS_SUCCESS) {
        printf("aflr4_initialize_param status = %d\n", status);
        goto cleanup;
    }

    if (prog_argv != NULL && prog_argc > 1) {
      status = ug_set_params_from_arg (prog_argv, prog_argc, 1, 0, AFLR4_Param_Struct_Ptr);
      if (status != CAPS_SUCCESS) {
          printf("ug_set_params_from_arg status = %d\n", status);
          goto cleanup;
      }
    }

    // Get message flag.
    (void ) ug_get_int_param ("Message_Flag", &Message_Flag, AFLR4_Param_Struct_Ptr);

    // Get an input parameter.
    // See any of the routines that have AFLR4_Param_Struct_Ptr as an argument.
    // This is not required. Shown only for example usage.

    // If there is a matching input parameter then iget will be set to 1.
    // If there is no matching input parameter then nothing is done and iget
    // will be set to 0.

    //      CHAR_UG_MAX char_var;
    //      INT_ iget, int_var;
    //      double double_var;
    //      iget = ug_get_char_param ("char_var", &char_var, AFLR4_Param_Struct_Ptr);
    //      iget = ug_get_int_param ("int_var", &int_var, AFLR4_Param_Struct_Ptr);
    //      iget = ug_get_double_param ("double_var", &double_var, AFLR4_Param_Struct_Ptr);

    // Set an input parameter.
    // See any of the routines that have AFLR4_Param_Struct_Ptr as an argument.
    // This is not required. Shown only for example usage.

    // If there is a matching input parameter then iset will be set to 1.
    // If there is no matching input parameter then nothing is done and iset
    // will be set to 0.
    // Note that the value set will be limited by the minimum and maximum
    // limits set within the parameter structure.

    // examples

    //      CHAR_UG_MAX char_var;
    //      INT_ iset, int_var;
    //      double double_var;
    //      iset = ug_set_char_param ("char_var", char_var, AFLR4_Param_Struct_Ptr);
    //      iset = ug_set_int_param ("int_var", int_var, AFLR4_Param_Struct_Ptr);
    //      iset = ug_set_double_param ("double_var", double_var, AFLR4_Param_Struct_Ptr);

    // Setup necessary AFLR4 geometry information from EGADS body

    status = body_to_AFLR4_Geom_Setup (bodyIn, tess, attrMap);
    if (status != CAPS_SUCCESS) {
        printf("body_to_AFLR4_Geom_Setup status = %d\n", status);
        goto cleanup;
    }

    // Read, save, and setup discrete geometry data from the discrete input file.
    // Modifications or replacement of this routine will be required for
    // alternative means of setting discrete data.
    // Note that this must be called after aflr4_cad_geom_setup.

    //status = aflr4_discrete_geom_setup (Message_Flag, AFLR4_Param_Struct_Ptr);
    //if (status != CAPS_SUCCESS) return -1; // Add error code

    // Register evaluation routine for discrete boundary edge grids.
    // A different evaluation routine may be needed for registration if an
    // alternative means of setting discrete data is required.
    // If there is no discrete data then this is never used and not required.
    dgeom_register_discrete_eval_bedge (aflr4_discrete_eval_bedge);
    //if (status != CAPS_SUCCESS) return -1; // Add error code

    // Set labels and flags for all CAD surface component, discrete surface
    // component, standard composite, and one glue-only composite definitions.
    // If alternative means of setting this data are needed then a modified
    // routine that accomplishes the same function must be created.
    status = aflr4_geom_setup (Message_Flag, AFLR4_Param_Struct_Ptr);
    if (status != CAPS_SUCCESS) {
        printf("aflr4_geom_setup status = %d\n", status);
        goto cleanup;
    }

    // Define discrete and/or CAD geometry component surfaces within the DGEOM
    // data structure. Discrete component surface grids are extracted from a
    // given complete discrete surface grid with multiple components and a
    // uv-map is generated. Those definitions that are part of a composite are
    // added to the composite and a uv-map is generated for each standard
    // composite definition.
    status = aflr4_geom_def (Message_Flag, AFLR4_Param_Struct_Ptr);
    if (status != CAPS_SUCCESS) {
        printf("aflr4_geom_def status = %d\n", status);
        goto cleanup;
    }

    // Generate optimal surface grids for all active discrete and/or CAD
    // geometry component surfaces along with all standard composite surfaces
    // and store data within the DGEOM data structure. Also, glue final
    // glue-only composite surfaces.

    status = aflr4_grid_gen (Message_Flag, AFLR4_Param_Struct_Ptr);
    if (status != CAPS_SUCCESS) {
        printf("aflr4_grid_gen status = %d\n", status);
        goto cleanup;
    }

    // Write final surface grid data file.
    // This is not required if an output file of the surface grid is not needed.
    // if (status == 0) status = aflr4_write_surf_grid_file (Message_Flag, AFLR4_Param_Struct_Ptr);

    // To extract a particular surface definition grid directly use the following
    // call. Note that you must free the returned arrays separately as they are

    // Get output id index (glue-only composite)
    status = aflr4_get_idef_output (&glueId);
    // If no definition is set then exit with error (i.e. multiple glue-only composites)
    if (status == 0) {

        printf("Multiple glue-only composites found for body (Possibly a non-closed mesh).\n");
        printf("aflr4_get_idef_output status = %d\n", status);
        goto cleanup;
    }

    // Build up the body tessellation object
    status = EG_openTessBody(surfaceMesh->bodyTessMap.egadsTess);
    if (status != EGADS_SUCCESS) {
        surfaceMesh->bodyTessMap.numTessFace  = 0;
        goto cleanup;
    }

    numSurface = dgeom_def_get_ndef(); // Get number of surfaces meshed

    if (numSurface > 0) {
        surfaceMesh->bodyTessMap.numTessFace = (numSurface-1); // Number of faces in the tessellation
        surfaceMesh->bodyTessMap.tessFaceQuadMap = (int *) EG_alloc((numSurface-1)*sizeof(int));
        if (surfaceMesh->bodyTessMap.tessFaceQuadMap == NULL) {
            surfaceMesh->bodyTessMap.numTessFace = 0;
            status = EGADS_MALLOC;
            goto cleanup;
        }
    } else {
        printf("Number of surfaces from dgeom_def_get_ndef = %d\n", numSurface);
        status = CAPS_BADVALUE;
        goto cleanup;
    }

    for (surf = 0; surf < numSurface ; surf++) {

        if (surf+1 == glueId) continue;

        status = aflr4_get_def (surf+1,
                                1, // If there are quads them get them.
                                &numTriFace,
                                &numNodes,
                                &numQuadFace,
                                &blFlag,
                                &bcFlag,
                                &triCon,
                                &quadCon,
                                &uv,
                                &xyz);
        if (status != CAPS_SUCCESS) goto cleanup; // Add error code

        surfaceMesh->bodyTessMap.tessFaceQuadMap[surf] = numQuadFace;

        status = buildTessellationObj(surf+1,
                                      numNodes,
                                      numTriFace,
                                      numQuadFace,
                                      uv,
                                      xyz,
                                      triCon,
                                      quadCon,
                                      &surfaceMesh->bodyTessMap.egadsTess);


        if (bcFlag != NULL) ug_free (bcFlag);
        bcFlag = NULL;
        if (blFlag != NULL) ug_free (blFlag);
        blFlag = NULL;
        if (triCon != NULL) ug_free (triCon);
        triCon = NULL;
        if (quadCon != NULL) ug_free (quadCon);
        quadCon = NULL;
        if (uv != NULL) ug_free (uv);
        uv = NULL;
        if (xyz != NULL) ug_free (xyz);
        xyz = NULL;

        if (status != CAPS_SUCCESS) goto cleanup;

    }

    status = EG_statusTessBody(surfaceMesh->bodyTessMap.egadsTess, &bodyIn, &i, &numNodes);
    if (status != EGADS_SUCCESS) {
        printf("\nTessellation object was not built correctly!!!\n");
        goto cleanup;
    }

    //printf(" Glue-ID = %d\n",glueId);

    status = mesh_surfaceMeshEGADSTess(surfaceMesh->bodyTessMap.egadsTess,
                                       &attrMap,
                                       surfaceMesh);
    if (status != CAPS_SUCCESS) goto cleanup;


    // Free all CAD related array space (and re-compress CAD input data file).
    // Modifications or replacement of this routine may be required for
    // alternative CAD systems and implementations.

    status = CAPS_SUCCESS;

    goto cleanup;

    cleanup:
        (void) EG_deleteObject(tess); //aflr4_CAD_Cleanup();

        aflr4_cad_geom_data_cleanup();

        // Free all other array space.
        if (prog_argv != NULL) {
            string_freeArray(prog_argc, &prog_argv);
        }

        ug_free_param(AFLR4_Param_Struct_Ptr);

        dftr3_eval_free();
        dgeom_def_free_all();

        return status;
}
