/*
 *      CAPS: Computational Aircraft Prototype Syntheses
 *
 *             AVL AIM
 *
 *      Copyright 2014-2018, Massachusetts Institute of Technology
 *      Licensed under The GNU Lesser General Public License, version 2.1
 *      See http://www.opensource.org/licenses/lgpl-2.1.php
 *
 *      Modified by Ryan Durscher AFLR/RQVC May 10, 2016
 */

#include <string.h>
#include <math.h>
#include "aimUtil.h"
#include "caps.h"

#include "miscUtils.h" // Bring in miscellaneous utilities
#include "vlmUtils.h"

#ifdef WIN32
#define getcwd   _getcwd
#define PATH_MAX _MAX_PATH

#define strcasecmp  stricmp
#define strncasecmp _strnicmp

#else

#include <unistd.h>
#include <limits.h>
#endif

#define NUMINPUT  11
#define NUMOUT    88
#define MAXPOINT  200
#define PI        3.1415926535897931159979635
#define NINT(A)         (((A) < 0)   ? (int)(A-0.5) : (int)(A+0.5))

//#define DEBUG

/*!\mainpage Introduction
 * \tableofcontents
 *
 * \section overviewAVL AVL AIM Overview
 * The use of lower-dimensional design tools is clearly desirable in a multidisciplinary/multi-fidelity aero
 * design optimization setting. This is the crux of the Computational Aircraft Prototype Syntheses (CAPS) program.
 * In many ways describing geometry appropriate for AVL (the Athena Vortex Lattice) code is more cumbersome than
 * higher fidelity codes that require an Outer Mold Line. The goal is to make a CAPS AIM (Analysis Input Module)
 * that directly feeds input to AVL and extracts the output quantities of interest from AVL's execution. This
 * needs to be consistent with a build description that is hierarchical and multi-fidelity. That is, the build
 * description that generates the geometric data at this level can be further enhanced to produce the complete
 * OML of the aircraft design under consideration. As for the geometric description, AVL requires airfoil section
 * data specified at the appropriate locations that describe the <em> skeleton </em> of the aircraft. These
 * sections when <em> lofted </em> as groups and finally <em> unioned </em> together builds the OML. Clearly,
 * intercepting the state of the geometry before these higher-level operations are applied provides the data
 * appropriate for AVL. This naturally constructs a hierarchical geometric view where a design can progress
 * into higher fidelities and feedback can be achieved where we can go back to this level of description when need be.
 *
 * An outline of the AIM's inputs and outputs are provided in \ref aimInputsAVL and \ref aimOutputsAVL, respectively. An
 * alternative to the AIM's outputs for retrieving sensitivity information is provided in \ref aimBackDoorAVL.
 *
 * The accepted and expected geometric representation and analysis intentions are detailed in \ref geomRepIntentAVL. Similarly, other geometric attribution
 * that the AIM makes use is provided in \ref attributeAVL.
 *
 * Upon running preAnalysis the AIM generates two files, 1) "avlInput.txt" which contains the input information and
 * control sequence for AVL to execute and 2) "caps.avl" which contains the geometry to be analyzed.
 * To populate output data the AIM expects files, "capsTotalForce.txt", "capsStatbilityDeriv.txt", "capsBodyAxisDeriv.txt", and
 * "capsHingeMoment.txt" to exist after running AVL
 * (see \ref aimOutputsAVL for additional information). An example execution for AVL looks like:
 *
 * \code{.sh}
 * avl caps < avlInput.txt > avlOutput.txt"
 * \endcode
 *
 * \section assumptionsAVL Assumptions
 * The AVL coordinate system assumption (X -- downstream, Y -- out the right wing, Z -- up) needs to be followed.
 *
 * Within <b> OpenCSM</b> there are a number of airfoil generation UDPs (User Defined Primitives). These include NACA 4
 * series, a more general NACA 4/5/6 series generator, Sobieczky's PARSEC parameterization and Kulfan's CST
 * parameterization. All of these UDPs generate <b> EGADS</b> <em> FaceBodies</em> where the <em>Face</em>'s underlying
 * <em>Surface</em>  is planar and the bounds of the <em>Face</em> is a closed set of <em>Edges</em> whose
 * underlying <em>Curves</em> contain the airfoil shape. In all cases, there is a <em>Node</em> that represents
 * the <em>Leading Edge</em> point and one or two <em>Nodes</em> at the <em>Trailing Edge</em> -- one if the
 * representation is for a sharp TE and the other if the definition is open or blunt. If there are 2 <em>Nodes</em>
 * at the back, then there are 3 <em>Edges</em> all together and closed, even though the airfoil definition
 * was left open at the TE. All of this information will be used to automatically fill in the AVL geometry description.
 *
 * The AVL Sections are automatically generated, one from each <em> FaceBody</em> and the details extracted from the geometry.
 * <b> Xle</b>, <b> Yle</b>, and <b> Zle</b>, are taken from the <em> Node</em> Associated with the <em> Leading Edge</em>.
 * The <b> Chord</b> is computed by getting the distance between the LE and TE (if there are 3 <em> Edges</em> in the
 * <em> FaceBody</em> the TE point is considered the mid-position on that third <em> Edge</em>). <b> Ainc</b> is computed by
 * registering the chordal direction of the <em> FaceBody</em> against the X-Z plane. The airfoil shapes are generated by
 * sampling the <em> Curves</em> and put directly in the input file via the <b> AIRFOIL</b> keyword after being normalized.
 *
 * It should be noted that general construction in either <b> OpenCSM</b> or even <b> EGADS</b> will be supported
 * as long as the topology described above is used. But care should be taken when constructing the airfoil shape
 * so that a discontinuity (i.e.,  simply <em>C<sup>0</sup></em>) is not generated at the <em>Node</em> representing
 * the <em>Leading Edge</em>. This can be done by splining the entire shape as one and then intersecting the single
 *  <em>Edge</em> to place the LE <em>Node</em>.
 *
 *  The rest of the information and options required to fill out the AVL geometry input file (<b> xxx.avl</b>)
 *  will be found in the attributes attached to the <em>FaceBody</em> itself. The conventions used will be
 *  described in the next section.
 *
 * Also note that this first implementation is not intended to provide complete control over AVL. In particular,
 * there is no mention above of the <b> BODY</b>, <b> DESIGN</b>, <b> CLAF</b>, or <b> CDCL</b> AVL keywords.
 *
 * \section avlExample Examples
 * An example problem using the AVL AIM may be found at \ref examplesAVL, which
 * contains example *.csm input files and pyCAPS scripts designed to make use of the AVL AIM.  These example
 * scripts make extensive use of the \ref attributeAVL, \ref aimInputsAVL, and \ref aimOutputsAVL.
 *
 */

/*! \page examplesAVL AIM Examples
 * This example contains a set of *.csm and pyCAPS (*.py) inputs that uses the AVL AIM.  A user should have knowledge
 * on the generation of parametric geometry in Engineering Sketch Pad (ESP) before attempting to integrate with any AIM.
 * Specifically, this example makes use of Design Parameters, Set Parameters, User Defined Primitive (UDP) and attributes
 * in ESP.
 *
 * The follow code details the process in a *.csm file that generates three airfoil sections to create a wing. Note
 * to execute in serveCSM a dictionary file must be included
 * "serveCSM -dict $ESP_ROOT/include/intent.dict avlDocExample1.csm"
 *
 * First step is to define the analysis intent that the geometry is intended support.
 * \snippet avlWing.csm capsIntent
 *
 * Next we will define the design parameters to define the wing cross section and planform.
 * \snippet avlWing.csm designParameters
 *
 * The design parameters will then be used to set parameters for use internally to create geometry.
 * \snippet avlWing.csm setParameters
 *
 * Finally, the airfoils are created using the User Defined Primitive (UDP) naca. The inputs used for this
 * example to the UDP are Thickness and Camber. Cross sections are in the X-Y plane and are rotated to the X-Z plane.
 * Reference quantities must exist on any body, otherwise AVL defaults to 1.0 for Area, Span,
 * Chord and 0.0 for X,Y,Z moment References
 *
 * \snippet avlWing.csm createAirfoils
 *
 * An example pyCAPS script that uses the above csm file to run AVL is as follows.
 *
 * First the pyCAPS and os module needs to be imported.
 * \snippet avl_PyTest.py import
 *
 * Note if your Python major version is less than 3 (i.e. Python 2.7). The following statement should also be included
 * so that print statements work correctly.
 *
 * \snippet avl_PyTest.py importPrint
 *
 * Once the modules have been loaded the problem needs to be initiated.
 *
 * \snippet avl_PyTest.py initateProblem
 *
 * Next the *.csm file is loaded and design parameter is changed - area in the geometry. Any despmtr from the
 * avlWing.csm file is available inside the pyCAPS script. They are: thick, camber, area, aspect, taper,
 * sweep, washout, dihedral...
 *
 * \snippet avl_PyTest.py geometry
 *
 * The AVL AIM is then loaded with the capsIntent set to LINEARAERO (this is consistent with the intent specified
 * above in the *.csm file.
 * \snippet avl_PyTest.py loadAIM
 *
 * After the AIM is loaded the Mach number and angle of attack are set, though all \ref aimInputsAVL are available.
 * \snippet avl_PyTest.py setInputs
 *
 * Once all the inputs have been set, preAnalysis needs to be executed. During this operation, all the necessary files to
 * run AVL are generated and placed in the analysis working directory (analysisDir)
 * \snippet avl_PyTest.py preAnalysis
 *
 * An OS system call is then made from Python to execute AVL.
 * \snippet avl_PyTest.py runAVL
 *
 * A call to postAnalysis is then made to check to see if AVL executed successfully and the expected files were generated.
 * \snippet avl_PyTest.py postAnalysis
 *
 * Similar to the AIM inputs, after the execution of AVL and postAnalysis any of the AIM's output
 * variables (\ref aimOutputsAVL) are readily available; for example,
 * \snippet avl_PyTest.py output
 *
 * results in
 * \code
 * CXtot   -0.00033
 * CYtot   1e-05
 * CZtot   -0.30016
 * Cltot   -0.0
 * Cmtot   -0.19468
 * Cntot   -1e-05
 * Cl'tot  -0.0
 * Cn'tot  -1e-05
 * CLtot   0.30011
 * CDtot   0.00557
 * CDvis   0.0
 * CLff    0.29968
 * CYff    0.0
 * CDind   0.00557
 * CDff    0.00492
 * e       0.9691
 * \endcode
 *
 * Additionally, besides making a call to the AIM outputs, sensitivity values may be obtained in the following manner,
 * \snippet avl_PyTest.py sensitivity
 *
 * The avlAIM supports the control surface modeling functionality inside AVL.  Trailing edge control surfaces can be
 * added to the above example by making use of the vlmControlName attribute (see \ref attributeAVL regarding the attribution
 * specifics).  To add a <b>RightFlap</b> and <b>LeftFlap</b> to the
 * previous example *.csm file the naca UDP entries are augmented with the following attributes.
 *
 * \code {.py}
 * # left tip
 * udprim    naca      Thickness thick     Camber    camber
 * attribute vlmControl_LeftFlap 80 # Hinge line is at 80% of the chord
 * ...
 *
 * # root
 * udprim    naca      Thickness thick     Camber    camber
 * attribute vlmControl_LeftFlap 80 # Hinge line is at 80% of the chord
 * attribute vlmControl_RightFlap 80 # Hinge line is at 80% of the chord
 * ...
 *
 * # right tip
 * udprim    naca      Thickness thick     Camber    camber
 * attribute vlmControl_RightFlap 80 # Hinge line is at 80% of the chord
 * ...
 * \endcode
 *
 * Note how the root airfoil contains two attributes for both the left and right flaps.
 *
 * In the pyCAPS script the \ref aimInputsAVL, <b>AVL_Control</b>, must be defined.
 * \code {.py}
 * flap = {"controlGain"     : 0.5,
 * 		  "deflectionAngle" : 10.0}
 *
 *  myAnalysis.setAnalysisVal("AVL_Control", [("LeftFlap", flap), ("RightFlap", flap)])
 \endcode
 *
 * Notice how the information defined in the <b>flap</b> variable is assigned to the vlmControl<b>Name</b> portion of the
 * attributes added to the *.csm file.
 *
 */

/*! \page attributeAVL AIM Attributes
 * The following list of attributes drives the AVL geometric definition. Each <em>FaceBody</em> which relates to AVL
 * <b> Sections</b> will be marked up in an appropriate manner to drive the input file construction. Many attributes
 * are required and those that are optional are marked so in the description:
 *
 * - <b> capsIntent</b> This attribute is a CAPS requirement to indicate the analysis intent the geometry
 * representation supports. Options are:  ALL, LINEARAERO
 *
 * - <b> capsReferenceArea</b>  [Optional: Default 1.0] This attribute may exist on any <em> Body</em>.  Its
 * value will be used as the SREF entry in the AVL input.
 *
 * - <b> capsReferenceChord</b>  [Optional: Default 1.0] This attribute may exist on any <em> Body</em>.  Its
 * value will be used as the CREF entry in the AVL input.
 *
 * - <b> capsReferenceSpan</b>  [Optional: Default 1.0] This attribute may exist on any <em> Body</em>.  Its
 * value will be used as the BREF entry in the AVL input.
 *
 * - <b> capsReferenceX</b>  [Optional: Default 0.0] This attribute may exist on any <em> Body</em>.  Its
 * value will be used as the Xref entry in the AVL input.
 *
 * - <b> capsReferenceY</b>  [Optional: Default 0.0] This attribute may exist on any <em> Body</em>.  Its
 * value will be used as the Yref entry in the AVL input.
 *
 * - <b> capsReferenceZ</b>  [Optional: Default 0.0] This attribute may exist on any <em> Body</em>.  Its
 *  value will be used as the Zref entry in the AVL input.
 *
 * - <b> capsGroup</b> This string attribute labels the <em> FaceBody</em> as to which AVL Surface the section
 *  is assigned. This should be something like: <em> Main\_Wing</em>, <em> Horizontal\_Tail</em>, etc. This informs
 *  the AVL AIM to collect all <em> FaceBodies</em> that match this attribute into a single AVL Surface.
 *
 * - <b> vlmControlName</b> This string attribute attaches a control surface to the <em> FaceBody</em>. The hinge
 * location is defined as the double value between 0 or 1.0.  The range as percentage from 0 to 100
 * will also work. The name of the
 * control surface is the string information after vlmControl (or vlmControl\_).  For Example, to define a control
 * surface named Aileron the following are identical (<em>attribute vlmControlAileron 0.8 </em> or
 * <em>attribute vlmControl\_Aileron 80</em>) . Multiple <em> vlmControl</em> attributes, with different
 * names, can be defined on a single <em> FaceBody</em>.
 * <br> <br>By default control surfaces with percentages less than 0.5 (< 50%) are considered leading edge flaps, while values
 * greater than or equal to 0.5 (>= 50%) are considered trailing edge flaps. This behavior may be overwritten when setting up
 * the control surface in "AVL_Control" (see \ref aimInputsAVL) with the keyword "leOrTe" (see \ref vlmControl for additional details).
 *
 * - <b> avlNumSpan </b> This attribute may be set on any given airfoil cross-section to overwrite the number of
 * spanwise horseshoe vortices placed on the surface (globally set - see keyword "numSpan" in \ref vlmSurface ) between two sections. Note, that the AIM internally sorts 
 * the sections in ascending y (or z) order, so care should be taken to select the correct section for the desired intent.
 *
 */


// To be changed
/*  - <b> avlComponent</em>  [optional]. This integer (or real) attribute is examined for the AVL Component index.
 *  This may be on any or all <em> FaceBodies</em> associated with the AVL Surface.

- <b> avlKeyword</em>  [optional]. This string attribute may contain the word (or words): {\tt NOWAKE</em>, {\tt NOALBE</em> and/or {\tt NOLOAD</em>. Multiple keywords need
to be delimited by a colon (`:'). This is associated with the AVL Surface.


\item {\bf avlSecValues.} This real valued attribute must be of at least 5 {\it doubles} in length and contain:
   \begin{enumerate}
   \item {\tt Nchord} -- The number of chordwise horseshoe vortices placed on the surface.
   \item {\tt Cspace} -- The chordwise vortex spacing parameter.
   \item {\tt Nspan} -- The number of spanwise horseshoe vortices placed on the surface.
   \item {\tt Sspace} -- The spanwise vortex spacing parameter.
   \item {\tt iYdup} -- The Y symmetry for this Surface (0 -- do nothing, 1 -- duplicate about the Y=0.0 plane).
   \end{enumerate}
   It should be noted that the first 2 (and last) values refer to the entire AVL Surface/capsGroup and therefore the first occurrence of the \textbf{\textit {capsGroup}} attribute in the list of {\it bodies} sets these values in the AVL input file (though {\tt iYdup} can be set in any Section).

\item {\bf avlCntrl}\textbf{\textit {Name}}  [optional]. This optional numerical attribute indicates that this {\it FaceBody} Section is a member of the control surface named as part of the attribute name (for example:  {\tt avlCntrlAileron} specifies that this Section is part of the {\it Aileron} control surface). This real valued attribute must be of at least 6 {\it doubles} in length and contain:
   \begin{enumerate}
   \item {\tt gain} -- The control deflection gain.
   \item {\tt Xhinge} -- The x/c location of hinge.
   \item {\tt Xhvec} -- The first component of the vector giving hinge axis about which surface rotates.
   \item {\tt Yhvec} -- The second component.
   \item {\tt Zhvec} -- The third component.
   \item {\tt SgnDup } -- The signed magnitude (or simply the scale factor) for the deflection of the duplicated surface.
   \end{enumerate}
\end{itemize}
 */


/*
  typedef struct {
    char   *name;
    double Cspace;
    int    Nchord;
    int    nowake;
    int    noalbe;
    int    noload;
    int    compon;
    int    iYdup;
    int    nsec;
    double xyz[3];        per body
    int    le;
    int    index;         per section
  } aimSurface;
 */


typedef struct {
    // Analysis file path/directory
    char *analysisPath;

    mapAttrToIndexStruct controlMap;

} aimStorage;

static aimStorage *avlInstance = NULL;
static int         numInstance  = 0;

//extern char *EG_strdup(const char *str);

/* ********************** AVL AIM Helper Functions ************************** */

static int writeSection(FILE *fp, ego body,
                        int nodeIndexLE, double *xyzLE,
                        int nodeIndexTE, double *xyzTE,
                        int Nspan, double Sspace)
{
    int status; // Function return status

    int i, j, k; // Indexing

    int found; // Boolean

    int edgeIndex, *edgeOrder=NULL, *tempOrder=NULL;

    int numPoint = 0;

    double       chord, ainc;
    double       xdot[3], ydot[3];

    // Attribute variables
    int          atype, alen;
    const int    *ints;
    const char   *string;
    const double *reals;

    int numEdge, numNode, numLoop;
    ego *nodes = NULL, *edges = NULL, *loops = NULL;

    int sense = 0;
    double xyz[3], xmin, trange[2], result[9];

    //EGADS returns
    int oclass, mtype, *sens = NULL, *edgeSense = NULL, numChildren;
    ego ref, *children = NULL, *temp = NULL;


    ego teObj = NULL;

    int numSpan = 0;

    /* get LE position, Chord and Ainc from the Body */
    status = EG_getBodyTopos(body, NULL, NODE, &numNode, &nodes);
    if (status != EGADS_SUCCESS) {
        printf("\tError in AVL writeSection, getBodyTopos Nodes = %d\n", status);
        goto cleanup;
    }

    status = EG_getBodyTopos(body, NULL, EDGE, &numEdge, &edges);
    if (status != EGADS_SUCCESS) {
        printf("\tError in AVL writeSection, getBodyTopos Edges = %d\n", status);
        goto cleanup;
    }

    status = EG_getBodyTopos(body, NULL, LOOP, &numLoop, &loops);
    if (status != EGADS_SUCCESS) {
        printf("\tError in AVL writeSection, getBodyTopos Loops = %d\n", status);
        goto cleanup;
    }

    numPoint = (int) MAXPOINT/numEdge;

    // if ((nEdge != nNode) || (nNode < 2) || (nNode > 3)) {

    if ((numEdge != numNode) || (numNode < 2) || (numLoop != 1)) { // Remove 3 node constraint
        printf("\tError in AVL writeSection, body has %d nodes, %d edges and %d loops!\n", numNode, numEdge, numLoop);
        status = CAPS_SOURCEERR;
        goto cleanup;
    }

    // Find TE Edge object, if any
    teObj = NULL;
    for (i = 0; i < numEdge; i++) {

        status = EG_getTopology(edges[i], &ref, &oclass, &mtype, trange, &numChildren, &children, &sens);

        if (status != EGADS_SUCCESS) {
            printf("\tError in AVL writeSection, Edge %d getTopology = %d!\n", i, status);
            goto cleanup;
        }

        if (numChildren != 2) {
            printf("\tError in AVL writeSection, Edge %d has %d nodes!\n", i, numChildren);
            goto cleanup;
        }

        if (numEdge <= 3) {
            if ( (children[0] != nodes[nodeIndexLE-1]) &&
                    (children[1] != nodes[nodeIndexLE-1])) {

                teObj = edges[i];
                // printf(" TE Edge = %d\n", i);
                break;
            }

        } else {

            // If the edge doesn't at least contain the TE node pass it by
            if (children[0] != nodes[nodeIndexTE-1] && children[1] != nodes[nodeIndexTE-1]) continue;

            // Lets compare the coordinates
            status = EG_getTopology(children[0], &ref, &oclass, &mtype, xyz, &numChildren, &temp, &sens);
            if (status != EGADS_SUCCESS) goto cleanup;

            xmin = xyz[0];

            status = EG_getTopology(children[1], &ref, &oclass, &mtype, xyz, &numChildren, &temp, &sens);
            if (status != EGADS_SUCCESS) goto cleanup;

            if (fabs(xyz[0] - xmin) < 10E-4) { // If we are at the same x-location

                teObj = edges[i];
                break;
            }
        }
    }

    //if ((numEdge == 3) && (teObj == NULL)) {
    //	printf("\tError in AVL writeSection, No TE edge found with number of edges = 3!\n");
    //	goto cleanup;
    //}

    if (teObj == NULL) {

        status = EG_getTopology(nodes[nodeIndexTE-1], &ref, &oclass, &mtype, result, &numChildren, &children, &sens);
        if (status != EGADS_SUCCESS) goto cleanup;

    } else {

        status = EG_getTopology(teObj, &ref, &oclass, &mtype, trange, &numChildren, &children, &sens);
        if (status != EGADS_SUCCESS) goto cleanup;

        xmin = 0.5*(trange[0]+trange[1]);

        status = EG_evaluate(teObj, &xmin, result);
        if (status != EGADS_SUCCESS) {
            printf("\tError in AVL writeSection, evaluate = %d!\n", status);
            goto cleanup;
        }
    }

    xdot[0]  =  result[0] - xyzLE[0];
    xdot[1]  =  result[1] - xyzLE[1];
    xdot[2]  =  result[2] - xyzLE[2];

    chord    =  sqrt(xdot[0]*xdot[0] + xdot[1]*xdot[1] + xdot[2]*xdot[2]);

    xdot[0] /=  chord;
    xdot[1] /=  chord;
    xdot[2] /=  chord;

    //	ainc     = -acos(xdot[0])*180./PI;
    ainc     = -atan2(xdot[2], xdot[0])*180./PI;

    ydot[0]  = -xdot[2];
    ydot[1]  =  xdot[1];
    ydot[2]  =  xdot[0];

    if (fabs(xdot[1]) > 1.e-5){
        printf("\tWarning AVL writeSection, Section not in Y-plane!\n");
    }

    numSpan = Nspan;
    status = EG_attributeRet(body, (const char *) "avlNumSpan", &atype, &alen, &ints, &reals, &string);
    if (status == EGADS_SUCCESS) {

        if (atype != ATTRINT && atype != ATTRREAL && alen != 1) {
            printf ("Error: Attribute %s should be followed by a single integer\n", "avlNumSpan");
        }

        if (atype == ATTRINT) {
            numSpan = ints[0];
        }

        if (atype == ATTRREAL) {
            numSpan = (int) reals[0];
        }
    }

    fprintf(fp, "SECTION\n%lf %lf %lf  %lf %lf  %d %lf\n\n",
            xyzLE[0], xyzLE[1], xyzLE[2], chord, ainc, numSpan, Sspace);

    fprintf(fp, "AIRFOIL\n");

    // Determine edge order
    tempOrder = (int *) EG_alloc(numEdge*sizeof(int));
    edgeOrder = (int *) EG_alloc(numEdge*sizeof(int));
    if ( (tempOrder == NULL) ||
            (edgeOrder == NULL)   ) {

        status = EGADS_MALLOC;
        goto cleanup;
    }


    for (i = 0; i < numEdge; i++) tempOrder[i] = CAPSMAGIC;

    status = EG_getTopology(loops[0], &ref, &oclass, &mtype, trange, &numChildren, &children, &edgeSense);
    if (status != EGADS_SUCCESS) goto cleanup;

    for (i = 0; i < numChildren; i++) { // Loop of edges in loop

        //printf("edgeIndex = %d\n", edgeIndex);
        edgeIndex = EG_indexBodyTopo(body, children[i]);

        if (edgeIndex < EGADS_SUCCESS) {
            status = CAPS_BADINDEX;
            goto cleanup;
        }

        //if (children[i] == teObj) continue;

        tempOrder[i] = edgeIndex;
    }

    // Reorder indexing
    found = (int) false;
    j = 0;
    if (teObj != NULL) { // If we have a trailing edge

        for (i = 0; i < numEdge; i++) {
            if (edges[tempOrder[i]-1] == teObj) found = (int) true;

            if (found == (int) true) { // Start saving edge indices
                edgeOrder[j] = tempOrder[i];
                j += 1;
            }
        }

    } else { // If we don't have a trailing edge - reorder based on trailing edge index

        for (i = 0; i < numEdge; i++) {

            status = EG_getTopology(edges[tempOrder[i]-1], &ref, &oclass, &mtype, trange, &numChildren, &children, &sens);
            if (status != EGADS_SUCCESS) goto cleanup;

            if (children[0] == nodes[nodeIndexTE-1]) found = (int) true;

            if (found == (int) true) { // Start saving edge indices
                edgeOrder[j] = tempOrder[i];
                j += 1;
            }
        }
    }

    // Fill in the rest of edge order
    i = 0;
    for (k = j; k < numEdge; k++) {
        edgeOrder[k] = tempOrder[i];
        i += 1;
    }

    // Write out points

    //fprintf(fp, "1.000000 %lf\n",
    //    				(ydot[0]*xyz[0]+ydot[1]*xyz[1]+ydot[2]*xyz[2])/chord);


    found = (int) false;

    // Loop through edges based on order
    for (i = 0; i < numEdge; i++) {
        //printf("Edge order %d\n", edgeOrder[i]);

        edgeIndex = edgeOrder[i] - 1; // -1 indexing

        if (edges[edgeIndex] == teObj) continue;

        // Get t- range for edge
        status = EG_getTopology(edges[edgeIndex], &ref, &oclass, &mtype, trange, &numChildren, &children, &sens);
        if (status != EGADS_SUCCESS) goto cleanup;

        // Get the sense of the edge from the loop
        for (j = 0; j < numEdge; j++) {
            sense = edgeSense[j];
            if (tempOrder[j] == edgeOrder[i]) break;
        }


        // Sense = 1
        //  LE <- TE ^ - Counter clockwise rotation
        //  LE -> TE

        // Sense = -1
        // ^ LE -> TE  - Clockwise rotation
        //   LE <- TE

        // Pick node based on sense
        if (sense == -1) {

            // There may be an issue if we have multiple edges with a sense of -1 - have fully tested

            // Start node
            if (children[0] == nodes[nodeIndexLE-1]) {
                fprintf(fp, "0.000000 0.000000\n");

            }

            if (found == (int) false) { // TE - first node occurrence of the trailing edge - enforce 1.0 start

                status = EG_getTopology(children[1], &ref, &oclass, &mtype, xyz, &numChildren, &temp, &sens);
                if (status != EGADS_SUCCESS) goto cleanup;

                xyz[0] -= xyzLE[0];
                xyz[1] -= xyzLE[1];
                xyz[2] -= xyzLE[2];

                fprintf(fp, "1.000000 %lf\n",(ydot[0]*xyz[0] +
                                              ydot[1]*xyz[1] +
                                              ydot[2]*xyz[2])/chord);
                found = (int) true;
            }

            // End node
            status = EG_getTopology(children[0], &ref, &oclass, &mtype, xyz, &numChildren, &temp, &sens);
            if (status != EGADS_SUCCESS) goto cleanup;


        } else {

            // Start end
            if (children[0] == nodes[nodeIndexLE-1]) {
                fprintf(fp, "0.000000 0.000000\n");
            }

            if (found == (int) false) { // TE - first node occurrence of the trailing edge - enforce 1.0 start

                status = EG_getTopology(children[0], &ref, &oclass, &mtype, xyz, &numChildren, &temp, &sens);
                if (status != EGADS_SUCCESS) goto cleanup;

                xyz[0] -= xyzLE[0];
                xyz[1] -= xyzLE[1];
                xyz[2] -= xyzLE[2];

                fprintf(fp, "1.000000 %lf\n",(ydot[0]*xyz[0] +
                        ydot[1]*xyz[1] +
                        ydot[2]*xyz[2])/chord);
                found = (int) true;
            }

            // End node
            status = EG_getTopology(children[1], &ref, &oclass, &mtype, xyz, &numChildren, &temp, &sens);
            if (status != EGADS_SUCCESS) goto cleanup;
        }

        // Subtract leading edge from end node - do be used as the last point
        xyz[0] -= xyzLE[0];
        xyz[1] -= xyzLE[1];
        xyz[2] -= xyzLE[2];

        // Write out in points along edge
        for (j = 1; j < numPoint; j++) {
            //if (sense == 1) {
            //	xmin = trange[0] + j*(trange[1]-trange[0])/numPoint;
            //} else {
            //	xmin = trange[1] - j*(trange[1]-trange[0])/numPoint;
            //}

            xmin = trange[0] + j*(trange[1]-trange[0])/numPoint;

            status = EG_evaluate(edges[edgeIndex], &xmin, result);
            if (status != EGADS_SUCCESS) goto cleanup;

            result[0] -= xyzLE[0];
            result[1] -= xyzLE[1];
            result[2] -= xyzLE[2];

            fprintf(fp, "%lf %lf\n", (xdot[0]*result[0]+xdot[1]*result[1]+xdot[2]*result[2])/chord,
                                     (ydot[0]*result[0]+ydot[1]*result[1]+ydot[2]*result[2])/chord);
        }

    }

    // Repeat the last point
    fprintf(fp, "1.000000 %lf\n", (ydot[0]*xyz[0]+ydot[1]*xyz[1]+ydot[2]*xyz[2])/chord);

    fprintf(fp, "\n");

    status = CAPS_SUCCESS;

    goto cleanup;

    cleanup:
        if (status != CAPS_SUCCESS) printf("Error: Premature exit in writeSection, status = %d\n", status);

        if (nodes != NULL) EG_free(nodes);
        if (edges != NULL) EG_free(edges);
        if (loops != NULL) EG_free(loops);

        if (tempOrder != NULL) EG_free(tempOrder);
        if (edgeOrder != NULL) EG_free(edgeOrder);

        return status;
}

static int read_Data(char *file, char *analysisPath, char *key, double *data) {

    int i; //Indexing

    size_t     linecap = 0;
    char       currentPath[PATH_MAX], *str = NULL, *valstr = NULL, *line = NULL;
    FILE       *fp = NULL;

    *data = 0.0;

    if (file == NULL) return CAPS_NULLVALUE;
    if (analysisPath == NULL) return CAPS_NULLVALUE;
    if (key == NULL) return CAPS_NULLVALUE;

    (void) getcwd(currentPath, PATH_MAX);

    if (chdir(analysisPath) != 0) return CAPS_DIRERR;

    // Open the AVL output file
    fp = fopen(file, "r");
    if (fp == NULL) {

        chdir(currentPath);

        return CAPS_DIRERR;
    }

    // Scan the file for the string
    valstr = NULL;

    while (getline(&line, &linecap, fp) >= 0) {

        if (line == NULL) continue;

        // printf("Key = %s\n", key);

        valstr = strstr(line, key);

        if (valstr != NULL) {

            // Check string for ***
            for (i = 0; i < strlen(valstr); i++) {
                if (valstr[i] == '*'){
                    *data = 0.0;

                    printf("Unreal value for variable %s - setting it to 0.0\n", str);
                    break;
                }
            }
            // Found it -- get the value
            sscanf(&valstr[strlen(key)+1], "%lf", data);
            break;
        }
    }

    // Restore the path we came in with
    chdir(currentPath);
    fclose(fp);

    if (line != NULL) EG_free(line);

    return CAPS_SUCCESS;
}

static int get_controlDeriv(void *aimInfo, char *analysisPath, int controlIndex, int outputIndex, double *data) {

    int status; // Function return status
    char *fileToOpen;

    char *coeff = NULL;
    char key[10];

    // Stability axis
    if        (outputIndex == aim_getIndex(aimInfo, "CLtot", ANALYSISOUT)) {
        fileToOpen = "capsStatbilityDeriv.txt";
        coeff = "CL";

    } else if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT)) {
        fileToOpen = "capsStatbilityDeriv.txt";
        coeff = "CY";

    } else if (outputIndex == aim_getIndex(aimInfo, "Cl'tot", ANALYSISOUT)) {
        fileToOpen = "capsStatbilityDeriv.txt";
        coeff = "Cl";

    } else if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT)) {
        fileToOpen = "capsStatbilityDeriv.txt";
        coeff = "Cm";
    } else if (outputIndex == aim_getIndex(aimInfo, "Cn'tot", ANALYSISOUT)) {
        fileToOpen = "capsStatbilityDeriv.txt";
        coeff = "Cn";

        // Body axis
    } else if (outputIndex == aim_getIndex(aimInfo, "CXtot", ANALYSISOUT)) {
        fileToOpen = "capsBodyAxisDeriv.txt";
        coeff = "CX";

    } else if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT)) {
        fileToOpen = "capsBodyAxisDeriv.txt";
        coeff = "CY";

    } else if (outputIndex == aim_getIndex(aimInfo, "CZtot", ANALYSISOUT)) {
        fileToOpen = "capsBodyAxisDeriv.txt";
        coeff = "CZ";

    } else if (outputIndex == aim_getIndex(aimInfo, "Cltot", ANALYSISOUT)) {
        fileToOpen = "capsBodyAxisDeriv.txt";
        coeff = "Cl";

    } else if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT)) {
        fileToOpen = "capsBodyAxisDeriv.txt";
        coeff = "Cm";

    } else if (outputIndex == aim_getIndex(aimInfo, "Cntot", ANALYSISOUT)) {
        fileToOpen = "capsBodyAxisDeriv.txt";
        coeff = "Cn";

    } else {
        printf("Unrecognized output variable for control derivatives!\n");
        status = CAPS_MISMATCH;
        goto cleanup;
    }

    sprintf(key, "%sd%d =", coeff, controlIndex);

    status = read_Data(fileToOpen, analysisPath, key, data);
    if (status != CAPS_SUCCESS) return status;

    status = CAPS_SUCCESS;
    goto cleanup;

    cleanup:
        return status;
}

static int parse_controlName(int iIndex, char string[], int *controlNumber) {

    int status; // Function return status

    char control[] = "AVL_Control";

    char *controlName = NULL;

    *controlNumber = CAPSMAGIC;

    //printf("String = %s\n", string);

    if (strncasecmp(string, control, strlen(control)) != 0 ) {
        status = CAPS_NOTFOUND;
        goto cleanup;
    }

    controlName = strstr(string, ":");
    if (controlName == NULL) {
        status = CAPS_NOTFOUND;
        goto cleanup;
    }

    // Increment pointer to remove ':'
    ++controlName;

    //printf("Temp = %s\n", temp);

    // Loop through and determine which control integer this name corresponds to?
    status = get_mapAttrToIndexIndex(&avlInstance[iIndex].controlMap,
            (const char *) controlName, controlNumber);
    if (status != CAPS_SUCCESS) goto cleanup;

    /*
	printf("ControlName = %s\n", controlName);
	printf("ControlIndex = %d\n", *controlNumber);
     */

    status = CAPS_SUCCESS;
    goto cleanup;

    cleanup:
        return status;
}

/* ********************** Exposed AIM Functions ***************************** */

int aimInitialize(/*@unused@*/ int ngIn, /*@unused@*/ /*@null@*/ capsValue *gIn,
        int *qeFlag, /*@null@*/ const char *unitSys, int *nIntent,
        int **intents, int *nIn, int *nOut, int *nFields, char ***fnames,
        int **ranks)
{
    int *ints, flag;

    aimStorage *tmp;

#ifdef DEBUG
    printf("\n avlAIM/aimInitialize   ngIn = %d!\n", ngIn);
#endif

    flag    = *qeFlag;
    *qeFlag = 0;


    /* specify the number of analysis input and out "parameters" */
    *nIn     = NUMINPUT;
    *nOut    = NUMOUT;
    if (flag == 1) return CAPS_SUCCESS;

    // Number of geometry fidelities that are required by the AIM and what they are
    *nIntent = 4;
    ints     = (int *) EG_alloc(*nIntent*sizeof(int));
    if (ints == NULL) return EGADS_MALLOC;
    ints[0]  = ALL | ALL;
    ints[1]  = ALL | LINEARAERO;
    ints[2]  = FACEBODY | ALL;
    ints[3]  = FACEBODY | LINEARAERO;
    *intents = ints;

    /*! \page geomRepIntentAVL Geometry Representation and Analysis Intent
     * The geometric representation for the AVL AIM requires that "body(ies)" [or cross-sections], be a face
     * body(ies) (FACEBODY) with the attribute <b> capsIntent</b> being set to <b>LINEARAERO</b> (intents of ALL also
     * accepted).
     */

    // Specify the field variables this analysis can generate
    *nFields = 0;
    *ranks   = NULL;
    *fnames  = NULL;

    // Allocate avlInstance
    if (avlInstance == NULL) {

        avlInstance = (aimStorage *) EG_alloc(sizeof(aimStorage));

        if (avlInstance == NULL) {
            EG_free(*intents);
            *intents = NULL;
            return EGADS_MALLOC;
        }

    } else {
        tmp = (aimStorage *) EG_reall(avlInstance, (numInstance+1)*sizeof(aimStorage));
        if (tmp == NULL) {
            EG_free(*intents);
            *intents = NULL;
            return EGADS_MALLOC;
        }

        avlInstance = tmp;
    }

    // Analysis file path/directory
    avlInstance[numInstance].analysisPath = NULL;

    // Intiate control map
    (void) initiate_mapAttrToIndexStruct(&avlInstance[numInstance].controlMap);

    // Increment number of instances
    numInstance += 1;

    return (numInstance -1);
}


int aimInputs(/*@unused@*/ int inst, /*@unused@*/ void *aimInfo, int index,
              char **ainame, capsValue *defval)
{
    /*! \page aimInputsAVL AIM Inputs
     * The following list outlines the AVL inputs along with their default value available
     * through the AIM interface.
     */

#ifdef DEBUG
    printf(" avlAIM/aimInputs instance = %d  index = %d!\n", inst, index);
#endif

    if (index == 1) {
        *ainame           = EG_strdup("Mach");
        defval->type      = Double;
        defval->vals.real = 0.0;
        defval->units     = NULL;

        /*! \page aimInputsAVL
         * - <B> Mach = 0.0 </B> <br>
         *  Mach number.
         */

    } else if (index == 2) {
        *ainame           = EG_strdup("Alpha");
        defval->type      = Double;
        defval->dim       = 0;
        defval->length    = 1;
        defval->nrow      = 1;
        defval->ncol      = 1;
        defval->units     = NULL;
        defval->vals.real = 0.0;
        defval->nullVal   = IsNull;
        defval->lfixed    = Change;
        defval->sfixed    = Change;
        defval->units     = EG_strdup("degree");

        /*! \page aimInputsAVL
         * - <B> Alpha = NULL </B> <br>
         *  Angle of attack [degree]. Either CL or Alpha must be defined but not both.
         */

    } else if (index == 3) {
        *ainame           = EG_strdup("Beta");
        defval->type      = Double;
        defval->vals.real = 0.0;
        defval->units     = NULL;
        defval->units     = EG_strdup("degree");

        /*! \page aimInputsAVL
         * - <B> Beta = 0.0 </B> <br>
         *  Sideslip angle [degree].
         */

    } else if (index == 4) {
        *ainame           = EG_strdup("RollRate");
        defval->type      = Double;
        defval->vals.real = 0.0;
        defval->units     = NULL;

        /*! \page aimInputsAVL
         * - <B> RollRate = 0.0 </B> <br>
         *  Non-dimensional roll rate.
         */

    } else if (index == 5) {
        *ainame           = EG_strdup("PitchRate");
        defval->type      = Double;
        defval->vals.real = 0.0;
        defval->units     = NULL;

        /*! \page aimInputsAVL
         * - <B> PitchRate = 0.0 </B> <br>
         *  Non-dimensional pitch rate.
         */

    } else if (index == 6) {
        *ainame           = EG_strdup("YawRate");
        defval->type      = Double;
        defval->vals.real = 0.0;
        defval->units     = NULL;

        /*! \page aimInputsAVL
         * - <B> YawRate = 0.0 </B> <br>
         *  Non-dimensional yaw rate.
         */

    } else if (index == 7) {
        *ainame           = EG_strdup("CDp");
        defval->type      = Double;
        defval->vals.real = 0.0;
        defval->units     = NULL;

        /*! \page aimInputsAVL
         * - <B> CDp = 0.0 </B> <br>
         *  A fixed value of profile drag to be added to all simulations.
         */

    } else if (index == 8) {
        *ainame              = EG_strdup("AVL_Surface");
        defval->type         = Tuple;
        defval->nullVal      = IsNull;
        //defval->units        = NULL;
        defval->dim          = 1;
        defval->lfixed       = Change;
        defval->vals.tuple   = NULL;

        /*! \page aimInputsAVL
         * - <B>AVL_Surface = NULL </B> <br>
         * See \ref vlmSurface for additional details.
         */

    } else if (index == 9) {
        *ainame              = EG_strdup("AVL_Control");
        defval->type         = Tuple;
        defval->nullVal      = IsNull;
        //defval->units        = NULL;
        defval->dim          = 1;
        defval->lfixed       = Change;
        defval->vals.tuple   = NULL;

        /*! \page aimInputsAVL
         * - <B>AVL_Control = NULL </B> <br>
         * See \ref vlmControl for additional details.
         */

    } else if (index == 10) {
        *ainame           = EG_strdup("CL");
        defval->type      = Double;
        defval->dim       = 1;
        defval->length    = 1;
        defval->nrow      = 1;
        defval->ncol      = 1;
        defval->units     = NULL;
        defval->vals.real = 0.0;
        defval->nullVal    = IsNull;
        defval->lfixed    = Change;

        /*! \page aimInputsAVL
         * - <B> CL = NULL </B> <br>
         *  Coefficient of Lift.  AVL will solve for Angle of Attack.  Either CL or Alpha must be defined but not both.
         */

    } else if (index == 11) {
        *ainame              = EG_strdup("Moment_Center");
        defval->type          = Double;
        defval->dim           = 1;
        defval->length        = 3;
        defval->nrow          = 3;
        defval->ncol          = 1;
        defval->units         = NULL;
        defval->vals.reals    = (double *) EG_alloc(defval->length*sizeof(double));
        if (defval->vals.reals == NULL) {
            return EGADS_MALLOC;
        } else {
            defval->vals.reals[0] = 0.0;
            defval->vals.reals[1] = 0.0;
            defval->vals.reals[2] = 0.0;
        }
        defval->nullVal       = IsNull;
        defval->lfixed        = Fixed;
        defval->sfixed        = Fixed;

        /*! \page aimInputsAVL
         * - <B>Moment_Center = NULL, [0.0, 0.0, 0.0]</B> <br>
         * Array values correspond to the Xref, Yref, and Zref variables.
         * Alternatively, the geometry (body) attributes "capsReferenceX", "capsReferenceY",
         * and "capsReferenceZ" may be used to specify the X-, Y-, and Z- reference centers, respectively
         * (note: values set through the AIM input will supersede the attribution values).
         */
    }

    /*else if (index == 11) {
     *ainame           = EG_strdup("Get_Stability");
		defval->type      = Boolean;
		defval->dim       = 1;
		defval->length    = 1;
		defval->nrow      = 1;
		defval->ncol      = 1;
		defval->units     = NULL;
		defval->vals.integer = (int) false;
		defval->nullVal    = NotNull;
		defval->lfixed    = Fixed;

		! \page aimInputsAVL
     * - <B> Get_Stability = False </B> <br>
     *  Execute the calculation of stability derivatives.

	}*/

    return CAPS_SUCCESS;
}


int aimPreAnalysis(int iIndex, /*@unused@*/ void *aimInfo,
                   const char *analysisPath, /*@null@*/ capsValue *aimInputs, capsErrs **errs)
{
    int status; // Function return status

    int i, j, k, surf, section, control; // Indexing

    int numBody;

    int found = (int) false; // Boolean tester

    int atype, alen;

    char currentPath[PATH_MAX] = "None";

    double 	     Sref, Cref, Bref, Xref, Yref, Zref;
    const int    *ints;
    const char   *string;
    const double *reals;
    ego          *bodies;

    int stringLength = 0; // String manipulation

    // File I/O
    FILE *fp = NULL;
    char inputFilename[] = "avlInput.txt";
    char avlFilename[] = "caps.avl";

    char totalForceFile[] = "capsTotalForce.txt";
    char stabilityFile[] = "capsStatbilityDeriv.txt";
    char bodyAxisFile[] = "capsBodyAxisDeriv.txt";
    char hingeMomentFile[] = "capsHingeMoment.txt";

    // AVL surface information
    int numAVLSurface = 0;
    vlmSurfaceStruct *avlSurface = NULL;

    // AVL control surface information
    int numAVLControl = 0;
    vlmControlStruct *avlControl = NULL;

    int numControlName = 0; // Unique control names
    char **controlName = NULL;

    mapAttrToIndexStruct attrMap; // Attribute to index map

    // Control related variables
    //double xyzHingeMag, xyzHingeVec[3];

#ifdef DEBUG
    printf(" avlAIM/aimPreAnalysis instance = %d\n", inst);
#endif

    // Initialize reference values
    Sref = 1.0;
    Cref = 1.0;
    Bref = 1.0;

    Xref = 0.0;
    Yref = 0.0;
    Zref = 0.0;

    // NULL out errs
    *errs = NULL;

    // Save a copy of analysisPath
    avlInstance[iIndex].analysisPath = (char *) analysisPath;

    if (aimInputs == NULL) {
#ifdef DEBUG
        printf(" avlAIM/aimPreAnalysis aimInputs == NULL!\n");
#endif

        return CAPS_NULLVALUE;
    }

    status = aim_getBodies(aimInfo, &numBody, &bodies);
    if (status != CAPS_SUCCESS) {
#ifdef DEBUG
        printf(" avlAIM/aimPreAnalysis getBodies = %d!\n", status);
#endif
        return status;
    }

    if (numBody == 0 || bodies == NULL) {
        printf(" avlAIM/aimPreAnalysis No Bodies!\n");
        return CAPS_SOURCEERR;
    }

    // Container for attribute to index map
    status = initiate_mapAttrToIndexStruct(&attrMap);
    if (status != CAPS_SUCCESS) return status;

    // Get capsGroup name and index mapping to make sure all bodies have a capsGroup value
    status = create_CAPSGroupAttrToIndexMap(numBody,
                                            bodies,
                                            0, // Only search down to the body level of the EGADS body
                                            &attrMap);
    if (status != CAPS_SUCCESS) goto cleanup;

    // Get AVL surface information
    if (aimInputs[aim_getIndex(aimInfo, "AVL_Surface", ANALYSISIN)-1].nullVal == NotNull) {

        status = get_vlmSurface(aimInputs[aim_getIndex(aimInfo, "AVL_Surface", ANALYSISIN)-1].length,
                                aimInputs[aim_getIndex(aimInfo, "AVL_Surface", ANALYSISIN)-1].vals.tuple,
                                &attrMap,
                                &numAVLSurface,
                                &avlSurface);

    } else {
        printf("No AVL_SURFACE tuple specified\n");
        status = CAPS_NOTFOUND;
    }
    if (status != CAPS_SUCCESS) goto cleanup;

    // Get AVL control surface information
    if (aimInputs[aim_getIndex(aimInfo, "AVL_Control", ANALYSISIN)-1].nullVal == NotNull) {

        status = get_vlmControl(aimInputs[aim_getIndex(aimInfo, "AVL_Control", ANALYSISIN)-1].length,
                                aimInputs[aim_getIndex(aimInfo, "AVL_Control", ANALYSISIN)-1].vals.tuple,
                                &numAVLControl,
                                &avlControl);

        if (status != CAPS_SUCCESS) goto cleanup;
    }

    // Accumulate section data
    status = vlm_getSection(numBody, bodies, (int) false, attrMap, numAVLSurface, &avlSurface);
    if (status != CAPS_SUCCESS) goto cleanup;

    // Loop through surfaces and transfer control surface data to sections
    for (surf = 0; surf < numAVLSurface; surf++) {

        status = get_ControlSurface(bodies,
                                    numAVLControl,
                                    avlControl,
                                    &avlSurface[surf]);
        if (status != CAPS_SUCCESS) goto cleanup;
    }

    // Order cross sections for each surface
    for (surf = 0; surf < numAVLSurface; surf++) vlm_orderSections(avlSurface[surf].numSection, avlSurface[surf].vlmSection);

    /*
   	// Determine the hinge line for any flaps
    for (surf = 0; surf < numAVLSurface; surf++) {

    	for (section = 0; section < avlSurface[surf].numSection; section++) {

    		for (control = 0; control < avlSurface[surf].vlmSection[section].numControl; control++) {

    			for (i = 0; i < avlSurface[surf].numSection; i++) {

    				if (i == section) continue;

    				for (j = 0; j < avlSurface[surf].vlmSection[i].numControl; j++) {

    					if( strcmp(avlSurface[surf].vlmSection[section].vlmControl[control].name,
    							   avlSurface[surf].vlmSection[i].vlmControl[j].name) == 0) {

    						xyzHingeVec[0] = avlSurface[surf].vlmSection[section].vlmControl[control].xyzHinge[0] -
											 avlSurface[surf].vlmSection[i].vlmControl[j].xyzHinge[0];

    						xyzHingeVec[1] = avlSurface[surf].vlmSection[section].vlmControl[control].xyzHinge[1] -
    										 avlSurface[surf].vlmSection[i].vlmControl[j].xyzHinge[1];

    						xyzHingeVec[2] = avlSurface[surf].vlmSection[section].vlmControl[control].xyzHinge[2] -
    										 avlSurface[surf].vlmSection[i].vlmControl[j].xyzHinge[2];

    						xyzHingeMag = sqrt(pow(xyzHingeVec[0], 2) +
    										   pow(xyzHingeVec[1], 2) +
											   pow(xyzHingeVec[2], 2));

    						avlSurface[surf].vlmSection[section].vlmControl[control].xyzHingeVec[0] = xyzHingeVec[0]/xyzHingeMag;
    						avlSurface[surf].vlmSection[section].vlmControl[control].xyzHingeVec[1] = xyzHingeVec[1]/xyzHingeMag;
    						avlSurface[surf].vlmSection[section].vlmControl[control].xyzHingeVec[2] = xyzHingeVec[2]/xyzHingeMag;

    					}
    				}
    			}
    		}
    	}
    }
     */

    // Get where we are and set the path to our input
    (void) getcwd(currentPath, PATH_MAX);

    if (chdir(analysisPath) != 0) {
        status = CAPS_DIRERR;
        goto cleanup;
    }

    // Open and write the input to control the AVL session
    fp = fopen(inputFilename, "w");
    if (fp == NULL) {
        printf("Unable to open file %s\n!", inputFilename);

        status = CAPS_IOERR;
        goto cleanup;
    }

    // Print the session file for the AVL run
    fprintf(fp, "OPER\n");

    if (aimInputs[aim_getIndex(aimInfo, "Alpha", ANALYSISIN)-1].nullVal ==  NotNull) {
        fprintf(fp, "A A ");//Alpha
        fprintf(fp, "%lf\n", aimInputs[1].vals.real);
    }

    if (aimInputs[aim_getIndex(aimInfo, "CL", ANALYSISIN)-1].nullVal ==  NotNull) {
        fprintf(fp, "A C ");//CL
        fprintf(fp, "%lf\n", aimInputs[9].vals.real);
    }

    fprintf(fp, "B B ");//Beta
    fprintf(fp, "%lf\n", aimInputs[2].vals.real);

    fprintf(fp, "R R ");//Roll Rate pb/2V
    fprintf(fp, "%lf\n", aimInputs[3].vals.real);

    fprintf(fp, "P P ");//Pitch Rate qc/2v
    fprintf(fp, "%lf\n", aimInputs[4].vals.real);

    fprintf(fp, "Y Y ");//Yaw Rate rb/2v
    fprintf(fp, "%lf\n", aimInputs[5].vals.real);

    // Check for control surface information
    j = 1;
    numControlName = 0;
    for (surf = 0; surf < numAVLSurface; surf++) {

        for (i = 0; i < avlSurface[surf].numSection; i++) {

            section = avlSurface[surf].vlmSection[i].sectionIndex;

            for (control = 0; control < avlSurface[surf].vlmSection[section].numControl; control++) {

                // Check to see if control surface hasn't already been written
                found = (int) false;

                if (numControlName == 0) {

                    numControlName += 1;

                    controlName = (char **) EG_alloc(numControlName * sizeof(char *));
                    if (controlName == NULL) {
                        numControlName -= 1;
                        status = EGADS_MALLOC;
                        goto cleanup;
                    }

                    found = (int) false;

                } else {

                    for (k = 0; k < numControlName; k++) {
                        if (strcmp(controlName[k],
                                avlSurface[surf].vlmSection[section].vlmControl[control].name) == 0) {

                            found = (int) true;
                            break;
                        }
                    }

                    if (found == (int) true) continue;

                    numControlName += 1;

                    controlName = (char **) EG_reall(controlName, numControlName * sizeof(char *));
                }

                stringLength = strlen(avlSurface[surf].vlmSection[section].vlmControl[control].name);

                controlName[numControlName -1] = (char *) EG_alloc((stringLength+1) * sizeof(char));

                if (controlName[numControlName -1] == NULL) {
                    status = EGADS_MALLOC;
                    goto cleanup;
                }

                strcpy(controlName[numControlName -1],
                        avlSurface[surf].vlmSection[section].vlmControl[control].name);

                // Store control map for later use
                status = increment_mapAttrToIndexStruct(&avlInstance[iIndex].controlMap,
                        (const char *) controlName[numControlName-1]);
                if (status != CAPS_SUCCESS) goto cleanup;

                fprintf(fp, "D%d D%d %f\n",j, j, avlSurface[surf].vlmSection[section].vlmControl[control].deflectionAngle);
                j += 1;
            }
        }
    }

    /*if(aimInputs[7].nullVal != IsNull) {
        // write control surface settings
        if (aimInputs[7].length == 1){
            fprintf(fp, "D1 D1 ");//Pitch Rate qc/2v
            fprintf(fp, "%lf\n", aimInputs[7].vals.real);
        } else {
            for (i=0; i < aimInputs[7].length; i++) {
                fprintf(fp,"D%d D%d %lf\n",i+1,i+1,aimInputs[7].vals.reals[i]);
            }
        }
    }*/

    fprintf(fp, "M\n"); // Modify parameters
    fprintf(fp, "MN\n");//Mach
    fprintf(fp, "%lf\n", aimInputs[0].vals.real);

    fprintf(fp, "\nX\n"); // Execute the calculation

    // Get Total forces
    fprintf(fp,"FT\n");
    fprintf(fp,"%s\n", totalForceFile);
    if (file_exist(totalForceFile) == (int) true) {
        fprintf(fp, "O\n");
    }

    // Get stability derivatives
    fprintf(fp,"ST\n");
    fprintf(fp,"%s\n", stabilityFile);
    if (file_exist(stabilityFile) == (int) true) {
        fprintf(fp, "O\n");
    }

    // Get stability (body axis) derivatives
    fprintf(fp,"SB\n");
    fprintf(fp,"%s\n", bodyAxisFile);
    if (file_exist(bodyAxisFile) == (int) true) {
        fprintf(fp, "O\n");
    }

    // Get hinge moments
    fprintf(fp,"HM\n");
    fprintf(fp,"%s\n", hingeMomentFile);
    if (file_exist(hingeMomentFile) == (int) true) {
        fprintf(fp, "O\n");
    }

    fprintf(fp, "\n");

    fprintf(fp, "Quit\n"); // Quit AVL
    fclose(fp);
    fp = NULL;

    /* open the file and write the avl data */
    fp = fopen(avlFilename, "w");
    if (fp == NULL) {
        printf("Unable to open file %s\n!", avlFilename);

        status =  CAPS_IOERR;
        goto cleanup;
    }

    // Loop over bodies and look for reference quantity attributes
    for (i=0; i < numBody; i++) {
        status = EG_attributeRet(bodies[i], "capsReferenceArea", &atype, &alen, &ints, &reals, &string);
        if (status == EGADS_SUCCESS) {

            if (atype == ATTRREAL && alen == 1) {
                Sref = (double) reals[0];
            } else {
                printf("capsReferenceArea should be followed by a single real value!\n");
                status = CAPS_BADVALUE;
                goto cleanup;
            }
        }

        status = EG_attributeRet(bodies[i], "capsReferenceChord", &atype, &alen, &ints, &reals, &string);
        if (status == EGADS_SUCCESS) {

            if (atype == ATTRREAL && alen == 1) {
                Cref = (double) reals[0];
            } else {
                printf("capsReferenceChord should be followed by a single real value!\n");
                status = CAPS_BADVALUE;
                goto cleanup;
            }
        }

        status = EG_attributeRet(bodies[i], "capsReferenceSpan", &atype, &alen, &ints, &reals, &string);
        if (status == EGADS_SUCCESS) {

            if (atype == ATTRREAL && alen == 1) {
                Bref = (double) reals[0];
            } else {
                printf("capsReferenceSpan should be followed by a single real value!\n");
                status = CAPS_BADVALUE;
                goto cleanup;
            }
        }

        status = EG_attributeRet(bodies[i], "capsReferenceX", &atype, &alen, &ints, &reals, &string);
        if (status == EGADS_SUCCESS) {

            if (atype == ATTRREAL && alen == 1) {
                Xref = (double) reals[0];
            } else {
                printf("capsReferenceX should be followed by a single real value!\n");
                status = CAPS_BADVALUE;
                goto cleanup;
            }
        }

        status = EG_attributeRet(bodies[i], "capsReferenceY", &atype, &alen, &ints, &reals, &string);
        if (status == EGADS_SUCCESS) {

            if (atype == ATTRREAL && alen == 1) {
                Yref = (double) reals[0];
            } else {
                printf("capsReferenceY should be followed by a single real value!\n");
                status = CAPS_BADVALUE;
                goto cleanup;
            }
        }

        status = EG_attributeRet(bodies[i], "capsReferenceZ", &atype, &alen, &ints, &reals, &string);
        if (status == EGADS_SUCCESS){

            if (atype == ATTRREAL && alen == 1) {
                Zref = (double) reals[0];
            } else {
                printf("capsReferenceZ should be followed by a single real value!\n");
                status = CAPS_BADVALUE;
                goto cleanup;
            }
        }
    }

    // Check for moment reference overwrites
    if (aimInputs[aim_getIndex(aimInfo, "Moment_Center", ANALYSISIN)-1].nullVal ==  NotNull) {

        Xref =aimInputs[aim_getIndex(aimInfo, "Moment_Center", ANALYSISIN)-1].vals.reals[0];
        Yref =aimInputs[aim_getIndex(aimInfo, "Moment_Center", ANALYSISIN)-1].vals.reals[1];
        Zref =aimInputs[aim_getIndex(aimInfo, "Moment_Center", ANALYSISIN)-1].vals.reals[2];
    }

    fprintf(fp, "CAPS generated Configuration\n");
    fprintf(fp, "0.0\n");                              /* Mach */
    fprintf(fp, "0 0 0.0\n");                          /* IYsym   IZsym   Zsym */
    fprintf(fp, "%lf %lf %lf\n", Sref, Cref, Bref); /* Sref    Cref    Bref */
    fprintf(fp, "%lf %lf %lf\n", Xref, Yref, Zref); /* Xref    Yref    Zref */
    fprintf(fp, "%lf\n", aimInputs[aim_getIndex(aimInfo, "CDp", ANALYSISIN)-1].vals.real); /* CDp */

    // Write out the Surfaces, one at a time
    for (surf = 0; surf < numAVLSurface; surf++) {

        printf("Writing surface - %s (ID = %d)\n", avlSurface[surf].name, surf);

        if (avlSurface[surf].numSection < 2) {
            printf("Surface %s only has %d Sections - it will be skipped!\n", avlSurface[surf].name,
                                                                              avlSurface[surf].numSection);
            continue;
        }

        fprintf(fp, "#\nSURFACE\n%s\n%d %lf\n\n", avlSurface[surf].name,
                                                  avlSurface[surf].Nchord,
                                                  avlSurface[surf].Cspace);

        if  (avlSurface[surf].compon != 0)
            fprintf(fp, "COMPONENT\n%d\n\n", avlSurface[surf].compon);

        if  (avlSurface[surf].iYdup  != 0)
            fprintf(fp, "YDUPLICATE\n0.0\n\n");

        if  (avlSurface[surf].nowake == (int) true) fprintf(fp, "NOWAKE\n");
        if  (avlSurface[surf].noalbe == (int) true) fprintf(fp, "NOALBE\n");
        if  (avlSurface[surf].noload == (int) true) fprintf(fp, "NOLOAD\n");

        if ( (avlSurface[surf].nowake == (int) true) ||
             (avlSurface[surf].noalbe == (int) true) ||
             (avlSurface[surf].noload == (int) true)) {

            fprintf(fp,"\n");
        }

        // Write the sections for each surface
        for (i = 0; i < avlSurface[surf].numSection; i++) {

            section = avlSurface[surf].vlmSection[i].sectionIndex;

            printf("\tSection %d of %d (ID = %d)\n", i+1, avlSurface[surf].numSection, section);

            // Write section data
            status = writeSection(fp, bodies[avlSurface[surf].vlmSection[section].bodyIndex],
                                             avlSurface[surf].vlmSection[section].nodeIndexLE,
                                             avlSurface[surf].vlmSection[section].xyzLE,
                                             avlSurface[surf].vlmSection[section].nodeIndexTE,
                                             avlSurface[surf].vlmSection[section].xyzTE,
                                             avlSurface[surf].Nspan,
                                             avlSurface[surf].Sspace);
            if (status != CAPS_SUCCESS) goto cleanup;

            // Write control information for each section
            for (control = 0; control < avlSurface[surf].vlmSection[section].numControl; control++) {

                printf("\t  Control surface %d of %d \n", control + 1, avlSurface[surf].vlmSection[section].numControl);

                fprintf(fp, "CONTROL\n");
                fprintf(fp, "%s ", avlSurface[surf].vlmSection[section].vlmControl[control].name);

                fprintf(fp, "%f ", avlSurface[surf].vlmSection[section].vlmControl[control].controlGain);

                if (avlSurface[surf].vlmSection[section].vlmControl[control].leOrTe == 0) { // Leading edge (-)

                    fprintf(fp, "%f ", -avlSurface[surf].vlmSection[section].vlmControl[control].percentChord);

                } else { // Trailing edge (+)

                    fprintf(fp, "%f ", avlSurface[surf].vlmSection[section].vlmControl[control].percentChord);
                }

                fprintf(fp, "%f %f %f ", avlSurface[surf].vlmSection[section].vlmControl[control].xyzHingeVec[0],
                        avlSurface[surf].vlmSection[section].vlmControl[control].xyzHingeVec[1],
                        avlSurface[surf].vlmSection[section].vlmControl[control].xyzHingeVec[2]);

                fprintf(fp, "%f\n", (double) avlSurface[surf].vlmSection[section].vlmControl[control].deflectionDup);
            }
            fprintf(fp, "\n");
        }
    }

    status = CAPS_SUCCESS;

    goto cleanup;

    cleanup:
        if (status != CAPS_SUCCESS) printf("Error: Premature exit in AVL preAnalysis() status = %d\n", status);

        // Restore the path we came in with and get out
        if (strcmp(currentPath, "None") != 0) chdir(currentPath);

        if (fp != NULL) fclose(fp);

        // Free array of control name strings
        if (numControlName != 0 && controlName != NULL){
            (void) string_freeArray(numControlName, &controlName);
        }

        // Attribute to index map
        (void ) destroy_mapAttrToIndexStruct(&attrMap);

        // Destroy avlSurfaces
        if (avlSurface != NULL) {
            for (i = 0; i < numAVLSurface; i++) {
                (void) destroy_vlmSurfaceStruct(&avlSurface[i]);
            }
        }

        if (avlSurface != NULL) EG_free(avlSurface);
        numAVLSurface = 0;

        // Destroy avlControl
        if (avlControl != NULL) {
            for (i = 0; i < numAVLControl; i++) {
                (void) destroy_vlmControlStruct(&avlControl[i]);
            }
        }

        if (avlControl != NULL) EG_free(avlControl);
        numAVLControl = 0;

        return status;
}


int aimOutputs(/*@unused@*/ int inst, /*@unused@*/ void *aimStruc, int index, char **aoname, capsValue *form)
{

    /*! \page aimOutputsAVL AIM Outputs
     * Optional outputs that echo the inputs.  These are parsed from the resulting output and can be used as a sanity check.
     */

#ifdef DEBUG
    printf(" avlAIM/aimOutputs instance = %d  index = %d!\n", inst, index);
#endif

    // Echo AVL flow conditions
    if (index == 1) {
        *aoname = EG_strdup("Alpha");

        /*! \page aimOutputsAVL
         * - <B> Alpha = </B> Angle of attack.
         */
    } else if (index == 2) {
        *aoname = EG_strdup("Beta");

        /*! \page aimOutputsAVL
         * - <B> Beta = </B> Sideslip angle.
         */
    } else if (index == 3) {
        *aoname = EG_strdup("Mach");

        /*! \page aimOutputsAVL
         * - <B> Mach = </B> Mach number.
         */

    } else if (index == 4) {
        *aoname = EG_strdup("pb/2V");

        /*! \page aimOutputsAVL
         * - <B> pb/2V = </B> Non-dimensional roll rate.
         */
    } else if (index == 5) {
        *aoname = EG_strdup("qc/2V");

        /*! \page aimOutputsAVL
         * - <B> qc/2V =</B> Non-dimensional pitch rate.
         */
    } else if (index == 6) {
        *aoname = EG_strdup("rb/2V");

        /*! \page aimOutputsAVL
         * - <B> rb/2V = </B> Non-dimensional yaw rate.
         */

    } else if (index == 7) {
        *aoname = EG_strdup("p'b/2V");

        /*! \page aimOutputsAVL
         * - <B> p'b/2V = </B> Non-dimensional roll acceleration.
         */

    } else if (index == 8) {
        *aoname = EG_strdup("r'b/2V");

        /*! \page aimOutputsAVL
         * - <B> r'b/2V = </B> Non-dimensional yaw acceleration.
         */

        // Body control derivatives
    } else if (index == 9) {
        *aoname = EG_strdup("CXtot");

        /*! \page aimOutputsAVL
         *
         * Calculated outputs.     See AVL documentation for a description of each quantity.
         * - <B> CXtot </B>
         */

    } else if (index == 10) {
        *aoname = EG_strdup("CYtot");

        /*! \page aimOutputsAVL
         *
         * - <B> CYtot </B>
         */
    } else if (index == 11) {
        *aoname = EG_strdup("CZtot");

        /*! \page aimOutputsAVL
         *
         * - <B> CZtot </B>
         */
    } else if (index == 12) {
        *aoname = EG_strdup("Cltot");

        /*! \page aimOutputsAVL
         *
         * - <B> Cltot </B>
         */
    } else if (index == 13) {
        *aoname = EG_strdup("Cmtot");

        /*! \page aimOutputsAVL
         *
         * - <B> Cmtot </B>
         */
    } else if (index == 14) {
        *aoname = EG_strdup("Cntot");

        /*! \page aimOutputsAVL
         *
         * - <B> Cntot </B>
         */
    } else if (index == 15) {
        *aoname = EG_strdup("Cl'tot");

        /*! \page aimOutputsAVL
         *
         * - <B> Cl'tot </B>
         */
    } else if (index == 16) {
        *aoname = EG_strdup("Cn'tot");

        /*! \page aimOutputsAVL
         *
         * - <B> Cn'tot </B>
         */
    } else if (index == 17) {
        *aoname = EG_strdup("CLtot");

        /*! \page aimOutputsAVL
         *
         * - <B> CLtot </B>
         */
    } else if (index == 18) {
        *aoname = EG_strdup("CDtot");

        /*! \page aimOutputsAVL
         *
         * - <B> CDtot </B>
         */
    } else if (index == 19) {
        *aoname = EG_strdup("CDvis");

        /*! \page aimOutputsAVL
         *
         * - <B> CDvis </B>
         */
    } else if (index == 20) {
        *aoname = EG_strdup("CLff");

        /*! \page aimOutputsAVL
         *
         * - <B> CLff </B>
         */
    } else if (index == 21) {
        *aoname = EG_strdup("CYff");

        /*! \page aimOutputsAVL
         *
         * - <B> CYff </B>
         */
    } else if (index == 22) {
        *aoname = EG_strdup("CDind");

        /*! \page aimOutputsAVL
         *
         * - <B> CDind </B>
         */
    } else if (index == 23) {
        *aoname = EG_strdup("CDff");

        /*! \page aimOutputsAVL
         *
         * - <B> CDff </B>
         */
    } else if (index == 24) {
        *aoname = EG_strdup("e");

        /*! \page aimOutputsAVL
         *
         * - <B> e </B>
         */
    } else if (index == 25) { // Alpha stability derivatives
        *aoname = EG_strdup("CLa");

    } else if (index == 26) {
        *aoname = EG_strdup("CYa");

    } else if (index == 27) {
        *aoname = EG_strdup("Cl'a");

    } else if (index == 28) {
        *aoname = EG_strdup("Cma");

    } else if (index == 29) {
        *aoname = EG_strdup("Cn'a");

        /*! \page aimOutputsAVL
         * Stability-axis derivatives - Alpha:
         * - <B>CLa</B> = z' force, CL, with respect to alpha.
         * - <B>CYa</B> = y force, CY, with respect to alpha.
         * - <B>Cl'a</B> = x' moment, Cl', with respect to alpha.
         * - <B>Cma</B> = y moment, Cm, with respect to alpha.
         * - <B>Cn'a</B> = z' moment, Cn', with respect to alpha.
         */

    } else if (index == 30) { // Beta stability derivatives
        *aoname = EG_strdup("CLb");

    } else if (index == 31) {
        *aoname = EG_strdup("CYb");

    } else if (index == 32) {
        *aoname = EG_strdup("Cl'b");

    } else if (index == 33) {
        *aoname = EG_strdup("Cmb");

    } else if (index == 34) {
        *aoname = EG_strdup("Cn'b");

        /*! \page aimOutputsAVL
         * Stability-axis derivatives - Beta:
         * - <B>CLb</B> = z' force, CL, with respect to beta.
         * - <B>CYb</B> = y force, CY, with respect to beta.
         * - <B>Cl'b</B> = x' moment, Cl', with respect to beta.
         * - <B>Cmb</B> = y moment, Cm, with respect to beta.
         * - <B>Cn'b</B> = z' moment, Cn', with respect to beta.
         */
    } else if (index == 35) { // Roll rate stability derivatives
        *aoname = EG_strdup("CLp'");

    } else if (index == 36) {
        *aoname = EG_strdup("CYp'");

    } else if (index == 37) {
        *aoname = EG_strdup("Cl'p'");

    } else if (index == 38) {
        *aoname = EG_strdup("Cmp'");

    } else if (index == 39) {
        *aoname = EG_strdup("Cn'p'");

        /*! \page aimOutputsAVL
         * Stability-axis derivatives - Roll rate, p':
         * - <B>CLp</B> = z' force, CL, with respect to roll rate, p'.
         * - <B>CYp</B> = y force, CY, with respect to roll rate, p'.
         * - <B>Cl'p</B> = x' moment, Cl', with respect to roll rate, p'.
         * - <B>Cmp</B> = y moment, Cm, with respect to roll rate, p'.
         * - <B>Cn'p</B> = z' moment, Cn', with respect to roll rate, p'.
         */
    } else if (index == 40) { // Pitch rate stability derivatives
        *aoname = EG_strdup("CLq'");

    } else if (index == 41) {
        *aoname = EG_strdup("CYq'");

    } else if (index == 42) {
        *aoname = EG_strdup("Cl'q'");

    } else if (index == 43) {
        *aoname = EG_strdup("Cmq'");

    } else if (index == 44) {
        *aoname = EG_strdup("Cn'q'");

        /*! \page aimOutputsAVL
         * Stability-axis derivatives - Pitch rate, q':
         * - <B>CLq</B> = z' force, CL, with respect to pitch rate, q'.
         * - <B>CYq</B> = y force, CY, with respect to pitch rate, q'.
         * - <B>Cl'q</B> = x' moment, Cl', with respect to pitch rate, q'.
         * - <B>Cmq</B> = y moment, Cm, with respect to pitch rate, q'.
         * - <B>Cn'q</B> = z' moment, Cn', with respect to pitch rate, q'.
         */
    } else if (index == 45) { // Yaw rate stability derivatives
        *aoname = EG_strdup("CLr'");

    } else if (index == 46) {
        *aoname = EG_strdup("CYr'");

    } else if (index == 47) {
        *aoname = EG_strdup("Cl'r'");

    } else if (index == 48) {
        *aoname = EG_strdup("Cmr'");

    } else if (index == 49) {
        *aoname = EG_strdup("Cn'r'");

        /*! \page aimOutputsAVL
         * Stability-axis derivatives - Yaw rate, r':
         * - <B>CLr</B> = z' force, CL, with respect to yaw rate, r'.
         * - <B>CYr</B> = y force, CY, with respect to yaw rate, r'.
         * - <B>Cl'r</B> = x' moment, Cl', with respect to yaw rate, r'.
         * - <B>Cmr</B> = y moment, Cm, with respect to yaw rate, r'.
         * - <B>Cn'r</B> = z' moment, Cn', with respect to yaw rate, r'.
         */
    } else if (index == 50) { // Axial velocity stability derivatives
        *aoname = EG_strdup("CXu");

    } else if (index == 51) {
        *aoname = EG_strdup("CYu");

    } else if (index == 52) {
        *aoname = EG_strdup("CZu");

    } else if (index == 53) {
        *aoname = EG_strdup("Clu");

    } else if (index == 54) {
        *aoname = EG_strdup("Cmu");

    } else if (index == 55) {
        *aoname = EG_strdup("Cnu");

        /*! \page aimOutputsAVL
         * Body-axis derivatives - Axial velocity, u:
         * - <B>CXu</B> = x force, CX, with respect to axial velocity, u.
         * - <B>CYu</B> = y force, CY, with respect to axial velocity, u.
         * - <B>CZu</B> = z force, CZ, with respect to axial velocity, u.
         * - <B>Clu</B> = x moment, Cl, with respect to axial velocity, u.
         * - <B>Cmu</B> = y moment, Cm, with respect to axial velocity, u.
         * - <B>Cnu</B> = z moment, Cn, with respect to axial velocity, u.
         */
    } else if (index == 56) { // Sideslip velocity stability derivatives
        *aoname = EG_strdup("CXv");

    } else if (index == 57) {
        *aoname = EG_strdup("CYv");

    } else if (index == 58) {
        *aoname = EG_strdup("CZv");

    } else if (index == 59) {
        *aoname = EG_strdup("Clv");

    } else if (index == 60) {
        *aoname = EG_strdup("Cmv");

    } else if (index == 61) {
        *aoname = EG_strdup("Cnv");

        /*! \page aimOutputsAVL
         * Body-axis derivatives - Sideslip velocity, v:
         * - <B>CXv</B> = x force, CX, with respect to sideslip velocity, v.
         * - <B>CYv</B> = y force, CY, with respect to sideslip velocity, v.
         * - <B>CZv</B> = z force, CZ, with respect to sideslip velocity, v.
         * - <B>Clv</B> = x moment, Cl, with respect to sideslip velocity, v.
         * - <B>Cmv</B> = y moment, Cm, with respect to sideslip velocity, v.
         * - <B>Cnv</B> = z moment, Cn, with respect to sideslip velocity, v.
         */
    } else if (index == 62) { // Normal velocity stability derivatives
        *aoname = EG_strdup("CXw");
    } else if (index == 63) {
        *aoname = EG_strdup("CYw");
    } else if (index == 64) {
        *aoname = EG_strdup("CZw");
    } else if (index == 65) {
        *aoname = EG_strdup("Clw");
    } else if (index == 66) {
        *aoname = EG_strdup("Cmw");
    } else if (index == 67) {
        *aoname = EG_strdup("Cnw");

        /*! \page aimOutputsAVL
         * Body-axis derivatives - Normal velocity, w:
         * - <B>CXw</B> = x force, CX, with respect to normal velocity, w.
         * - <B>CYw</B> = y force, CY, with respect to normal velocity, w.
         * - <B>CZw</B> = z force, CZ, with respect to normal velocity, w.
         * - <B>Clw</B> = x moment, Cl, with respect to normal velocity, w.
         * - <B>Cmw</B> = y moment, Cm, with respect to normal velocity, w.
         * - <B>Cnw</B> = z moment, Cn, with respect to normal velocity, w.
         */
    } else if (index == 68) { // Roll rate stability derivatives
        *aoname = EG_strdup("CXp");
    } else if (index == 69) {
        *aoname = EG_strdup("CYp");
    } else if (index == 70) {
        *aoname = EG_strdup("CZp");
    } else if (index == 71) {
        *aoname = EG_strdup("Clp");
    } else if (index == 72) {
        *aoname = EG_strdup("Cmp");
    } else if (index == 73) {
        *aoname = EG_strdup("Cnp");

        /*! \page aimOutputsAVL
         * Body-axis derivatives - Roll rate, p:
         * - <B>CXp</B> = x force, CX, with respect to roll rate, p.
         * - <B>CYp</B> = y force, CY, with respect to roll rate, p.
         * - <B>CZp</B> = z force, CZ, with respect to roll rate, p.
         * - <B>Clp</B> = x moment, Cl, with respect to roll rate, p.
         * - <B>Cmp</B> = y moment, Cm, with respect to roll rate, p.
         * - <B>Cnp</B> = z moment, Cn, with respect to roll rate, p.
         */
    } else if (index == 74) { // Pitch rate stability derivatives
        *aoname = EG_strdup("CXq");
    } else if (index == 75) {
        *aoname = EG_strdup("CYq");
    } else if (index == 76) {
        *aoname = EG_strdup("CZq");
    } else if (index == 77) {
        *aoname = EG_strdup("Clq");
    } else if (index == 78) {
        *aoname = EG_strdup("Cmq");
    } else if (index == 79) {
        *aoname = EG_strdup("Cnq");

        /*! \page aimOutputsAVL
         * Body-axis derivatives - Pitch rate, q:
         * - <B>CXq</B> = x force, CX, with respect to pitch rate, q.
         * - <B>CYq</B> = y force, CY, with respect to pitch rate, q.
         * - <B>CZq</B> = z force, CZ, with respect to pitch rate, q.
         * - <B>Clq</B> = x moment, Cl, with respect to pitch rate, q.
         * - <B>Cmq</B> = y moment, Cm, with respect to pitch rate, q.
         * - <B>Cnq</B> = z moment, Cn, with respect to pitch rate, q.
         */
    } else if (index == 80) { // Yaw rate stability derivatives
        *aoname = EG_strdup("CXr");
    } else if (index == 81) {
        *aoname = EG_strdup("CYr");
    } else if (index == 82) {
        *aoname = EG_strdup("CZr");
    } else if (index == 83) {
        *aoname = EG_strdup("Clr");
    } else if (index == 84) {
        *aoname = EG_strdup("Cmr");
    } else if (index == 85) {
        *aoname = EG_strdup("Cnr");

        /*! \page aimOutputsAVL
         * Body-axis derivatives - Yaw rate, r:
         * - <B>CXr</B> = x force, CX, with respect to yaw rate, r.
         * - <B>CYr</B> = y force, CY, with respect to yaw rate, r.
         * - <B>CZr</B> = z force, CZ, with respect to yaw rate, r.
         * - <B>Clr</B> = x moment, Cl, with respect to yaw rate, r.
         * - <B>Cmr</B> = y moment, Cm, with respect to yaw rate, r.
         * - <B>Cnr</B> = z moment, Cn, with respect to yaw rate, r.
         */
    } else if (index == 86) {
        *aoname = EG_strdup("ControlStability");

        /*! \page aimOutputsAVL
         * Controls:
         * - <B>ControlStability</B> = a (or an array) of tuple(s) with a
         * structure of ("Control Surface Name", "JSON Dictionary") for all control surfaces in the stability axis frame.
         * The JSON dictionary has the form = {"CLtot":value,"CYtot":value,"Cl'tot":value,"Cmtot":value,"Cn'tot":value}
         */
    } else if (index == 87) {
        *aoname = EG_strdup("ControlBody");

        /*! \page aimOutputsAVL
         * - <B>ControlBody</B> = a (or an array) of tuple(s) with a
         * structure of ("Control Surface Name", "JSON Dictionary") for all control surfaces in the body axis frame.
         * The JSON dictionary has the form = {"CXtot":value,"CYtot":value,"CZtot":value,"Cltot":value,"Cmtot":value,"Cntot":value}
         */
    } else if (index == 88) {
        *aoname = EG_strdup("HingeMoment");

        /*! \page aimOutputsAVL
         * - <B>HingeMoment</B> = a (or an array) of tuple(s) with a
         * structure of ("Control Surface Name", "HingeMoment")
         */
    } else {

        return CAPS_NOTFOUND;
    }

    if (index >= 86) {
        form->type = Tuple;
        form->units = NULL;
        form->vals.tuple = NULL;
        form->lfixed  = form->sfixed = Change;

    } else {

        form->type = Double;
        form->units = NULL;
        form->vals.reals = NULL;
        form->vals.real = 0;

    }

    return CAPS_SUCCESS;
}


int aimCalcOutput(int iIndex, void *aimInfo,
                  const char *analysisPath, int index, capsValue *val, capsErrs **errors)
{

    int status; // Function return status

    int i; // Indexing

    char *key = NULL;
    char *fileToOpen = "capsTotalForce.txt";

    double tempVal[6];

    char jsonOut[200];

    /*
	fileToOpen= "capsTotalForce.txt";
	fileToOpen = "capsStatbilityDeriv.txt";
	fileToOpen= "capsBodyAxisDeriv.txt";
	fileToOpen = "capsHingeMoment.txt";
     */

#ifdef DEBUG
    const char *name;

    status = aim_getName(aimInfo, index, ANALYSISOUT, &name);
    printf(" avlAIM/aimCalcOutput instance = %d  index = %d  %s %d!\n", iIndex, index, name, status);
#endif

    *errors        = NULL;
    val->vals.real = 0.0;

    switch (index) {
    case 1:
        fileToOpen= "capsTotalForce.txt";
        key = "Alpha =";
        break;
    case 2:
        fileToOpen= "capsTotalForce.txt";
        key = "Beta  =";
        break;
    case 3:
        fileToOpen= "capsTotalForce.txt";
        key = "Mach  =";
        break;
    case 4:
        fileToOpen= "capsTotalForce.txt";
        key = "pb/2V =";
        break;
    case 5:
        fileToOpen= "capsTotalForce.txt";
        key = "qc/2V =";
        break;
    case 6:
        fileToOpen= "capsTotalForce.txt";
        key = "rb/2V =";
        break;
    case 7:
        fileToOpen= "capsTotalForce.txt";
        key = "p'b/2V =";
        break;
    case 8:
        fileToOpen= "capsTotalForce.txt";
        key = "r'b/2V =";
        break;
    case 9:
        fileToOpen= "capsTotalForce.txt";
        key = "CXtot =";
        break;
    case 10:
        fileToOpen= "capsTotalForce.txt";
        key = "CYtot =";
        break;
    case 11:
        fileToOpen= "capsTotalForce.txt";
        key = "CZtot =";
        break;
    case 12:
        fileToOpen= "capsTotalForce.txt";
        key = "Cltot =";
        break;
    case 13:
        fileToOpen= "capsTotalForce.txt";
        key = "Cmtot =";
        break;
    case 14:
        fileToOpen= "capsTotalForce.txt";
        key = "Cntot =";
        break;
    case 15:
        fileToOpen= "capsTotalForce.txt";
        key = "Cl'tot =";
        break;
    case 16:
        fileToOpen= "capsTotalForce.txt";
        key = "Cn'tot =";
        break;
    case 17:
        fileToOpen= "capsTotalForce.txt";
        key = "CLtot =";
        break;
    case 18:
        fileToOpen= "capsTotalForce.txt";
        key = "CDtot =";
        break;
    case 19:
        fileToOpen= "capsTotalForce.txt";
        key = "CDvis =";
        break;
    case 20:
        fileToOpen= "capsTotalForce.txt";
        key = "CLff  =";
        break;
    case 21:
        fileToOpen= "capsTotalForce.txt";
        key = "CYff  =";
        break;
    case 22:
        fileToOpen= "capsTotalForce.txt";
        key = "CDind =";
        break;
    case 23:
        fileToOpen= "capsTotalForce.txt";
        key = "CDff  =";
        break;
    case 24:
        fileToOpen= "capsTotalForce.txt";
        key = "e =";
        break;

        // Alpha stability derivatives
    case 25:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "CLa =";
        break;

    case 26:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "CYa =";
        break;

    case 27:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "Cla =";
        break;

    case 28:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "Cma =";
        break;

    case 29:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "Cna =";
        break;

        // Beta stability derivatives
    case 30:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "CLb =";
        break;

    case 31:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "CYb =";
        break;

    case 32:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "Clb =";
        break;

    case 33:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "Cmb =";
        break;

    case 34:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "Cnb =";
        break;

        // Roll rate p' stability derivatives
    case 35:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "CLp =";
        break;

    case 36:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "CYp =";
        break;

    case 37:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "Clp =";
        break;

    case 38:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "Cmp =";
        break;

    case 39:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "Cnp =";
        break;

        // Pitch rate q' stability derivatives
    case 40:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "CLq =";
        break;

    case 41:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "CYq =";
        break;

    case 42:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "Clq =";
        break;

    case 43:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "Cmq =";
        break;

    case 44:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "Cnq =";
        break;

        // Yaw rate r' stability derivatives
    case 45:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "CLr =";
        break;

    case 46:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "CYr =";
        break;

    case 47:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "Clr =";
        break;

    case 48:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "Cmr =";
        break;

    case 49:
        fileToOpen = "capsStatbilityDeriv.txt";
        key = "Cnr =";
        break;


        // Axial vel (body axis)  stability derivatives
    case 50:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "CXu =";
        break;

    case 51:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "CYu =";
        break;

    case 52:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "CZu =";
        break;

    case 53:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "Clu =";
        break;

    case 54:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "Cmu =";
        break;

    case 55:fileToOpen= "capsBodyAxisDeriv.txt";
    key = "Cnu =";
    break;


    // Slidslip vel (body axis) stability derivatives
    case 56:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "CXv =";
        break;

    case 57:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "CYv =";
        break;

    case 58:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "CZv =";
        break;

    case 59:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "Clv =";
        break;

    case 60:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "Cmv =";
        break;

    case 61:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "Cnv =";
        break;

        // Normal vel (body axis)  stability derivatives
    case 62:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "CXw =";
        break;

    case 63:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "CYw =";
        break;

    case 64:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "CZw =";
        break;

    case 65:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "Clw =";
        break;

    case 66:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "Cmw =";
        break;

    case 67:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "Cnw =";
        break;

        // Roll rate (body axis)  stability derivatives
    case 68:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "CXp =";
        break;

    case 69:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "CYp =";
        break;

    case 70:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "CZp =";
        break;

    case 71:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "Clp =";
        break;

    case 72:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "Cmp =";
        break;

    case 73:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "Cnp =";
        break;

        // Pitch rate (body axis)  stability derivatives
    case 74:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "CXq =";
        break;

    case 75:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "CYq =";
        break;

    case 76:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "CZq =";
        break;

    case 77:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "Clq =";
        break;

    case 78:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "Cmq =";
        break;

    case 79:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "Cnq =";
        break;

        // Yaw rate (body axis)  stability derivatives
    case 80:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "CXr =";
        break;

    case 81:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "CYr =";
        break;

    case 82:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "CZr =";
        break;

    case 83:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "Clr =";
        break;

    case 84:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "Cmr =";
        break;

    case 85:
        fileToOpen= "capsBodyAxisDeriv.txt";
        key = "Cnr =";
        break;

    case 88:
        fileToOpen = "capsHingeMoment.txt";
        break;
    }

    if (index >= 86) { // Need to build something for control output

        // Clear initial memory
        if (val->vals.tuple != NULL) EG_free(val->vals.tuple);

        // Initiate tuple base on number of control surfaces
        val->length = val->nrow = avlInstance[iIndex].controlMap.numAttribute;

        val->vals.tuple = (capsTuple *) EG_alloc(val->length*sizeof(capsTuple));
        if (val->vals.tuple == NULL) {
            status = EGADS_MALLOC;
            goto cleanup;
        }

        for (i = 0; i < val->length; i++) val->vals.tuple[i].name = val->vals.tuple[i].value = NULL;

        // Loop through control surfaces
        for (i = 0; i < avlInstance[iIndex].controlMap.numAttribute; i++) {

            val->vals.tuple[i].name = EG_strdup(avlInstance[iIndex].controlMap.attributeName[i]);

            // Stability axis
            if (index == 86) {
                status = get_controlDeriv(aimInfo, (char *) analysisPath, avlInstance[iIndex].controlMap.attributeIndex[i],
                                          aim_getIndex(aimInfo, "CLtot", ANALYSISOUT), &tempVal[0]);
                if (status != CAPS_SUCCESS) goto cleanup;

                status = get_controlDeriv(aimInfo, (char *) analysisPath, avlInstance[iIndex].controlMap.attributeIndex[i],
                                          aim_getIndex(aimInfo, "CYtot", ANALYSISOUT), &tempVal[1]);
                if (status != CAPS_SUCCESS) goto cleanup;


                status = get_controlDeriv(aimInfo, (char *) analysisPath, avlInstance[iIndex].controlMap.attributeIndex[i],
                                          aim_getIndex(aimInfo, "Cl'tot", ANALYSISOUT), &tempVal[2]);
                if (status != CAPS_SUCCESS) goto cleanup;


                status = get_controlDeriv(aimInfo, (char *) analysisPath, avlInstance[iIndex].controlMap.attributeIndex[i],
                                          aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT), &tempVal[3]);
                if (status != CAPS_SUCCESS) goto cleanup;


                status = get_controlDeriv(aimInfo, (char *) analysisPath, avlInstance[iIndex].controlMap.attributeIndex[i],
                                          aim_getIndex(aimInfo, "Cn'tot", ANALYSISOUT), &tempVal[4]);
                if (status != CAPS_SUCCESS) goto cleanup;

                sprintf(jsonOut,"{\"%s\":%7.6f,\"%s\":%7.6f,\"%s\":%7.6f,\"%s\":%7.6f,\"%s\":%7.6f}", "CLtot",  tempVal[0],
                                                                                                      "CYtot",  tempVal[1],
                                                                                                      "Cl'tot", tempVal[2],
                                                                                                      "Cmtot",  tempVal[3],
                                                                                                      "Cn'tot", tempVal[4]);
                val->vals.tuple[i].value = EG_strdup(jsonOut);

            } else if (index == 87) {

                // Body axis
                status = get_controlDeriv(aimInfo, (char *) analysisPath, avlInstance[iIndex].controlMap.attributeIndex[i],
                                          aim_getIndex(aimInfo, "CXtot", ANALYSISOUT), &tempVal[0]);
                if (status != CAPS_SUCCESS) goto cleanup;

                status = get_controlDeriv(aimInfo, (char *) analysisPath, avlInstance[iIndex].controlMap.attributeIndex[i],
                                          aim_getIndex(aimInfo, "CYtot", ANALYSISOUT), &tempVal[1]);
                if (status != CAPS_SUCCESS) goto cleanup;

                status = get_controlDeriv(aimInfo, (char *) analysisPath, avlInstance[iIndex].controlMap.attributeIndex[i],
                                          aim_getIndex(aimInfo, "CZtot", ANALYSISOUT), &tempVal[2]);
                if (status != CAPS_SUCCESS) goto cleanup;


                status = get_controlDeriv(aimInfo, (char *) analysisPath, avlInstance[iIndex].controlMap.attributeIndex[i],
                                          aim_getIndex(aimInfo, "Cltot", ANALYSISOUT), &tempVal[3]);
                if (status != CAPS_SUCCESS) goto cleanup;


                status = get_controlDeriv(aimInfo, (char *) analysisPath, avlInstance[iIndex].controlMap.attributeIndex[i],
                                          aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT), &tempVal[4]);
                if (status != CAPS_SUCCESS) goto cleanup;


                status = get_controlDeriv(aimInfo, (char *) analysisPath, avlInstance[iIndex].controlMap.attributeIndex[i],
                                          aim_getIndex(aimInfo, "Cntot", ANALYSISOUT), &tempVal[5]);
                if (status != CAPS_SUCCESS) goto cleanup;;

                sprintf(jsonOut,"{\"%s\":%7.6f,\"%s\":%7.6f,\"%s\":%7.6f,\"%s\":%7.6f,\"%s\":%7.6f,\"%s\":%7.6f}", "CXtot", tempVal[0],
                                                                                                                   "CYtot", tempVal[1],
                                                                                                                   "CZtot", tempVal[2],
                                                                                                                   "Cltot", tempVal[3],
                                                                                                                   "Cmtot", tempVal[4],
                                                                                                                   "Cntot", tempVal[5]);
                val->vals.tuple[i].value = EG_strdup(jsonOut);

            } else if (index == 88) {

                status = read_Data(fileToOpen, (char *) analysisPath, val->vals.tuple[i].name, &tempVal[0]);
                if (status != CAPS_SUCCESS) goto cleanup;

                sprintf(jsonOut, "%5.4e", tempVal[0]);

                val->vals.tuple[i].value = EG_strdup(jsonOut);

            } else {
                status = CAPS_NOTFOUND;
            }

        }

    } else {

        if (key == NULL) {
            printf("No string key found!\n");
            status = CAPS_NOTFOUND;
            goto cleanup;
        }

        status = read_Data(fileToOpen, (char *) analysisPath, key, &val->vals.real);
        if (status != CAPS_SUCCESS) goto cleanup;
    }

    status = CAPS_SUCCESS;
    goto cleanup;

    cleanup:
        return status;
}

/*
 * since this AIM does not support field variables or CAPS bounds, the other
 * AIM functions are not required to be filled in except for aimFreeDiscr
 * that does some cleanup when CAPS terminates
 */

int aimFreeDiscr(/*@null@*/ capsDiscr *discr)
{

    int iIndex;

#ifdef DEBUG
    if (discr == NULL) {
        printf(" avlAIM/aimFreeDiscr w/ NULL!\n");
    } else {
        printf(" avlAIM/aimFreeDiscr instance = %d!\n", discr->instance);
    }
#endif

    if (discr == NULL) {

        for (iIndex = 0; iIndex < numInstance; iIndex++) {
            printf(" Cleaning up avlInstance - %d\n", iIndex);

            // Attribute to index map
            (void) destroy_mapAttrToIndexStruct(&avlInstance[iIndex].controlMap);

        }

        if (avlInstance != NULL) EG_free(avlInstance);
        avlInstance = NULL;
        numInstance = 0;
    }

    return CAPS_SUCCESS;
}


// Back function used to calculate sensitivities
int aimBackdoor(int iIndex, void *aimInfo, char *JSONin, char **JSONout) {

    /*! \page aimBackDoorAVL AIM Back Door
     * The back door function of this AIM may be used as an alternative to retrieve sensitivity information produced by the
     * AIM. The JSONin string should be of the following form '{"mode": "sensitivity", "inputVar": "Name of Input Variable",
     * "outputVar": "Name of Output Variable"}', while the JSONout string will look like '{"sensitivity": value}'. Important: the
     * JSONout string is freeable! Invalid combinations of input and output variables returns a CAPS_MISMATCH error code.
     *
     * Acceptable values for the "Name of Input Variable" are as follows (definitions are consistent, where appropriate,
     *  with \ref aimInputsAVL):
     * - "Alpha"
     * - "Beta"
     * - "RollRate"
     * - "PitchRate"
     * - "YawRate"
     * - "AxialVelocity"
     * - "SideslipVelocity"
     * - "NormalVelocity"
     * - "AVL_Control:Name_of_Control_Surface", where Name_of_Control_Surface should be replaced with name of the desired
     * control surface
     *
     * Acceptable values for the "Name of Output Variable" are as follows (definitions are consistent with \ref aimOutputsAVL):
     * - "CLtot"
     * - "CYtot"
     * - "Cl'tot"
     * - "Cmtot"
     * - "Cn'tot"
     * - "CXtot"
     * - "CZtot"
     * - "Cltot"
     * - "Cntot"
     */

    int status; // Function return status

    int index;
    double data;
    capsValue val;
    capsErrs *errors;

    char *keyValue = NULL;
    char *keyWord = NULL;

    char *tempString = NULL; // Freeable

    int inputIndex = 0, outputIndex = 0, controlIndex;

    char *outputJSON = NULL;

    index = -99;

    *JSONout = NULL;

    if (avlInstance[iIndex].analysisPath == NULL) {
        printf("Analysis path hasn't been set - Need to run preAnalysis first!");
        return CAPS_DIRERR;
    }

    keyWord = "mode";
    status = search_jsonDictionary(JSONin, keyWord, &keyValue);
    if (status != CAPS_SUCCESS) goto cleanup;


    // Type of acceptable modes for this AIM backdoor
    if (strcasecmp(keyValue, "\"Sensitivity\"") != 0) {

        printf("Error: A valid mode wasn't found for AIMBackDoor!\n");
        status = CAPS_NOTFOUND;
        goto cleanup;
    }

    // We are using this to get sensitivities
    if (strcasecmp(keyValue, "\"Sensitivity\"") == 0) {

        if (keyValue != NULL) EG_free(keyValue);
        keyValue = NULL;

        // Input variable
        keyWord = "inputVar";
        status = search_jsonDictionary(JSONin, keyWord, &keyValue);
        if (status != CAPS_SUCCESS) goto cleanup;

        tempString = string_removeQuotation(keyValue);

        inputIndex = aim_getIndex(aimInfo, tempString, ANALYSISIN);
        if (inputIndex <= 0) {
            status = inputIndex;

            if (strcasecmp(tempString, "AxialVelocity") == 0) {
                inputIndex = NUMINPUT+1;

            } else if (strcasecmp(tempString, "SideslipVelocity") == 0) {
                inputIndex = NUMINPUT+2;

            } else if (strcasecmp(tempString, "NormalVelocity") == 0) {
                inputIndex = NUMINPUT+3;

                // Try to see if it is a control parameter
            } else if (parse_controlName(iIndex, tempString, &controlIndex) == CAPS_SUCCESS) {
                inputIndex = CAPSMAGIC;

            } else {

                printf("Error: Unable to get index for inputVar = %s\n", tempString);
                goto cleanup;
            }
        }

        if (tempString != NULL) EG_free(tempString);
        tempString = NULL;

        // Outout variable
        keyWord = "outputVar";
        status = search_jsonDictionary(JSONin, keyWord, &keyValue);
        if (status != CAPS_SUCCESS) goto cleanup;

        tempString = string_removeQuotation(keyValue);

        outputIndex = aim_getIndex(aimInfo, tempString, ANALYSISOUT);
        if (outputIndex <= 0) {
            status = outputIndex;
            printf("Error: Unable to get index for outputVar = %s\n", tempString);
            goto cleanup;
        }

        if (tempString != NULL) EG_free(tempString);
        tempString = NULL;

        //printf("InputIndex = %d\n", inputIndex);
        //printf("OutputIndex = %d\n", outputIndex);

        // Stability axis - Alpha
        if (outputIndex == aim_getIndex(aimInfo, "CLtot", ANALYSISOUT) &&
            inputIndex == aim_getIndex(aimInfo, "Alpha", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"CLa", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "Alpha", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"CYa", ANALYSISOUT);

        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cl'tot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "Alpha", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Cl'a", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "Alpha", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Cma", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cn'tot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "Alpha", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Cn'a", ANALYSISOUT);
        }

        // Stability axis - Beta
        else if (outputIndex == aim_getIndex(aimInfo, "CLtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "Beta", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"CLb", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "Beta", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"CYb", ANALYSISOUT);

        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cl'tot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "Beta", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Cl'b", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "Beta", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Cmb", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cn'tot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "Beta", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Cn'b", ANALYSISOUT);
        }

        // Stability axis - RollRate
        else if (outputIndex == aim_getIndex(aimInfo, "CLtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "RollRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"CLp'", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "RollRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"CYp'", ANALYSISOUT);

        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cl'tot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "RollRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Cl'p'", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "RollRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Cmp'", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cn'tot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "RollRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Cn'p'", ANALYSISOUT);
        }

        // Stability axis - PitchRate
        else if (outputIndex == aim_getIndex(aimInfo, "CLtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "PitchRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"CLq'", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "PitchRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"CYq'", ANALYSISOUT);

        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cl'tot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "PitchRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Cl'q'", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "PitchRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Cmq'", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cn'tot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "PitchRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Cn'q'", ANALYSISOUT);
        }


        // Stability axis - YawRate
        else if (outputIndex == aim_getIndex(aimInfo, "CLtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "YawRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"CLr'", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "YawRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"CYr'", ANALYSISOUT);

        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cl'tot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "YawRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Cl'r'", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "YawRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Cmr'", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cn'tot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "YawRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Cn'r'", ANALYSISOUT);
        }


        //////////////////////////////////////////////////////////

        // Body axis - AxialVelocity
        else if (outputIndex == aim_getIndex(aimInfo, "CXtot", ANALYSISOUT) &&
                 inputIndex  == NUMINPUT+1) {

            index = aim_getIndex(aimInfo,"CXu", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT) &&
                 inputIndex  == NUMINPUT+1) {

            index = aim_getIndex(aimInfo,"CYu", ANALYSISOUT);

        }

        else if (outputIndex == aim_getIndex(aimInfo, "CZtot", ANALYSISOUT) &&
                 inputIndex  == NUMINPUT+1) {

            index = aim_getIndex(aimInfo,"CZu", ANALYSISOUT);

        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cltot", ANALYSISOUT) &&
                 inputIndex  == NUMINPUT+1) {

            index = aim_getIndex(aimInfo,"Clu", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT) &&
                 inputIndex  == NUMINPUT+1) {

            index = aim_getIndex(aimInfo,"Cmu", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cntot", ANALYSISOUT) &&
                 inputIndex  == NUMINPUT+1) {

            index = aim_getIndex(aimInfo,"Cnu", ANALYSISOUT);
        }

        // Body axis - Sideslip
        else if (outputIndex == aim_getIndex(aimInfo, "CXtot", ANALYSISOUT) &&
                 inputIndex  == NUMINPUT+2) {

            index = aim_getIndex(aimInfo,"CXv", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT) &&
                 inputIndex  == NUMINPUT+2) {

            index = aim_getIndex(aimInfo,"CYv", ANALYSISOUT);

        }

        else if (outputIndex == aim_getIndex(aimInfo, "CZtot", ANALYSISOUT) &&
                 inputIndex  == NUMINPUT+2) {

            index = aim_getIndex(aimInfo,"CZv", ANALYSISOUT);

        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cltot", ANALYSISOUT) &&
                 inputIndex  == NUMINPUT+2) {

            index = aim_getIndex(aimInfo,"Clv", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT) &&
                 inputIndex  == NUMINPUT+2) {

            index = aim_getIndex(aimInfo,"Cmv", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cntot", ANALYSISOUT) &&
                 inputIndex  == NUMINPUT+2) {

            index = aim_getIndex(aimInfo,"Cnv", ANALYSISOUT);
        }

        // Body axis - Normal
        else if (outputIndex == aim_getIndex(aimInfo, "CXtot", ANALYSISOUT) &&
                 inputIndex  == NUMINPUT+3) {

            index = aim_getIndex(aimInfo,"CXw", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT) &&
                 inputIndex  == NUMINPUT+3) {

            index = aim_getIndex(aimInfo,"CYw", ANALYSISOUT);

        }

        else if (outputIndex == aim_getIndex(aimInfo, "CZtot", ANALYSISOUT) &&
                 inputIndex  == NUMINPUT+3) {

            index = aim_getIndex(aimInfo,"CZw", ANALYSISOUT);

        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cltot", ANALYSISOUT) &&
                 inputIndex  == NUMINPUT+3) {

            index = aim_getIndex(aimInfo,"Clw", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT) &&
                 inputIndex  == NUMINPUT+3) {

            index = aim_getIndex(aimInfo,"Cmw", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cntot", ANALYSISOUT) &&
                 inputIndex  == NUMINPUT+3) {

            index = aim_getIndex(aimInfo,"Cnw", ANALYSISOUT);
        }

        // Body axis - RollRate
        else if (outputIndex == aim_getIndex(aimInfo, "CXtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "RollRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"CXp", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "RollRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"CYp", ANALYSISOUT);

        }

        else if (outputIndex == aim_getIndex(aimInfo, "CZtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "RollRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"CZp", ANALYSISOUT);

        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cltot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "RollRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Clp", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "RollRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Cmp", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cntot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "RollRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Cnp", ANALYSISOUT);
        }

        // Body axis - PitchRate
        else if (outputIndex == aim_getIndex(aimInfo, "CXtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "PitchRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"CXq", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "PitchRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"CYq", ANALYSISOUT);

        }

        else if (outputIndex == aim_getIndex(aimInfo, "CZtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "PitchRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"CZq", ANALYSISOUT);

        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cltot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "PitchRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Clq", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "PitchRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Cmq", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cntot", ANALYSISOUT) &&
                 inputIndex == aim_getIndex(aimInfo, "PitchRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Cnq", ANALYSISOUT);
        }

        // Body axis - YawRate
        else if (outputIndex == aim_getIndex(aimInfo, "CXtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "YawRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"CXr", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "CYtot", ANALYSISOUT) &&
                 inputIndex == aim_getIndex(aimInfo, "YawRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"CYr", ANALYSISOUT);

        }

        else if (outputIndex == aim_getIndex(aimInfo, "CZtot", ANALYSISOUT) &&
                 inputIndex == aim_getIndex(aimInfo, "YawRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"CZr", ANALYSISOUT);

        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cltot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "YawRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Clr", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cmtot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "YawRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Cmr", ANALYSISOUT);
        }

        else if (outputIndex == aim_getIndex(aimInfo, "Cntot", ANALYSISOUT) &&
                 inputIndex  == aim_getIndex(aimInfo, "YawRate", ANALYSISIN)) {

            index = aim_getIndex(aimInfo,"Cnr", ANALYSISOUT);
        }
        else if (inputIndex == CAPSMAGIC) { // Control variable found
            index = CAPSMAGIC;
        }

        else {
            printf("Invalid combination of input and output variables.\n");
            status = CAPS_MISMATCH;
            goto cleanup;
        }

        // If we have a control variable
        if (index == CAPSMAGIC) {

            status = get_controlDeriv(aimInfo, avlInstance[iIndex].analysisPath, controlIndex, outputIndex, &data);
            if (status != CAPS_SUCCESS) goto cleanup;

            val.vals.real = data;


        } else { // Make the standard call

            status = aimCalcOutput(iIndex, aimInfo, avlInstance[iIndex].analysisPath, index, &val, &errors);
            if (status != CAPS_SUCCESS) goto cleanup;
        }

        outputJSON = (char *) EG_alloc(50*sizeof(char));

        sprintf(outputJSON, "{\"sensitivity\": %7.6f}", val.vals.real);
    }

    *JSONout = outputJSON;

    status = CAPS_SUCCESS;
    goto cleanup;

    cleanup:
        if (status != CAPS_SUCCESS) printf("Error: Premature exit in aimBackdoor, status = %d\n", status);

        if (keyValue != NULL) EG_free(keyValue);

        if (tempString != NULL) EG_free(tempString);

        return status;
}
