/*
 *      EGADS: Electronic Geometry Aircraft Design System
 *
 *      Tesselate EGADS bodies and write:
 *         CART3D unstructured surface triangulation tri file
 *                with FaceID in Component field
 *         TESS   information about vertex/geometry ownership
 *         PLOT3D structured surface grid file from quadding algorithm
 *         PLOT3D structured surface grid file from direct uv parameters space 
 *                evaluation (untrimmed)
 *
 *      Modified from egads2cart by William M. Chan (May, 2016)
 *
 *      === Log ===
 *      April, 2017 (Bob and William):
 *        - Added cell size (maxe) control to uv plot3d file output (uv file can now
 *          be controlled with maxe, maxa, maxc).
 *        - Added iblank to uv plot3d file output to mark if vertex is inside (ib=1),
 *          or outside (ib=0) of BRep face (may need to add a tolerance for vertices
 *          with uv's close to the boundaries).
 *        - Hard coded different qparam values to try to influence quadding.
 *
 *      Copyright 2011-2018, Massachusetts Institute of Technology
 *      Licensed under The GNU Lesser General Public License, version 2.1
 *      See http://www.opensource.org/licenses/lgpl-2.1.php
 *
 */

#include "egads.h"
#include <math.h>
#include <string.h>

#ifdef WIN32
#define snprintf _snprintf
#endif


#define PI     3.1415926535897931159979635
#define MAXUVS 1024                        /* max Us or Vs for untrimmed data */
#define CROSS(a,b,c)  a[0] = (b[1]*c[2]) - (b[2]*c[1]);\
                      a[1] = (b[2]*c[0]) - (b[0]*c[2]);\
                      a[2] = (b[0]*c[1]) - (b[1]*c[0])


typedef struct {
  int    nu;
  int    nv;
  double us[MAXUVS];
  double vs[MAXUVS];
} UVmap;


/* Global vars */

static char   version[5] = "1.11";
static int    wrtqud = 0;
static int    wrtuv  = 0;
static double mxang  = 0.0;
static double mxedg  = 0.0;
static double mxchd  = 0.0;


/*
 * 	calculates and returns a complete Body tessellation
 *                 note that this is consistent with the EGADS global numbers
 *
 * 	where:	body	- ego of a body tessellation
 *              nface   - number of faces in the body
 *              face    - the EGADS Face objects
 * 		nvert	- Number of vertices (returned)
 * 		verts	- coordinates (returned) 3*nverts in len -- freeable
 * 		ntriang	- number of triangles (returned)
 * 		triang	- triangle indices (returned) 3*ntriang in len
 *			  freeable
 *              comp    - Cart3D component ID per triangle -- freeable
 */

static int
bodyTessellation(ego tess, int nface, int *nvert, double **verts,
                 int *ntriang, int **triang, int **comp)

{
  int          status, i, j, k, base, npts, ntri, *tri, *table, *compID;
  int          plen, tlen;
  const int    *tris, *tric, *ptype, *pindex;
  double       *xyzs;
  const double *points, *uv;

  *nvert  = *ntriang = 0;
  *verts  = NULL;
  *triang = NULL;
  *comp   = NULL;
  npts = ntri = 0;

  for (i = 1; i <= nface; i++) {
    status = EG_getTessFace(tess, i, &plen, &points, &uv, &ptype, &pindex,
                            &tlen, &tris, &tric);
    if (status != EGADS_SUCCESS) {
      printf(" Face %d: EG_getTessFace status = %d (bodyTessellation)!\n",
             i, status);
    } else {
      npts += plen;
      ntri += tlen;
    }
  }

  /* get the memory associated with the points */

  table = (int *) EG_alloc(2*npts*sizeof(int));
  if (table == NULL) {
    printf(" Error: Can not allocate node table (bodyTessellation)!\n");
    return EGADS_MALLOC;
  }

  xyzs = (double *) EG_alloc(3*npts*sizeof(double));
  if (xyzs == NULL) {
    printf(" Error: Can not allocate XYZs (bodyTessellation)!\n");
    EG_free(table);
    return EGADS_MALLOC;
  }

  /* zipper up the edges -- a Face at a time */

  npts = 0;
  for (j = 1; j <= nface; j++) {
    status = EG_getTessFace(tess, j, &plen, &points, &uv, &ptype, &pindex,
                            &tlen, &tris, &tric);
    if (status != EGADS_SUCCESS) continue;

    for (i = 0; i < plen; i++) {
      table[2*npts  ] = ptype[i];
      table[2*npts+1] = pindex[i];
      xyzs[3*npts  ]  = points[3*i  ];
      xyzs[3*npts+1]  = points[3*i+1];
      xyzs[3*npts+2]  = points[3*i+2];

      /* for non-interior pts -- try to match with others */

      if (ptype[i] != -1)
        for (k = 0; k < npts; k++)
          if ((table[2*k]==table[2*npts]) && (table[2*k+1]==table[2*npts+1])) {
            table[2*npts  ] = k;
            table[2*npts+1] = 0;
            break;
          }
      
      npts++;
    }

  }

  /* fill up the whole triangle list -- a Face at a time */

  tri = (int *) EG_alloc(3*ntri*sizeof(int));
  if (tri == NULL) {
    printf(" Error: Can not allocate triangles (bodyTessellation)!\n");
    EG_free(xyzs);
    EG_free(table);
    return EGADS_MALLOC;
  }
  compID = (int *) EG_alloc(ntri*sizeof(int));
  if (compID == NULL) {
    printf(" Error: Can not allocate components (bodyTessellation)!\n");
    EG_free(tri);
    EG_free(xyzs);
    EG_free(table);
    return EGADS_MALLOC;
  }

  ntri = base = 0;
  for (j = 1; j <= nface; j++) {

    /* get the face tessellation and store it away */
    status = EG_getTessFace(tess, j, &plen, &points, &uv, &ptype, &pindex,
                            &tlen, &tris, &tric);
    if (status != EGADS_SUCCESS) continue;

    for (i = 0; i < tlen; i++, ntri++) {

      k = tris[3*i  ] + base;

      if (table[2*k-1] == 0) {
        tri[3*ntri  ] = table[2*k-2] + 1;
      } else {
        tri[3*ntri  ] = k;
      }

      k = tris[3*i+1] + base;

      if (table[2*k-1] == 0) {
        tri[3*ntri+1] = table[2*k-2] + 1;
      } else {
        tri[3*ntri+1] = k;
      }

      k = tris[3*i+2] + base;

      if (table[2*k-1] == 0) {
        tri[3*ntri+2] = table[2*k-2] + 1;
      } else {
        tri[3*ntri+2] = k;
      }

      compID[ntri] = j;
    }
    base += plen;
  }


  /*  remove the unused points -- crunch the point list
   *  NOTE: the returned pointer verts has the full length (not realloc'ed)
   */
  for (i = 0; i <   npts; i++) table[i] = 0;
  for (i = 0; i < 3*ntri; i++) table[tri[i]-1]++;
  for (plen = i = 0; i < npts; i++) {
    if (table[i] == 0) continue;
    xyzs[3*plen  ] = xyzs[3*i  ];
    xyzs[3*plen+1] = xyzs[3*i+1];
    xyzs[3*plen+2] = xyzs[3*i+2];
    plen++;
    table[i] = plen;
  }

  /* reset the triangle indices */
  for (i = 0; i < 3*ntri; i++) {
    k      = tri[i]-1;
    tri[i] = table[k];
  }

  EG_free(table);

  *nvert   = plen;
  *verts   = xyzs;
  *ntriang = ntri;
  *triang  = tri;
  *comp    = compID;

  return EGADS_SUCCESS;

}


static int strroot(char *s)
{

  /* Return string index location in string s by locating last . */

  int i, iroot, len;

  len   = strlen(s);
  iroot = len-1;

  i = len-1;
  while ( s[i] != '.' && i > 0 ) {
    i--;
  }
  if ( i >= 0 ) iroot = i;

  return iroot;
}


static void writeAttr(FILE *fp, ego obj, /*@null@*/ char *filter)
{
  int          i, j, n, stat, nattr, atype, alen, flen;
  const int    *ints;
  const double *reals;
  const char   *name, *str;
  
  stat = EG_attributeNum(obj, &nattr);
  if (stat != EGADS_SUCCESS) return;
  
  for (n = i = 0; i < nattr; i++) {
    stat = EG_attributeGet(obj, i+1, &name, &atype, &alen, &ints, &reals, &str);
    if (stat != EGADS_SUCCESS) continue;
    if ((atype != ATTRINT) && (atype != ATTRREAL) && (atype != ATTRSTRING))
      continue;
    if (filter != NULL) {
      flen = strlen(filter);
      if (flen > strlen(name)) continue;
      for (j = 0; j < flen; j++)
        if (filter[j] != name[j]) break;
      if (j != flen) continue;
    }
    n++;
  }
  fprintf(fp, " %6d\n", n);
  if (n == 0) return;

  for (i = 0; i < nattr; i++) {
    stat = EG_attributeGet(obj, i+1, &name, &atype, &alen, &ints, &reals, &str);
    if (stat != EGADS_SUCCESS) continue;
    if ((atype != ATTRINT) && (atype != ATTRREAL) && (atype != ATTRSTRING))
      continue;
    if (filter != NULL) {
      flen = strlen(filter);
      if (flen > strlen(name)) continue;
      for (j = 0; j < flen; j++)
        if (filter[j] != name[j]) break;
      if (j != flen) continue;
    }
    if (atype == ATTRSTRING) alen = strlen(str);
    fprintf(fp, " %6d %6d %s\n", atype, alen, name);
    if (atype == ATTRSTRING) {
      fprintf(fp, " %s\n", str);
    } else if (atype == ATTRREAL) {
      for (j = 0; j < alen; j++) {
        fprintf(fp, " %20.13le", reals[j]);
        if ((j+1)%4 == 0) fprintf(fp,"\n");
      }
      if (j%4 != 0) fprintf(fp, "\n");
    } else {
      for (j = 0; j < alen; j++) {
        fprintf(fp, " %10d", ints[j]);
        if ((j+1)%8 == 0) fprintf(fp,"\n");
      }
      if (j%8 != 0) fprintf(fp, "\n");
    }
  }

}


static void updateMap(ego face, double mxedg, double sag, double angle, UVmap *map)
{
  int    i, j, max, stat, cnt;
  double d, dist, dot, last, uv[2], xyz[3], *u1, *v1, result[18], xyzs[3*MAXUVS];
  
  dot = cos(PI*angle/180.0);
  u1  = &result[3];
  v1  = &result[6];
  
  /* sag -- look in U direction */
  do {
    max  = -1;
    dist = 0.0;
    for (j = 0; j < map->nv; j++) {
      uv[1] = map->vs[j];
      for (i = 0; i < map->nu; i++) {
        uv[0] = map->us[i];
        stat  = EG_evaluate(face, uv, result);
        if (stat != EGADS_SUCCESS) {
          printf(" updateMap: Fill U EG_evaluate = %d!\n", stat);
          return;
        }
        xyzs[3*i  ] = result[0];
        xyzs[3*i+1] = result[1];
        xyzs[3*i+2] = result[2];
      }
      for (i = 1; i < map->nu; i++) {
        xyz[0] = 0.5*(xyzs[3*i-3]  + xyzs[3*i  ]);
        xyz[1] = 0.5*(xyzs[3*i-2]  + xyzs[3*i+1]);
        xyz[2] = 0.5*(xyzs[3*i-1]  + xyzs[3*i+2]);
        uv[0]  = 0.5*(map->us[i-1] + map->us[i]);
        stat   = EG_evaluate(face, uv, result);
        if (stat != EGADS_SUCCESS) {
          printf(" updateMap: Half U EG_evaluate = %d!\n", stat);
          return;
        }
        d = sqrt((xyz[0]-result[0])*(xyz[0]-result[0]) +
                 (xyz[1]-result[1])*(xyz[1]-result[1]) +
                 (xyz[2]-result[2])*(xyz[2]-result[2]));
        if (d <= sag) continue;
        if (d > dist) {
          max  = i;
          dist = d;
        }
      }
    }
    if (map->nu == MAXUVS) break;
    if (max != -1) {
      for (i = map->nu; i >= max; i--) map->us[i] = map->us[i-1];
      map->us[max] = 0.5*(map->us[max+1] + map->us[max-1]);
      map->nu++;
    }
  } while (max != -1);
  
  /* sag -- look in V direction */
  do {
    max  = -1;
    dist = 0.0;
    for (i = 0; i < map->nu; i++) {
      uv[0] = map->us[i];
      for (j = 0; j < map->nv; j++) {
        uv[1] = map->vs[j];
        stat  = EG_evaluate(face, uv, result);
        if (stat != EGADS_SUCCESS) {
          printf(" updateMap: Fill V EG_evaluate = %d!\n", stat);
          return;
        }
        xyzs[3*j  ] = result[0];
        xyzs[3*j+1] = result[1];
        xyzs[3*j+2] = result[2];
      }
      for (j = 1; j < map->nv; j++) {
        xyz[0] = 0.5*(xyzs[3*j-3]  + xyzs[3*j  ]);
        xyz[1] = 0.5*(xyzs[3*j-2]  + xyzs[3*j+1]);
        xyz[2] = 0.5*(xyzs[3*j-1]  + xyzs[3*j+2]);
        uv[1]  = 0.5*(map->vs[j-1] + map->vs[j]);
        stat   = EG_evaluate(face, uv, result);
        if (stat != EGADS_SUCCESS) {
          printf(" updateMap: Half V EG_evaluate = %d!\n", stat);
          return;
        }
        d = sqrt((xyz[0]-result[0])*(xyz[0]-result[0]) +
                 (xyz[1]-result[1])*(xyz[1]-result[1]) +
                 (xyz[2]-result[2])*(xyz[2]-result[2]));
        if (d <= sag) continue;
        if (d > dist) {
          max  = j;
          dist = d;
        }
      }
    }
    if (map->nv == MAXUVS) break;
    if (max != -1) {
      for (j = map->nv; j >= max; j--) map->vs[j] = map->vs[j-1];
      map->vs[max] = 0.5*(map->vs[max+1] + map->vs[max-1]);
      map->nv++;
    }
  } while (max != -1);
  
  /* angle -- look in U direction */
  last = -1.0;
  cnt  = 0;
  do {
    max  = -1;
    dist = 1.0;
    for (j = 0; j < map->nv; j++) {
      uv[1] = map->vs[j];
      for (i = 0; i < map->nu; i++) {
        uv[0] = map->us[i];
        stat  = EG_evaluate(face, uv, result);
        if (stat != EGADS_SUCCESS) {
          printf(" updateMap: Fill Ua EG_evaluate = %d!\n", stat);
          return;
        }
        CROSS(xyz, u1, v1);
        d = sqrt(xyz[0]*xyz[0] + xyz[1]*xyz[1] + xyz[2]*xyz[2]);
        if (d == 0) d = 1.0;
        xyzs[3*i  ] = xyz[0]/d;
        xyzs[3*i+1] = xyz[1]/d;
        xyzs[3*i+2] = xyz[2]/d;
      }
      for (i = 1; i < map->nu; i++) {
        if (sqrt(xyzs[3*i-3]*xyzs[3*i-3] + xyzs[3*i-2]*xyzs[3*i-2] +
                 xyzs[3*i-1]*xyzs[3*i-1]) == 0.0) continue;
        if (sqrt(xyzs[3*i  ]*xyzs[3*i  ] + xyzs[3*i+1]*xyzs[3*i+1] +
                 xyzs[3*i+2]*xyzs[3*i+2]) == 0.0) continue;
        d = xyzs[3*i  ]*xyzs[3*i-3] + xyzs[3*i+1]*xyzs[3*i-2] +
            xyzs[3*i+2]*xyzs[3*i-1];
        if (d >= dot) continue;
        if (d < -0.2) continue;     /* there is some problem - C0? */
        if (d < dist) {
          max  = i;
          dist = d;
        }
      }
    }
    if (map->nu == MAXUVS) break;
    if (max != -1) {
/*    printf(" Umax = %d (%d)  %lf (%lf)\n", max, map->nu, dist, dot);  */
      for (i = map->nu; i >= max; i--) map->us[i] = map->us[i-1];
      map->us[max] = 0.5*(map->us[max+1] + map->us[max-1]);
      map->nu++;
    }
    if (dist >= last) cnt++;
    last = dist;
  } while ((max != -1) && (cnt < 10)) ;
  
  /* angle -- look in V direction */
  last = -1.0;
  cnt  = 0;
  do {
    max  = -1;
    dist = 1.0;
    for (i = 0; i < map->nu; i++) {
      uv[0] = map->us[i];
      for (j = 0; j < map->nv; j++) {
        uv[1] = map->vs[j];
        stat  = EG_evaluate(face, uv, result);
        if (stat != EGADS_SUCCESS) {
          printf(" updateMap: Fill Va EG_evaluate = %d!\n", stat);
          return;
        }
        CROSS(xyz, u1, v1);
        d = sqrt(xyz[0]*xyz[0] + xyz[1]*xyz[1] + xyz[2]*xyz[2]);
        if (d == 0) d = 1.0;
        xyzs[3*j  ] = xyz[0]/d;
        xyzs[3*j+1] = xyz[1]/d;
        xyzs[3*j+2] = xyz[2]/d;
      }
      for (j = 1; j < map->nv; j++) {
        if (sqrt(xyzs[3*j-3]*xyzs[3*j-3] + xyzs[3*j-2]*xyzs[3*j-2] +
                 xyzs[3*j-1]*xyzs[3*j-1]) == 0.0) continue;
        if (sqrt(xyzs[3*j  ]*xyzs[3*j  ] + xyzs[3*j+1]*xyzs[3*j+1] +
                 xyzs[3*j+2]*xyzs[3*j+2]) == 0.0) continue;
        d = xyzs[3*j  ]*xyzs[3*j-3] + xyzs[3*j+1]*xyzs[3*j-2] +
            xyzs[3*j+2]*xyzs[3*j-1];
        if (d >= dot) continue;
        if (d < -0.2) continue;     /* there is some problem - C0? */
        if (d < dist) {
          max  = j;
          dist = d;
        }
      }
    }
    if (map->nv == MAXUVS) break;
    if (max != -1) {
/*    printf(" Vmax = %d (%d)  %lf (%lf)\n", max, map->nv, dist, dot);  */
      for (j = map->nv; j >= max; j--) map->vs[j] = map->vs[j-1];
      map->vs[max] = 0.5*(map->vs[max+1] + map->vs[max-1]);
      map->nv++;
    }
    if (dist >= last) cnt++;
    last = dist;
  } while ((max != -1) && (cnt < 10)) ;
  
  /* maxedge -- look in U direction */
  do {
    max  = -1;
    dist = 0.0;
    for (j = 0; j < map->nv; j++) {
      uv[1] = map->vs[j];
      for (i = 0; i < map->nu; i++) {
        uv[0] = map->us[i];
        stat  = EG_evaluate(face, uv, result);
        if (stat != EGADS_SUCCESS) {
          printf(" updateMap: Fill U EG_evaluate = %d!\n", stat);
          return;
        }
        xyzs[3*i  ] = result[0];
        xyzs[3*i+1] = result[1];
        xyzs[3*i+2] = result[2];
      }
      for (i = 1; i < map->nu; i++) {
        xyz[0] = xyzs[3*i-3] - xyzs[3*i  ];
        xyz[1] = xyzs[3*i-2] - xyzs[3*i+1];
        xyz[2] = xyzs[3*i-1] - xyzs[3*i+2];
        d = sqrt(xyz[0]*xyz[0] + xyz[1]*xyz[1] + xyz[2]*xyz[2]);
        if (d <= mxedg) continue;
        if (d > dist) {
          max  = i;
          dist = d;
        }
      }
    }
    if (map->nu == MAXUVS) break;
    if (max != -1) {
      for (i = map->nu; i >= max; i--) map->us[i] = map->us[i-1];
      map->us[max] = 0.5*(map->us[max+1] + map->us[max-1]);
      map->nu++;
    }
  } while (max != -1);
  
  /* sag -- look in V direction */
  do {
    max  = -1;
    dist = 0.0;
    for (i = 0; i < map->nu; i++) {
      uv[0] = map->us[i];
      for (j = 0; j < map->nv; j++) {
        uv[1] = map->vs[j];
        stat  = EG_evaluate(face, uv, result);
        if (stat != EGADS_SUCCESS) {
          printf(" updateMap: Fill V EG_evaluate = %d!\n", stat);
          return;
        }
        xyzs[3*j  ] = result[0];
        xyzs[3*j+1] = result[1];
        xyzs[3*j+2] = result[2];
      }
      for (j = 1; j < map->nv; j++) {
        xyz[0] = xyzs[3*j-3] - xyzs[3*j  ];
        xyz[1] = xyzs[3*j-2] - xyzs[3*j+1];
        xyz[2] = xyzs[3*j-1] - xyzs[3*j+2];
        d = sqrt(xyz[0]*xyz[0] + xyz[1]*xyz[1] + xyz[2]*xyz[2]);
        if (d <= mxedg) continue;
        if (d > dist) {
          max  = j;
          dist = d;
        }
      }
    }
    if (map->nv == MAXUVS) break;
    if (max != -1) {
      for (j = map->nv; j >= max; j--) map->vs[j] = map->vs[j-1];
      map->vs[max] = 0.5*(map->vs[max+1] + map->vs[max-1]);
      map->nv++;
    }
  } while (max != -1);

}


static int parse_args(const int argc, char *argv[], char *ifile)
{

    int i;

    printf("********** egads2cgt version %s **********\n", version);

    for (i = 1; i <= argc-1; i++) {

      if (strcmp("-i", argv[i]) == 0 ) {
	 i++;
         strcpy(ifile, argv[i]);
      } else if (strcmp("-q", argv[i]) == 0 ) {
 	 wrtqud = 1;
      } else if (strcmp("-uv", argv[i]) == 0 ) {
 	 wrtuv = 1;
      } else if (strcmp("-maxa", argv[i]) == 0 ) {
	 i++;
         sscanf(argv[i], "%lf", &mxang);
      } else if (strcmp("-maxe", argv[i]) == 0 ) {
	 i++;
         sscanf(argv[i], "%lf", &mxedg);
      } else if (strcmp("-maxc", argv[i]) == 0 ) {
	 i++;
         sscanf(argv[i], "%lf", &mxchd);
      } else {
	printf("Usage: egads2cgt [argument list] (defaults are in parenthesis)\n");
        printf("   -i input geometry filename <*.egads, *.stp, *.igs> ()\n");
        printf("   -q write structured patches from quadding scheme to plot3d surface grid file\n");
        printf("   -uv write structured patches from uv evaluation to plot3d surface grid file\n");
        printf("   -maxa mxang <Max allow dihedral angle (deg)> (15.0)\n");
        printf("   -maxe mxedg <Max allow edge length (0.025 * size)\n");
        printf("   -maxc mxchd <Max allow chord-height tolerance (0.001 * size)\n");
        return 1;
      }
    }

    return 0;
}


int main(int argc, char *argv[])

{
  int          i, j, k, m, mm, n, status, oclass, mtype, nbody, nvert, ntriang;
  int          nnode, nedge, nloop, nface, dim, ptype, pindex;
  int          qnverts, npatch, npatchtot, nn, ip, ii, jj, iroot, iper, count;
  const int    *qptype, *qpindex, *pvindex, *pvbounds;
  int          *triang, *comp, *nipatch, *njpatch, *senses, *ivec;
  char         ifile[120], filename[120], rootname[120], trifilename[120];
  char         p3dfilename[120], uvfilename[120];
  const char   *OCCrev;
  double       params[3], box[6], size, *verts, range[4], uv[2], result[18];
  double       qparams[3], alen, *rvec;
  const double *qxyz, *quv;
  FILE         *fp;
  ego          context, model, geom, *bodies, tess, body, ref;
  ego          *dum, *nodes, *edges, *loops, *faces;
  UVmap        *map;

  /* look at EGADS revision */
  EG_revision(&i, &j, &OCCrev);
  printf("\n Using EGADS %2d.%02d with %s\n\n", i, j, OCCrev);
  
  /* get arguments */
  status = parse_args(argc, argv, ifile);
  if (status != 0) exit(status);
  
  printf("mxang:    %lf\n", mxang);
  printf("mxedg:    %lf\n", mxedg);
  printf("mxchd:    %lf\n", mxchd);

  /* Build filenames from rootname */
  iroot = strroot(ifile);
  printf("iroot:    %d\n", iroot);
  strncpy(rootname, ifile, iroot);
  rootname[iroot] = 0;
  printf("rootname: %s\n", rootname);

  /* Initialize */
  status = EG_open(&context);
  if (status != EGADS_SUCCESS) {
    printf(" EG_open = %d!\n\n", status);
    return 1;
  }
  status = EG_loadModel(context, 0, ifile, &model);
  if (status != EGADS_SUCCESS) {
    printf(" EG_loadModel = %d\n\n", status);
    EG_close(context);
    return 1;
  }
  status = EG_getBoundingBox(model, box);
  if (status != EGADS_SUCCESS) {
    printf(" EG_getBoundingBox = %d\n\n", status);
    EG_deleteObject(model);
    EG_close(context);
    return 1;
  }

  size = sqrt((box[0]-box[3])*(box[0]-box[3]) + (box[1]-box[4])*(box[1]-box[4]) +
              (box[2]-box[5])*(box[2]-box[5]));

  /* Get all bodies */

  status = EG_getTopology(model, &geom, &oclass, &mtype, NULL, &nbody, &bodies,
                          &triang);
  if (status != EGADS_SUCCESS) {
    printf(" EG_getTopology = %d\n\n", status);
    EG_deleteObject(model);
    EG_close(context);
    return 1;
  }
  printf(" Number of Bodies = %d\n\n", nbody);

/* Set tesselation parameters */

  if        (mxedg == 0.0) {
     params[0] = 0.025*size;
  } else if (mxedg >  0.0) {
     params[0] = mxedg;
  } else if (mxedg <  0.0) {
     params[0] = fabs(mxedg*size);
  }

  if        (mxchd == 0.0) {
     params[1] =  0.001*size;
  } else if (mxchd >  0.0) {
     params[1] = mxchd;
  } else if (mxchd <  0.0) {
     params[1] = fabs(mxchd*size);
  }
  if        (mxang == 0.0) {
     params[2] = 15.0;
  } else {
     params[2] = fabs(mxang);
  }
  printf(" Tess params: %lf %lf %lf\n\n", params[0], params[1], params[2]);

  /* ---------------- Tesselate each body ---------------- */

  for (i = 0; i < nbody; i++) {
    snprintf(filename, 120, "%s.%3.3d.p3d", rootname, i+1);

    /* Add attribute */

    qparams[0] = 0.15;
    qparams[1] = 10.0;
    qparams[2] = 0.0;
    status = EG_attributeAdd( bodies[i], ".qParams", ATTRREAL, 3, NULL, qparams, NULL);
    if (status != EGADS_SUCCESS) {
      printf(" Body %d: EG_attributeAdd = %d\n", i+1, status);
    }

    mtype  = 0;
    status = EG_getTopology(bodies[i], &geom, &oclass, &mtype, NULL, &j,
                            &dum, &triang);
    if (status != EGADS_SUCCESS) {
      printf(" Body %d: EG_getTopology = %d\n", i+1, status);
      continue;
    }

    if        (mtype == WIREBODY) {
      printf(" Body %2d: Type = WireBody\n", i+1);
    } else if (mtype == FACEBODY) {
      printf(" Body %2d: Type = FaceBody\n", i+1);
    } else if (mtype == SHEETBODY) {
      printf(" Body %2d: Type = SheetBody\n", i+1);
    } else {
      printf(" Body %2d: Type = SolidBody\n", i+1);
    }

    status = EG_makeTessBody(bodies[i], params, &tess);
    if (status != EGADS_SUCCESS) {
      printf(" EG_makeTessBody %d = %d\n", i, status);
      continue;
    }
    printf("          Tessellation completed!\n");
    
    status = EG_getBodyTopos(bodies[i], NULL, NODE, &nnode, &nodes);
    if (status != EGADS_SUCCESS) {
      printf(" EG_getBodyTopos N %d = %d\n", i, status);
      EG_deleteObject(tess);
      continue;
    }
    status = EG_getBodyTopos(bodies[i], NULL, EDGE, &nedge, &edges);
    if (status != EGADS_SUCCESS) {
      printf(" EG_getBodyTopos E %d = %d\n", i, status);
      EG_free(nodes);
      EG_deleteObject(tess);
      continue;
    }
    status = EG_getBodyTopos(bodies[i], NULL, LOOP, &nloop, &loops);
    if (status != EGADS_SUCCESS) {
      printf(" EG_getBodyTopos L %d = %d\n", i, status);
      EG_free(edges);
      EG_free(nodes);
      EG_deleteObject(tess);
      continue;
    }
    status = EG_getBodyTopos(bodies[i], NULL, FACE, &nface, &faces);
    if (status != EGADS_SUCCESS) {
      printf(" EG_getBodyTopos F %d = %d\n", i, status);
      EG_free(loops);
      EG_free(edges);
      EG_free(nodes);
      EG_deleteObject(tess);
      continue;
    }
    
    /* ------------------- Write out tess owner file ----------------------- */

    status = EG_statusTessBody(tess, &body, &n, &nvert);
    if (status != EGADS_SUCCESS) {
      printf(" EG_statusTessBody = %d!\n", status);
      EG_free(faces);
      EG_free(loops);
      EG_free(edges);
      EG_free(nodes);
      EG_deleteObject(tess);
      continue;
    }
    comp = (int *) EG_alloc(nnode*sizeof(int));
    if (comp == NULL) {
      printf(" Cannot Allocate %d ints!\n", nnode);
      EG_free(faces);
      EG_free(loops);
      EG_free(edges);
      EG_free(nodes);
      EG_deleteObject(tess);
      continue;
    }
    for (n = 1; n <= nvert; n++) {
      status = EG_getGlobal(tess, n, &ptype, &pindex, NULL);
      if (status != EGADS_SUCCESS) {
        printf(" Vert %d: EG_getGlobal = %d\n", n, status);
        continue;
      }
      if (ptype != 0) continue;
      comp[pindex-1] = n;
    }

    /* write it out */
    snprintf(trifilename, 120, "%s.%3.3d.tess", rootname, i+1);
    fp = fopen(trifilename, "w");
    if (fp == NULL) {
      printf(" Cannot Open file %s -- NO FILE WRITTEN\n", trifilename);
      EG_free(comp);
      EG_free(faces);
      EG_free(loops);
      EG_free(edges);
      EG_free(nodes);
      EG_deleteObject(tess);
      continue;
    }
    printf(" Writing EGADS tess file: %s\n", trifilename);
    
    /* header */
    fprintf(fp, " %6d %6d %6d %6d %6d\n", mtype, nnode, nedge, nloop, nface);
    writeAttr(fp, bodies[i], NULL);

    /* nodes */
    for (n = 1; n <= nnode; n++) {
      fprintf(fp, " %6d %6d\n", n, comp[n-1]);
      writeAttr(fp, nodes[n-1], NULL);
    }
    EG_free(comp);
    EG_free(nodes);
    
    /* edges */
    for (n = 1; n <= nedge; n++) {
      status = EG_getTopology(edges[n-1], &geom, &oclass, &mtype, range, &j,
                              &dum, &triang);
      if (status != EGADS_SUCCESS) {
        printf(" %d: EG_getTopology = %d!\n", n, status);
        fprintf(fp, " %6d %6d\n", n, 0);
        fprintf(fp, " %6d\n", 0);
        continue;
      }
      if (mtype == DEGENERATE) {
        fprintf(fp, " %6d %6d\n", n, 0);
        writeAttr(fp, edges[n-1], NULL);
        continue;
      }

      status = EG_getTessEdge(tess, n, &m, &qxyz, &quv);
      if (status != EGADS_SUCCESS) {
        printf(" %d: EG_getTessEdge = %d!\n", n, status);
        fprintf(fp, " %6d %6d\n", n, 0);
        writeAttr(fp, edges[n-1], NULL);
        continue;
      }

      status = EG_arcLength(  edges[n-1], range[0], range[1], &alen);
      if (status != EGADS_SUCCESS) {
        printf(" %d: EG_arcLength = %d!\n", n, status);
      }
      else {
	if ( alen < 0.01 ) {
	  printf(" edge # %d: arc length = %le, number of pts = %d\n", n,alen,m);
        }
      }

      fprintf(fp, " %6d %6d\n", n, m);
      for (j = 1; j <= m; j++) {
        dim = 0;
        if (j == 1) {
          status = EG_indexBodyTopo(bodies[i], dum[0]);
          if (status > EGADS_SUCCESS) dim = status;
        }
        if (j == m) {
          if (mtype == TWONODE) {
            status = EG_indexBodyTopo(bodies[i], dum[1]);
          } else {
            status = EG_indexBodyTopo(bodies[i], dum[0]);
          }
          if (status > EGADS_SUCCESS) dim = status;
        }
        status = EG_localToGlobal(tess, -n, j, &k);
        if (status != EGADS_SUCCESS)
          printf(" %d/%d: EG_localToGlobal Edge = %d!\n", n, j, status);
        fprintf(fp, " %6d %20.13le %6d  ", k, quv[j-1], dim);
        if (j%2 == 0) fprintf(fp,"\n");
      }
      if ((j-1)%2 != 0) fprintf(fp,"\n");
      writeAttr(fp, edges[n-1], NULL);
    }
    EG_free(edges);
    
    /* loops */
    for (j = 1; j <= nloop; j++) {
      status = EG_getTopology(loops[j-1], &geom, &oclass, &mtype, NULL, &m,
                              &dum, &senses);
      if (status != EGADS_SUCCESS) {
        printf(" Body %d: EG_getTopology L %d = %d\n", i+1, j, status);
        continue;
      }
      fprintf(fp, " %6d %6d\n", j, m);
      for (k = 0; k < m; k++)
        fprintf(fp, " %6d %6d\n",
                EG_indexBodyTopo(bodies[i], dum[k]), senses[k]);
      writeAttr(fp, loops[j-1], NULL);
    }
    EG_free(loops);
    
    /* faces */
    for (j = 1; j <= nface; j++) {
      status = EG_getTopology(faces[j-1], &geom, &oclass, &mtype, NULL, &mm,
                              &dum, &senses);
      if (status != EGADS_SUCCESS) {
        printf(" Body %d: EG_getTopology F %d = %d\n", i+1, j, status);
        continue;
      }
      status = EG_getGeometry(geom, &k, &m, &ref, &ivec, &rvec);
      if (status != EGADS_SUCCESS) {
        printf(" %d: EG_getGeometry = %d!\n", j, status);
        continue;
      } else {
        printf(" face %d: surface type = %d\n",j, m);
      }
      status = EG_getTessFace(tess, j, &m, &qxyz, &quv, &qptype, &qpindex,
                              &n, &pvindex, &pvbounds);
      if (status != EGADS_SUCCESS) {
        printf(" %d: EG_getTessFace = %d!\n", j, status);
        continue;
      }
      fprintf(fp, " %6d %6d %6d %6d\n", j, mm, mtype, m);
      for (k = 0; k < mm; k++)
        fprintf(fp, " %6d", EG_indexBodyTopo(bodies[i], dum[k]));
      fprintf(fp, "\n");
      for (k = 1; k <= m; k++) {
        status = EG_localToGlobal(tess, j, k, &n);
        if (status != EGADS_SUCCESS)
          printf(" %d/%d: EG_localToGlobal Face = %d!\n", j, k, status);
        fprintf(fp, " %6d %20.13le %20.13le %6d %6d\n",
                n, quv[2*k-2], quv[2*k-1], qptype[k-1], qpindex[k-1]);
      }
      writeAttr(fp, faces[j-1], NULL);
    }
    fclose(fp);
    
    /* ---------- Write quads file (plot3d format) for each body ----------- */

    if (wrtqud == 1) {
      
      qparams[0] = qparams[1] = qparams[2] = 0.0;
      
      /* First count total number of quad patches to write and write to file */
      npatchtot = 0;
      EG_setOutLevel(context, 0);
      for (n = 0; n < nface; n++) {
        status = EG_makeQuads(tess, qparams, n+1);
        if (EGADS_SUCCESS == status) {
          status    = EG_getQuads(tess, n+1, &qnverts, &qxyz, &quv, &qptype,
                                  &qpindex, &npatch);
          npatchtot = npatchtot + npatch;
        }
      }
      EG_setOutLevel(context, 1);
      printf(" Total number of quad patches %d \n", npatchtot);
      
      if (npatchtot > 0) {
        snprintf(p3dfilename, 120, "%s.%3.3d.p3d", rootname, i+1);
        printf(" Writing PLOT3D Quadded file: %s\n", p3dfilename);
        
        nipatch = njpatch = NULL;
        fp      = fopen(p3dfilename, "w");
        if (fp == NULL) {
          printf(" Error Opening %s!\n", p3dfilename);
          goto bailQuad;
        }
        fprintf(fp, "%d\n", npatchtot);
        
        /* Now get dimensions of all patches and write to file */
        
        nipatch = (int *) EG_alloc(npatchtot*sizeof(int));
        njpatch = (int *) EG_alloc(npatchtot*sizeof(int));
        if ((nipatch == NULL) || (njpatch == NULL)) {
          printf(" Malloc ERROR on %d Patches!\n", npatchtot);
          goto bailQuad;
        }
        
        EG_setOutLevel(context, 0);
        for (nn = n = 0; n < nface; n++) {
          status = EG_makeQuads(tess, qparams, n+1) ;
          if (EGADS_SUCCESS == status) {
            status = EG_getQuads(tess, n+1, &qnverts, &qxyz, &quv, &qptype,
                                 &qpindex, &npatch);
            for (ip = 0; ip < npatch; ip++, nn++) {
              status = EG_getPatch(tess, n+1, ip+1, &nipatch[nn], &njpatch[nn],
                                   &pvindex, &pvbounds);
            }
          }
        }
        
        for (ip = 0; ip < npatchtot; ip++)
          fprintf(fp, "%d %d %d\n", nipatch[ip], njpatch[ip],1);
        
        /* Write x,y,z of each quad patch */
        for (nn = n = 0; n < nface; n++) {
          
          status = EG_makeQuads(tess, qparams, n+1);
          if (status == EGADS_SUCCESS) {
            status = EG_getQuads(tess, n+1, &qnverts, &qxyz, &quv, &qptype,
                                 &qpindex, &npatch);
            for (ip = 0; ip < npatch; ip++, nn++) {
              status = EG_getPatch(tess, n+1, ip+1, &nipatch[nn], &njpatch[nn],
                                   &pvindex, &pvbounds);
              for (count = jj = 0; jj < njpatch[nn]; jj++) {
                for (ii = 0; ii < nipatch[nn]; ii++) {
                  m = pvindex[jj*nipatch[nn] + ii] - 1;
                  fprintf(fp, "%20.13le ", qxyz[3*m]);
                  count++;
                  if (count%5 == 0) fprintf(fp, "\n");
                }
              }
              fprintf(fp, "\n");
              
              for (count = jj = 0; jj < njpatch[nn]; jj++) {
                for (ii = 0; ii < nipatch[nn]; ii++) {
                  m = pvindex[jj*nipatch[nn] + ii] - 1;
                  fprintf(fp, "%20.13le ", qxyz[3*m+1]);
                  count++;
                  if (count%5 == 0) fprintf(fp, "\n");
                }
              }
              fprintf(fp, "\n");
              
              for (count = jj = 0; jj < njpatch[nn]; jj++) {
                for (ii = 0; ii < nipatch[nn]; ii++) {
                  m = pvindex[jj*nipatch[nn] + ii] - 1;
                  fprintf(fp, "%20.13le ", qxyz[3*m+2]);
                  count++;
                  if (count%5 == 0) fprintf(fp, "\n");
                }
              }
              fprintf(fp, "\n");
              
            }
          }
        }
        EG_setOutLevel(context, 1);
      bailQuad:
        if (nipatch != NULL) EG_free(nipatch);
        if (njpatch != NULL) EG_free(njpatch);
        if (fp      != NULL) fclose(fp);
      }
    }

   /* ----------- Write uv file (plot3d format) for each body ------------ */

    if (wrtuv == 1) {

      map = (UVmap *) EG_alloc(nface*sizeof(UVmap));
      if (map == NULL) {
        printf(" ERROR Allocating %d Maps!\n", nface);
      } else {
        snprintf(uvfilename, 120, "%s.%3.3d.uv", rootname, i+1);
        printf(" Writing PLOT3D Untrimmed file: %s\n", uvfilename);
        
        fp = fopen(uvfilename, "w");
        if (fp == NULL) {
          printf(" Error Opening %s!\n", uvfilename);
        } else {
          
          for (n = 0; n < nface; n++) {
            status = EG_getRange(faces[n], range, &iper);
            if (status != EGADS_SUCCESS)
              printf(" %d: EG_getRange = %d\n", n+1, status);
            map[n].nu = 9;
            map[n].nv = 9;
            for (jj = 0; jj < map[n].nv; jj++)
              map[n].vs[jj] =  range[2] +
                              (range[3] - range[2])*(double)jj/(map[n].nv-1);
            for (ii = 0; ii < map[n].nu; ii++)
              map[n].us[ii] =  range[0] +
                              (range[1] - range[0])*(double)ii/(map[n].nu-1);
            updateMap(faces[n], params[0], params[1], params[2], &map[n]);
          }
          fprintf(fp, "%d\n", nface);
          for (n = 0; n < nface; n++)
            fprintf(fp,"%d %d %d \n", map[n].nu, map[n].nv, 1);
          
          for (n = 0; n < nface; n++) {

            status = EG_getTopology(faces[n], &geom, &oclass, &mtype, NULL, &mm,
                                    &dum, &senses);
            if (status != EGADS_SUCCESS) {
               printf(" Body %d: EG_getTopology F %d = %d\n", i+1, n+1, status);
               continue;
            }
          
            for (dim = 0; dim < 3; dim++) {
              for (count = jj = 0; jj < map[n].nv; jj++) {
                uv[1] = map[n].vs[jj];
                for (ii = 0; ii < map[n].nu; ii++) {
                  if ( mtype == SFORWARD ) {
                     uv[0] = map[n].us[ii];
                  } else {
                     uv[0] = map[n].us[map[n].nu-ii-1];
                  }
                  status = EG_evaluate(faces[n], uv, result);
                  if (status != EGADS_SUCCESS)
                    printf(" %d: EG_evaluate = %d\n", n+1, status);
                  fprintf(fp, "%20.13le ", result[dim]);
                  count++;
                  if (count%5 == 0) fprintf(fp, "\n");
                }

                if (count%5 != 0) fprintf(fp, "\n");
              }
            }

              for (count = jj = 0; jj < map[n].nv; jj++) {
                uv[1] = map[n].vs[jj];
                for (ii = 0; ii < map[n].nu; ii++) {
                  if ( mtype == SFORWARD ) {
                     uv[0] = map[n].us[ii];
                  } else {
                     uv[0] = map[n].us[map[n].nu-ii-1];
                  }
                  status = EG_inFace(faces[n], uv);
                  if ( status < 0 ) {
		    printf(" face,ii,jj %d %d %d: inFace = %d\n",n+1,ii,jj,status);
                  }
                  fprintf(fp, "%d ", 1-status);
                  count++;
                  if (count%15 == 0) fprintf(fp, "\n");
                }
                if (count%15 != 0) fprintf(fp, "\n");
              }
            
          }
          fclose(fp);
        }
        EG_free(map);
      }
    }
    EG_free(faces);

   /* --------- Zip up the tessellation and write in CART3D format --------- */

    status = bodyTessellation(tess, nface, &nvert, &verts,
                              &ntriang, &triang, &comp);
    EG_deleteObject(tess);
    if (status != EGADS_SUCCESS) continue;

    /* write it out */
    snprintf(trifilename, 120, "%s.%3.3d.tri", rootname, i+1);

    fp = fopen(trifilename, "w");
    if (fp == NULL) {
      printf(" Can not Open file %s -- NO FILE WRITTEN\n", trifilename);
      EG_free(verts);
      EG_free(triang);
      continue;
    }
    printf(" Writing CART3D tri file: %s\n", trifilename);

    /* header */
    fprintf(fp, "%d  %d\n", nvert, ntriang);

    /* ...vertList     */
    for (j = 0; j < nvert; j++)
      fprintf(fp, " %20.13le %20.13le %20.13le\n",
              verts[3*j  ], verts[3*j+1], verts[3*j+2]);

    /* ...Connectivity */
    for (j = 0; j < ntriang; j++)
      fprintf(fp, "%6d %6d %6d\n", triang[3*j], triang[3*j+1], triang[3*j+2]);

    /* ...Component list*/
    for (j = 0; j < ntriang; j++)
      fprintf(fp, "%6d\n", comp[j]);

    fclose(fp);

    printf("      # verts = %d,  # tris = %d\n\n", nvert, ntriang);

    EG_free(verts);
    EG_free(triang);
    EG_free(comp);
  }

  status = EG_deleteObject(model);
  if (status != EGADS_SUCCESS) printf(" EG_deleteObject = %d\n", status);

  EG_close(context);

  return 0;
}
