#
# Database utilities
# 
# procedures for accessing database attributes
#
#

set wd 80
set fmt "  | %-20.20s | %-10.10s | %-${wd}.${wd}s |"
set skipEntsWithoutAttrs 0
set showExtraEntInfo 1

proc toLink { ent } {
  set title "Entity is [expr {[$ent getEnabled] ? "Visible" : "hidden"}]"
  return "<a href='pwent://pwi.app?ents=$ent' title='$title'>[$ent getName]</a>"
}

proc toNames { ents } {
  set ret {}
  foreach ent $ents {
    if { [llength $ent] > 1 } {
      lappend ret [toNames $ent]
    } elseif { ![string match {::pw::*} $ent] } {
      lappend ret $ent
    } else {
      lappend ret [toLink $ent]
    }
  }
  return $ret
}

proc dumpInfoItem { pfx val } {
  if { 0 != [llength [set val [string trim $val]]] } {
    puts [format "  %-6.6s: %s" $pfx [list $val]]
  }
}

proc dumpExtraInfo { dbEnt } {
  dumpInfoItem type [$dbEnt getType]
  dumpInfoItem isUV [$dbEnt isParametric]
  dumpInfoItem ref [toNames [$dbEnt getReferencingEntities]]
  dumpInfoItem sup [toNames [$dbEnt getSupportEntities]]

  switch -- [$dbEnt getDescription] {
  Edge -
  LinearSpline -
  TrimmedSurface {
    # NOP
  }
  BSplineSurface -
  Quilt {
    set cnt [$dbEnt getBoundaryCount]
    set bndrys {}
    for {set ii 1} {$ii <= $cnt} {incr ii} {
      lappend bndrys [$dbEnt getBoundary $ii]
    }
    dumpInfoItem bndry [toNames $bndrys]
  }
  SolidModel {
    set cnt [$dbEnt getBoundaryCount]
    set bndrys {}
    for {set ii 1} {$ii <= $cnt} {incr ii} {
      lappend bndrys [$dbEnt getBoundary $ii]
    }
    dumpInfoItem bndry [toNames $bndrys]
  }
  Model {
    set cnt [$dbEnt getBoundaryCount]
    set bndrys {}
    for {set ii 1} {$ii <= $cnt} {incr ii} {
      lappend bndrys [$dbEnt getBoundary $ii]
    }
    dumpInfoItem bndry [toNames $bndrys]
  }
  default {
  }}
}

proc dumpAttr { dbEnt name } {
  global fmt
  lassign [$dbEnt getImportedAttribute $name] aType aVal
  puts [format $fmt $name $aType $aVal ]
}

proc dumpAttrs { dbEnt } {
  set attrNames [$dbEnt getImportedAttributeNames]
  if { 0 == [llength $attrNames] && $::skipEntsWithoutAttrs } {
    return
  }
  global fmt wd
  set dashes [string repeat - $wd]
  puts "\n[$dbEnt getDescription] [toLink $dbEnt]:"
  if { $::showExtraEntInfo } {
    dumpExtraInfo $dbEnt
  }
  if { 0 != [llength $attrNames] } {
    puts [format $fmt {Attribute Name} Type Value]
    puts [format $fmt $dashes $dashes $dashes]
    foreach attrName $attrNames {
      dumpAttr $dbEnt $attrName
    }
    puts [format $fmt $dashes $dashes $dashes]
  }
}

proc attributeValue { dbEnt entDesc phrase} {
  set value ""

  # Check entity
  if { $entDesc == [$dbEnt getDescription] } {
    set attrNames [$dbEnt getImportedAttributeNames]
    if { 0 != [llength $attrNames] } {
      foreach attrName $attrNames {
        if { "" == $value && $attrName == $phrase } {
          lassign [$dbEnt getImportedAttribute $attrName] aType aVal
          set value $aVal
        }
      }
    }
  }

  # Check support entities
  if { "" == $value } {
    set sups [$dbEnt getSupportEntities]
    if { 0 != [llength $sups] } {
      foreach sup $sups {
        if { "" == $value && $entDesc == [$sup getDescription] } {
          set attrNames [$sup getImportedAttributeNames]
          if { 0 != [llength $attrNames] } {
            foreach attrName $attrNames {
              if { "" == $value && $attrName == $phrase } {
                lassign [$sup getImportedAttribute $attrName] aType aVal
                set value $aVal
              }
            }
          }
        }
      }
    }
  }

  # Check reference entities
  #if { "" == $value } {
  #  set refs [$dbEnt getReferencingEntities]
  #  if { 0 != [llength $refs] } {
  #    foreach ref $refs {
  #      if { "" == $value && $entDesc == [$ref getDescription] } {
  #        set attrNames [$ref getImportedAttributeNames]
  #        if { 0 != [llength $attrNames] } {
  #          foreach attrName $attrNames {
  #            if { "" == $value && $attrName == $phrase } {
  #              lassign [$ref getImportedAttribute $attrName] aType aVal
  #              puts "aVal = $aVal"
  #              set value $aVal
  #            }
  #          }
  #        }
  #      }
  #    }
  #  }
  #}

  return $value
}

proc assembleTaggedQuilts { } {
  #
  # retrieve quilt names from geometry
  #

  puts "Looking for quilt names in geometry."

  set QuiltList {}
  set dbEnts [pw::Database getAll]
  foreach dbEnt $dbEnts {
    switch -- [$dbEnt getDescription] {
    Quilt {
      lappend QuiltList [pw::DatabaseEntity getByName [$dbEnt getName]]
    }
    default {
    }}
  }
  puts "Quilt list has [llength $QuiltList] entries."
  
  set QuiltNames {}

  #Search geometry for quilt names
  set i 0
  foreach qlt $QuiltList {

    set dbEnts [$qlt getSupportEntities]

    foreach dbEnt $dbEnts {
      set qname [attributeValue $dbEnt "TrimmedSurface" "PW:QuiltName" ]
      if {"" != $qname } {
        puts "Quilt [expr $i+1] of [llength $QuiltList] has name $qname."
        set nflag 0
        foreach name $QuiltNames {
          if {$qname == $name} {
            set nflag 1
            puts "Already in QuiltNames list."
            break
          }
        }
        if { 0 == $nflag } {
          lappend QuiltNames $qname
        }
      }
    }
    incr i 1
  }

  if { 0 < [llength $QuiltNames] } {

    foreach qname $QuiltNames {
      puts "Assembling quilts tagged with $qname"
      set quilts {}
      foreach qlt $QuiltList {
        set dbEnts [$qlt getSupportEntities]

        foreach dbEnt $dbEnts {
          set name [attributeValue $dbEnt "TrimmedSurface" "PW:QuiltName" ]
          if { $qname == $name } {
            lappend quilts [pw::DatabaseEntity getByName [$qlt getName]]
            break;
          }
        }
      }

      if { 1 < [llength $quilts] } {
        set assembleMode [pw::Application begin Modify $quilts]
        set tmp [pw::Quilt assemble $quilts]
        $assembleMode end
      }
      unset quilts

    }
  }

  unset QuiltList
  unset QuiltNames

}

