// Replace "Pointwise" and "pointwise" the "Name" and "name" of AIM respectfully!

/*
 *      CAPS: Computational Aircraft Prototype Syntheses
 *
 *             Pointwise AIM
 *
 *      Modified from code written by Author, Affiliation
 *
 */


/*!\mainpage Introduction
 *
 * \section overviewPointwise Pointwise AIM Overview
 * A module in the Computational Aircraft Prototype Syntheses (CAPS) has been developed to interact with the
 * ......
 *
 * The Pointwise AIM provides the CAPS users with the ability to generate .......
 *
 * An outline of the AIM's inputs and outputs are provided in \ref aimInputsPointwise and \ref aimOutputsPointwise, respectively.
 *
 * The accepted and expected geometric representation and analysis intentions are detailed in \ref geomRepIntentPointwise.
 *
 * Details of the AIM's shareable data structures are outlined in \ref sharableDataPointwise if connecting this AIM to other AIMs in a
 * parent-child like manner.
 *
 * Files output:
 *  - capsInput.txt - Global glyph inputs set by CAPS
 *  - caps.nmb - Pointwise nmb file generated
 *
 *
 * Pointwise attributes:
 *
 * | Key                 | Value                                            | Geometry Location  |  Description |
 * | :-----------------  | :----------------------------------------------: | :----------------: | :----------- |
 * | PW:Name             |                                                  | Face | Boundary name for domain or collection of domains. |
 * | PW:SurfaceCellType  | $Triangle, $TriangleQuad                         | Face | Surface cell type. Global default is Triangle. |
 * | PW:SurfaceAlgorithm | $Delaunay, $AdvancingFront, $AdvancingFrontOrtho | Face | Surface meshing algorithm. Global default is AdvancingFront. |
 * | PW:WallSpacing      | > 0.0                                            | Face | Viscous normal spacing for T-Rex extrusion. |
 * | PW:QuiltName        |                                                  | Face | Name to give one or more quilts that are assembled into a single quilt. No angle test is performed. |
 * | PW:Baffle           | $Baffle or $Intersect                            | Face | Either a true baffle surface or a surface intersected by a baffle |
 *
 */

// Header functions
#include <string.h>
#include <math.h>
#include "capsTypes.h"
#include "aimUtil.h"

#include "meshUtils.h"       // Collection of helper functions for meshing
#include "miscUtils.h"       // Collection of helper functions for miscellaneous analysis

#include "egads.h"
#include "nmb.h"

// Windows aliasing
#ifdef WIN32
#define getcwd     _getcwd
#define strcasecmp stricmp
#define PATH_MAX   _MAX_PATH
#else
#include <unistd.h>
#include <limits.h>
#endif

//#define DEBUG

// Additional storage values for the instance of the AIM
typedef struct {

    int numBody;

} aimStorage;

static aimStorage *pointwiseInstance = NULL;
static int         numInstance  = 0;

static int destroy_aimStorage(int iIndex) {


    pointwiseInstance[iIndex].numBody = 0;

    return CAPS_SUCCESS;
}

char *EG_strdup(const char *str);

static int writeGlobalGlyph(void *aimInfo, capsValue *aimInputs) {

    int status;

    FILE *fp = NULL;
    char filename[] = "capsInput.txt";

    // Assumed we are currently already in the correct directory
    fp = fopen(filename, "w");
    if (fp == NULL) {
        status = CAPS_IOERR;
        goto cleanup;
    }


    fprintf(fp, "set boundaryDecay %f\n",     aimInputs[aim_getIndex(aimInfo, "Boundary_Decay", ANALYSISIN)-1].vals.real);
    fprintf(fp, "set collisionBuffer %f\n",   aimInputs[aim_getIndex(aimInfo, "Collision_Buffer", ANALYSISIN)-1].vals.real);
    fprintf(fp, "set conInitDim %d\n",        aimInputs[aim_getIndex(aimInfo, "Connector_Initial_Dim", ANALYSISIN)-1].vals.integer);
    fprintf(fp, "set conMaxDim %d\n",         aimInputs[aim_getIndex(aimInfo, "Connector_Max_Dim", ANALYSISIN)-1].vals.integer);
    fprintf(fp, "set conMinDim %d\n",         aimInputs[aim_getIndex(aimInfo, "Connector_Min_Dim", ANALYSISIN)-1].vals.integer);
    fprintf(fp, "set conTurnAngle %f\n",      aimInputs[aim_getIndex(aimInfo, "Connector_Turn_Angle", ANALYSISIN)-1].vals.real);
    fprintf(fp, "set conSplitAngle %f\n",     aimInputs[aim_getIndex(aimInfo, "Connector_Split_Angle", ANALYSISIN)-1].vals.real);
    fprintf(fp, "set edgeMaxGrowthRate %f\n", aimInputs[aim_getIndex(aimInfo, "Edge_Max_Growth_Rate", ANALYSISIN)-1].vals.real);
    fprintf(fp, "set domainFullLayers %d\n",  aimInputs[aim_getIndex(aimInfo, "Domain_Full_Layers", ANALYSISIN)-1].vals.integer);
    fprintf(fp, "set domainMaxLayers %d\n",   aimInputs[aim_getIndex(aimInfo, "Domain_Max_Layers", ANALYSISIN)-1].vals.integer);
    fprintf(fp, "set domainGrowthRate %f\n",  aimInputs[aim_getIndex(aimInfo, "Domain_Growth_Rate", ANALYSISIN)-1].vals.real);
    fprintf(fp, "set fullLayers %d\n",        aimInputs[aim_getIndex(aimInfo, "Full_Layers", ANALYSISIN)-1].vals.integer);
    fprintf(fp, "set maxLayers %d\n",         aimInputs[aim_getIndex(aimInfo, "Max_Layers", ANALYSISIN)-1].vals.integer);
    fprintf(fp, "set growthRate %f\n",        aimInputs[aim_getIndex(aimInfo, "Growth_Rate", ANALYSISIN)-1].vals.real);
    fprintf(fp, "set proxGrowthRate %f\n",    aimInputs[aim_getIndex(aimInfo, "Proximity_Growth_Rate", ANALYSISIN)-1].vals.real);
    fprintf(fp, "set TRexPrisms %d\n",        aimInputs[aim_getIndex(aimInfo, "TRex_Prisms_Flag", ANALYSISIN)-1].vals.integer);
    fprintf(fp, "set IsoTriQuad %d\n",        aimInputs[aim_getIndex(aimInfo, "IsoTriQuad_Flag", ANALYSISIN)-1].vals.integer);
    fprintf(fp, "set volInitialize %d\n",     aimInputs[aim_getIndex(aimInfo, "Volume_Initialize_Flag", ANALYSISIN)-1].vals.integer);
    fprintf(fp, "set CAESolver %s\n", "\"UGRID\"");

    status = CAPS_SUCCESS;
    goto cleanup;

    cleanup:
        if (status != CAPS_SUCCESS) printf("Error: Premature exit in writeGlobalGlyph, status %d\n", status);

        if (fp != NULL) fclose(fp);

        return status;
}

static int set_autoQuiltAttr(ego *body) {

    int status; // Function return status

    int i, j; // Indexing

    int numBodyFace, numChildrenLoop;
    ego *bodyFace = NULL, *childrenLoop=NULL;

    ego geomRef, geomRef2;
    int oclass, mtypeFace, *senses;
    double data[18];

    char pwQuiltAttr[] = "PW:QuiltName";

    const char *face1Attr = NULL, *face2Attr=NULL;

    char quiltNamePrefix[] = "capsAutoQuilt_";
    char quiltName[30];
    int quiltIndex = 1;

    status = EG_getBodyTopos(*body, NULL, FACE, &numBodyFace, &bodyFace);
    if (status != EGADS_SUCCESS) goto cleanup;

    // Loop through the faces and find what edges
    for (i = 0; i < numBodyFace; i++) {

        status = EG_getTopology(bodyFace[i], &geomRef, &oclass, &mtypeFace, data,
                                &numChildrenLoop, &childrenLoop, &senses);
        if (status != EGADS_SUCCESS) goto cleanup;

        status = retrieve_stringAttr(bodyFace[i], pwQuiltAttr, &face1Attr);
        if (status == EGADS_NOTFOUND) {
            face1Attr = NULL;
        } else if (status != CAPS_SUCCESS) {
            goto cleanup;
        }

        for (j = 0; j < numBodyFace; j++) {

            // If faces match
            if (i == j) continue;

            status = EG_getTopology(bodyFace[j], &geomRef2, &oclass, &mtypeFace, data,
                                    &numChildrenLoop, &childrenLoop, &senses);
            if (status != EGADS_SUCCESS) goto cleanup;

            status = EG_isSame(geomRef, geomRef2);
            if (status < CAPS_SUCCESS ) goto cleanup;

            if (status == EGADS_OUTSIDE) continue;

            // I have the same piece of geometry - should be continuous - let's quilt
            status = retrieve_stringAttr(bodyFace[j], pwQuiltAttr, &face2Attr);
            if (status == EGADS_NOTFOUND) {
                face2Attr = NULL;
            } else if (status != CAPS_SUCCESS) {
                goto cleanup;
            }

            // If neither face has an attribute
            if (face1Attr == NULL && face2Attr == NULL) {

                sprintf(quiltName, "%s%d", quiltNamePrefix, quiltIndex);

                // Face 1
                status = EG_attributeAdd(bodyFace[i], pwQuiltAttr, ATTRSTRING, 0, NULL, NULL, quiltName);
                if (status != EGADS_SUCCESS) goto cleanup;

                // Face 2
                status = EG_attributeAdd(bodyFace[j], pwQuiltAttr, ATTRSTRING, 0, NULL, NULL, quiltName);
                if (status != EGADS_SUCCESS) goto cleanup;

                quiltIndex += 1;

            } else if (face1Attr == NULL) {

                status = EG_attributeAdd(bodyFace[i], pwQuiltAttr, ATTRSTRING, 0, NULL, NULL, face2Attr);
                if (status != EGADS_SUCCESS) goto cleanup;

            } else if (face2Attr == NULL) {

                status = EG_attributeAdd(bodyFace[i], pwQuiltAttr, ATTRSTRING, 0, NULL, NULL, face1Attr);
                if (status != EGADS_SUCCESS) goto cleanup;
            } else {
                // Both Face 1 and Face 2 already have attributes
                continue;
            }
        } // Second face loop
    } // Face loop

    status = CAPS_SUCCESS;
    goto cleanup;

    cleanup:
        if (status != CAPS_SUCCESS) printf("Error: Premature exit in set_autoQuiltAttr, status %d\n", status);

        if (bodyFace != NULL) EG_free(bodyFace);

        return status;
}

/* ********************** Exposed AIM Functions ***************************** */

int aimInitialize(int ngIn, capsValue *gIn,
                  int *qeFlag, const char *unitSys, int *nIntent,
                  int **intents, int *nIn, int *nOut, int *nFields,
                  char ***fnames, int **ranks)
{
    int flag;  // query/execute flag
    int        *ints;
    aimStorage *tmp;

#ifdef DEBUG
    printf("\n pointwiseAIM/aimInitialize   ngIn = %d!\n", ngIn);
#endif

    flag    = *qeFlag;

    // Does the AIM execute itself (i.e. no external executable is called)
    *qeFlag = 0; // 1 = AIM executes itself, 0 otherwise

    // specify the number of analysis input and out "parameters"
    *nIn     = 21;
    *nOut    = 1;
    if (flag == 1) return CAPS_SUCCESS;

    *nIntent = 9;
    ints     = (int *) EG_alloc(*nIntent*sizeof(int));
    if (ints == NULL) return EGADS_MALLOC;

    /*! \page geomRepIntentPointwise Geometry Representation and Analysis Intent
     * The geometric representation for the Pointwise AIM requires the body(ies) be either face body(ies) (FACEBODY), solid body(ies)
     * (SOLIDBODY) or  non- and manifold sheet body(ies) (SHEETBODY). Furthermore, the attribute capsIntent should
     * be set to CFD and STRUCTURE analyses, or ALL.
     */

    ints[0]  = ALL | ALL;
    ints[1]  = ALL | STRUCTURE;
    ints[2]  = ALL | CFD;
    ints[3]  = SOLIDBODY | STRUCTURE;
    ints[4]  = SOLIDBODY | CFD;
    ints[5]  = SHEETBODY | STRUCTURE;
    ints[6]  = SHEETBODY | CFD;
    ints[7]  = FACEBODY | STRUCTURE;
    ints[8]  = FACEBODY | CFD;

    *intents = ints;

    // Specify the field variables this analysis can generate
    *nFields = 0;
    *ranks   = NULL;
    *fnames  = NULL;

    // Allocate pointwiseInstance
    if (pointwiseInstance == NULL) {
        pointwiseInstance = (aimStorage *) EG_alloc(sizeof(aimStorage));

        if (pointwiseInstance == NULL) {
            EG_free(*intents);
            *intents = NULL;
            return EGADS_MALLOC;
        }

    } else {
        tmp = (aimStorage *) EG_reall(pointwiseInstance, (numInstance+1)*sizeof(aimStorage));
        if (tmp == NULL) {
            EG_free(*intents);
            *intents = NULL;
            return EGADS_MALLOC;
        }
        pointwiseInstance = tmp;
    }

    // Set initial values for pointwiseInstance

    pointwiseInstance[numInstance].numBody = 0;

    // Increment number of instances
    numInstance += 1;

    return (numInstance-1);
}

// Available AIM inputs
int aimInputs(int iIndex, void *aimInfo, int index, char **ainame,
        capsValue *defval)
{

    /*! \page aimInputsPointwise AIM Inputs
     * The following list outlines the Pointwise options along with their default value available
     * through the AIM interface.
     */

#ifdef DEBUG
    printf(" pointwiseAIM/aimInputs instance = %d  index = %d!\n", iIndex, index);
#endif

    // Inputs
    if (index == 1) {
        *ainame              = EG_strdup("Proj_Name"); // If NULL a volume grid won't be written by the AIM
        defval->type         = String;
        defval->nullVal      = IsNull;
        defval->vals.string  = NULL;
        //defval->vals.string  = EG_strdup("CAPS");
        defval->lfixed       = Change;

        /*! \page aimInputsPointwise
         * - <B> Proj_Name = NULL</B> <br>
         * This corresponds to the output name of the mesh. If left NULL, the mesh is not written to a file.
         */
    } else if (index == 2) {
        *ainame              = EG_strdup("Boundary_Decay");
        defval->type         = Double;
        defval->nullVal      = NotNull;
        defval->units        = NULL;
        defval->lfixed       = Fixed;
        defval->dim          = 0;
        defval->vals.real    = 0.75;

        /*! \page aimInputsPointwise
         * - <B> Boundary_Decay = 0.75</B> <br>
         * Volumetric boundary decay.
         */
    } else if (index == 3) {
        *ainame              = EG_strdup("Collision_Buffer");
        defval->type         = Double;
        defval->nullVal      = NotNull;
        defval->units        = NULL;
        defval->lfixed       = Fixed;
        defval->dim          = 0;
        defval->vals.real    = 1.0;

        /*! \page aimInputsPointwise
         * - <B> Collision_Buffer = 1.0</B> <br>
         * Collision buffer for colliding fronts.
         */
    } else if (index == 4) {
        *ainame              = EG_strdup("Connector_Initial_Dim");
        defval->type         = Integer;
        defval->nullVal      = NotNull;
        defval->units        = NULL;
        defval->lfixed       = Fixed;
        defval->dim          = 0;
        defval->vals.integer = 11;

        /*! \page aimInputsPointwise
         * - <B> Connector_Initial_Dim = 11</B> <br>
         * Initial connector dimension.
         */
    } else if (index == 5) {
        *ainame              = EG_strdup("Connector_Max_Dim");
        defval->type         = Integer;
        defval->nullVal      = NotNull;
        defval->units        = NULL;
        defval->lfixed       = Fixed;
        defval->dim          = 0;
        defval->vals.integer = 512;

        /*! \page aimInputsPointwise
         * - <B> Connector_Max_Dim = 512</B> <br>
         * Maximum connector dimension.
         */
    } else if (index == 6) {
        *ainame              = EG_strdup("Connector_Min_Dim");
        defval->type         = Integer;
        defval->nullVal      = NotNull;
        defval->units        = NULL;
        defval->lfixed       = Fixed;
        defval->dim          = 0;
        defval->vals.integer = 11;

        /*! \page aimInputsPointwise
         * - <B> Connector_Min_Dim = 11</B> <br>
         * Minimum connector dimension.
         */
    } else if (index == 7) {
        *ainame              = EG_strdup("Connector_Turn_Angle");
        defval->type         = Double;
        defval->nullVal      = NotNull;
        defval->units        = NULL;
        defval->lfixed       = Fixed;
        defval->dim          = 0;
        defval->vals.real    = 0.0;

        /*! \page aimInputsPointwise
         * - <B> Connector_Turn_Angle = 0.0</B> <br>
         * Maximum turning angle on connectors (0 - not used).
         */
    } else if (index == 8) {
        *ainame              = EG_strdup("Connector_Split_Angle");
        defval->type         = Double;
        defval->nullVal      = NotNull;
        defval->units        = NULL;
        defval->lfixed       = Fixed;
        defval->dim          = 0;
        defval->vals.real    = 0.0;

        /*! \page aimInputsPointwise
         * - <B> Connector_Split_Angle = 0.0</B> <br>
         * Turning angle on connectors to split (0 - not used).
         */

    } else if (index == 9) {
        *ainame              = EG_strdup("Edge_Max_Growth_Rate");
        defval->type         = Double;
        defval->nullVal      = NotNull;
        defval->units        = NULL;
        defval->lfixed       = Fixed;
        defval->dim          = 0;
        defval->vals.real    = 1.3;

        /*! \page aimInputsPointwise
         * - <B> Edge_Max_Growth_Rate = 1.3</B> <br>
         * Volumetric edge ratio.
         */
    } else if (index == 10) {
        *ainame              = EG_strdup("Domain_Full_Layers");
        defval->type         = Integer;
        defval->nullVal      = NotNull;
        defval->units        = NULL;
        defval->lfixed       = Fixed;
        defval->dim          = 0;
        defval->vals.integer = 0;

        /*! \page aimInputsPointwise
         * - <B> Domain_Full_Layers = 0</B> <br>
         * Domain full layers (0 for multi-normals, >= 1 for single normal).
         */
    } else if (index == 11) {
        *ainame              = EG_strdup("Domain_Max_Layers");
        defval->type         = Integer;
        defval->nullVal      = NotNull;
        defval->units        = NULL;
        defval->lfixed       = Fixed;
        defval->dim          = 0;
        defval->vals.integer = 10;

        /*! \page aimInputsPointwise
         * - <B> Domain_Max_Layers = 10</B> <br>
         * Domain maximum layers.
         */
    } else if (index == 12) {
        *ainame              = EG_strdup("Domain_Growth_Rate");
        defval->type         = Double;
        defval->nullVal      = NotNull;
        defval->units        = NULL;
        defval->lfixed       = Fixed;
        defval->dim          = 0;
        defval->vals.real    = 1.2;

        /*! \page aimInputsPointwise
         * - <B> Domain_Growth_Rate = 1.2</B> <br>
         * Domain maximum layers.
         */
    } else if (index == 13) {
        *ainame              = EG_strdup("Domain_Growth_Rate");
        defval->type         = Double;
        defval->nullVal      = NotNull;
        defval->units        = NULL;
        defval->lfixed       = Fixed;
        defval->dim          = 0;
        defval->vals.real    = 1.2;

        /*! \page aimInputsPointwise
         * - <B> Domain_Growth_Rate = 1.2</B> <br>
         * Domain growth rate for extrusion.
         */
    } else if (index == 14) {
        *ainame              = EG_strdup("Full_Layers");
        defval->type         = Integer;
        defval->nullVal      = NotNull;
        defval->units        = NULL;
        defval->lfixed       = Fixed;
        defval->dim          = 0;
        defval->vals.integer = 1;

        /*! \page aimInputsPointwise
         * - <B> Full_Layers = 1</B> <br>
         * Full layers (0 for multi-normals, >= 1 for single normaln.
         */
    } else if (index == 15) {
        *ainame              = EG_strdup("Max_Layers");
        defval->type         = Integer;
        defval->nullVal      = NotNull;
        defval->units        = NULL;
        defval->lfixed       = Fixed;
        defval->dim          = 0;
        defval->vals.integer = 100;

        /*! \page aimInputsPointwise
         * - <B> Max_Layers = 100</B> <br>
         * Maximum layers.
         */
    } else if (index == 16) {
        *ainame              = EG_strdup("Growth_Rate");
        defval->type         = Double;
        defval->nullVal      = NotNull;
        defval->units        = NULL;
        defval->lfixed       = Fixed;
        defval->dim          = 0;
        defval->vals.real    = 1.2;

        /*! \page aimInputsPointwise
         * - <B> Growth_Rate = 1.2</B> <br>
         * Growth rate for boundary layer extrusion.
         */
    } else if (index == 17) {
        *ainame              = EG_strdup("Proximity_Growth_Rate");
        defval->type         = Double;
        defval->nullVal      = NotNull;
        defval->units        = NULL;
        defval->lfixed       = Fixed;
        defval->dim          = 0;
        defval->vals.real    = 1.3;

        /*! \page aimInputsPointwise
         * - <B> Proximity_Growth_Rate = 1.3</B> <br>
         * Connector proximity growth rate.
         */
    } else if (index == 18) {
        *ainame              = EG_strdup("TRex_Prisms_Flag");
        defval->type         = Boolean;
        defval->vals.integer = (int) false;
        defval->nullVal      = NotNull;

        /*! \page aimInputsPointwise
         * - <B> TRex_Prisms_Flag = False</B> <br>
         * Allow conversion to prisms.
         */
    } else if (index == 19) {
        *ainame              = EG_strdup("IsoTriQuad_Flag");
        defval->type         = Boolean;
        defval->vals.integer = (int) false;
        defval->nullVal      = NotNull;

        /*! \page aimInputsPointwise
         * - <B> IsoTriQuad_Flag = False</B> <br>
         * Allow triangles and quads in domains.
         */
    } else if (index == 20) {
        *ainame              = EG_strdup("Volume_Initialize_Flag");
        defval->type         = Boolean;
        defval->vals.integer = (int) true;
        defval->nullVal      = NotNull;

        /*! \page aimInputsPointwise
         * - <B> Volume_Initialize_Flag = True</B> <br>
         * Initialize block after setup.
         */
    } else if (index == 21) {
        *ainame              = EG_strdup("Auto_Quilt_Flag");
        defval->type         = Boolean;
        defval->vals.integer = (int) true;
        defval->nullVal      = NotNull;

        /*! \page aimInputsPointwise
         * - <B> Auto_Quilt_Flag = True</B> <br>
         * Automatically quilt faces that have the same geometric references.
         */
    }

    return CAPS_SUCCESS;
}

// Shareable data for the AIM - typically optional
int aimData(int iIndex, const char *name, enum capsvType *vtype,
        int *rank, int *nrow, int *ncol, void **data, char **units)
{
    /*! \page sharableDataPointwise AIM Shareable Data
     * The Pointwise AIM has the following shareable data types/values with its children AIMs if they are so inclined.
     * - <B> Body_Count</B> <br>
     * Example shareable data returned in **** format.
     */

#ifdef DEBUG
    printf(" pointwiseAIM/aimData instance = %d  name = %s!\n", inst, name);
#endif

    // The shareable data
    if (strcasecmp(name, "Body_Count") == 0){
        *vtype = Value;
        *rank = *ncol = 1;
        *nrow  = 1;
        *data  = (void *) &pointwiseInstance[iIndex].numBody;
        *units = NULL;

        return CAPS_SUCCESS;
    }

    return CAPS_NOTFOUND;

}

// AIM preAnalysis function
int aimPreAnalysis(int iIndex, void *aimInfo, char *analysisPath, capsValue *aimInputs, capsErrs **errs)
{
    int status; // Status return
    int i; // Index

    char currentPath[PATH_MAX]; // Current directory path

    char nmbFileName[] = "caps.nmb";
    int asciiOut = 0;
    int verbose = 0;

    int numBody = 0; // Number of bodies
    ego *bodies = NULL; // EGADS body objects
    ego *bodyCopy = NULL;

    ego context, model;

    // NULL out errors
    *errs = NULL;

    // Get AIM bodies
    status = aim_getBodies(aimInfo, &numBody, &bodies);
    if (status != CAPS_SUCCESS) return status;

    #ifdef DEBUG
        printf(" pointwiseAIM/aimPreAnalysis instance = %d  numBody = %d!\n", iIndex, numBody);
    #endif

    if (numBody <= 0 || bodies == NULL) {
        #ifdef DEBUG
            printf(" pointwiseAIM/aimPreAnalysis No Bodies!\n");
        #endif

        return CAPS_SOURCEERR;
    }


    (void) getcwd(currentPath, PATH_MAX);
    if (chdir(analysisPath) != 0) return CAPS_DIRERR;


    // Cleanup previous aimStorage for the instance in case this is the second time through preAnalysis for the
    // same instance
    status = destroy_aimStorage(iIndex);
    if (status != CAPS_SUCCESS) {
        printf("Status = %d, pointwiseAIM instance %d, aimStorage cleanup!!!\n", status, iIndex);
        goto cleanup;
    }

    bodyCopy = (ego *) EG_alloc(numBody*sizeof(ego));
    if (bodyCopy == NULL) {
        status = EGADS_MALLOC;
        goto cleanup;
    }

    for (i = 0; i < numBody; i++) bodyCopy[i] = NULL;

    // Get context
    status = EG_getContext(bodies[0], &context);
    if (status != EGADS_SUCCESS) goto cleanup;

    // Make a copy of the bodies
    for (i = 0; i < numBody; i++) {
        status = EG_copyObject(bodies[i], NULL, &bodyCopy[i]);
        if (status != EGADS_SUCCESS) goto cleanup;
    }

    // Auto quilt faces
    if (aimInputs[aim_getIndex(aimInfo, "Auto_Quilt_Flag", ANALYSISIN)-1].vals.integer == (int) true) {
        printf("Automatically quilting faces...\n");
        for (i = 0; i < numBody; i++) {
            status = set_autoQuiltAttr(&bodyCopy[i]);
            if (status != CAPS_SUCCESS) goto cleanup;
        }
    }

    // Create a model from the copied bodies
    status = EG_makeTopology(context, NULL, MODEL, 0, NULL, numBody, bodyCopy, NULL, &model);
    if (status != EGADS_SUCCESS) goto cleanup;

    printf("Writing global Glyph inputs...\n");
    status = writeGlobalGlyph(aimInfo, aimInputs);
    if (status != CAPS_SUCCESS) goto cleanup;

    printf("Writing NMB file....\n");
    status = NMB_write(model, nmbFileName, asciiOut, verbose);
    if (status != EGADS_SUCCESS) {
        printf(" NMB_write status = %d\n", status);
        goto cleanup;
    }

    status = CAPS_SUCCESS;
    goto cleanup;

    cleanup:

        if (status != CAPS_SUCCESS) printf("Error: pointwiseAIM (instance = %d) status %d\n", iIndex, status);

        chdir(currentPath);

        if (bodyCopy != NULL) {
            for (i = 0;  i < numBody; i++) {
                if (bodyCopy[i] != NULL)  {
                    (void) EG_deleteObject(bodyCopy[i]);
                }
            }
            EG_free(bodyCopy);
        }

        return status;
}

// Available AIM outputs
int aimOutputs(int iIndex, void *aimInfo,
               int index, char **aoname, capsValue *form)
{
    /*! \page aimOutputsPointwise AIM Outputs
     * The following list outlines the Pointwise AIM outputs available through the AIM interface.
     *
     * - <B>OutputVariable</B> = True if a ....., False if not.
     */

#ifdef DEBUG
    printf(" pointwiseAIM/aimOutputs instance = %d  index = %d!\n", inst, index);
#endif

    if (index == 1) {
        *aoname = EG_strdup("OutputVariable");
        form->type = Boolean;
        form->vals.integer = (int) false;
    }

    return CAPS_SUCCESS;
}

// Get value for a given output variable
int aimCalcOutput(int iIndex, void *aimInfo,
        char *analysisPath, int index,
        capsValue *val, capsErrs **errors)
{

#ifdef DEBUG
    printf(" pointwiseAIM/aimCalcOutput instance = %d  index = %d!\n", iIndex, index);
#endif

    *errors = NULL;

    // Fill in to populate output variable = index

    if (index == 1){

        val->vals.integer = (int) false;
    }

    return CAPS_SUCCESS;
}

/************************************************************************/

/*
 * since this AIM does not support field variables or CAPS bounds, the
 * following functions are not required to be filled in except for aimDiscr
 * which just stores away our bodies and aimFreeDiscr that does some cleanup
 * when CAPS terminates
 */


int aimFreeDiscr(/*@null@*/ capsDiscr *discr)
{
    int iIndex; // Indexing

    int status; // Returning status

#ifdef DEBUG
    if (discr == NULL) {
        printf(" pointwiseAIM/aimFreeDiscr w/ NULL!\n");
    } else {
        printf(" pointwiseAIM/aimFreeDiscr instance = %d!\n", discr->instance);
    }
#endif

    if (discr == NULL) {

        // Clean up pointwiseInstance data
        for ( iIndex = 0; iIndex < numInstance; iIndex++) {

            printf(" Cleaning up pointwiseInstance - %d\n", iIndex);

            status = destroy_aimStorage(iIndex);
            if (status != CAPS_SUCCESS) printf("Status = %d, pointwiseAIM instance %d, aimStorage cleanup!!!\n", status, iIndex);
        }

        if (pointwiseInstance != NULL) EG_free(pointwiseInstance);
        pointwiseInstance = NULL;
        numInstance = 0;

    } else {

        // Clean up discr object
    }

    return CAPS_SUCCESS;

}


int aimDiscr(/*@unused@*/ char *tname, capsDiscr *discr)
{
    int stat, inst;

    inst = discr->instance;

#ifdef DEBUG
    printf(" pointwiseAIM/aimDiscr: tname = %s, instance = %d!\n", tname, inst);
#endif

    if ((inst < 0) || (inst >= numInstance)) return CAPS_BADINDEX;

    stat = aimFreeDiscr(discr);
    if (stat != CAPS_SUCCESS) return stat;

    return CAPS_SUCCESS;
}


int aimLocateElement(/*@unused@*/ capsDiscr *discr, /*@unused@*/ double *params,
        /*@unused@*/ double *param,    /*@unused@*/ int *eIndex,
        /*@unused@*/ double *bary)
{
#ifdef DEBUG
    printf(" pointwiseAIM/aimLocateElement instance = %d!\n", discr->instance);
#endif

    return CAPS_SUCCESS;
}

int aimTransfer(/*@unused@*/ capsDiscr *discr, /*@unused@*/ char *name,
        /*@unused@*/ int npts, /*@unused@*/ int rank,
        /*@unused@*/ double *data, /*@unused@*/ char **units)
{
#ifdef DEBUG
    printf(" pointwiseAIM/aimTransfer name = %s  instance = %d  npts = %d/%d!\n",
            name, discr->instance, npts, rank);
#endif

    return CAPS_SUCCESS;
}

int aimInterpolation(/*@unused@*/ capsDiscr *discr, /*@unused@*/ char *name,
        /*@unused@*/ int eIndex, /*@unused@*/ double *bary,
        /*@unused@*/ int rank, /*@unused@*/ double *data,
        /*@unused@*/ double *result)
{
#ifdef DEBUG
    printf(" pointwiseAIM/aimInterpolation  %s  instance = %d!\n",
            name, discr->instance);
#endif

    return CAPS_SUCCESS;
}


int aimInterpolateBar(/*@unused@*/ capsDiscr *discr, /*@unused@*/ char *name,
        /*@unused@*/ int eIndex, /*@unused@*/ double *bary,
        /*@unused@*/ int rank, /*@unused@*/ double *r_bar,
        /*@unused@*/ double *d_bar)
{
#ifdef DEBUG
    printf(" pointwiseAIM/aimInterpolateBar  %s  instance = %d!\n",
            name, discr->instance);
#endif

    return CAPS_SUCCESS;
}


int aimIntegration(/*@unused@*/ capsDiscr *discr, /*@unused@*/ char *name,
        /*@unused@*/ int eIndex, /*@unused@*/ int rank,
        /*@unused@*/ /*@null@*/ double *data, /*@unused@*/ double *result)
{
#ifdef DEBUG
    printf(" pointwiseAIM/aimIntegration  %s  instance = %d!\n",
            name, discr->instance);
#endif

    return CAPS_SUCCESS;
}


int aimIntegrateBar(/*@unused@*/ capsDiscr *discr, /*@unused@*/ char *name,
        /*@unused@*/ int eIndex, /*@unused@*/ int rank,
        /*@unused@*/ double *r_bar, /*@unused@*/ double *d_bar)
{
#ifdef DEBUG
    printf(" pointwiseAIM/aimIntegrateBar  %s  instance = %d!\n",
            name, discr->instance);
#endif

    return CAPS_SUCCESS;
}
