// Tetgen interface functions - Written by Dr. Ryan Durscher AFRL/RQVC

#include "tetgen.h"
#include "meshTypes.h"
#include "egads.h"
#include "capsTypes.h"

#include "meshUtils.h"

static int tetgen_to_MeshStruct(tetgenio *mesh, meshStruct *genUnstrMesh)  {
    int status; // Function return status

    int i, j, elementIndex; // Indexing

    int numPoint;
    int defaultVolID = 1; // Default volume ID

    meshAnalysisTypeEnum analysisType;

    analysisType = genUnstrMesh->analysisType;

    // Cleanup existing node and elements
    (void) destroy_meshNodes(genUnstrMesh);

    (void) destroy_meshElements(genUnstrMesh);

    (void) destroy_meshQuickRefStruct(&genUnstrMesh->meshQuickRef);

    genUnstrMesh->meshType = VolumeMesh;

    // Numbers
    genUnstrMesh->numNode = mesh->numberofpoints;
    genUnstrMesh->numElement = mesh->numberoftrifaces + mesh->numberoftetrahedra;

    genUnstrMesh->meshQuickRef.useStartIndex = (int) true;
    genUnstrMesh->meshQuickRef.numTriangle = mesh->numberoftrifaces;
    genUnstrMesh->meshQuickRef.numTetrahedral = mesh->numberoftetrahedra;
    genUnstrMesh->meshQuickRef.startIndexTriangle = 0;
    genUnstrMesh->meshQuickRef.startIndexTetrahedral = mesh->numberoftrifaces;

    // Nodes - allocate
    genUnstrMesh->node = (meshNodeStruct *) EG_alloc(genUnstrMesh->numNode*sizeof(meshNodeStruct));
    if (genUnstrMesh->node == NULL) return EGADS_MALLOC;

    // Nodes - set
    for (i = 0; i < genUnstrMesh->numNode; i++) {

        // Initiate node
        status = initiate_meshNodeStruct(&genUnstrMesh->node[i], analysisType);
        if (status != CAPS_SUCCESS) return status;

        // Copy node data
        genUnstrMesh->node[i].nodeID = i+1;

        genUnstrMesh->node[i].xyz[0] = mesh->pointlist[3*i+0];
        genUnstrMesh->node[i].xyz[1] = mesh->pointlist[3*i+1];
        genUnstrMesh->node[i].xyz[2] = mesh->pointlist[3*i+2];

        /*
        if (mesh->numberofpointattributes != 0) {
        mesh->pointmarkerlist != NULL
        }
         */
    }

    // Elements - allocate
    genUnstrMesh->element = (meshElementStruct *) EG_alloc(genUnstrMesh->numElement*sizeof(meshElementStruct));
    if (genUnstrMesh->element == NULL) return EGADS_MALLOC;

    elementIndex = 0;
    numPoint = 0;
    // Elements -Set triangles
    for (i = 0; i < mesh->numberoftrifaces; i++) {
        status = initiate_meshElementStruct(&genUnstrMesh->element[elementIndex], analysisType);
        if (status != CAPS_SUCCESS) return status;

        genUnstrMesh->element[elementIndex].elementType = Triangle;
        genUnstrMesh->element[elementIndex].elementID   = elementIndex+1;
        genUnstrMesh->element[elementIndex].markerID = mesh->trifacemarkerlist[i];

        status = mesh_allocMeshElementConnectivity(&genUnstrMesh->element[elementIndex]);
        if (status != CAPS_SUCCESS) return status;

        if (i == 0) { // Only need this once
            numPoint = mesh_numMeshElementConnectivity(&genUnstrMesh->element[elementIndex]);
        }

        for (j = 0; j < numPoint; j++ ) {
            genUnstrMesh->element[elementIndex].connectivity[j] = mesh->trifacelist[numPoint*i+j];
        }

        elementIndex += 1;
    }

    //Set tetrahedral
    numPoint = 0;
    for (i = 0; i < mesh->numberoftetrahedra; i++) {
        status = initiate_meshElementStruct(&genUnstrMesh->element[elementIndex], analysisType);
        if (status != CAPS_SUCCESS) return status;

        genUnstrMesh->element[elementIndex].elementType = Tetrahedral;
        genUnstrMesh->element[elementIndex].elementID   = elementIndex+1;

        if (mesh->numberoftetrahedronattributes != 0) {
            genUnstrMesh->element[elementIndex].markerID = (int) mesh->tetrahedronattributelist[mesh->numberoftetrahedronattributes*i + 0];
        } else {
            genUnstrMesh->element[elementIndex].markerID = defaultVolID;
        }

        status = mesh_allocMeshElementConnectivity(&genUnstrMesh->element[elementIndex]);
        if (status != CAPS_SUCCESS) return status;

        if (i == 0) { // Only need this once
            numPoint = mesh_numMeshElementConnectivity(&genUnstrMesh->element[elementIndex]);
        }

        for (j = 0; j < numPoint; j++ ) {
            genUnstrMesh->element[elementIndex].connectivity[j] = mesh->tetrahedronlist[numPoint*i+j];
        }

        elementIndex += 1;
    }

    return 0;
}

//#ifdef __cplusplus
extern "C" {
//#endif

int tetgen_VolumeMesh(int numHoles,
                      double *holes,
                      meshInputStruct meshInput,
                      meshStruct *surfaceMesh,
                      meshStruct *volumeMesh)
{
    // Initialize variables
    int status = 0; //Function status return
    int i, j; // Indexing
    char *inputString = NULL; // Input string to tetgen

    // Create Tetgen input string
    char temp[120] = "p"; // Tetrahedralize a piecewise linear complex flag
    char q[80];


    // TetGen variables
    tetgenio in, out;
    tetgenio::facet *f;
    tetgenio::polygon *p;

    printf("\nGenerating volume mesh using TetGen.....\n");

    // First index - either 0 or 1
    in.firstnumber = 1;

    // Check inputs
    if (surfaceMesh->numNode    == 0) return CAPS_BADVALUE;
    if (surfaceMesh->numElement == 0) return CAPS_BADVALUE;

    // Set the number of surface nodes
    in.numberofpoints = surfaceMesh->numNode;

    // Create surface point array
    in.pointlist = new REAL[in.numberofpoints *3];

    // Transfer input coordinates to tetgen array
    for(i = 0; i < surfaceMesh->numNode; i++) {
        in.pointlist[3*i  ] = surfaceMesh->node[i].xyz[0];
        in.pointlist[3*i+1] = surfaceMesh->node[i].xyz[1];
        in.pointlist[3*i+2] = surfaceMesh->node[i].xyz[2];
    }

    // Set the number of surface tri
    in.numberoffacets = surfaceMesh->numElement;

    // Create surface tri arrays
    in.facetlist = new tetgenio::facet[in.numberoffacets];

    // Create surface marker/BC arrays
    in.facetmarkerlist = new int[in.numberoffacets];

    // Transfer input surfaceMesh->localTriFaceList array to tetgen array
    for(i = 0; i < surfaceMesh->numElement; i++) {
        f =&in.facetlist[i];
        f->numberofpolygons = 1;
        f->polygonlist = new tetgenio::polygon[f->numberofpolygons];
        f->numberofholes = 0;
        f->holelist = NULL;
        p = &f->polygonlist[0];

        p->numberofvertices = mesh_numMeshElementConnectivity(&surfaceMesh->element[i]);
        p->vertexlist = new int[p->numberofvertices];

        for (j = 0; j < p->numberofvertices; j++) {
            p->vertexlist[j] = surfaceMesh->element[i].connectivity[j];
        }
    }

    // Transfer input BC array to tetgen array
    for(i = 0; i < surfaceMesh->numElement; i++) {
        in.facetmarkerlist[i] = surfaceMesh->element[i].markerID;
    }

    // Add hole information
    if ((holes != NULL) && (numHoles != 0)) {
        //printf("\n WE HAVE HOLES\n");

        // Initialize variables
        in.numberofholes = numHoles;
        in.holelist = new REAL[in.numberofholes *3];

        // Transfer hole points to tetgen array
        for (i = 0; i < in.numberofholes; i++) {
            in.holelist[3*i  ] = holes[3*i  ];
            in.holelist[3*i+1] = holes[3*i+1];
            in.holelist[3*i+2] = holes[3*i+2];
        }
    }

    // If no input string is provided create a simple one based on exposed parameters
    if (meshInput.tetgenInput.meshInputString == NULL) {

        if (meshInput.preserveSurfMesh !=0) strcat(temp, "Y"); // Preserve surface mesh flag

        if ((meshInput.tetgenInput.meshQuality_rad_edge != 0) ||
                (meshInput.tetgenInput.meshQuality_angle != 0)) {

            strcat(temp, "q"); // Mesh quality flag

            if (meshInput.tetgenInput.meshQuality_rad_edge >= 0) {

                sprintf(q, "%.3f",meshInput.tetgenInput.meshQuality_rad_edge);
                strcat(temp,q);

            } else if (meshInput.tetgenInput.meshQuality_rad_edge < 0)  {

                printf("Not setting meshQuality radius-edge ratio. Value needs to be positive\n");
            }

            if (meshInput.tetgenInput.meshQuality_angle >= 0) {

                sprintf(q, "/%.3f",meshInput.tetgenInput.meshQuality_angle);
                strcat(temp,q);

            } else if (meshInput.tetgenInput.meshQuality_angle < 0) {

                printf("Not setting meshQuality dihedral angle. Value needs to be positive\n");
            }

        }

        if ((meshInput.quiet != 0) && (meshInput.tetgenInput.verbose == 0)) strcat(temp, "Q"); // Quiet: No terminal output except errors
        if (meshInput.tetgenInput.verbose != 0) strcat(temp, "V"); // Verbose: Detailed information, more terminal output.

        if (meshInput.tetgenInput.meshTolerance > 0) {
            sprintf(q, "T%.2e",meshInput.tetgenInput.meshTolerance);
            strcat(temp,q);
        }

        // Transfer temp character array to input
        inputString = temp;

    } else {

        inputString = meshInput.tetgenInput.meshInputString;
        // Dont think this is needed - maybe p is the default
        // Check to make sure a 'p' is included in the input string provided by a user
        /*if (strchr(input,'p') == NULL) {
            printf("Warning: No \'p\' was found in the Tetgen input string. Input into Tetgen is currently based on PLC's\n");
        }*/
    }

    printf("\nTetgen input string = %s""\n", inputString);

    // Create volume mesh
    try {
        tetrahedralize(inputString, &in, &out);
    } catch (...){
        printf("Tetgen failed to generate a volume mesh......!!!\n");
        return -335;
    }

    // Save data
    //in.save_nodes((char *)"TETGEN_Test");
    //in.save_poly((char *) "TETGEN_Test");
    //out.save_faces((char *) "TETGEN_Test");

    // Transfer tetgen mesh structure to genUnstrMesh format
    status = tetgen_to_MeshStruct(&out, volumeMesh);
    if (status != 0) return status;


    // Populate surface mesh
    /* if (meshInput.preserveSurfMesh == (int) false &&
        meshInput.tetgenInput.meshInputString == NULL &&
        meshInput.tetgenInput.ignoreSurfaceExtract == (int) false) {

        printf("Extracting surface mesh from volume\n");

        int nodeCount = 0;
        int node;
        bool found = false;

        int *unique_Nodelist;

        unique_Nodelist = new int [3*out.numberoftrifaces]; // At most 3*numTriFaces

        for (i = 0; i < 3*out.numberoftrifaces; i++) {

            node = out.trifacelist[i];

            found = false;
            for (j = 0; j < nodeCount; j++){

                if (unique_Nodelist[j] == node) {
                    found = true;
                    break;
                }
            }

            if (found == false) {
                unique_Nodelist[nodeCount] = node;
                nodeCount += 1;
            }
        }

        // Populate local node list
        //printf("Num. nodes = %d\n",nodeCount);
        genUnstrMesh->surfaceMesh.numNodes = nodeCount;

        if (genUnstrMesh->surfaceMesh.nodeLocal2Global != NULL) EG_free(genUnstrMesh->surfaceMesh.nodeLocal2Global);

        genUnstrMesh->surfaceMesh.nodeLocal2Global = (int *) EG_alloc(genUnstrMesh->surfaceMesh.numNodes * sizeof(int));
        if (genUnstrMesh->surfaceMesh.nodeLocal2Global == NULL) return EGADS_MALLOC;

        if (genUnstrMesh->surfaceMesh.xyz != NULL) EG_free(genUnstrMesh->surfaceMesh.xyz);

        genUnstrMesh->surfaceMesh.xyz  = (double *) EG_alloc(3*genUnstrMesh->surfaceMesh.numNodes *sizeof(double));
        if (genUnstrMesh->surfaceMesh.xyz == NULL) return EGADS_MALLOC;

        for (i = 0; i < genUnstrMesh->surfaceMesh.numNodes; i++) {
            genUnstrMesh->surfaceMesh.xyz[3*i + 0] = out.pointlist[(unique_Nodelist[i]-1) + 0]; // -1 -> 1 bias of node numbering
            genUnstrMesh->surfaceMesh.xyz[3*i + 1] = out.pointlist[(unique_Nodelist[i]-1) + 1];
            genUnstrMesh->surfaceMesh.xyz[3*i + 2] = out.pointlist[(unique_Nodelist[i]-1) + 2];

            genUnstrMesh->surfaceMesh.nodeLocal2Global[i] = unique_Nodelist[i];
        }

        // Populate local node connectivity
        //genUnstrMesh->surfaceMesh.numTriFace = out.numberoftrifaces;
        if (genUnstrMesh->surfaceMesh.localTriFaceList != NULL) EG_free(genUnstrMesh->surfaceMesh.localTriFaceList);

        genUnstrMesh->surfaceMesh.localTriFaceList = (int *) EG_alloc(3*genUnstrMesh->surfaceMesh.numTriFace * sizeof(int));
        if (genUnstrMesh->surfaceMesh.localTriFaceList == NULL) return EGADS_MALLOC;

        for (i = 0; i < 3*genUnstrMesh->surfaceMesh.numTriFace; i++) {

            for (j = 0; j < genUnstrMesh->surfaceMesh.numNodes; j++) {

                if (out.trifacelist[i] == genUnstrMesh->surfaceMesh.nodeLocal2Global[j]) {
                    genUnstrMesh->surfaceMesh.localTriFaceList[i] = j+1; // Indexing starts at 1 -> 1 bias of node numbering
                    break;
                }
            }
        }

        // Clean up unique node list
        delete [] unique_Nodelist;
        unique_Nodelist = NULL;

    } else if (meshInput.preserveSurfMesh == (int) true &&
               meshInput.tetgenInput.meshInputString == NULL) {
        printf("Surface mesh as been preserved getting global node connectivity and local-2-global node mapping\n");

        if (genUnstrMesh->surfaceMesh.nodeLocal2Global != NULL) EG_free(genUnstrMesh->surfaceMesh.nodeLocal2Global);

        genUnstrMesh->surfaceMesh.nodeLocal2Global = (int *) EG_alloc(genUnstrMesh->surfaceMesh.numNodes * sizeof(int));
        if (genUnstrMesh->surfaceMesh.nodeLocal2Global == NULL) return EGADS_MALLOC;

        //Get local to global node numbering
        for (i = 0; i < genUnstrMesh->surfaceMesh.numNodes; i++) {
            for(j = 0; j < out.numberofpoints; j++){
                if ( genUnstrMesh->surfaceMesh.xyz[3*i + 0] == out.pointlist[3*j + 0] &&
                     genUnstrMesh->surfaceMesh.xyz[3*i + 1] == out.pointlist[3*j + 1] &&
                     genUnstrMesh->surfaceMesh.xyz[3*i + 2] == out.pointlist[3*j + 2]) {
                    break;
                }
            }

            genUnstrMesh->surfaceMesh.nodeLocal2Global[i] = j+in.firstnumber; // Indexing starts at 1 -> 1 bias of node numbering
        }
    }
     */
    // Release TetGen memory - get released when function goes out of scope
    //in.deinitialize();
    //out.deinitialize();

    printf("Done meshing using TetGen!\n");

    return CAPS_SUCCESS;
}

//#ifdef __cplusplus
}
//#endif


