// Miscellaneous related utility functions - Written by Dr. Ryan Durscher AFRL/RQVC

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <math.h>
#ifdef WIN32
#include <io.h>
#else
#include <unistd.h>
#endif

#include "egads.h"
#include "capsTypes.h"  // Bring in CAPS types
#include "miscTypes.h"  // Bring in misc. structures

#ifdef WIN32
#define snprintf _snprintf
#define strcasecmp stricmp
#define access _access
#endif

// Does a file exist?
int file_exist(char *file) {

    #ifdef WIN32
        if (access((const char *) file, 0) == 0) return (int) true;

    #else
        if (access((const char *) file, F_OK) == 0) return (int) true;

    #endif

    return (int) false;
}

// Simple json dictionary parser - currently doesn't support nested arrays for keyValue
int search_jsonDictionary(char *stringToSearch, char *keyWord, char **keyValue) {

    int  i; // Indexing for loop

    int  keyIndexStart; // Starting index of keyValue

    int  keyLength = 0; // Length of keyValue

    char pattern[255]; // Substring - keyWord

    char *patternMatch = NULL; // Mathced substring - keyWord

    // Build pattern
    sprintf(pattern, "\"%s\":", keyWord);

    // Debug function
    //printf("Pattern - [%s]\n", pattern);
    //printf("Search string - [%s]\n", stringToSearch);

    // Search string for pattern - return pointer to match
    patternMatch = strstr(stringToSearch, pattern);

    // If pattern is found - get keyValue
    if (patternMatch != NULL) {

        // Get the starting index of the keyValue (using pointer arithmetic)
        keyIndexStart = (patternMatch - stringToSearch) + strlen(pattern);

        if (stringToSearch[keyIndexStart] == ' ') keyIndexStart += 1; // Skip whitespace after :, some JSON writers have spaces others don't

        // See how long our keyValue is - at most length of incoming string
        for (i = 0; i < (int) strlen(stringToSearch); i++) {

            // Array - currently does not support nested arrays
            if (stringToSearch[keyIndexStart] == '[') {
                keyLength += 1;

                if (stringToSearch[keyIndexStart+keyLength] == ']') {
                    keyLength += 1;
                    break;
                }

            } else {

                keyLength += 1;
                if (stringToSearch[keyIndexStart+keyLength] == ',' ||
                    stringToSearch[keyIndexStart+keyLength] == '}') break;
            }
        }

        // Debug function
        //printf("Key start %d\n"    , keyIndexStart);
        //printf("Size of keyLength - %d\n", keyLength);

        // Free keyValue (if not already null) and allocate
        if (*keyValue != NULL) EG_free(*keyValue);

        if (keyLength > 0) {
            *keyValue = (char *) EG_alloc((keyLength+1) * sizeof(char));
            if (*keyValue == NULL) return EGADS_MALLOC;

        } else {
            //printf("No match found for %s\n",keyWord);
            *keyValue = NULL; //"\0";
            return CAPS_NOTFOUND;
        }

        // Copy matched expression to keyValue
        strncpy((*keyValue), stringToSearch + keyIndexStart, keyLength);
        (*keyValue)[keyLength] = '\0';

    } else {

        //printf("No match found for %s\n",keyWord);
        *keyValue = NULL; //"\0";
        return CAPS_NOTFOUND;
    }

    // Debug function
    // printf("keyValue = [%s]\n", *keyValue);

    return CAPS_SUCCESS;
}

// Free an array strings
int string_freeArray(int numString, char **strings[]) {

    int i; // Indexing

    // Debugging
    // printf("NumStrings = %d\n", numString);

    if (numString == 0) return CAPS_SUCCESS;

    if ((*strings) != NULL) {

        // Free string one by one
        for (i = 0; i < numString; i++) {

        	// Debugging
            // printf("Strings %d = %s\n", i, (*strings)[i]);

            if ((*strings)[i] != NULL) EG_free((*strings)[i]);
        }

        EG_free(*strings);
        *strings = NULL;
    }

    return CAPS_SUCCESS;
}

// Remove quotation marks around a string
char * string_removeQuotation(char *string) {

    int startIndex = 0, endIndex = 0; // Indexing

    int length; // String length with quotes removed

    char *stringNoQuotation;

    if (string == NULL) return NULL;

    for (startIndex = 0; startIndex < strlen(string); startIndex++) {
        if (string[startIndex] != ' ') break;
    }

    for (endIndex = strlen(string)-1; endIndex == 0; endIndex--) {
        if (string[endIndex] != ' ') break;
    }

    // Debug
    //printf("Start Index = %d\n", startIndex);
    //printf("End Index = %d\n", endIndex);
    //printf("Incoming string - %s\n", string);

    // If string doesn't have quatation marks simply copy the string
    if ((string[startIndex] == '\"' || string[startIndex] == '\'') &&
        (string[endIndex]   == '\"' || string[endIndex]   == '\'')) {

        length = endIndex - startIndex -1;

        stringNoQuotation = (char *) EG_alloc((length + 1) * sizeof(char));
        if (stringNoQuotation == NULL) {
            printf("Memory allocation error in string_removeQuotation\n");
            return stringNoQuotation;
        }

        memcpy(stringNoQuotation, &string[startIndex+1], length);
        stringNoQuotation[length] = '\0';

    } else {
        length = strlen(string);

        stringNoQuotation = (char *) EG_alloc((length + 1) * sizeof(char));
        if (stringNoQuotation == NULL) {
            printf("Memory allocation error in string_removeQuotation\n");
            return stringNoQuotation;
        }

        memcpy(stringNoQuotation, string, length);
        stringNoQuotation[length] = '\0';
    }

    // Debug
    //printf("Outgoing string - %s\n", stringNoQuotation);

    return stringNoQuotation;
}

// Convert a string to double
int string_toDouble(char *string, double *number) {

    sscanf(string, "%lf", number);

    return CAPS_SUCCESS;
}

// Convert a string to double array - array is assumed to be bounded by '[' and ']' and comma separated
//  for example [3.0, 41, -4.53E2]
int string_toDoubleArray(char *stringToSearch, int arraySize, double numberArray[]) {

    int i, j;         // Array indexing

    int matchLength;  // String length of matching pattern
    int startIndex; // Keep track of where we are in the string array

    int arrayIndex;
    char *numberString = NULL; // Temporary string to hold the found number

    // Debug function
    // printf("Search string - [%s]\n", stringToSearch);

    // Check to make sure first and last of the incoming string denote an array
    if (stringToSearch[0] != '[' &&
        stringToSearch[strlen(stringToSearch)] != ']') {

        printf("Error (string_toDoubleArray): incoming string should be bounded by '[' and ']' and comma separated\n");
        return CAPS_BADVALUE;
    }

    // Set default array values
    for (i = 0; i < arraySize; i++) numberArray[i] = 0;

    startIndex = 1;
    arrayIndex = 0;
    // Parse string based on defined pattern
    for (i = 1; i < strlen(stringToSearch); i++) {

        if(stringToSearch[i] == ',' ||
           stringToSearch[i] == ']') {

            matchLength = i-startIndex;

            // Debug function
            // printf("MatchLength = %d\n", matchLength);

            // Make sure we aren't exceeding our arrays bounds
            if (arrayIndex >= arraySize) {
                printf("Warning (string_toDoubleArray): Array size mismatch - to many values found!!\n");
                return CAPS_MISMATCH;
            }

            // If not just a blank space
            if (matchLength > 0) {

                if (numberString != NULL) EG_free(numberString);

                numberString = (char *) EG_alloc( (matchLength+1) * sizeof(char));

                // Check for malloc error
                if (numberString == NULL) {

                    // If no match was found fill in the rest of the array with 0
                    for (j = arrayIndex; j < arraySize; j++) numberArray[j] = 0;

                    printf("string_toDoubleArray: MALLOC error!\n");
                    return EGADS_MALLOC;
                }

                // Copy sub-string, that is the elements of the array
                strncpy(numberString, stringToSearch + startIndex, matchLength);
                numberString[matchLength] = '\0';

                //Debug function
                //printf("Number string = [%s]\n",numberString);

                // Convert array element to double
                sscanf(numberString, "%lf", &numberArray[arrayIndex]);

                // Increment start indexes
                arrayIndex += 1;
                startIndex = i+1; // +1 to skip the commas
            }
        }
    }

    // Check to see if our array is under-sized
    if (arrayIndex != arraySize) {
        printf("Warning (string_toDoubleArray): Array size mismatch - remaining values will be 0\n");
        return CAPS_MISMATCH;
    }

    // Debug function - print out number array
    //for (i = 0; i < arraySize; i++) printf("Value = %f\n", numberArray[i]);

    // Free allocated numberString
    if (numberString != NULL) EG_free(numberString);

    return CAPS_SUCCESS;
}

// Convert a string to an array of doubles - array is assumed to be bounded by '[' and ']' and comma separated
//  for example [3.0, 1.0, 9.0]. If the string coming in is not an array the an array of one is created.
int string_toDoubleDynamicArray(char *stringToSearch, int *arraySize, double *numberArray[]) {

    int status;       // Function return
    int i, j;         // Array indexing

    int matchLength;  // String length of matching pattern
    int startIndex; // Keep track of where we are in the string array

    int arrayIndex;
    int haveArray = (int) false;
    char *numberString = NULL; // Temporary string to hold the found element of array

    // Debug function
    //printf("Search string - %s\n", stringToSearch);

    if (stringToSearch == NULL) return CAPS_BADVALUE;

    // Check to make sure first and last of the incoming string denote an array
    if (stringToSearch[0] != '[' &&
            stringToSearch[strlen(stringToSearch)] != ']') {
        *arraySize = 1;

    } else { // Lets count how many commas we have
        haveArray = (int) true;
        *arraySize = 1;
        for (i = 1; i < strlen(stringToSearch); i++) {
            if (stringToSearch[i] == ',') *arraySize = *arraySize + 1;
        }
    }

    // Debug function
    //printf("Number of strings = %d\n", *arraySize);

    // Free array if already set
    //if (*numberArray != NULL) EG_free(*numberArray);

    // Allocate numberArray
    *numberArray = (double *) EG_alloc(*arraySize*sizeof(double));
    if (*numberArray == NULL) {
        printf("string_toDoubleDynamicArray: MALLOC error!\n");
        return EGADS_MALLOC;
    }

    if (*arraySize == 1 && haveArray == (int) false) {

        // Debug function
        //printf("We don't have an array\n");

        status = string_toDouble(stringToSearch, &(*numberArray)[0]);
        if (status != CAPS_SUCCESS) return status;


    } else {

        startIndex = 1;
        arrayIndex = 0;
        // Parse string based on defined pattern
        for (i = 1; i < strlen(stringToSearch); i++) {

            if(stringToSearch[i] == ',' ||
                    stringToSearch[i] == ']') {

                matchLength = i-startIndex;

                // Debug function
                //printf("MatchLength = %d\n", matchLength);

                // Make sure we aren't exceeding our arrays bounds
                if (arrayIndex >= *arraySize) {
                    return CAPS_MISMATCH;
                }

                if (matchLength > 0) {

                    if (numberString != NULL) EG_free(numberString);

                    numberString = (char *) EG_alloc( (matchLength+1) * sizeof(char));

                    // Check for malloc error
                    if (numberString == NULL) {

                        // If we have a malloc error fill in the rest of the array with 0
                        for (j = arrayIndex; j < *arraySize; j++) (*numberArray)[j] = 0;

                        printf("string_toDoubleDynamicArray: MALLOC error!\n");
                        return EGADS_MALLOC;
                    }

                    // Copy sub-string, that is the elements of the array
                    strncpy(numberString, stringToSearch + startIndex, matchLength);
                    numberString[matchLength] = '\0';

                    //Debug function
                    //printf("Number string = [%s]\n",numberString);

                    // Convert array element to double
                    status = string_toDouble(numberString, &(*numberArray)[arrayIndex]);
                    if (status != CAPS_SUCCESS) return status;

                    // Increment start indexes
                    arrayIndex += 1;
                    startIndex = i+1; // +1 to skip the commas
                }
            }
        }
    }

    // Debug function - print out number array
    //for (i = 0; i < arraySize; i++) printf("Value = %f\n", numberArray[i]);

    // Free allocated numberString
    if (numberString != NULL) EG_free(numberString);

    return CAPS_SUCCESS;
}

// Convert a string to an array of strings - array is assumed to be bounded by '[' and ']' and comma separated
//  for example ["3.0", "hey", "foo"] - note the strings should NOT contain commas. If the
//  string coming in is not an array the string is simply copied. Also quotations are removed
int string_toStringDynamicArray(char *stringToSearch, int *arraySize, char **stringArray[]) {

    int i, j;         // Array indexing

    int matchLength;  // String length of matching pattern
    int startIndex; // Keep track of where we are in the string array

    int arrayIndex;
    int haveArray = (int) false;
    char *quoteString = NULL; // Temporary string to hold the found string
    char *noQuoteString = NULL;  // Temporary string to hold the found string with quotation marks removed

    // Debug function
    //printf("Search string - %s length %d\n", stringToSearch, strlen(stringToSearch));

    if (stringToSearch == NULL) return CAPS_BADVALUE;

    // Check to make sure first and last of the incoming string denote an array
    if (stringToSearch[0] == '[' ) {

        // Debug function
        //printf("[ found\n");

        if(stringToSearch[strlen(stringToSearch)-1] == ']') { // Lets count how many commas we have

            // Debug function
            //printf("] found\n");

            haveArray = (int) true;
            *arraySize = 1;
            for (i = 1; i < strlen(stringToSearch); i++) {

                if (stringToSearch[i] == ',') {

                    *arraySize = *arraySize + 1;
                }
            }

        } else {

            *arraySize = 1;
        }

    } else {

        *arraySize = 1;
    }

    // Debug function
    //printf("Number of strings (arraySize) = %d\n", *arraySize);

    // Allocate stringArray
    *stringArray = (char **) EG_alloc(*arraySize*sizeof(char **));
    if (*stringArray == NULL) return EGADS_MALLOC;

    if (*arraySize == 1 && haveArray == (int) false) {

        // Debug function
        //printf("We don't have an array\n");

        noQuoteString = string_removeQuotation(stringToSearch);

        (*stringArray)[0] = (char *) EG_alloc((strlen(noQuoteString)+1) * sizeof(char));

        // Check for malloc error
        if ((*stringArray)[0] == NULL) {

            // Free string array
            if (*stringArray != NULL) EG_free(*stringArray);
            *stringArray = NULL;

            // Free no quote array
            if (noQuoteString != NULL) EG_free(noQuoteString);

            return EGADS_MALLOC;
        }

        strncpy((*stringArray)[0], noQuoteString, strlen(noQuoteString));
        (*stringArray)[0][strlen(noQuoteString)] = '\0';

        if (noQuoteString != NULL) {
            EG_free(noQuoteString);
            noQuoteString = NULL;
        }

    } else {

        startIndex = 1;
        arrayIndex = 0;
        // Parse string based on defined pattern
        for (i = 1; i < strlen(stringToSearch); i++) {
            matchLength = 0;
            if(stringToSearch[i] == ',' ||
                    stringToSearch[i] == ']') {

                matchLength = i-startIndex;

                // Debug function
                //printf("MatchLength = %d\n", matchLength);

                // Make sure we aren't exceeding our arrays bounds
                if (arrayIndex >= *arraySize) {
                    return CAPS_MISMATCH;
                }

                // If not just a blank space
                if (matchLength > 0) {

                    // Allocate the quoted string array
                    quoteString = (char *) EG_alloc( (matchLength+1) * sizeof(char));
                    if (quoteString == NULL) return EGADS_MALLOC;

                    strncpy(quoteString, stringToSearch + startIndex, matchLength);
                    quoteString[matchLength] = '\0';

                    // Remove quotations
                    noQuoteString = string_removeQuotation(quoteString);

                    // Free quote string array
                    if (quoteString != NULL) {
                        EG_free(quoteString);
                        quoteString = NULL;
                    }

                    // Allocate string array element based on no quote string
                    (*stringArray)[arrayIndex] = (char *) EG_alloc( (strlen(noQuoteString)+1) * sizeof(char));

                    // Check for malloc error
                    if ((*stringArray)[arrayIndex] == NULL) {

                        // Free string array
                        for (j = 0; j <  arrayIndex; j++) EG_free((*stringArray)[j]);
                        EG_free(*stringArray);

                        // Free no quote string
                        if (noQuoteString != NULL) EG_free(noQuoteString);

                        return EGADS_MALLOC;
                    }

                    // Copy no quote string into array
                    strncpy((*stringArray)[arrayIndex], noQuoteString, strlen(noQuoteString));
                    (*stringArray)[arrayIndex][strlen(noQuoteString)] = '\0';

                    // Free no quote array
                    if (noQuoteString != NULL) {
                        EG_free(noQuoteString);
                        noQuoteString = NULL;
                    }

                    // Increment start indexes
                    arrayIndex += 1;
                    startIndex = i+1; // +1 to skip the commas
                }
            }
        }
    }

    // Debug function - print out number array
    //for (i = 0; i < *arraySize; i++) printf("Value = %s\n", (*stringArray)[i]);

    // Free quote array
    if (quoteString != NULL) EG_free(quoteString);

    // Free no quote array
    if (noQuoteString != NULL) EG_free(noQuoteString);


    return CAPS_SUCCESS;
}

// Convert a string to integer
int string_toInteger(char *string, int *number) {

    sscanf(string, "%d", number);

    return CAPS_SUCCESS;
}

// Convert a string to an array of doubles - array is assumed to be bounded by '[' and ']' and comma separated
//  for example [3, 1, 9]. If the string coming in is not an array the an array of one is created.
int string_toIntegerDynamicArray(char *stringToSearch, int *arraySize, int *numberArray[]) {

    int status;       // Function return
    int i, j;         // Array indexing

    int matchLength;  // String length of matching pattern
    int startIndex; // Keep track of where we are in the string array

    int arrayIndex;
    int haveArray = (int) false;
    char *numberString = NULL; // Temporary string to hold the found element of array

    // Debug function
    //printf("Search string - %s\n", stringToSearch);

    if (stringToSearch == NULL) return CAPS_BADVALUE;

    // Check to make sure first and last of the incoming string denote an array
    if (stringToSearch[0] != '[' &&
            stringToSearch[strlen(stringToSearch)] != ']') {
        *arraySize = 1;

    } else { // Lets count how many commas we have
        haveArray = (int) true;
        *arraySize = 1;
        for (i = 1; i < strlen(stringToSearch); i++) {
            if (stringToSearch[i] == ',') *arraySize = *arraySize + 1;
        }
    }

    // Debug function
    //printf("Number of strings = %d\n", *arraySize);

    // Free array if already set
    //if (*numberArray != NULL) EG_free(*numberArray);

    // Allocate numberArray
    *numberArray = (int *) EG_alloc(*arraySize*sizeof(int));
    if (*numberArray == NULL) {
        printf("string_toIntegerDynamicArray: MALLOC error!\n");
        return EGADS_MALLOC;
    }

    if (*arraySize == 1 && haveArray == (int) false) {

        // Debug function
        //printf("We don't have an array\n");

        status = string_toInteger(stringToSearch, &(*numberArray)[0]);
        if (status != CAPS_SUCCESS) return status;


    } else {

        startIndex = 1;
        arrayIndex = 0;
        // Parse string based on defined pattern
        for (i = 1; i < strlen(stringToSearch); i++) {

            if(stringToSearch[i] == ',' ||
                    stringToSearch[i] == ']') {

                matchLength = i-startIndex;

                // Debug function
                //printf("MatchLength = %d\n", matchLength);

                // Make sure we aren't exceeding our arrays bounds
                if (arrayIndex >= *arraySize) {
                    return CAPS_MISMATCH;
                }

                if (matchLength > 0) {

                    if (numberString != NULL) EG_free(numberString);

                    numberString = (char *) EG_alloc( (matchLength+1) * sizeof(char));

                    // Check for malloc error
                    if (numberString == NULL) {

                        // If we have a malloc error fill in the rest of the array with 0
                        for (j = arrayIndex; j < *arraySize; j++) (*numberArray)[j] = 0;

                        printf("string_toIntegerDynamicArray: MALLOC error!\n");
                        return EGADS_MALLOC;
                    }

                    // Copy sub-string, that is the elements of the array
                    strncpy(numberString, stringToSearch + startIndex, matchLength);
                    numberString[matchLength] = '\0';

                    //Debug function
                    //printf("Number string = [%s]\n",numberString);

                    // Convert array element to double
                    status = string_toInteger(numberString, &(*numberArray)[arrayIndex]);
                    if (status != CAPS_SUCCESS) return status;

                    // Increment start indexes
                    arrayIndex += 1;
                    startIndex = i+1; // +1 to skip the commas
                }
            }
        }
    }

    // Debug function - print out number array
    //for (i = 0; i < arraySize; i++) printf("Value = %f\n", numberArray[i]);

    // Free allocated numberString
    if (numberString != NULL) EG_free(numberString);

    return CAPS_SUCCESS;
}

// Convert a string to boolean
int string_toBoolean(char *string, int *number) {

    if (strncmp(string, "T", 1) == 0 ||
        strncmp(string, "t", 1) == 0 ||
        strncmp(string, "1", 1) == 0) {

        *number = (int) true;

    } else if (strncmp(string, "F", 1) == 0 ||
               strncmp(string, "f", 1) == 0 ||
               strncmp(string, "0", 1) == 0) {

        *number = (int) false;

    } else {

        printf("Error: Unrecognized boolean string - %s\n", string);
        return CAPS_BADVALUE;
    }

    return CAPS_SUCCESS;
}

// Convert a single string to a prog_argv array
int string_toProgArgs(char *meshInputString, int *prog_argc, char ***prog_argv)
{
    // Input:
    //		meshInputstring - (single) string as would be input at the command line
    // Output:
    //		prog_argc - pointer to number of command line inputs
    //		prog_argv - pointer to array of strings
    // Note: For a command line input the first argument is the program call itself - space is allocated
    //		 	to account for this but is isn't used.

    int startIndex, endIndex, stringIndex; //Indexing;
    int stringLength;

    //printf("Mesh inputString = %s\n", meshInputString);

    if (meshInputString != NULL) {

        // Count out how many space for allocation purposes
        *prog_argc = 2; // Argue 1 = program, Last argument will not have a space after it
        //*prog_argc = 1; // Last argument will not have a space after it

        for (startIndex = 0; startIndex < strlen(meshInputString); startIndex++) {
            if (meshInputString[startIndex] == ' ') *prog_argc += 1;
        }

        // Size argv array based on number of found spaces
        *prog_argv = (char **) EG_alloc(*prog_argc*sizeof(char *));
        if (*prog_argv == NULL) return EGADS_MALLOC;

        for (stringIndex = 0; stringIndex < *prog_argc; stringIndex++) (*prog_argv)[stringIndex] = NULL;

        // Loop through input string again and break at spaces
        startIndex = 0; // Index of start of string

        stringIndex = 1; // First index [0] is nothing

        for (endIndex = 0; endIndex < strlen(meshInputString); endIndex++) {

            if (meshInputString[endIndex] == ' ') {
                stringLength = endIndex-startIndex;

                (*prog_argv)[stringIndex] = (char *) EG_alloc((stringLength+1)*sizeof(char)); // Allocate array
                if ((*prog_argv)[stringIndex] == NULL) return EGADS_MALLOC;

                memcpy((*prog_argv)[stringIndex],
                        meshInputString+startIndex,
                        stringLength*sizeof(char));

                (*prog_argv)[stringIndex][stringLength] = '\0';

                stringIndex += 1; // Increment prog_argv index
                startIndex = endIndex+1; //Skip space, ' '
            }
        }

        // Get last value
        stringLength = endIndex-startIndex;

        (*prog_argv)[stringIndex] = (char *) EG_alloc((stringLength+1)*sizeof(char)); // Allocate array
        if ((*prog_argv)[stringIndex] == NULL) return EGADS_MALLOC;

        memcpy((*prog_argv)[stringIndex],
                meshInputString+startIndex,
                stringLength*sizeof(char));

        (*prog_argv)[stringIndex][stringLength] = '\0';

    } else {
        *prog_argc = 0;
        *prog_argv = NULL;
    }

    //printf("Number of args = %d\n",*prog_argc);
    //for (i = 0; i <*prog_argc ; i++) printf("Arg %d = %s\n", i, (*prog_argv)[i]);

    return CAPS_SUCCESS;

}

// Force a string to upper case
void string_toUpperCase ( char *sPtr )
{
    while ( *sPtr != '\0' ) {
        *sPtr = toupper ( ( unsigned char ) *sPtr );
        ++sPtr;
    }
}

// The max x,y,z coordinates where P(3*i + 0) = x_i, P(3*i + 1) = y_i, and P(3*i + 2) = z_i
void maxCoords(int sizeP, double *P, double *x, double *y, double *z) {

    //printf("Getting max bounds\n");
    int i;
    *x = 0;
    *y = 0;
    *z = 0;
    for (i = 0; i < sizeP ; i++){
        if (P[3*i  ] > *x) *x = P[3*i  ];
        if (P[3*i+1] > *y) *y = P[3*i+1];
        if (P[3*i+2] > *z) *z = P[3*i+2];
    }
}

// The min x,y,z coordinates where P(3*i + 0) = x_i, P(3*i + 1) = y_i, and P(3*i + 2) = z_i
void minCoords(int sizeP, double *P, double *x, double *y, double *z) {

    //printf("Getting min bounds\n");
    int i;
    *x = 0;
    *y = 0;
    *z = 0;
    for (i = 0; i < sizeP ; i++){
        if (P[3*i  ] < *x) *x = P[3*i  ];
        if (P[3*i+1] < *y) *y = P[3*i+1];
        if (P[3*i+2] < *z) *z = P[3*i+2];
    }
}

// Return the endianness of the machine we are working on
int get_MachineENDIANNESS(void) {
    // Return = 0 - little
    // Return = 1 - big

    int num = 1;

    if (*(char *)&num == 1) {

        //printf("\nLittle-Endian\n");
        return 0;

    } else {

        //printf("Big-Endian\n");
        return 1;
    }
}

#ifdef WIN32
int getline(char **linep, size_t *linecapp, FILE *stream)
{
    int  i = 0, cnt = 0;
  
    char *buffer;

    buffer = (char *) EG_reall(*linep, 1026*sizeof(char));
    if (buffer == NULL) return EGADS_MALLOC;

    *linep    = buffer;
    *linecapp = 1024;
    while (cnt < 1024 && !ferror(stream)) {
        i = fread(&buffer[cnt], sizeof(char), 1, stream);
        if (i == 0) break;
        cnt++;
        if ((buffer[cnt-1] == 10) || (buffer[cnt-1] == 13) || // 10 = new line, 13 = carriage return, 0 = NULL
            (buffer[cnt-1] == 0)) break;
    }

    buffer[cnt] = 0;
    return cnt == 0 ? -1 : cnt;
}
#endif

// Return the max value of two doubles
double max_DoubleVal(double x1, double x2)
{
    if (x1 >= x2) return x1;
    else	      return x2;
}
// Return the min value of two doubles
double min_DoubleVal(double x1, double x2)
{
    if (x1 <= x2) return x1;
    else	      return x2;
}

// Cross product axb = c
void cross_DoubleVal(double a[], double b[], double c[]){

    c[0] = (a[1]*b[2]) - (a[2]*b[1]);
    c[1] = (a[2]*b[0]) - (a[0]*b[2]);
    c[2] = (a[0]*b[1]) - (a[1]*b[0]);
}

// Dot product a*b = c
double dot_DoubleVal(double a[], double b[]){
    double c;

    c = (a[0]*b[0] + a[1]*b[1] + a[2]*b[2]);

    return c;
}

// Convert an integer to a string of a given field width and justification
char * convert_integerToString(int integerVal, int fieldWidth, int leftOrRight)
{

    // Input:
    //  integerVal  - Integer to convert to string
    //  fieldWidth  - Desired fieldwidth of string must be at least 1
    //  leftofRight - Left justified = 0, Right justified = anything else

    // Ouput:
    //  stringVal - Returned integer in string format. Returned as a const char *

    int i; // Indexing

    int numSpaces = 0; // Number of blank spaces counter
    int inputTest = 0; // Input check

    char *stringVal = NULL; // Returned string array
    char *blankSpaces = NULL; // Blank space array


    // First check input parameters
    if( fieldWidth > 15 ) {

        inputTest = 1;
        printf("Error in convert_integerToString: fieldWidth > 15 not verified\n");
    }
    if ( (integerVal >= 0) &&  // Positive values
            ( (long) integerVal >= (long) pow(10.0,(double) fieldWidth)) ) {

        inputTest = 1;
    }
    if ( (integerVal      < 0) &&  // Negative values
         ( (long) abs(integerVal) >= (long) pow(10.0,(double) fieldWidth-1)) ) {

        inputTest = 1;
    }

    if (fieldWidth <= 0) inputTest = 1;

    // If checks failed report error and return an empty string
    if (inputTest == 1) {

        printf("Error in convert_integerToString: Input %i is too large for the requested fieldWidth of %d\n", integerVal,
                                                                                                               fieldWidth);
        printf("\tReturning an empty string\n");
        return (char *) EG_alloc(sizeof(char));
    }

    // Find out how many blank spaces there will be
    if (integerVal >= 0) numSpaces = fieldWidth - 1;
    else numSpaces = fieldWidth - 2;
    for (i = 1; i < fieldWidth; i++) {

        if (integerVal >= 0 &&     (long) integerVal  < (long) pow(10.0,(double) i  )) break;
        if (integerVal  < 0 && (long) abs(integerVal) < (long) pow(10.0,(double) i-1)) break;

        numSpaces -= 1;
    }

    // Alloc blank spaces array
    blankSpaces = (char *) EG_alloc((numSpaces+1)*sizeof(char));
    for (i = 0; i < numSpaces; i++) blankSpaces[i] = ' ';
    blankSpaces[numSpaces] = '\0';

    // Alloc output string array
    stringVal = (char *) EG_alloc((fieldWidth+1)*sizeof(char));

    // Populate output string array with blank spaces and integer depending on justification
    if (leftOrRight == 0) {
        if      (integerVal >= 0) sprintf(stringVal,"%i%s",     integerVal , blankSpaces);
        else if (integerVal  < 0) sprintf(stringVal,"-%i%s",abs(integerVal), blankSpaces);

    } else {
        if      (integerVal >= 0) sprintf(stringVal,"%s%i" ,blankSpaces,    integerVal);
        else if (integerVal  < 0) sprintf(stringVal,"%s-%i",blankSpaces,abs(integerVal));
    }

    // Free blanks space array
    EG_free(blankSpaces);

    // Return string array
    return stringVal;
}

// Convert a double to 8 field numeric string, used by convert_doubleToString for 8 fieldWidth requests
// function NOT exposed to miscUtils.h only used internally
static void d2s8fld(double d, int format, char *out)
{
    // Example 123.4567
    // format 0
    //         |
    // 1.234E+2
    // format 1
    // 1.2345+2
    //         |

    int scimag, scidir;
    double sign, absval;
    char sci[2], tmp[20];
    char twoSigFigs[4], threeSigFigs[5], fourSigFigs[6], fiveSigFigs[7];

    // If zero
    if (d ==   0 ||
        d == -0.0 ||
        d == +0.0) {
        sprintf(out, "%s", "0.000000");
        return;
    }

    // Determine components of Scientific Notation Double
    absval = fabs(d);
    sign = absval / d;
    sprintf(tmp,"%1.10E",absval);
    scimag = log10(absval);
    if (absval > 1.0) {
        sprintf(sci , "+");
        scidir = 1;
    } else {
        sprintf(sci , "-");
        scidir = 0;
        scimag = -scimag+1;
    }

    strncpy(twoSigFigs,tmp,3); twoSigFigs[3]='\0';
    strncpy(threeSigFigs,tmp,4); threeSigFigs[4]='\0';
    strncpy(fourSigFigs,tmp,5); fourSigFigs[5]='\0';
    strncpy(fiveSigFigs,tmp,6); fiveSigFigs[6]='\0';

    if (scimag == 0) {
        if (sign > 0.0) {
            sprintf(out,"%5.6f",absval);
        } else {
            sprintf(out,"-%5.5f",absval);
        }
    }

    else if (scimag == 1) {
        if (sign > 0.0 && scidir == 1) {
            sprintf(out,"%5.5f",absval);
        } else if (sign > 0.0 && scidir == 0) {
            sprintf(out,"%5.6f",absval);
        } else if (sign < 0.0 && scidir == 1) {
            sprintf(out,"-%5.4f",absval);
        } else if (sign < 0.0 && scidir == 0) {
            sprintf(out,"-%5.5f",absval);
        }
    }

    else if (scimag == 2) {
        if (sign > 0.0 && scidir == 1) {
            sprintf(out,"%5.4f",absval);
        } else if (sign > 0.0 && scidir == 0) {
            if (format == 1) {
                sprintf(out,"%s%s%d",fiveSigFigs,sci,scimag);
            } else {
                sprintf(out,"%sE%s%d",fourSigFigs,sci,scimag);
            }
        } else if (sign < 0.0 && scidir == 1) {
            sprintf(out,"-%5.3f",absval);
        } else if (sign < 0.0 && scidir == 0) {
            if (format == 1) {
                sprintf(out,"-%s%s%d",fourSigFigs,sci,scimag);
            } else {
                sprintf(out,"-%sE%s%d",threeSigFigs,sci,scimag);
            }
        }
    }

    else if (scimag == 3) {
        if (sign > 0.0 && scidir == 1) {
            sprintf(out,"%5.3f",absval);
        } else if (sign > 0.0 && scidir == 0) {
            if (format == 1) {
                sprintf(out,"%s%s%d",fiveSigFigs,sci,scimag);
            } else {
                sprintf(out,"%sE%s%d",fourSigFigs,sci,scimag);
            }
        } else if (sign < 0.0 && scidir == 1) {
            sprintf(out,"-%5.2f",absval);
        } else if (sign < 0.0 && scidir == 0) {
            if (format == 1) {
                sprintf(out,"-%s%s%d",fourSigFigs,sci,scimag);
            } else {
                sprintf(out,"-%sE%s%d",threeSigFigs,sci,scimag);
            }
        }
    }

    else if (scimag == 4) {
        if (sign > 0.0 && scidir == 1) {
            sprintf(out,"%5.2f",absval);
        } else if (sign > 0.0 && scidir == 0) {
            if (format == 1) {
                sprintf(out,"%s%s%d",fiveSigFigs,sci,scimag);
            } else {
                sprintf(out,"%sE%s%d",fourSigFigs,sci,scimag);
            }
        } else if (sign < 0.0 && scidir == 1) {
            sprintf(out,"-%5.1f",absval);
        } else if (sign < 0.0 && scidir == 0) {
            if (format == 1) {
                sprintf(out,"-%s%s%d",fourSigFigs,sci,scimag);
            } else {
                sprintf(out,"-%sE%s%d",threeSigFigs,sci,scimag);
            }
        }
    }

    else if (scimag > 4 && scimag < 10) {
        if (sign > 0.0) {
            if (format == 1) {
                sprintf(out,"%s%s%d",fiveSigFigs,sci,scimag);
            } else {
                sprintf(out,"%sE%s%d",fourSigFigs,sci,scimag);
            }
        } else {
            if (format == 1) {
                sprintf(out,"-%s%s%d",fourSigFigs,sci,scimag);
            } else {
                sprintf(out,"-%sE%s%d",threeSigFigs,sci,scimag);
            }
        }
    }

    else if (scimag >= 10 && scimag < 100) {
        if (sign > 0.0) {
            if (format == 1) {
                sprintf(out,"%s%s%d",fourSigFigs,sci,scimag);
            } else {
                sprintf(out,"%sE%s%d",threeSigFigs,sci,scimag);
            }
        } else {
            if (format == 1) {
                sprintf(out,"-%s%s%d",threeSigFigs,sci,scimag);
            } else {
                sprintf(out,"-%sE%s%d",twoSigFigs,sci,scimag);
            }
        }
    }

    else {
        printf("FATAL Error: number with exponent greater than 100!\n");
        printf("File: %s: line %d", __FILE__, __LINE__);
        exit(1);
    }
}

// Convert a double to 8 field numeric string, used by convert_doubleToString for 7 fieldWidth requests
// function NOT exposed to mscUtils.h only used internally
static void
d2s7fld(double d, int format, char *out)
{
    // Example 123.456
    // format 0
    //         |
    // 1.23E+2
    // format 1
    // 1.234+2
    //         |

    int scimag, scidir;
    double sign, absval;
    char sci[2], tmp[20];
    char oneSigFigs[3], twoSigFigs[4], threeSigFigs[5], fourSigFigs[6], fiveSigFigs[7];

    // If zero
    if (d ==    0 ||
        d == -0.0 ||
        d == +0.0) {
        sprintf(out, "%s", "0.00000");
        return;
    }

    // Determine components of Scientific Notation Double
    absval = fabs(d);
    sign = absval / d;
    sprintf(tmp,"%1.10E",absval);
    scimag = log10(absval);
    if (absval > 1.0) {
        sprintf(sci , "+");
        scidir = 1;
    } else {
        sprintf(sci , "-");
        scidir = 0;
        scimag = -scimag+1;
    }

    strncpy(oneSigFigs,tmp,2); oneSigFigs[2]='\0';
    strncpy(twoSigFigs,tmp,3); twoSigFigs[3]='\0';
    strncpy(threeSigFigs,tmp,4); threeSigFigs[4]='\0';
    strncpy(fourSigFigs,tmp,5); fourSigFigs[5]='\0';
    strncpy(fiveSigFigs,tmp,6); fiveSigFigs[6]='\0';

    if (scimag == 0) {
        if (sign > 0.0) {
            sprintf(out,"%5.5f",absval);
        } else {
            sprintf(out,"-%5.4f",absval);
        }
    }

    else if (scimag == 1) {
        if (sign > 0.0 && scidir == 1) {
            sprintf(out,"%5.4f",absval);
        } else if (sign > 0.0 && scidir == 0) {
            sprintf(out,"%5.5f",absval);
        } else if (sign < 0.0 && scidir == 1) {
            sprintf(out,"-%5.3f",absval);
        } else if (sign < 0.0 && scidir == 0) {
            sprintf(out,"-%5.4f",absval);
        }
    }

    else if (scimag == 2) {
        if (sign > 0.0 && scidir == 1) {
            sprintf(out,"%5.3f",absval);
        } else if (sign > 0.0 && scidir == 0) {
            if (format == 1) {
                sprintf(out,"%s%s%d",fourSigFigs,sci,scimag);
            } else {
                sprintf(out,"%sE%s%d",threeSigFigs,sci,scimag);
            }
        } else if (sign < 0.0 && scidir == 1) {
            sprintf(out,"-%5.2f",absval);
        } else if (sign < 0.0 && scidir == 0) {
            if (format == 1) {
                sprintf(out,"-%s%s%d",threeSigFigs,sci,scimag);
            } else {
                sprintf(out,"-%sE%s%d",twoSigFigs,sci,scimag);
            }
        }
    }

    else if (scimag == 3) {
        if (sign > 0.0 && scidir == 1) {
            sprintf(out,"%5.2f",absval);
        } else if (sign > 0.0 && scidir == 0) {
            if (format == 1) {
                sprintf(out,"%s%s%d",fourSigFigs,sci,scimag);
            } else {
                sprintf(out,"%sE%s%d",threeSigFigs,sci,scimag);
            }
        } else if (sign < 0.0 && scidir == 1) {
            sprintf(out,"-%5.1f",absval);
        } else if (sign < 0.0 && scidir == 0) {
            if (format == 1) {
                sprintf(out,"-%s%s%d",threeSigFigs,sci,scimag);
            } else {
                sprintf(out,"-%sE%s%d",twoSigFigs,sci,scimag);
            }
        }
    }

    else if (scimag == 4) {
        if (sign > 0.0 && scidir == 1) {
            sprintf(out,"%5.1f",absval);
        } else if (sign > 0.0 && scidir == 0) {
            if (format == 1) {
                sprintf(out,"%s%s%d",fourSigFigs,sci,scimag);
            } else {
                sprintf(out,"%sE%s%d",threeSigFigs,sci,scimag);
            }
        } else if (sign < 0.0 && scidir == 1) {
            sprintf(out,"-%5.0f.",absval);
        } else if (sign < 0.0 && scidir == 0) {
            if (format == 1) {
                sprintf(out,"-%s%s%d",threeSigFigs,sci,scimag);
            } else {
                sprintf(out,"-%sE%s%d",twoSigFigs,sci,scimag);
            }
        }
    }

    else if (scimag > 4 && scimag < 10) {
        if (sign > 0.0) {
            if (format == 1) {
                sprintf(out,"%s%s%d",fourSigFigs,sci,scimag);
            } else {
                sprintf(out,"%sE%s%d",threeSigFigs,sci,scimag);
            }
        } else {
            if (format == 1) {
                sprintf(out,"-%s%s%d",threeSigFigs,sci,scimag);
            } else {
                sprintf(out,"-%sE%s%d",twoSigFigs,sci,scimag);
            }
        }
    }

    else if (scimag >= 10 && scimag < 100) {
        if (sign > 0.0) {
            if (format == 1) {
                sprintf(out,"%s%s%d",threeSigFigs,sci,scimag);
            } else {
                sprintf(out,"%sE%s%d",twoSigFigs,sci,scimag);
            }
        } else {
            if (format == 1) {
                sprintf(out,"-%s%s%d",twoSigFigs,sci,scimag);
            } else {
                sprintf(out,"-%sE%s%d",oneSigFigs,sci,scimag);
            }
        }
    }

    else {
        printf("FATAL Error: number with exponent greater than 100!\n");
        printf("File: %s: line %d", __FILE__, __LINE__);
        exit(1);
    }

}

// Convert an double to a string (scientific notation is used depending on the fieldwidth and value) of a given field width
// Note: if fieldwidth isn't large enough values will be truncated (not rounded)
char * convert_doubleToString(double doubleVal, int fieldWidth, int leftOrRight)
{

    // Input:
    //  integerVal  - Integer to convert to string
    //  fieldWidth  - Desired fieldwidth of string must be at least 5
    //  leftofRight - Left justified = 0, Right justified = anything else

    // Ouput:
    //  stringVal - Returned integer in string format. Returned as a const char *

    int i; // Index

    int inputTest = 0; // Input check
    int minfieldWidth = 0; // Minimum field width for input

    int powerExpUpper, powerExpLower; // Exponent powers for max width number comparison

    char *stringVal = NULL;
    //char *numString = NULL;
    char numString[255];

    // If checks inputs and return an empty string
    if (fabs(doubleVal) < 1 && fabs(doubleVal) != 0 && doubleVal < 0  && fieldWidth < 3 ) {
        minfieldWidth = 2;
        inputTest = 1;

    } else if ( fabs(doubleVal) >= 1  && doubleVal < 0 && fieldWidth < 3 ) {
        minfieldWidth = 2;
        inputTest = 1;

    } else	if (fabs(doubleVal) < 1 && fabs(doubleVal) != 0 && doubleVal > 0  && fieldWidth < 3 ) {
        minfieldWidth = 2;
        inputTest = 1;

    } else if ( fabs(doubleVal) >= 1  && doubleVal > 0 && fieldWidth < 2 ) {
        minfieldWidth = 1;
        inputTest = 1;

    } else if ( fieldWidth < 2 ) {
        minfieldWidth = 1;
        inputTest = 1;
    }

    if (inputTest == 1) {
        printf("Error in convert_doubleToString: Input fieldWidth of %d must be greater than %d for the input value %E\n", fieldWidth, minfieldWidth,doubleVal);
        printf("\tReturning an empty string.\n");
        return (char *) EG_alloc(sizeof(char));
    }

    stringVal = (char *) EG_alloc((fieldWidth+1)*sizeof(char));
    //numString = (char *) EG_alloc((fieldWidth+1)*sizeof(char));

    if (fieldWidth == 8){
        d2s8fld(doubleVal, 0, stringVal);
    } else if (fieldWidth == 7) {
        d2s7fld(doubleVal, 0, stringVal);
    } else {
        if (doubleVal < 0) {
            powerExpUpper = fieldWidth - 2;
            powerExpLower = fieldWidth - 3;
        } else {
            powerExpUpper = fieldWidth - 1;
            powerExpLower = fieldWidth - 2;
        }

        /*
		// Testing information
		printf("\n\nInput double = %f, input fieldwidth = %d\n", doubleVal, fieldWidth);
		if (doubleVal < 0) {
			printf("Power Max = -%.10f\n",pow(10,(double)     powerExpUpper));
			printf("Power Min = -%.10f\n",pow(10,(double) -1* powerExpLower));
		} else {
			printf("Power Max = %.10f\n",pow(10,(double)     powerExpUpper));
			printf("Power Min = %.10f\n",pow(10,(double) -1* powerExpLower));
		}
         */

        if ((fabs(doubleVal) < pow(10,(double)     powerExpUpper)  &&
                fabs(doubleVal) >= pow(10,(double) -1*powerExpLower)) ||
                doubleVal == 0.0)  {

            if (doubleVal < 0) {
                // Checks fieldWidth requirement and return an empty string if failed
                sprintf(numString, "%+.*f", fieldWidth-2, doubleVal);

            } else if (doubleVal == 0 || doubleVal == -0.0 || doubleVal == +0.0) {

                sprintf(numString, "%s", "0.0");

            } else {
                sprintf(numString, "%.*f", fieldWidth-1,doubleVal);
            }

            // Testing information
            //printf("String Length %d, %s, width = %d\n", (int) strlen(numString), numString, fieldWidth );

            // Populate output string array with blank spaces and number string depending on justification
            if (leftOrRight == 0) {

                strncpy(stringVal, numString, fieldWidth);
                for (i = 0; i < fieldWidth - (int) strlen(numString); i++) {
                    stringVal[strlen(numString)+i] = ' ';
                }

            } else {

                if (fieldWidth - (int) strlen(numString) > 0 ) {
                    for (i = 0; i < fieldWidth - (int) strlen(numString); i++) {
                        stringVal[i] = ' ';
                    }

                    for (i = 0; i < (int) strlen(numString); i++) {
                        stringVal[fieldWidth - (int) strlen(numString) + i] = numString[i];
                    }

                } else {

                    strncpy(stringVal, numString, fieldWidth);
                }
            }

        } else {

            if (doubleVal < 0) {
                // If checks inputs and return an empty string
                if (fieldWidth < 7 ) {
                    printf("Error in convert_doubleToString: Input fieldWidth of %d must be greater than 7 for the input value %E\n", fieldWidth,doubleVal);
                    printf("\tReturning an empty string\n");

                    EG_free(stringVal);
                    //EG_free(numString);
                    return "\0";
                }

                sprintf(stringVal, "%+*.*E",fieldWidth,fieldWidth-7, doubleVal);

            } else {
                // If checks inputs and return an empty string
                if (fieldWidth < 6 ) {
                    printf("Error in convert_doubleToString: Input fieldWidth of %d must be greater than 6 for the input value %E\n", fieldWidth,doubleVal);
                    printf("\tReturning an empty string\n");

                    EG_free(stringVal);
                    //EG_free(numString);
                    return "\0";
                }

                sprintf(stringVal, "%*.*E",fieldWidth,fieldWidth-6, doubleVal);
            }

        }

        // Add termination character to the end of the string
        stringVal[fieldWidth] = '\0';

        // Testing information
        //printf("String (%s)\n", stringVal);

        //EG_free(numString);

        // Return string array

    } // end if fieldWidth if statement

    return stringVal;
}

// Search a mapAttrToIndex structure for a given keyword and set/return the corresponding index
int get_mapAttrToIndexIndex(mapAttrToIndexStruct *attrMap, const char *keyWord, int *index) {

    // If the keyword is not found a CAPS_NOTFOUND is returned

    int i; // Indexing

    *index = CAPSMAGIC; // Default value of return

    if (attrMap == NULL) return CAPS_NULLVALUE;
    if (keyWord == NULL) return CAPS_NULLVALUE;

    for (i = 0; i < attrMap->numAttribute; i++) {

        if (strcmp(attrMap->attributeName[i], (char *) keyWord ) == 0) {

            *index = attrMap->attributeIndex[i];
            return CAPS_SUCCESS;
        }
    }


    return CAPS_NOTFOUND;
}

// Search a mapAttrToIndex structure for a given index and return the corresponding keyword
int get_mapAttrToIndexKeyword(mapAttrToIndexStruct *attrMap, int index, char **keyWord) {

    // If the keyword is not found a CAPS_NOTFOUND is returned

    int i; // Indexing

    *keyWord = NULL; // Default value of return

    if (attrMap == NULL) return CAPS_NULLVALUE;

    for (i = 0; i < attrMap->numAttribute; i++) {

        if (index == attrMap->attributeIndex[i]) {

            *keyWord = (char *) attrMap->attributeName[i];
            return CAPS_SUCCESS;
        }
    }

    return CAPS_NOTFOUND;
}

// Set the index of a given keyword in a mapAttrToIndex structure
int set_mapAttrToIndexStruct(mapAttrToIndexStruct *attrMap, const char *keyWord, int index) {

    int i = 0; // Indexing

    if (attrMap == NULL) return CAPS_NULLVALUE;
    if (keyWord == NULL) return CAPS_NULLVALUE;

    for (i = 0; i < attrMap->numAttribute; i++) {

        if (strcmp(attrMap->attributeName[i], (char *) keyWord ) == 0) {

            attrMap->attributeIndex[i] = index;
            return CAPS_SUCCESS;
        }
    }

    return CAPS_NOTFOUND;
}

// Increment a mapAttrToIndex structure with the given keyword and set the default index (= numAttribute)
int increment_mapAttrToIndexStruct(mapAttrToIndexStruct *attrMap, const char *keyWord) {

    int dummyIndex; // Dummy index
    int status = 0; // Returning status code
    int i; // Indexing

    if (attrMap == NULL) return CAPS_NULLVALUE;
    if (keyWord == NULL) return CAPS_NULLVALUE;

    status = get_mapAttrToIndexIndex(attrMap, keyWord, &dummyIndex);
    if (status == CAPS_NOTFOUND){
        attrMap->numAttribute += 1;
    } else {
        //printf("Attribute name already exists in attrMap structure !! \n");
        return EGADS_EXISTS;
    }

    if (attrMap->numAttribute == 1 ) {

        /*
	    printf("Size of char** - %d\n", (int) sizeof(char **));
	    printf("Size of const char* - %d\n", (int) sizeof(const char *));
	    printf("Size of char* - %d\n", (int) sizeof(char *));
	    printf("Size of char - %d\n", (int) sizeof(char));
	    printf("Size of int* - %d\n", (int) sizeof(int *));
	    printf("Size of int - %d\n", (int) sizeof(int));
         */

        attrMap->attributeName  = (char **) EG_alloc(attrMap->numAttribute*sizeof(char *));
        attrMap->attributeIndex = (int   *) EG_alloc(attrMap->numAttribute*sizeof(int   ));
    } else {

        attrMap->attributeName  = (char **) EG_reall(attrMap->attributeName , attrMap->numAttribute*sizeof(char *));
        attrMap->attributeIndex = (int   *) EG_reall(attrMap->attributeIndex, attrMap->numAttribute*sizeof(int   ));
    }

    if (attrMap->attributeName  == NULL) return EGADS_MALLOC;
    if (attrMap->attributeIndex == NULL) return EGADS_MALLOC;

    attrMap->attributeName[attrMap->numAttribute-1] = (char *) EG_alloc((strlen(keyWord) + 1)*sizeof(char));
    if (attrMap->attributeName[attrMap->numAttribute-1] ==  NULL ) {
        for (i = 0; i < attrMap->numAttribute; i++) {
            EG_free(attrMap->attributeName[i]);
        }
        attrMap->attributeName = NULL;
        return EGADS_MALLOC;
    }

    //printf("KEY WORD = %s\n",keyWord);
    sprintf(attrMap->attributeName[attrMap->numAttribute-1], "%s", keyWord);

    status = set_mapAttrToIndexStruct(attrMap, keyWord, attrMap->numAttribute);
    if (status != CAPS_SUCCESS) return status;

    return CAPS_SUCCESS;
}

// Initiate (0 out all values and NULL all pointers) an attribute map in the mapAttrToIndexStruct structure format
int initiate_mapAttrToIndexStruct(mapAttrToIndexStruct *attrMap) {

    if (attrMap == NULL) return CAPS_NULLVALUE;

    attrMap->mapName = NULL;

    attrMap->numAttribute = 0;
    attrMap->attributeName = NULL;
    attrMap->attributeIndex = NULL;

    return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) an attribute map in the mapAttrToIndexStruct structure format
int destroy_mapAttrToIndexStruct(mapAttrToIndexStruct *attrMap) {

    int i; // Indexing

    if (attrMap == NULL) return CAPS_NULLVALUE;

    if (attrMap->mapName != NULL) EG_free(attrMap->mapName);
    attrMap->mapName = NULL;

    if (attrMap->attributeName != NULL) {

        for (i = 0; i < attrMap->numAttribute; i++) {

            if (attrMap->attributeName[i] != NULL) EG_free(attrMap->attributeName[i]);
        }

        EG_free(attrMap->attributeName);
    }

    if (attrMap->attributeIndex != NULL) EG_free(attrMap->attributeIndex);

    attrMap->attributeName  = NULL;
    attrMap->attributeIndex = NULL;

    attrMap->numAttribute = 0;

    return CAPS_SUCCESS;
}

// Make a copy of attribute map (attrMapIn)
int copy_mapAttrToIndexStruct(mapAttrToIndexStruct *attrMapIn, mapAttrToIndexStruct *attrMapOut) {

    int status; // Function return status
    int i, j; // Indexing
    int stringLength;
    char *keyWord = NULL;

    if (attrMapIn  == NULL) return CAPS_NULLVALUE;
    if (attrMapOut == NULL) return CAPS_NULLVALUE;

    // Destroy attrMapOut in case it is already allocated - this implies that it must have at least been initiated
    status =  destroy_mapAttrToIndexStruct(attrMapOut);
    if (status != CAPS_SUCCESS) return status;

    stringLength = strlen(attrMapIn->mapName);

    attrMapOut->mapName = (char *) EG_alloc((stringLength+1)*sizeof(char));
    if (attrMapOut->mapName == NULL) return EGADS_MALLOC;

    memcpy(attrMapOut->mapName,
            attrMapIn->mapName,
            stringLength*sizeof(char));
    attrMapOut->mapName[stringLength] = '\0';

    attrMapOut->numAttribute = attrMapIn->numAttribute;

    attrMapOut->attributeIndex = (int *) EG_alloc(attrMapOut->numAttribute*sizeof(int));
    if (attrMapOut->attributeIndex == NULL) return EGADS_MALLOC;

    memcpy(attrMapOut->attributeIndex,
            attrMapIn->attributeIndex,
            attrMapOut->numAttribute*sizeof(int));

    attrMapOut->attributeName  = (char **) EG_alloc(attrMapOut->numAttribute*sizeof(char *));
    if (attrMapOut->attributeName == NULL) return EGADS_MALLOC;

    for (i = 0; i < attrMapOut->numAttribute; i++){

        keyWord = attrMapIn->attributeName[i];

        attrMapOut->attributeName[i] = (char *) EG_alloc((strlen(keyWord) + 1)*sizeof(char));

        if (attrMapOut->attributeName[i] ==  NULL ) {

            for (j = 0; j < i; j++) {
                EG_free(attrMapOut->attributeName[j]);
            }

            EG_free(attrMapOut->attributeName);
            EG_free(attrMapOut->attributeIndex);

            attrMapOut->attributeName = NULL;
            attrMapOut->attributeIndex = NULL;
            return EGADS_MALLOC;
        }

        sprintf(attrMapOut->attributeName[i], "%s", keyWord);
    }


    return CAPS_SUCCESS;
}

// Retrieve the string following a generic tag (given by attributeKey)
int retrieve_stringAttr(ego geomEntity, char *attributeKey, const char **string) {

    int status; // Status return

    int atype, alen; // EGADS return variables
    const int    *ints;
    const double *reals;


    status = EG_attributeRet(geomEntity, (const char *) attributeKey, &atype, &alen, &ints, &reals, string);
    if (status != EGADS_SUCCESS) return status;

    if (atype != ATTRSTRING) {
        printf ("Error: Attribute %s should be followed by a single string\n", attributeKey);
        return EGADS_ATTRERR;
    }

    return CAPS_SUCCESS;
}

// Retrieve the string following a capsGroup tag
int retrieve_CAPSGroupAttr(ego geomEntity, const char **string) {

    int status; // Status return

    char *attributeKey = "capsGroup";

    status = retrieve_stringAttr(geomEntity, attributeKey, string);
    return status;
}

// Retrieve the string following a capsConstraint tag
int retrieve_CAPSConstraintAttr(ego geomEntity, const char **string) {

    int status; // Status return

    char *attributeKey = "capsConstraint";

    status = retrieve_stringAttr(geomEntity, attributeKey, string);
    return status;
}

// Retrieve the string following a capsLoad tag
int retrieve_CAPSLoadAttr(ego geomEntity, const char **string) {

    int status; // Status return

    char *attributeKey = "capsLoad";

    status = retrieve_stringAttr(geomEntity, attributeKey, string);
    return status;
}

// Retrieve the string following a capsBound tag
int retrieve_CAPSBoundAttr(ego geomEntity, const char **string) {

    int status; // Status return

    char *attributeKey = "capsBound";

    status = retrieve_stringAttr(geomEntity, attributeKey, string);
    return status;
}

// Retrieve the string following a capsIgnore tag
int retrieve_CAPSIgnoreAttr(ego geomEntity, const char **string) {

    int status;
    char *attributeKey = "capsIgnore";

    status = retrieve_stringAttr(geomEntity, attributeKey, string);
    return status;
}

// Retrieve the string following a capsConnect tag
int retrieve_CAPSConnectAttr(ego geomEntity, const char **string) {

    int status;
    char *attributeKey = "capsConnect";

    status = retrieve_stringAttr(geomEntity, attributeKey, string);
    return status;
}


// Retrieve the string following a capsConnectLink tag
int retrieve_CAPSConnectLinkAttr(ego geomEntity, const char **string) {

    int status;
    char *attributeKey = "capsConnectLink";

    status = retrieve_stringAttr(geomEntity, attributeKey, string);
    return status;
}

// Retrieve the value following a capsIntent
int retrieve_CAPSIntent(ego geomEntity, int *intent) {

    int status; // Status return

    int atype, alen; // EGADS return variables
    const int    *ints;
    const double *reals;
    const char *string;

    const char *attributeKey = "capsIntent";

    *intent = (int) 0;
    status = EG_attributeRet(geomEntity, attributeKey, &atype, &alen, &ints, &reals, &string);
    if (status != EGADS_SUCCESS) return status;

    if (atype != ATTRREAL && alen == 1) {
        printf ("Error: Attribute %s should be followed by an integer value\n", attributeKey);
        return EGADS_ATTRERR;
    }

    *intent = (int) reals[0];

    return CAPS_SUCCESS;
}

/*
// Retrieve the string following a CoordSystem tag
int retrieve_CoordSystemAttr(ego geomEntity, const char **name) {

    int status; // Status return
    int i; // Indexing

    int atype, alen; // EGADS return variables
    const int    *ints;
    const double *reals;
    const char *string;

    int numAttr;

    status = EG_attributeNum(geomEntity, &numAttr);
	if (status != EGADS_SUCCESS) return EGADS_NOTFOUND;

	for (i = 0; i < numAttr; i++) {
		status = EG_attributeGet(geomEntity, i+1, name, &atype, &alen, &ints, &reals, &string);

		if (atype == ATTRCSYS) return CAPS_SUCCESS;
	}

	return EGADS_NOTFOUND;
}
*/

// Create a mapping between unique, generic (specified via mapName) attribute names and an index value
int create_GenericAttrToIndexMap(int numBody, ego bodies[], int attrLevelIn, char *mapName, mapAttrToIndexStruct *attrMap) {

    // In:
    // 		numBody   = Number of incoming bodies
    //		bodies    = Array of ego bodies
    // 		attrLevel = Level of depth to traverse the body:  0 - search just body attributes
    // 														  1 - search the body and all the faces
    //														  2 - search the body, faces, and all the edges
    // 														 >2 - search the body, faces, edges, and all the nodes
    // Out:
    // 		attrMap = A filled mapAttrToIndex structure

    int attr, body, face, edge, node; // Indexing variables

    int status; // Function return integer

    int stringLength;
    const char *groupName; // Attribute name

    int numFace = 0, numEdge = 0, numNode = 0; // Number of egos
    ego *faces = NULL, *edges = NULL, *nodes = NULL; // Geometry

    int attrLevel;

    // EGADS function returns
    ego  eref, *echilds;
    int  oclass, nchild, *senses, bodySubType = 0; // Body classification
    double data[4];

    // Destroy attrMap in case it is already allocated - this implies that is must have at least been initiated
    status =  destroy_mapAttrToIndexStruct(attrMap);
    if (status != CAPS_SUCCESS) return status;

    if (mapName == NULL) return CAPS_NULLVALUE;

    printf("Mapping %s attributes ................\n", mapName);

    stringLength = strlen(mapName);
    attrMap->mapName = (char *) EG_alloc((stringLength+1)*sizeof(char));
    if (attrMap->mapName == NULL) return EGADS_MALLOC;

    memcpy(attrMap->mapName,
            mapName,
            stringLength*sizeof(char));
    attrMap->mapName[stringLength] = '\0';

    // Search through bodies
    for (body = 0; body < numBody; body++) {

        attrLevel = attrLevelIn;

        status = EG_getTopology(bodies[body], &eref, &oclass, &bodySubType, data, &nchild, &echilds, &senses);
        if (status < EGADS_SUCCESS) goto cleanup;
        if (oclass == NODE) attrLevel = 0; // If we have a node body - change attrLevel to just the body

        // Get groupName following capsBound
        status = retrieve_stringAttr(bodies[body], mapName, &groupName);
        if (status != EGADS_SUCCESS && status != EGADS_NOTFOUND) goto cleanup;

        // Set attribute map
        if (status == CAPS_SUCCESS) {
            status = increment_mapAttrToIndexStruct(attrMap, groupName);
            if (status != CAPS_SUCCESS && status != EGADS_EXISTS) goto cleanup;
        }

        // Search through faces
        if (attrLevel > 0) {

            // Determine the number of faces
            status = EG_getBodyTopos(bodies[body], NULL, FACE, &numFace, &faces);
            if (status != EGADS_SUCCESS) goto cleanup;

            // Loop through faces
            for (face = 0; face < numFace; face++) {

                // Get groupName following capsBound
                status = retrieve_stringAttr(faces[face], mapName, &groupName);
                if (status != EGADS_SUCCESS && status != EGADS_NOTFOUND) goto cleanup;

                // Set attribute map
                if (status == CAPS_SUCCESS) {
                    status = increment_mapAttrToIndexStruct(attrMap, groupName);
                    if (status != CAPS_SUCCESS && status != EGADS_EXISTS) goto cleanup;
                }
            }
        } // End face loop

        // Search through edges
        if (attrLevel > 1) {
            status = EG_getBodyTopos(bodies[body], NULL, EDGE, &numEdge, &edges);
            if (status != EGADS_SUCCESS) goto cleanup;

            // Loop through edges
            for (edge = 0; edge < numEdge; edge++) {

                // Get groupName following capsBound
                status = retrieve_stringAttr(edges[edge], mapName, &groupName);
                if (status != EGADS_SUCCESS && status != EGADS_NOTFOUND) goto cleanup;

                // Set attribute map
                if (status == CAPS_SUCCESS) {
                    status = increment_mapAttrToIndexStruct(attrMap, groupName);
                    if (status != CAPS_SUCCESS && status != EGADS_EXISTS) goto cleanup;
                }
            }
        } // End edge loop

        // Search through nodes
        if (attrLevel > 2) {
            status = EG_getBodyTopos(bodies[body], NULL, NODE, &numNode, &nodes);
            if (status != EGADS_SUCCESS) return status;

            // Loop through nodes
            for (node = 0; node < numNode; node++) {

                // Get groupName following capsBound
                status = retrieve_stringAttr(nodes[node], mapName, &groupName);
                if (status != EGADS_SUCCESS && status != EGADS_NOTFOUND) goto cleanup;

                // Set attribute map
                if (status == CAPS_SUCCESS) {
                    status = increment_mapAttrToIndexStruct(attrMap, groupName);
                    if (status != CAPS_SUCCESS && status != EGADS_EXISTS) goto cleanup;
                }
            }
        } // End node loop


        if (faces != NULL) EG_free(faces);
        if (edges != NULL) EG_free(edges);
        if (nodes != NULL) EG_free(nodes);

        faces = NULL;
        edges = NULL;
        nodes = NULL;

    } // End body loop

    printf("\tNumber of unique %s attributes = %d\n", mapName, attrMap->numAttribute);
    for (attr = 0; attr < attrMap->numAttribute; attr++) {
        printf("\tName = %s, index = %d\n", attrMap->attributeName[attr], attrMap->attributeIndex[attr]);
    }

    status = CAPS_SUCCESS;

    goto cleanup;

    cleanup:

        if (faces != NULL) EG_free(faces);
        if (edges != NULL) EG_free(edges);
        if (nodes != NULL) EG_free(nodes);

        return status;
}

// Create a mapping between unique capsGroup attribute names and an index value
int create_CAPSGroupAttrToIndexMap(int numBody, ego bodies[], int attrLevel, mapAttrToIndexStruct *attrMap) {

    // In:
    //      numBody   = Number of incoming bodies
    //      bodies    = Array of ego bodies
    //      attrLevel = Level of depth to traverse the body:  0 - search just body attributes
    //                                                        1 - search the body and all the faces
    //                                                        2 - search the body, faces, and all the edges
    //                                                       >2 - search the body, faces, edges, and all the nodes
    // Out:
    // 		attrMap = A filled mapAttrToIndex structure

    int status; // Function return integer

    char *mapName = "capsGroup";

    status = create_GenericAttrToIndexMap(numBody, bodies, attrLevel, mapName, attrMap);

    return status;
}

// Create a mapping between unique CoordSystem attribute names and an index value
int create_CoordSystemAttrToIndexMap(int numBody, ego bodies[], int attrLevelIn, mapAttrToIndexStruct *attrMap) {

    // In:
    //      numBody   = Number of incoming bodies
    //      bodies    = Array of ego bodies
    //      attrLevel = Level of depth to traverse the body:  0 - search just body attributes
    //                                                        1 - search the body and all the faces
    //                                                        2 - search the body, faces, and all the edges
    //                                                       >2 - search the body, faces, edges, and all the nodes
    // Out:
    //      attrMap = A filled mapAttrToIndex structure

    int attr, body, face, edge, node; // Indexing variables

    int status; // Function return integer

    const char *csysName; // Attribute name

    int numFace = 0, numEdge = 0, numNode = 0; // Number of egos
    ego *faces = NULL, *edges = NULL, *nodes = NULL; // Geometry

    int attrLevel;

    int atype, alen; // EGADS return variables
    const int    *ints;
    const double *reals;
    const char *string;
    int numAttr;

    ego  eref, *echilds;
    int  oclass, nchild, *senses, bodySubType = 0; // Body classification
    double data[4];


    // Destroy attrMap in case it is already allocated - this implies that is must have at least been initiated
    status =  destroy_mapAttrToIndexStruct(attrMap);
    if (status != CAPS_SUCCESS) return status;

    printf("Mapping Csys attributes ................\n");

    // Search through bodies
    for (body = 0; body < numBody; body++) {

        attrLevel = attrLevelIn;

        status = EG_getTopology(bodies[body], &eref, &oclass, &bodySubType, data, &nchild, &echilds, &senses);
        if (status < EGADS_SUCCESS) goto cleanup;
        if (bodySubType == NOMTYPE) attrLevel = 0; // If we have a node body - change attrLevel to just the body

        // Get csysName following CoordSystem
        status = EG_attributeNum(bodies[body], &numAttr);
        if (status != EGADS_SUCCESS) return EGADS_NOTFOUND;

        for (attr = 0; attr < numAttr; attr++) {

            status = EG_attributeGet(bodies[body], attr+1, &csysName, &atype, &alen, &ints, &reals, &string);
            if (status != EGADS_SUCCESS) goto cleanup;

            if (atype != ATTRCSYS) continue;

            // Set attribute map
            status = increment_mapAttrToIndexStruct(attrMap, csysName);
            if (status != CAPS_SUCCESS && status != EGADS_EXISTS) goto cleanup;
        }

        // Search through faces
        if (attrLevel > 0) {

            // Determine the number of faces
            status = EG_getBodyTopos(bodies[body], NULL, FACE, &numFace, &faces);
            if (status != EGADS_SUCCESS) goto cleanup;

            // Loop through faces
            for (face = 0; face < numFace; face++) {

                status = EG_attributeNum(faces[face], &numAttr);
                if (status != EGADS_SUCCESS) return EGADS_NOTFOUND;

                for (attr = 0; attr < numAttr; attr++) {

                    status = EG_attributeGet(faces[face], attr+1, &csysName, &atype, &alen, &ints, &reals, &string);
                    if (status != EGADS_SUCCESS) goto cleanup;

                    if (atype != ATTRCSYS) continue;

                    // Set attribute map
                    status = increment_mapAttrToIndexStruct(attrMap, csysName);
                    if (status != CAPS_SUCCESS && status != EGADS_EXISTS) goto cleanup;
                }
            }
        } // End face loop

        // Search through edges
        if (attrLevel > 1) {
            status = EG_getBodyTopos(bodies[body], NULL, EDGE, &numEdge, &edges);
            if (status != EGADS_SUCCESS) goto cleanup;

            // Loop through edges
            for (edge = 0; edge < numEdge; edge++) {

                status = EG_attributeNum(edges[edge], &numAttr);
                if (status != EGADS_SUCCESS) return EGADS_NOTFOUND;

                for (attr = 0; attr < numAttr; attr++) {

                    status = EG_attributeGet(edges[edge], attr+1, &csysName, &atype, &alen, &ints, &reals, &string);
                    if (status != EGADS_SUCCESS) goto cleanup;

                    if (atype != ATTRCSYS) continue;

                    // Set attribute map
                    status = increment_mapAttrToIndexStruct(attrMap, csysName);
                    if (status != CAPS_SUCCESS && status != EGADS_EXISTS) goto cleanup;
                }
            }
        } // End edge loop

        // Search through nodes
        if (attrLevel > 2) {
            status = EG_getBodyTopos(bodies[body], NULL, NODE, &numNode, &nodes);
            if (status != EGADS_SUCCESS) return status;

            // Loop through nodes
            for (node = 0; node < numNode; node++) {

                status = EG_attributeNum(nodes[node], &numAttr);
                if (status != EGADS_SUCCESS) return EGADS_NOTFOUND;

                for (attr = 0; attr < numAttr; attr++) {

                    status = EG_attributeGet(nodes[node], attr+1, &csysName, &atype, &alen, &ints, &reals, &string);
                    if (status != EGADS_SUCCESS) goto cleanup;

                    if (atype != ATTRCSYS) continue;

                    // Set attribute map
                    status = increment_mapAttrToIndexStruct(attrMap, csysName);
                    if (status != CAPS_SUCCESS && status != EGADS_EXISTS) goto cleanup;
                }
            }
        } // End node loop


        if (faces != NULL) EG_free(faces);
        if (edges != NULL) EG_free(edges);
        if (nodes != NULL) EG_free(nodes);

        faces = NULL;
        edges = NULL;
        nodes = NULL;

    } // End body loop

    printf("\tNumber of unique Csys attributes = %d\n", attrMap->numAttribute);
    for (attr = 0; attr < attrMap->numAttribute; attr++) {
        printf("\tName = %s, index = %d\n", attrMap->attributeName[attr], attrMap->attributeIndex[attr]);
    }

    status = CAPS_SUCCESS;

    goto cleanup;

    cleanup:
        if (faces != NULL) EG_free(faces);
        if (edges != NULL) EG_free(edges);
        if (nodes != NULL) EG_free(nodes);

        return status;
}

// Create a mapping between unique capsConstraint attribute names and an index value
int create_CAPSConstraintAttrToIndexMap(int numBody, ego bodies[], int attrLevel, mapAttrToIndexStruct *attrMap) {

    // In:
    //      numBody   = Number of incoming bodies
    //      bodies    = Array of ego bodies
    //      attrLevel = Level of depth to traverse the body:  0 - search just body attributes
    //                                                        1 - search the body and all the faces
    //                                                        2 - search the body, faces, and all the edges
    //                                                       >2 - search the body, faces, edges, and all the nodes
    // Out:
    //      attrMap = A filled mapAttrToIndex structure

    int status; // Function return integer

    char *mapName = "capsConstraint";

    status = create_GenericAttrToIndexMap(numBody, bodies, attrLevel, mapName, attrMap);

    return status;
}

// Create a mapping between unique capsLoad attribute names and an index value
int create_CAPSLoadAttrToIndexMap(int numBody, ego bodies[], int attrLevel, mapAttrToIndexStruct *attrMap) {

    // In:
    //      numBody   = Number of incoming bodies
    //      bodies    = Array of ego bodies
    //      attrLevel = Level of depth to traverse the body:  0 - search just body attributes
    //                                                        1 - search the body and all the faces
    //                                                        2 - search the body, faces, and all the edges
    //                                                       >2 - search the body, faces, edges, and all the nodes
    // Out:
    //      attrMap = A filled mapAttrToIndex structure

    int status; // Function return integer

    char *mapName = "capsLoad";

    status = create_GenericAttrToIndexMap(numBody, bodies, attrLevel, mapName, attrMap);

    return status;
}

// Create a mapping between unique capsBound attribute names and an index value
int create_CAPSBoundAttrToIndexMap(int numBody, ego bodies[], int attrLevel, mapAttrToIndexStruct *attrMap) {

    // In:
    //      numBody   = Number of incoming bodies
    //      bodies    = Array of ego bodies
    //      attrLevel = Level of depth to traverse the body:  0 - search just body attributes
    //                                                        1 - search the body and all the faces
    //                                                        2 - search the body, faces, and all the edges
    //                                                       >2 - search the body, faces, edges, and all the nodes
    // Out:
    //      attrMap = A filled mapAttrToIndex structure

    int status; // Function return integer

    char *mapName = "capsBound";

    status = create_GenericAttrToIndexMap(numBody, bodies, attrLevel, mapName, attrMap);

    return status;
}

// Create a mapping between unique capsConnect attribute names and an index value
int create_CAPSConnectAttrToIndexMap(int numBody, ego bodies[], int attrLevel, mapAttrToIndexStruct *attrMap) {

    // In:
    //      numBody   = Number of incoming bodies
    //      bodies    = Array of ego bodies
    //      attrLevel = Level of depth to traverse the body:  0 - search just body attributes
    //                                                        1 - search the body and all the faces
    //                                                        2 - search the body, faces, and all the edges
    //                                                       >2 - search the body, faces, edges, and all the nodes
    // Out:
    //      attrMap = A filled mapAttrToIndex structure

    int status; // Function return integer

    char *mapName = "capsConnect";

    status = create_GenericAttrToIndexMap(numBody, bodies, attrLevel, mapName, attrMap);

    return status;
}

// Check capsLength consistency in the bodies and return the value. No check is done to make sure ALL bodies have
// a capsLength, just that if present it is consistent.
int check_CAPSLength(int numBody, ego bodies[], const char **lengthString) {

    int i;
    int status;

    int found = (int) false;

    // EGADS return values
    const char   *string;

    *lengthString = NULL;

    for (i =0; i < numBody; i++) {

        status = retrieve_stringAttr(bodies[i], "capsLength", &string);
        if (status != CAPS_SUCCESS)	continue;

        // Save the string
        if (found == (int) false) {
            found = true;
            *lengthString = string;

        } else { // Compare the strings

            if (strcasecmp(*lengthString, string) != 0) {
                printf("Inconsistent length units on bodies, capsLength %s found on one body, "
                        "while %s found on another\n", *lengthString, string);
                return CAPS_MISMATCH;
            }
        }
    }

    if (found == (int) false) return CAPS_NOTFOUND;

    return CAPS_SUCCESS;
}


// Check capsIntent consistency in the bodies and return the value.
// All bodies must either have or not have a capsInent.
int check_CAPSIntent(int numBody, ego bodies[], int *intent) {

    int i;
    int status;

    int currentIntent = 0;

    *intent = CAPSMAGIC;

    // Check bodies
    for (i = 0; i < numBody; i++) {

        status = retrieve_CAPSIntent(bodies[i], &currentIntent);
        if (status != EGADS_SUCCESS) continue;

        if (*intent == CAPSMAGIC) {
            *intent = currentIntent;
        } else {

            if (*intent != currentIntent) {
                printf("All bodies don't have the same capsIntent value - one body "
                        "found with = %d, while another has %d!\n", *intent, currentIntent);

                *intent = CAPSMAGIC;
                return CAPS_MISMATCH;
            }
        }
    }

    // If we found at least one intent, re-check bodies because the first bodies checked might not have had an intent
    if (*intent != CAPSMAGIC) {
        for (i = 0; i < numBody; i++) {

            status = retrieve_CAPSIntent(bodies[i], &currentIntent);

            if (status != CAPS_SUCCESS) {
                printf("A capsIntent value of %d was found, all bodies must have this value as well!\n", *intent);
                *intent = CAPSMAGIC;
                return status;
            }
        }
    }

    return CAPS_SUCCESS;
}
