// Miscellaneous related utility functions - Written by Dr. Ryan Durscher AFRL/RQVC

#include "meshTypes.h"  // Bring in mesh structures
#include "capsTypes.h"  // Bring in CAPS types
#include "cfdTypes.h"   // Bring in cfd structures
#include "miscTypes.h"  // Bring in miscellaneous structures

#ifdef __cplusplus
extern "C" {
#endif

// Does a file exist?
int file_exist(char *file);

// Simple json dictionary parser - currently doesn't support nested arrays for keyValue
int search_jsonDictionary(char *stringToSearch, char *keyWord, char **keyValue);

// Convert a single string to a prog_argv array
int string_toProgArgs(char *meshInputString, int *prog_argc, char ***prog_argv);

// Convert a string to double
int string_toDouble(char *string, double *number);

// Convert a string to double array
int string_toDoubleArray(char *string, int arraySize, double numberArray[]);

// Convert a string to an array of doubles
int string_toDoubleDynamicArray(char *stringToSearch, int *arraySize, double *numberArray[]);

// Convert a string to an array of strings
int string_toStringDynamicArray(char *stringToSearch, int *arraySize, char **stringArray[]);

// Convert a string to boolean
int string_toBoolean(char *string, int *number);

// Convert a string to integer
int string_toInteger(char *string, int *number);

// Convert a string to an array of integer
int string_toIntegerDynamicArray(char *stringToSearch, int *arraySize, int *numberArray[]);

// Free an array strings
int string_freeArray(int numString, char **strings[]);

// Remove quotation marks around a string - Returning char * should be free'd after use
char * string_removeQuotation(char *string);

// Force a string to upper case
void string_toUpperCase ( char *sPtr );

// The max x,y,z coordinates where P(3*i + 0) = x_i, P(3*i + 1) = y_i, and P(3*i + 2) = z_i
void maxCoords(int sizeP, double *P, double *x, double *y, double *z);

// The min x,y,z coordinates where P(3*i + 0) = x_i, P(3*i + 1) = y_i, and P(3*i + 2) = z_i
void minCoords(int sizeP, double *P, double *x, double *y, double *z);

// Return the endianness of the machine we are working on - 0 = little, 1 = big
int get_MachineENDIANNESS(void);

#ifdef WIN32
int getline(char **linep, size_t *linecapp, FILE *stream);
#endif

// Return the max value of two doubles
double max_DoubleVal(double x1, double x2);

// Return the max value of two doubles
double min_DoubleVal(double x1, double x2);

// Cross product axb = c
void cross_DoubleVal(double a[], double b[], double c[]);

// Dot product a*b = c
double dot_DoubleVal(double a[], double b[]);

// Convert an integer to a string of a given field width and justifacation - Returning char * should be free'd after use
char * convert_integerToString(int integerVal, int fieldWidth, int leftOrRight);

// Convert an double to a string (scientific notation is used depending on the fieldwidth and value) of a given field width
//      - Returning char * should be free'd after use
char * convert_doubleToString(double doubleVal, int fieldWidth, int leftOrRight);

// Initiate (0 out all values and NULL all pointers) an attribute map in the mapAttrToIndexStruct structure format
int initiate_mapAttrToIndexStruct(mapAttrToIndexStruct *attrMap);

// Destroy (0 out all values and NULL all pointers) an attribute map in the mapAttrToIndexStruct structure format
int destroy_mapAttrToIndexStruct(mapAttrToIndexStruct *attrMap);

// Make a copy of attribute map (attrMapIn)
int copy_mapAttrToIndexStruct(mapAttrToIndexStruct *attrMapIn, mapAttrToIndexStruct *attrMapOut);

// Search a mapAttrToIndex structure for a given keyword and set/return the corresponding index
int get_mapAttrToIndexIndex(mapAttrToIndexStruct *attrMap, const char *keyWord, int *index);

// Search a mapAttrToIndex structure for a given index and return the corresponding keyword
int get_mapAttrToIndexKeyword(mapAttrToIndexStruct *attrMap, int index, char **keyWord);

// Set the index of a given keyword in a mapAttrToIndex structure
int set_mapAttrToIndexStruct(mapAttrToIndexStruct *attrMap, const char *keyWord, int index);

// Increment a mapAttrToIndex structure with the given keyword and set the default index (= numAttribute)
int increment_mapAttrToIndexStruct(mapAttrToIndexStruct *attrMap, const char *keyWord);

// Retrieve the string following a generic tag (given by attributeKey)
int retrieve_stringAttr(ego geomEntity, char *attributeKey, const char **string);

// Retrieve the string following a capsGroup tag
int retrieve_CAPSGroupAttr(ego geomEntity, const char **string);

// Retrieve the string following a capsConstraint tag
int retrieve_CAPSConstraintAttr(ego geomEntity, const char **string);

// Retrieve the string following a capsLoad tag
int retrieve_CAPSLoadAttr(ego geomEntity, const char **string);

// Retrieve the string following a capsBound tag
int retrieve_CAPSBoundAttr(ego geomEntity, const char **string);

// Retrieve the string following a capsIgnore tag
int retrieve_CAPSIgnoreAttr(ego geomEntity, const char **string);

// Retrieve the string following a capsConnect tag
int retrieve_CAPSConnectAttr(ego geomEntity, const char **string);

// Retrieve the string following a capsConnectLink tag
int retrieve_CAPSConnectLinkAttr(ego geomEntity, const char **string);

// Retrieve the value following a capsIntent
int retrieve_CAPSIntent(ego geomEntity, int *intent);

/*
// Retrieve the string following a CoordSystem tag
int retrieve_CoordSystemAttr(ego geomEntity, const char **string);
*/

// Create a mapping between unique, generic (specified via mapName) attribute names and an index value
int create_GenericAttrToIndexMap(int numBody, ego bodies[], int attrLevel, char *mapName, mapAttrToIndexStruct *attrMap);

// Create a mapping between unique capsGroup attribute names and an index value
int create_CAPSGroupAttrToIndexMap(int numBody, ego bodies[], int attrLevel, mapAttrToIndexStruct *attrMap);

// Create a mapping between unique CoordSystem attribute names and an index value
int create_CoordSystemAttrToIndexMap(int numBody, ego bodies[], int attrLevel, mapAttrToIndexStruct *attrMap);

// Create a mapping between unique capsConstraint attribute names and an index value
int create_CAPSConstraintAttrToIndexMap(int numBody, ego bodies[], int attrLevel, mapAttrToIndexStruct *attrMap);

// Create a mapping between unique capsLoad attribute names and an index value
int create_CAPSLoadAttrToIndexMap(int numBody, ego bodies[], int attrLevel, mapAttrToIndexStruct *attrMap);

// Create a mapping between unique capsBound attribute names and an index value
int create_CAPSBoundAttrToIndexMap(int numBody, ego bodies[], int attrLevel, mapAttrToIndexStruct *attrMap);

// Create a mapping between unique capsConnect attribute names and an index value
int create_CAPSConnectAttrToIndexMap(int numBody, ego bodies[], int attrLevel, mapAttrToIndexStruct *attrMap);

// Check capsLength consistency in the bodies and return the value. No check is done to make sure ALL bodies have
// a capsLength, just that if present it is consistent.
int check_CAPSLength(int numBody, ego bodies[], const char **lengthString);

// Check capsIntent consistency in the bodies and return the value.
// All bodies must either have or not have a capsInent.
int check_CAPSIntent(int numBody, ego bodies[], int *intent);

#ifdef __cplusplus
}
#endif
