// Vortex lattice analysis related utility functions - Written by Dr. Ryan Durscher AFRL/RQVC

#include <string.h>
#include <math.h>
#include "capsTypes.h"  // Bring in CAPS types
#include "vlmTypes.h"   // Bring in Vortex Lattice Method structures
#include "vlmUtils.h"
#include "miscUtils.h"  // Bring in miscellaneous utilities

#ifdef WIN32
#define strcasecmp  stricmp
#endif

// Fill vlmSurface in a vlmSurfaceStruct format with vortex lattice information
// from an incoming surfaceTuple
int get_vlmSurface(int numTuple,
                   capsTuple surfaceTuple[],
                   mapAttrToIndexStruct *attrMap,
                   int *numVLMSurface,
                   vlmSurfaceStruct *vlmSurface[]) {

    /*! \page vlmSurface Vortex Lattice Surface
     * Structure for the Vortex Lattice Surface tuple  = ("Name of Surface", "Value").
     * "Name of surface defines the name of the surface in which the data should be applied.
     *  The "Value" can either be a JSON String dictionary (see Section \ref jsonStringVLMSurface)
     *  or a single string keyword string (see Section \ref keyStringVLMSurface).
     */

    int status; //Function return

    int i, groupIndex, attrIndex; // Indexing

    char *keyValue = NULL; // Key values from tuple searches
    char *keyWord = NULL; // Key words to find in the tuples

    int numGroupName = 0;
    char **groupName = NULL;

    int stringLength = 0;

    // Clear out vlmSurface is it has already been allocated
    if (*vlmSurface != NULL) {
        for (i = 0; i < (*numVLMSurface); i++) {
            status = destroy_vlmSurfaceStruct(&(*vlmSurface)[i]);
            if (status != CAPS_SUCCESS) printf("destroy_vlmSurfaceStruct status = %d\n", status);
        }
    }

    printf("Getting vortex lattice surface data\n");

    if (numTuple <= 0){
        printf("\tNumber of VLM Surface tuples is %d\n", numTuple);
        return CAPS_NOTFOUND;
    }

    *numVLMSurface = numTuple;
    *vlmSurface = (vlmSurfaceStruct *) EG_alloc((*numVLMSurface) * sizeof(vlmSurfaceStruct));
    if (*vlmSurface == NULL) {
        *numVLMSurface = 0;
        return EGADS_MALLOC;
    }

    // Initiate vlmSurfaces
    for (i = 0; i < (*numVLMSurface); i++) {
        status = initiate_vlmSurfaceStruct(&(*vlmSurface)[i]);
        if (status != CAPS_SUCCESS) goto cleanup;
    }

    for (i = 0; i < (*numVLMSurface); i++) {

        printf("\tVLM surface name - %s\n", surfaceTuple[i].name);

        // Copy surface name
        stringLength = strlen(surfaceTuple[i].name);
        (*vlmSurface)[i].name = (char *) EG_alloc((stringLength + 1)*sizeof(char));
        if ((*vlmSurface)[i].name == NULL) return EGADS_MALLOC;

        memcpy((*vlmSurface)[i].name,
                surfaceTuple[i].name,
                stringLength);
        (*vlmSurface)[i].name[stringLength] = '\0';

        // Do we have a json string?
        if (strncmp(surfaceTuple[i].value, "{", 1) == 0) {

            //printf("JSON String - %s\n",surfaceTuple[i].value);

            /*! \page vlmSurface
             * \section jsonStringVLMSurface JSON String Dictionary
             *
             * If "Value" is a JSON string dictionary (eg. "Value" = {"numChord": 5, "spaceChord": 1.0, "numSpan": 10, "spaceSpan": 0.5}) the following keywords ( = default values) may be used:
             *
             * \if (AVL)
             * <ul>
             *  <li> <B>groupName = "(no default)"</B> </li> <br>
             *  Single or list of <em>capsGroup</em> names used to define the surface (e.g. "Name1" or ["Name1","Name2",...].
             *  If no groupName variable is provided an attempted will be made to use the tuple name instead;
             * </ul>
             * \else
             * <ul>
             *  <li> <B>groupName = "(no default)"</B> </li> <br>
             *  Single or list of <em>capsGroup</em> names used to define the surface (e.g. "Name1" or ["Name1","Name2",...].
             *  If no groupName variable is provided an attempted will be made to use the tuple name instead;
             * </ul>
             * \endif
             *
             */

            // Get surface variables
            keyWord = "groupName";
            status = search_jsonDictionary( surfaceTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toStringDynamicArray(keyValue, &numGroupName, &groupName);
                if (status != CAPS_SUCCESS) goto cleanup;

                if (keyValue != NULL) EG_free(keyValue);
                keyValue = NULL;

                // Determine how many capsGroups go into making the surface
                for (groupIndex = 0; groupIndex < numGroupName; groupIndex++) {

                    status = get_mapAttrToIndexIndex(attrMap, (const char *) groupName[groupIndex], &attrIndex);

                    if (status == CAPS_NOTFOUND) {
                        printf("\tgroupName name %s not found in attribute map of capsGroups!!!!\n", groupName[groupIndex]);
                        continue;

                    } else if (status != CAPS_SUCCESS) goto cleanup;

                    (*vlmSurface)[i].numAttr += 1;
                    if ((*vlmSurface)[i].numAttr == 1) {
                        (*vlmSurface)[i].attrIndex = (int *) EG_alloc(1*sizeof(int));

                    } else{
                        (*vlmSurface)[i].attrIndex = (int *) EG_reall((*vlmSurface)[i].attrIndex,
                                (*vlmSurface)[i].numAttr*sizeof(int));
                    }

                    if ((*vlmSurface)[i].attrIndex == NULL) {
                        status = EGADS_MALLOC;
                        goto cleanup;
                    }

                    (*vlmSurface)[i].attrIndex[(*vlmSurface)[i].numAttr-1] = attrIndex;
                }

                status = string_freeArray(numGroupName, &groupName);
                if (status != CAPS_SUCCESS) goto cleanup;
                groupName = NULL;
                numGroupName = 0;

            } else {
                printf("\tNo \"groupName\" variable provided or no matches found, going to use tuple name\n");
            }

            if ((*vlmSurface)[i].numAttr == 0) {

                status = get_mapAttrToIndexIndex(attrMap, (const char *) (*vlmSurface)[i].name, &attrIndex);
                if (status == CAPS_NOTFOUND) {
                    printf("\tTuple name %s not found in attribute map of capsGroups!!!!\n", (*vlmSurface)[i].name);
                    goto cleanup;
                }

                (*vlmSurface)[i].numAttr += 1;
                if ((*vlmSurface)[i].numAttr == 1) {
                    (*vlmSurface)[i].attrIndex = (int *) EG_alloc(1*sizeof(int));
                } else{
                    (*vlmSurface)[i].attrIndex = (int *) EG_reall((*vlmSurface)[i].attrIndex,
                            (*vlmSurface)[i].numAttr*sizeof(int));
                }

                if ((*vlmSurface)[i].attrIndex == NULL) {
                    status = EGADS_MALLOC;
                    goto cleanup;
                }

                (*vlmSurface)[i].attrIndex[(*vlmSurface)[i].numAttr-1] = attrIndex;
            }

            if (keyValue != NULL) EG_free(keyValue);
            keyValue = NULL;

            /*! \page vlmSurface
             * \if (AVL)
             * <ul>
             * <li> <B>noKeyword = "(no default)"</B> </li> <br>
             *      "No" type. Options: NOWAKE, NOALBE, NOLOAD.
             * </ul>
             * \endif
             */

            keyWord = "noKeyword";
            status = search_jsonDictionary( surfaceTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                if      (strcasecmp(keyValue, "\"NOWAKE\"") == 0) (*vlmSurface)[i].nowake = (int) true;
                else if (strcasecmp(keyValue, "\"NOALBE\"") == 0) (*vlmSurface)[i].noalbe = (int) true;
                else if (strcasecmp(keyValue, "\"NOLOAD\"") == 0) (*vlmSurface)[i].noload = (int) true;
                else {

                    printf("\tUnrecognized \"%s\" specified (%s) for VLM Section tuple %s, current options are "
                            "\" NOWAKE, NOALBE, or  NOLOAD\"\n", keyWord, keyValue, surfaceTuple[i].name);
                }

                if (keyValue != NULL) EG_free(keyValue);
                keyValue = NULL;
            }

            /*! \page vlmSurface
             * <ul>
             * <li> <B>numChord = 10</B> </li> <br>
             *  The number of chordwise horseshoe vortices placed on the surface.
             * </ul>
             *
             */
            keyWord = "numChord";
            status = search_jsonDictionary( surfaceTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toInteger(keyValue, &(*vlmSurface)[i].Nchord);
                if (status != CAPS_SUCCESS) goto cleanup;

                if (keyValue != NULL) EG_free(keyValue);
                keyValue = NULL;
            }

            /*! \page vlmSurface
             * <ul>
             * <li> <B>spaceChord = 0.0</B> </li> <br>
             *  The chordwise vortex spacing parameter.
             * </ul>
             *
             */
            keyWord = "spaceChord";
            status = search_jsonDictionary( surfaceTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &(*vlmSurface)[i].Cspace);
                if (status != CAPS_SUCCESS) goto cleanup;

                if (keyValue != NULL) EG_free(keyValue);
                keyValue = NULL;
            }

            /*! \page vlmSurface
             * <ul>
             * <li> <B>numSpan = 10</B> </li> <br>
             *  The number of spanwise horseshoe vortices placed on the surface.
             * </ul>
             *
             */
            keyWord = "numSpan";
            status = search_jsonDictionary( surfaceTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toInteger(keyValue, &(*vlmSurface)[i].Nspan);
                if (status != CAPS_SUCCESS) goto cleanup;

                if (keyValue != NULL) EG_free(keyValue);
                keyValue = NULL;
            }

            /*! \page vlmSurface
             * <ul>
             * <li> <B>spaceSpan = 0.0</B> </li> <br>
             *  The spanwise vortex spacing parameter.
             * </ul>
             *
             */
            keyWord = "spaceSpan";
            status = search_jsonDictionary( surfaceTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &(*vlmSurface)[i].Sspace);
                if (status != CAPS_SUCCESS) goto cleanup;

                if (keyValue != NULL) EG_free(keyValue);
                keyValue = NULL;
            }

            /*! \page vlmSurface
             * <ul>
             * <li> <B>yMirror = False</B> </li> <br>
             *  Mirror surface about the y-direction.
             * </ul>
             *
             */
            keyWord = "yMirror";
            status = search_jsonDictionary( surfaceTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toBoolean(keyValue, &(*vlmSurface)[i].iYdup);
                if (status != CAPS_SUCCESS) goto cleanup;

                if (keyValue != NULL) EG_free(keyValue);
                keyValue = NULL;
            }
        } else {

            /*! \page vlmSurface
             * \section keyStringVLMSurface Single Value String
             *
             * If "Value" is a single string the following options maybe used:
             * - (NONE Currently)
             *
             */
            printf("\tNo current defaults for get_vlmSurface, tuple value must be a JSON string\n");
            status = CAPS_BADVALUE;
            goto cleanup;
        }
    }

    printf("\tDone getting vortex lattice surface data\n");

    status = CAPS_SUCCESS;
    goto cleanup;

    cleanup:
        if (status != CAPS_SUCCESS) printf("Error: Premature exit in get_vlmSurface, status = %d\n", status);

        if (keyValue != NULL) EG_free(keyValue);

        if (numGroupName != 0 && groupName != NULL){
            (void) string_freeArray(numGroupName, &groupName);
        }
        return status;
}


// Fill vlmControl in a vlmControlStruct format with vortex lattice information
// from an incoming controlTuple
int get_vlmControl(int numTuple,
                   capsTuple controlTuple[],
                   int *numVLMControl,
                   vlmControlStruct *vlmControl[]) {

    /*! \page vlmControl Vortex Lattice Control Surface
     * Structure for the Vortex Lattice Control Surface tuple  = ("Name of Control Surface", "Value").
     * "Name of control surface defines the name of the control surface in which the data should be applied.
     *  The "Value" must be a JSON String dictionary (see Section \ref jsonStringVLMSection).
     */

    int status; //Function return

    int i; // Indexing

    char *keyValue = NULL; // Key values from tuple searches
    char *keyWord = NULL; // Key words to find in the tuples

    int stringLength = 0;

    // Clear out vlmSurface is it has already been allocated
    if (*vlmControl != NULL) {
        for (i = 0; i < (*numVLMControl); i++) {
            status = destroy_vlmControlStruct(&(*vlmControl)[i]);
            if (status != CAPS_SUCCESS) printf("destroy_vlmControlStruct status = %d\n", status);
        }
    }

    printf("Getting vortex lattice control surface data\n");

    if (numTuple <= 0){
        printf("\tNumber of VLM Surface tuples is %d\n", numTuple);
        return CAPS_NOTFOUND;
    }

    *numVLMControl = numTuple;
    *vlmControl = (vlmControlStruct *) EG_alloc((*numVLMControl) * sizeof(vlmControlStruct));
    if (*vlmControl == NULL) {
        *numVLMControl = 0;
        return EGADS_MALLOC;
    }

    // Initiate vlmSurfaces
    for (i = 0; i < (*numVLMControl); i++) {
        status = initiate_vlmControlStruct(&(*vlmControl)[i]);
        if (status != CAPS_SUCCESS) goto cleanup;
    }

    for (i = 0; i < (*numVLMControl); i++) {

        printf("\tVLM control surface name - %s\n", controlTuple[i].name);

        // Copy surface name
        stringLength = strlen(controlTuple[i].name);
        (*vlmControl)[i].name = (char *) EG_alloc((stringLength + 1)*sizeof(char));
        if ((*vlmControl)[i].name == NULL) return EGADS_MALLOC;

        memcpy((*vlmControl)[i].name,
                controlTuple[i].name,
                stringLength);
        (*vlmControl)[i].name[stringLength] = '\0';

        // Do we have a json string?
        if (strncmp(controlTuple[i].value, "{", 1) == 0) {

            //printf("JSON String - %s\n",surfaceTuple[i].value);

            /*! \page vlmControl
             * \section jsonStringVLMSection JSON String Dictionary
             *
             * If "Value" is a JSON string dictionary (e.g. "Value" = {"deflectionAngle": 10.0}) the following keywords ( = default values) may be used:
             *
             */

            /*! \page vlmControl
             * \if (AVL)
             * <ul>
             * <li> <B>deflectionAngle = 0.0</B> </li> <br>
             *      Deflection angle of the control surface.
             * </ul>
             * \endif
             */
            keyWord = "deflectionAngle";
            status = search_jsonDictionary( controlTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &(*vlmControl)[i].deflectionAngle);
                if (status != CAPS_SUCCESS) goto cleanup;

                if (keyValue != NULL) EG_free(keyValue);
                keyValue = NULL;
            }

            /*
            ! \page vlmControl
             * \if (AVL)
             * <ul>
             * <li> <B>percentChord = 0.0</B> </li> <br>
             *      Percentage along the airfoil chord the control surface's hinge line resides.
             * </ul>
             * \endif

			keyWord = "percentChord";
			status = search_jsonDictionary( controlTuple[i].value, keyWord, &keyValue);
			if (status == CAPS_SUCCESS) {

				status = string_toDouble(keyValue, &(*vlmControl)[i].percentChord);
				if (status != CAPS_SUCCESS) goto cleanup;

				if (keyValue != NULL) EG_free(keyValue);
				keyValue = NULL;
			}
             */

            /*! \page vlmControl
             * \if (AVL)
             * <ul>
             * <li> <B>leOrTe = (no default) </B> </li> <br>
             *      Is the control surface a leading ( = 0) or trailing (> 0) edge effector? Overrides
             *      the assumed default value set by the geometry: If the percentage along
             *      the airfoil chord is < 50% a leading edge flap is assumed, while >= 50% indicates a
             *      trailing edge flap.
             * </ul>
             * \endif
             */
            keyWord = "leOrTe";
            status = search_jsonDictionary( controlTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                (*vlmControl)[i].leOrTeOverride = (int) true;
                status = string_toInteger(keyValue, &(*vlmControl)[i].leOrTe);
                if (status != CAPS_SUCCESS) goto cleanup;

                if (keyValue != NULL) EG_free(keyValue);
                keyValue = NULL;
            }

            /*! \page vlmControl
             * \if (AVL)
             * <ul>
             * <li> <B>controlGain = 1.0</B> </li> <br>
             *      Control deflection gain, units:  degrees deflection / control variable
             * </ul>
             * \endif
             */
            keyWord = "controlGain";
            status = search_jsonDictionary( controlTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDouble(keyValue, &(*vlmControl)[i].controlGain);
                if (status != CAPS_SUCCESS) goto cleanup;

                if (keyValue != NULL) EG_free(keyValue);
                keyValue = NULL;
            }

            /*! \page vlmControl
             * \if (AVL)
             * <ul>
             * <li> <B>hingeLine = [0.0 0.0 0.0]</B> </li> <br>
             *      Alternative vector giving hinge axis about which surface rotates
             * </ul>
             * \endif
             */
            keyWord = "hingeLine";
            status = search_jsonDictionary( controlTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toDoubleArray(keyValue, 3, (*vlmControl)[i].xyzHingeVec);
                if (status != CAPS_SUCCESS) goto cleanup;

                if (keyValue != NULL) EG_free(keyValue);
                keyValue = NULL;
            }

            /*! \page vlmControl
             * \if (AVL)
             * <ul>
             * <li> <B>deflectionDup = 0 </B> </li> <br>
             *      Sign of deflection for duplicated surface
             * </ul>
             * \endif
             */
            keyWord = "deflectionDup";
            status = search_jsonDictionary( controlTuple[i].value, keyWord, &keyValue);
            if (status == CAPS_SUCCESS) {

                status = string_toInteger(keyValue, &(*vlmControl)[i].deflectionDup);
                if (status != CAPS_SUCCESS) goto cleanup;

                if (keyValue != NULL) EG_free(keyValue);
                keyValue = NULL;
            }



        } else {

            /*! \page vlmControl
             * \section keyStringVLMControl Single Value String
             *
             * If "Value" is a single string, the following options maybe used:
             * - (NONE Currently)
             *
             */
            printf("\tNo current defaults for get_vlmControl, tuple value must be a JSON string\n");
            status = CAPS_BADVALUE;
            goto cleanup;
        }
    }

    printf("\tDone getting vortex lattice control surface data\n");

    status = CAPS_SUCCESS;
    goto cleanup;

    cleanup:
        if (status != CAPS_SUCCESS) printf("Error: Premature exit in get_vlmSurface, status = %d\n", status);

        if (keyValue != NULL) EG_free(keyValue);

        return status;
}


// Initiate (0 out all values and NULL all pointers) of a control in the vlmcontrol structure format
int initiate_vlmControlStruct(vlmControlStruct *control) {

    control->name = NULL; // Control surface name

    control->deflectionAngle = 0.0; // Deflection angle of the control surface
    control->controlGain = 1.0; //Control deflection gain, units:  degrees deflection / control variable

    control->percentChord = 0.0; // Percentage along chord

    control->xyzHinge[0] = 0.0; // xyz location of the hinge
    control->xyzHinge[1] = 0.0;
    control->xyzHinge[2] = 0.0;

    control->xyzHingeVec[0] = 0.0; // Vector of hinge line at xyzHinge
    control->xyzHingeVec[1] = 0.0;
    control->xyzHingeVec[2] = 0.0;


    control->leOrTeOverride = (int) false; // Does the user want to override the geometry set value?
    control->leOrTe = 0; // Leading = 0 or trailing > 0 edge control surface

    control->deflectionDup = 0; // Sign of deflection for duplicated surface

    return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) of a control in the vlmcontrol structure format
int destroy_vlmControlStruct(vlmControlStruct *control) {

    if (control->name != NULL) EG_free(control->name);
    control->name = NULL; // Control surface name

    control->deflectionAngle = 0.0; // Deflection angle of the control surface

    control->controlGain = 1.0; //Control deflection gain, units:  degrees deflection / control variable

    control->percentChord = 0.0; // Percentage along chord

    control->xyzHinge[0] = 0.0; // xyz location of the hinge
    control->xyzHinge[1] = 0.0;
    control->xyzHinge[2] = 0.0;

    control->xyzHingeVec[0] = 0.0; // Vector of hinge line at xyzHinge
    control->xyzHingeVec[1] = 0.0;
    control->xyzHingeVec[2] = 0.0;

    control->leOrTeOverride = (int) false; // Does the user want to override the geometry set value?
    control->leOrTe = 0; // Leading = 0 or trailing > 0 edge control surface

    control->deflectionDup = 0; // Sign of deflection for duplicated surface

    return CAPS_SUCCESS;
}


// Initiate (0 out all values and NULL all pointers) of a section in the vlmSection structure format
int initiate_vlmSectionStruct(vlmSectionStruct *section) {

    section->name = NULL;

    section->bodyIndex = 0;
    section->sectionIndex = 0;

    section->xyzLE[0] = 0.0;
    section->xyzLE[1] = 0.0;
    section->xyzLE[2] = 0.0;
    section->nodeIndexLE = 0;

    section->xyzTE[0] = 0.0;
    section->xyzTE[1] = 0.0;
    section->xyzTE[2] = 0.0;
    section->nodeIndexTE = 0;

    section->chord = 0.0;

    section->numControl = 0;
    section->vlmControl = NULL;

    return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) of a section in the vlmSection structure format
int destroy_vlmSectionStruct(vlmSectionStruct *section) {

    int status; // Function return status
    int i; // Indexing

    if (section->name != NULL) EG_free(section->name);
    section->name = NULL;

    section->bodyIndex = 0;
    section->sectionIndex = 0;

    section->xyzLE[0] = 0.0;
    section->xyzLE[1] = 0.0;
    section->xyzLE[2] = 0.0;
    section->nodeIndexLE = 0;

    section->xyzTE[0] = 0.0;
    section->xyzTE[1] = 0.0;
    section->xyzTE[2] = 0.0;
    section->nodeIndexTE = 0;

    section->chord = 0.0;

    if (section->vlmControl != NULL) {

        for (i = 0; i < section->numControl; i++) {
            status = destroy_vlmControlStruct(&section->vlmControl[i]);
            if (status != CAPS_SUCCESS) printf("destroy_vlmControlStruct %d\n", status);
        }

        EG_free(section->vlmControl);
    }

    section->vlmControl = NULL;
    section->numControl = 0;

    return CAPS_SUCCESS;
}

// Initiate (0 out all values and NULL all pointers) of a surface in the vlmSurface structure format
int initiate_vlmSurfaceStruct(vlmSurfaceStruct  *surface) {

    // Surface information
    surface->name   = NULL;

    surface->numAttr = 0;    // Number of capsGroup/attributes used to define a given surface
    surface->attrIndex = NULL; // List of attribute map integers that correspond given capsGroups

    surface->Nchord = 10;
    surface->Cspace = 0.0;

    surface->Nspan = 10;
    surface->Sspace = 0.0;

    surface->nowake = (int) false;
    surface->noalbe = (int) false;;
    surface->noload = (int) false;;
    surface->compon = 0;
    surface->iYdup  = (int) false;

    // Section storage
    surface->numSection = 0;
    surface->vlmSection = NULL;

    return CAPS_SUCCESS;
}

// Destroy (0 out all values and NULL all pointers) of a surface in the vlmSurface structure format
int destroy_vlmSurfaceStruct(vlmSurfaceStruct  *surface) {

    int status; // Function return status
    int i; // Indexing

    // Surface information
    if (surface->name != NULL) EG_free(surface->name);
    surface->name = NULL;

    surface->numAttr = 0;    // Number of capsGroup/attributes used to define a given surface
    if (surface->attrIndex != NULL) EG_free(surface->attrIndex);
    surface->attrIndex = NULL; // List of attribute map integers that correspond given capsGroups


    surface->Nchord = 0;
    surface->Cspace = 0.0;

    surface->Nspan = 0;
    surface->Sspace = 0.0;

    surface->nowake = (int) false;
    surface->noalbe = (int) false;;
    surface->noload = (int) false;;
    surface->compon = 0;
    surface->iYdup  = (int) false;

    // Section storage
    if (surface->vlmSection != NULL) {
        for (i = 0; i < surface->numSection; i++ ) {
            status = destroy_vlmSectionStruct(&surface->vlmSection[i]);
            if (status != CAPS_SUCCESS) printf("destroy_vlmSectionStruct status = %d", status);
        }

        if (surface->vlmSection != NULL) EG_free(surface->vlmSection);
    }

    surface->vlmSection = NULL;
    surface->numSection = 0;

    return CAPS_SUCCESS;
}

// Populate vlmSurface-section control surfaces from geometry attributes, modify control properties based on
// incoming vlmControl structures
int get_ControlSurface(ego bodies[],
                       int numControl,
                       vlmControlStruct vlmControl[],
                       vlmSurfaceStruct *vlmSurface) {

    int status = 0; // Function status return

    int section, control, attr, index; // Indexing

    int atype, alen; // EGADS return variables
    const int    *ints;
    const double *reals;
    const char *string;

    const char *attributeKey = "vlmControl", *aName = NULL;
    const char *attrName = NULL;

    int numAttr = 0; // Number of attributes

    int stringLength = 0; // Length of a string

    // Control related variables
    double chordPercent, chordVec[3]; //chordLength

    for (section = 0; section < vlmSurface->numSection; section++) {

        vlmSurface->vlmSection[section].numControl = 0;

        status = EG_attributeNum(bodies[vlmSurface->vlmSection[section].bodyIndex], &numAttr);
        if (status != EGADS_SUCCESS)  return status;

        // Control attributes
        for (attr = 0; attr < numAttr; attr++) {

            status = EG_attributeGet(bodies[vlmSurface->vlmSection[section].bodyIndex],
                                     attr+1,
                                     &aName,
                                     &atype,
                                     &alen,
                                     &ints,
                                     &reals, &string);


            if (status != EGADS_SUCCESS) continue;
            if (atype  != ATTRREAL)      continue;

            if (strncmp(aName, attributeKey, strlen(attributeKey)) != 0) continue;

            if (alen == 0) {
                printf( "Warning: %s should be followed by a single value corresponding to the flap location "
                        "as a function of the chord. 0 - 1 (fraction - %% / 100), 1-100 (%%)\n", attributeKey);
                continue;
            }

            if (reals[0] > 100) {
                printf( "Warning: %s value (%f) must be less than 100\n", attributeKey, reals[0]);
                continue;
            }

            //printf("Attribute name = %s\n", aName);

            if (aName[strlen(attributeKey)] == '\0')

                attrName = "Flap";

            else if	(aName[strlen(attributeKey)] == '_') {

                attrName = &aName[strlen(attributeKey)+1];

            } else {

                attrName = &aName[strlen(attributeKey)];
            }

            //printf("AttrName = %s\n", attrName);

            vlmSurface->vlmSection[section].numControl += 1;

            index = vlmSurface->vlmSection[section].numControl-1; // Make copy to shorten the following lines of code

            if (vlmSurface->vlmSection[section].numControl == 1) {
                vlmSurface->vlmSection[section].vlmControl = (vlmControlStruct *) EG_alloc(vlmSurface->vlmSection[section].numControl*sizeof(vlmControlStruct));

            } else {

                vlmSurface->vlmSection[section].vlmControl = (vlmControlStruct *) EG_reall(vlmSurface->vlmSection[section].vlmControl,
                        vlmSurface->vlmSection[section].numControl*sizeof(vlmControlStruct));
            }

            if (vlmSurface->vlmSection[section].vlmControl == NULL) return EGADS_MALLOC;

            status = initiate_vlmControlStruct(&vlmSurface->vlmSection[section].vlmControl[index]);
            if (status != CAPS_SUCCESS) return status;

            // Get name of control surface
            if (vlmSurface->vlmSection[section].vlmControl[index].name != NULL) EG_free(vlmSurface->vlmSection[section].vlmControl[index].name);

            stringLength = strlen(attrName);

            vlmSurface->vlmSection[section].vlmControl[index].name = (char *) EG_alloc((stringLength+1)*sizeof(char));
            if (vlmSurface->vlmSection[section].vlmControl[index].name == NULL) return EGADS_MALLOC;

            memcpy(vlmSurface->vlmSection[section].vlmControl[index].name,
                    attrName,
                    stringLength*sizeof(char));

            vlmSurface->vlmSection[section].vlmControl[index].name[stringLength] = '\0';

            // Loop through control surfaces from input Tuple and see if defaults have be augmented
            for (control = 0; control < numControl; control++) {

                if (strcasecmp(vlmControl[control].name, attrName) != 0) continue;

                status = copy_vlmControlStruct(&vlmControl[control],
                        &vlmSurface->vlmSection[section].vlmControl[index]);
                if (status != CAPS_SUCCESS) return status;

                break;
            }

            if (control == numControl) {
                printf("Warning: Control %s not found in controls tuple! Only defaults will be used.\n", attrName);
            }

            // Get percent of chord from attribute
            if (reals[0] < 0.0) {
                printf("Warning: Percent chord must < 0, converting to a positive number.\n");
                vlmSurface->vlmSection[section].vlmControl[index].percentChord = -1.0* reals[0];

            } else {

                vlmSurface->vlmSection[section].vlmControl[index].percentChord = reals[0];
            }

            // Was value given as a percentage or fraction
            if (vlmSurface->vlmSection[section].vlmControl[index].percentChord  >= 1.0) {
                vlmSurface->vlmSection[section].vlmControl[index].percentChord  = vlmSurface->vlmSection[section].vlmControl[index].percentChord / 100;
            }

            if (vlmSurface->vlmSection[section].vlmControl[index].leOrTeOverride == (int) false) {

                if (vlmSurface->vlmSection[section].vlmControl[index].percentChord  < 0.5) {
                    vlmSurface->vlmSection[section].vlmControl[index].leOrTe = 0;
                } else {
                    vlmSurface->vlmSection[section].vlmControl[index].leOrTe = 1;
                }
            }

            // Get xyz of hinge location
            chordVec[0] = vlmSurface->vlmSection[section].xyzTE[0] - vlmSurface->vlmSection[section].xyzLE[0];
            chordVec[1] = vlmSurface->vlmSection[section].xyzTE[1] - vlmSurface->vlmSection[section].xyzLE[1];
            chordVec[2] = vlmSurface->vlmSection[section].xyzTE[2] - vlmSurface->vlmSection[section].xyzLE[2];
//            chordLength = sqrt(pow(chordVec[0], 2) +
//                               pow(chordVec[1], 2) +
//                               pow(chordVec[2], 2));
            chordPercent = vlmSurface->vlmSection[section].vlmControl[index].percentChord;

            vlmSurface->vlmSection[section].vlmControl[index].xyzHinge[0] = chordPercent*chordVec[0] +
                                                                            vlmSurface->vlmSection[section].xyzLE[0];

            vlmSurface->vlmSection[section].vlmControl[index].xyzHinge[1] = chordPercent*chordVec[1] +
                                                                            vlmSurface->vlmSection[section].xyzLE[1];

            vlmSurface->vlmSection[section].vlmControl[index].xyzHinge[2] = chordPercent*chordVec[2] +
                                                                            vlmSurface->vlmSection[section].xyzLE[2];

            /*
			printf("\nCheck hinge line \n");

			printf("ChordLength = %f\n", chordLength);
			printf("ChordPercent = %f\n", chordPercent);
			printf("LeadingEdge = %f %f %f\n", vlmSurface->vlmSection[section].xyzLE[0],
											   vlmSurface->vlmSection[section].xyzLE[1],
											   vlmSurface->vlmSection[section].xyzLE[2]);

			printf("TrailingEdge = %f %f %f\n", vlmSurface->vlmSection[section].xyzTE[0],
												vlmSurface->vlmSection[section].xyzTE[1],
												vlmSurface->vlmSection[section].xyzTE[2]);

			printf("Hinge location = %f %f %f\n", vlmSurface->vlmSection[section].vlmControl[index].xyzHinge[0],
												  vlmSurface->vlmSection[section].vlmControl[index].xyzHinge[1],
												  vlmSurface->vlmSection[section].vlmControl[index].xyzHinge[2]);
             */

            /*// Set the hinge vector to the plane normal - TOBE ADDED
			vlmSurface->vlmSection[section].vlmControl[index].xyzHingeVec[0] =
			vlmSurface->vlmSection[section].vlmControl[index].xyzHingeVec[1] =
			vlmSurface->vlmSection[section].vlmControl[index].xyzHingeVec[2] =
             */
        }
    }

    return CAPS_SUCCESS;
}


// Make a copy of vlmControlStruct (it is assumed controlOut has already been initialized)
int copy_vlmControlStruct(vlmControlStruct *controlIn, vlmControlStruct *controlOut) {

    int stringLength = 0;

    if (controlIn == NULL) return CAPS_NULLVALUE;
    if (controlOut == NULL) return CAPS_NULLVALUE;

    if (controlIn->name != NULL) {
        stringLength = strlen(controlIn->name);

        if (controlOut->name != NULL) EG_free(controlOut->name);

        controlOut->name = (char *) EG_alloc((stringLength+1)*sizeof(char));
        if (controlOut->name == NULL) return EGADS_MALLOC;

        memcpy(controlOut->name,
                controlIn->name,
                stringLength*sizeof(char));

        controlOut->name[stringLength] = '\0';
    }

    controlOut->deflectionAngle = controlIn->deflectionAngle; // Deflection angle of the control surface
    controlOut->controlGain = controlIn->controlGain; //Control deflection gain, units:  degrees deflection / control variable


    controlOut->percentChord = controlIn->percentChord; // Percentage along chord

    controlOut->xyzHinge[0] = controlIn->xyzHinge[0]; // xyz location of the hinge
    controlOut->xyzHinge[1] = controlIn->xyzHinge[1];
    controlOut->xyzHinge[2] = controlIn->xyzHinge[2];

    controlOut->xyzHingeVec[0] = controlIn->xyzHingeVec[0]; // xyz location of the hinge
    controlOut->xyzHingeVec[1] = controlIn->xyzHingeVec[1];
    controlOut->xyzHingeVec[2] = controlIn->xyzHingeVec[2];


    controlOut->leOrTeOverride = controlIn->leOrTeOverride; // Does the user want to override the geometry set value?
    controlOut->leOrTe = controlIn->leOrTe; // Leading = 0 or trailing > 0 edge control surface

    controlOut->deflectionDup = controlIn->deflectionDup; // Sign of deflection for duplicated surface

    return CAPS_SUCCESS;
}

// Make a copy of vlmSectionStruct (it is assumed sectionOut has already been initialized)
int copy_vlmSectionStruct(vlmSectionStruct *sectionIn, vlmSectionStruct *sectionOut) {

    int status; // Function return status

    int i; // Indexing

    int stringLength = 0;

    if (sectionIn == NULL) return CAPS_NULLVALUE;
    if (sectionOut == NULL) return CAPS_NULLVALUE;

    if (sectionIn->name != NULL) {
        stringLength = strlen(sectionIn->name);

        if (sectionOut->name != NULL) EG_free(sectionOut->name);

        sectionOut->name = (char *) EG_alloc((stringLength+1)*sizeof(char));
        if (sectionOut->name == NULL) return EGADS_MALLOC;

        memcpy(sectionOut->name,
                sectionIn->name,
                stringLength*sizeof(char));

        sectionOut->name[stringLength] = '\0';
    }

    sectionOut->bodyIndex = sectionIn->bodyIndex;  // Body index - 0 bias
    sectionOut->sectionIndex = sectionIn->sectionIndex; // Section index - 0 bias


    sectionOut->xyzLE[0] = sectionIn->xyzLE[0]; // xyz coordinates for leading edge
    sectionOut->xyzLE[1] = sectionIn->xyzLE[1];
    sectionOut->xyzLE[2] = sectionIn->xyzLE[2];

    sectionOut->nodeIndexLE = sectionIn->nodeIndexLE; // Leading edge node index with reference to xyzLE

    sectionOut->xyzTE[0] = sectionIn->xyzTE[0]; // xyz location of the hinge
    sectionOut->xyzTE[1] = sectionIn->xyzTE[1];
    sectionOut->xyzTE[2] = sectionIn->xyzTE[2];

    sectionOut->nodeIndexTE = sectionIn->nodeIndexTE; // Trailing edge node index with reference to xyzTE

    sectionOut->chord = sectionIn->chord; // Chord

    sectionOut->numControl = sectionIn->numControl;

    if (sectionOut->numControl != 0) {

        sectionOut->vlmControl = (vlmControlStruct *) EG_alloc(sectionOut->numControl*sizeof(vlmControlStruct));
        if (sectionOut->vlmControl == NULL) return EGADS_MALLOC;

        for (i = 0; i < sectionOut->numControl; i++) {

            status = initiate_vlmControlStruct(&sectionOut->vlmControl[i]);
            if (status != CAPS_SUCCESS) return status;

            status = copy_vlmControlStruct(&sectionIn->vlmControl[i], &sectionOut->vlmControl[i]);
            if (status != CAPS_SUCCESS) return status;
        }
    }

    return CAPS_SUCCESS;
}

// Make a copy of vlmSectionStruct (it is assumed sectionOut has already been initialized)
// Also the section in vlmSurface are reordered based on a vlm_orderSections() function call
int copy_vlmSurfaceStruct(vlmSurfaceStruct *surfaceIn, vlmSurfaceStruct *surfaceOut) {

    int status; // Function return status

    int i, sectionIndex; // Indexing

    int stringLength;

    if (surfaceIn == NULL) return CAPS_NULLVALUE;
    if (surfaceOut == NULL) return CAPS_NULLVALUE;


    if (surfaceIn->name != NULL) {
        stringLength = strlen(surfaceIn->name);

        if (surfaceOut->name != NULL) EG_free(surfaceOut->name);

        surfaceOut->name = (char *) EG_alloc((stringLength+1)*sizeof(char));
        if (surfaceOut->name == NULL) return EGADS_MALLOC;

        memcpy(surfaceOut->name,
                surfaceIn->name,
                stringLength*sizeof(char));

        surfaceOut->name[stringLength] = '\0';
    }

    surfaceOut->numAttr = surfaceIn->numAttr;

    if (surfaceIn->attrIndex != NULL) {

        surfaceOut->attrIndex = (int *) EG_alloc(surfaceOut->numAttr *sizeof(int));
        if (surfaceOut->attrIndex == NULL) return EGADS_MALLOC;

        memcpy(surfaceOut->attrIndex,
                surfaceIn->attrIndex,
                surfaceOut->numAttr*sizeof(int));
    }

    surfaceOut->Cspace = surfaceIn->Cspace;
    surfaceOut->Sspace = surfaceIn->Sspace;

    surfaceOut->Nchord = surfaceIn->Nchord;
    surfaceOut->Nspan  = surfaceIn->Nspan;

    surfaceOut->nowake = surfaceIn->nowake;
    surfaceOut->noalbe = surfaceIn->noalbe;
    surfaceOut->noload = surfaceIn->noload;

    surfaceOut->compon = surfaceIn->compon;
    surfaceOut->iYdup = surfaceIn->iYdup;

    surfaceOut->numSection = surfaceIn->numSection;


    if (surfaceIn->vlmSection != NULL) {

        surfaceOut->vlmSection = (vlmSectionStruct *) EG_alloc(surfaceOut->numSection*sizeof(vlmSectionStruct));

        if (surfaceOut->vlmSection == NULL) return EGADS_MALLOC;

        (void) vlm_orderSections(surfaceIn->numSection, surfaceIn->vlmSection);

        for (i = 0; i < surfaceOut->numSection; i++) {

            status = initiate_vlmSectionStruct(&surfaceOut->vlmSection[i]);
            if (status != CAPS_SUCCESS) return status;

            // Sections aren't necessarily stored in order coming out of vlm_GetSection, however sectionIndex is (after a
            // call to vlm_orderSection()) !

            sectionIndex = surfaceIn->vlmSection[i].sectionIndex;

            status = copy_vlmSectionStruct(&surfaceIn->vlmSection[sectionIndex], &surfaceOut->vlmSection[i]);
            if (status != CAPS_SUCCESS) return status;

            // Reset the sectionIndex that is keeping track of the section order.
            surfaceOut->vlmSection[i].sectionIndex = i;
        }
    }

    return CAPS_SUCCESS;
}


static int findLeadingEdge(ego body, int *nodeIndexLE, double *xyzLE )
{
    int status; // Function return
    int  i; // Indexing

    // Node variables
    ego    *nodes = NULL;
    int    numNode;
    double xmin=1E6, xyz[3]={0.0, 0.0, 0.0};

    //EGADS returns (unused)
    int oclass, mtype, *sens = NULL, numChildren;
    ego ref, *children = NULL;

    *nodeIndexLE = 0;

    // Assume the LE position is the most forward Node in X
    status = EG_getBodyTopos(body, NULL, NODE, &numNode, &nodes);
    if (status != EGADS_SUCCESS) goto cleanup;


    for (i = 0; i < numNode; i++) {
        status = EG_getTopology(nodes[i], &ref, &oclass, &mtype, xyz,
                                &numChildren, &children, &sens);
        if (status != EGADS_SUCCESS) goto cleanup;

        if (*nodeIndexLE == 0) {
            *nodeIndexLE = i+1;
            xmin = xyz[0];
        } else {

            if (xyz[0] < xmin) {
                *nodeIndexLE = i+1;
                xmin = xyz[0];
            }
        }
    }

    if (*nodeIndexLE == 0) {
        printf(" findLeadingEdge: Body has no LE!\n");
        status = CAPS_NOTFOUND;
        goto cleanup;
    }

    status = EG_getTopology(nodes[*nodeIndexLE-1], &ref, &oclass, &mtype, xyzLE, &numChildren, &children, &sens);

    goto cleanup;

    cleanup:
        if (status != CAPS_SUCCESS) printf("Error in findLeadingEdge - status %d\n", status);
        if (nodes != NULL) EG_free(nodes);

        return status;
}

static int findTrailingEdge(ego body, int *nodeIndexTE, double *xyzTE )
{
    int status; // Function return
    int  i; // Indexing

    // Node variables
    ego    *nodes = NULL;
    int    numNode;
    double xmax=-1E6, xyz[3]={0.0, 0.0, 0.0};

    //EGADS returns (unused)
    int oclass, mtype, *sens = NULL, numChildren;
    ego ref, *children = NULL;

    *nodeIndexTE = 0;

    // Assume the TE position is the most rear Node in X
    status = EG_getBodyTopos(body, NULL, NODE, &numNode, &nodes);
    if (status != EGADS_SUCCESS) goto cleanup;


    for (i = 0; i < numNode; i++) {
        status = EG_getTopology(nodes[i], &ref, &oclass, &mtype, xyz,
                &numChildren, &children, &sens);
        if (status != EGADS_SUCCESS) goto cleanup;

        if (*nodeIndexTE == 0) {
            *nodeIndexTE = i+1;
            xmax = xyz[0];
        } else {

            if (xyz[0] > xmax) {
                *nodeIndexTE = i+1;
                xmax = xyz[0];
            }
        }
    }

    if (*nodeIndexTE == 0) {
        printf(" findTrailingEdge: Body has no TE!\n");
        status = CAPS_NOTFOUND;
        goto cleanup;
    }
    status = EG_getTopology(nodes[*nodeIndexTE-1], &ref, &oclass, &mtype, xyzTE, &numChildren, &children, &sens);

    goto cleanup;

    cleanup:
        if (status != CAPS_SUCCESS) printf("Error in findTrailingEdge - status %d\n", status);
        if (nodes != NULL) EG_free(nodes);

        return status;
}

static int getChord(ego body, double *chord ) {

    int status; // Function return status

    double xdot[3];

    int numNode;
    ego *nodes = NULL;

    double xyzLE[3], xyz[3], xmin, trange[2], result[9];

    //EGADS returns
    int oclass, mtype, *sens = NULL, numChildren;
    ego ref, *children = NULL;

    int nodeIndexLE, nodeIndexTE;

    ego teObj = NULL;

    status = vlm_findTEObj(body, &teObj);
    if (status != CAPS_SUCCESS) goto cleanup;


    status = findLeadingEdge(body, &nodeIndexLE, xyzLE);
    if (status != EGADS_SUCCESS) {
        printf("\tError in findLeadingEdge, getBodyTopos Nodes = %d\n", status);
        goto cleanup;
    }

    status = findTrailingEdge(body, &nodeIndexTE, xyz);
    if (status != EGADS_SUCCESS) {
        printf("\tError in findTrailingEdge, getBodyTopos Nodes = %d\n", status);
        goto cleanup;
    }

    status = EG_getBodyTopos(body, NULL, NODE, &numNode, &nodes);
    if (status != EGADS_SUCCESS) {
        printf("\tError in AVL writeSection, getBodyTopos Nodes = %d\n", status);
        goto cleanup;
    }

    if (teObj == NULL) {

        status = EG_getTopology(nodes[nodeIndexTE-1], &ref, &oclass, &mtype, result, &numChildren, &children, &sens);
        if (status != EGADS_SUCCESS) goto cleanup;

    } else {

        status = EG_getTopology(teObj, &ref, &oclass, &mtype, trange, &numChildren, &children, &sens);
        if (status != EGADS_SUCCESS) goto cleanup;

        xmin = 0.5*(trange[0]+trange[1]);

        status = EG_evaluate(teObj, &xmin, result);
        if (status != EGADS_SUCCESS) {
            printf("\tError in getChord, evaluate = %d!\n", status);
            goto cleanup;
        }
    }

    xdot[0]  =  result[0] - xyzLE[0];
    xdot[1]  =  result[1] - xyzLE[1];
    xdot[2]  =  result[2] - xyzLE[2];

    *chord    =  sqrt(xdot[0]*xdot[0] + xdot[1]*xdot[1] + xdot[2]*xdot[2]);

    status = CAPS_SUCCESS;
    goto cleanup;

    cleanup:
        if (status != CAPS_SUCCESS) printf("Error: Premature exit in getChord, status = %d\n", status);

        if (nodes != NULL) EG_free(nodes);

        return status;

}

// Find the EGO object pertaining the to trailing edge
int vlm_findTEObj(ego body, ego *teObj) {

    int status; // Function return status

    int i; // Indexing

    int numNode, numEdge;
    ego *nodes = NULL, *edges = NULL;

    double xyz[3], xmin, trange[2];

    //EGADS returns
    int oclass, mtype, *sens = NULL, numChildren;
    ego ref, *children = NULL, *temp = NULL;

    int nodeIndexLE, nodeIndexTE;

    *teObj = NULL;

    status = findLeadingEdge(body, &nodeIndexLE, xyz);
    if (status != EGADS_SUCCESS) {
        printf("\tError in findLeadingEdge, getBodyTopos Nodes = %d\n", status);
        goto cleanup;
    }

    status = findTrailingEdge(body, &nodeIndexTE, xyz);
    if (status != EGADS_SUCCESS) {
        printf("\tError in findTrailingEdge, getBodyTopos Nodes = %d\n", status);
        goto cleanup;
    }

    status = EG_getBodyTopos(body, NULL, NODE, &numNode, &nodes);
    if (status != EGADS_SUCCESS) {
        printf("\tError in vlm_findTEObj, getBodyTopos Nodes = %d\n", status);
        goto cleanup;
    }

    status = EG_getBodyTopos(body, NULL, EDGE, &numEdge, &edges);
    if (status != EGADS_SUCCESS) {
        printf("\tError in vlm_findTEObj, getBodyTopos Edges = %d\n", status);
        goto cleanup;
    }

    for (i = 0; i < numEdge; i++) {

        status = EG_getTopology(edges[i], &ref, &oclass, &mtype, trange, &numChildren, &children, &sens);

        if (status != EGADS_SUCCESS) {
            printf("\tError in vlm_findTEObj, Edge %d getTopology = %d!\n", i, status);
            goto cleanup;
        }

        if (numChildren != 2) {
            printf("\tError in vlm_findTEObj, Edge %d has %d nodes!\n", i, numChildren);
            goto cleanup;
        }

        if (numEdge <= 3) {
            if ( (children[0] != nodes[nodeIndexLE-1]) &&
                 (children[1] != nodes[nodeIndexLE-1])) {

                *teObj = edges[i];
                // printf(" TE Edge = %d\n", i);
                break;
            }

        } else {

            // If the edge doesn't at least contain the TE node pass it by
            if (children[0] != nodes[nodeIndexTE-1] && children[1] != nodes[nodeIndexTE-1]) continue;

            // Lets compare the coordinates
            status = EG_getTopology(children[0], &ref, &oclass, &mtype, xyz, &numChildren, &temp, &sens);
            if (status != EGADS_SUCCESS) goto cleanup;

            xmin = xyz[0];

            status = EG_getTopology(children[1], &ref, &oclass, &mtype, xyz, &numChildren, &temp, &sens);
            if (status != EGADS_SUCCESS) goto cleanup;

            if (fabs(xyz[0] - xmin) < 10E-4) { // If we are at the same x-location

                *teObj = edges[i];
                break;
            }
        }
    }

    status = CAPS_SUCCESS;
    goto cleanup;

    cleanup:
        if (status != CAPS_SUCCESS) printf("Error: Premature exit in vlm_findTEObj, status = %d\n", status);

        if (nodes != NULL) EG_free(nodes);
        if (edges != NULL) EG_free(edges);

        return status;

}


// Accumulate VLM section data from a set of bodies
int vlm_getSection(int numBody,
                   ego bodies[],
                   int ignoreBodyIntent,
                   mapAttrToIndexStruct attrMap,
                   int numSurface,
                   vlmSurfaceStruct *vlmSurface[]) {

    int status; // Function return status

    int i, body, surf, section; // Indexing
    int attrIndex;

    int found = (int) false; // Boolean tester

    const char *groupName = NULL;

    int currentIntent = CAPSMAGIC;

    // Loop through bodies
    for (body = 0; body < numBody; body++) {

        if (ignoreBodyIntent != (int) true) {

            status = retrieve_CAPSIntent(bodies[body], &currentIntent);
            if (status == CAPS_SUCCESS) {

                if (currentIntent != LINEARAERO) continue;
            }
        }

        // Loop through surfaces
        for (surf = 0; surf < numSurface; surf++) {

            status = retrieve_CAPSGroupAttr(bodies[body], &groupName);
            if (status != CAPS_SUCCESS) {
                printf("No capsGroup value found on body %d, body will not be used\n", body+1);
                continue;
            }

            status = get_mapAttrToIndexIndex(&attrMap, groupName, &attrIndex);
            if (status == CAPS_NOTFOUND) {
                printf("VLM Surface name \"%s\" not found in attrMap\n", groupName);
                goto cleanup;
            }

            found = (int) false;

            // See if attrIndex is in the attrIndex array for the surface
            for (i = 0; i < (*vlmSurface)[surf].numAttr; i++) {

                if (attrIndex == (*vlmSurface)[surf].attrIndex[i]) {
                    found = (int) true;
                    break;
                }
            }

            // If attrIndex isn't in the array the; body doesn't belong in this surface
            if (found == (int) false) continue;

            // Increment the number of sections
            (*vlmSurface)[surf].numSection += 1;

            // Get section index
            section = (*vlmSurface)[surf].numSection -1;

            if ((*vlmSurface)[surf].numSection == 1) {
                (*vlmSurface)[surf].vlmSection = (vlmSectionStruct *) EG_alloc(1*sizeof(vlmSectionStruct));
            } else {

                (*vlmSurface)[surf].vlmSection = (vlmSectionStruct *) EG_reall((*vlmSurface)[surf].vlmSection,
                                                                               (*vlmSurface)[surf].numSection*
                                                                               sizeof(vlmSectionStruct));
            }
            if ((*vlmSurface)[surf].vlmSection == NULL) {
                status = EGADS_MALLOC;
                goto cleanup;
            }

            status = initiate_vlmSectionStruct(&(*vlmSurface)[surf].vlmSection[section]);
            if (status != CAPS_SUCCESS) goto cleanup;

            // Store the section index
            (*vlmSurface)[surf].vlmSection[section].sectionIndex = section;

            // Store the body index that the section is coming from
            (*vlmSurface)[surf].vlmSection[section].bodyIndex = body;

            // Get leading edge coordinates
            status = findLeadingEdge(bodies[body],
                                     &(*vlmSurface)[surf].vlmSection[section].nodeIndexLE,
                                     (*vlmSurface)[surf].vlmSection[section].xyzLE);
            if (status != CAPS_SUCCESS) {

                if ((*vlmSurface)[surf].vlmSection[section].nodeIndexLE == 0) {
                    printf(" vlm_getSection: Body %d has no LE!\n", body +1);
                    status = CAPS_NOTFOUND;
                }

                goto cleanup;
            }

            // Get trailing edge coordinates
            status = findTrailingEdge(bodies[body],
                                      &(*vlmSurface)[surf].vlmSection[section].nodeIndexTE,
                                      (*vlmSurface)[surf].vlmSection[section].xyzTE);
            if (status != CAPS_SUCCESS) {

                if ((*vlmSurface)[surf].vlmSection[section].nodeIndexTE == 0) {
                    printf(" vlm_getSection: Body %d has no TE!\n", body +1);
                    status = CAPS_NOTFOUND;
                }
                goto cleanup;
            }

            // Get chord
            status = getChord(bodies[body],
                    &(*vlmSurface)[surf].vlmSection[section].chord);
            if (status != CAPS_SUCCESS) {
                goto cleanup;
            }
        }
    }

    status = CAPS_SUCCESS;
    goto cleanup;

    cleanup:
        if (status != CAPS_SUCCESS) printf("Error: Status %d during vlm_GetSection", status);
        return status;
}

// Order VLM sections increasing order
void vlm_orderSections(int numSection, vlmSectionStruct section[])
{
    int    i1, i2, j, k, hit;
    double box[6], lScale;

    j = section[0].sectionIndex;

    box[0] = box[3] = section[j].xyzLE[0];
    box[1] = box[4] = section[j].xyzLE[1];
    box[2] = box[5] = section[j].xyzLE[2];

    for (k = 0; k < numSection; k++) {
        j = section[k].sectionIndex;
        if (section[j].xyzLE[0] < box[0]) box[0] = section[j].xyzLE[0];
        if (section[j].xyzLE[0] > box[3]) box[3] = section[j].xyzLE[0];
        if (section[j].xyzLE[1] < box[1]) box[1] = section[j].xyzLE[1];
        if (section[j].xyzLE[1] > box[4]) box[4] = section[j].xyzLE[1];
        if (section[j].xyzLE[2] < box[2]) box[2] = section[j].xyzLE[2];
        if (section[j].xyzLE[2] > box[5]) box[5] = section[j].xyzLE[2];
    }

    lScale = box[3] - box[0];
    if (box[4] - box[1] > lScale) lScale = box[4] - box[1];
    if (box[5] - box[2] > lScale) lScale = box[5] - box[2];

    if ((box[4]-box[1])/lScale > 1.e-5) {
        /* Y ordering */
        do {
            for (hit = k = 0; k < numSection-1; k++) {
                i1 = section[k  ].sectionIndex;
                i2 = section[k+1].sectionIndex;

                if (section[i1].xyzLE[1] <= section[i2].xyzLE[1]) continue;

                section[k  ].sectionIndex = i2;
                section[k+1].sectionIndex = i1;
                hit++;
            }
        } while (hit != 0);

    } else if ((box[5]-box[2])/lScale > 1.e-5) {
        /* assume Vertical Tail and do Z ordering */
        do {
            for (hit = k = 0; k < numSection-1; k++) {
                i1 = section[k  ].sectionIndex;
                i2 = section[k+1].sectionIndex;

                if (section[i1].xyzLE[2] <= section[i2].xyzLE[2]) continue;

                section[k  ].sectionIndex = i2;
                section[k+1].sectionIndex = i1;
                hit++;
            }
        } while (hit != 0);

    } else {
        printf(" AIM Warning: No Y or Z Variation in Surface LE Bounding Box!\n");
        printf("              %lf %lf   %lf %lf   %lf %lf  - lScale = %lf\n", box[0],
                                                                              box[3],
                                                                              box[1],
                                                                              box[4],
                                                                              box[2],
                                                                              box[5],
                                                                              lScale);
    }

    /*
     printf(" Sections =");
     for (j = 0; j < surfaces[i].nsec; j++) printf(" %d", surfaces[j].sectionIndex);
     printf("\n");
     */
}
