from __future__ import print_function

# Import pyCAPS class file
from pyCAPS import capsProblem

# Import os module   
try: 
    import os
except:
    print ("Unable to import os module")
    raise SystemExit

# Import SU2 python environment
from parallel_computation import parallel_computation as su2Run

# Initialize capsProblem object
myProblem = capsProblem()

# Create working directory variable 
workDir = "AeroelasticSimple_Pressure"

# Create projectName vairbale
projectName = "aeroelasticSimple_Pressure"

# Load CSM file 
myProblem.loadCAPS("./csmData/aeroelasticDataTransferSimple.csm")

# Load AIMs 
myMesh = myProblem.loadAIM(aim = "tetgenAIM", 
                           altName= "tetgen",
                           analysisDir = workDir + "_SU2",
                           capsIntent = "CFD")

  
su2 = myProblem.loadAIM(aim = "su2AIM", 
                        altName = "su2",
                        analysisDir = workDir + "_SU2", 
                        parents = ["tetgen"],
                        capsIntent = "CFD")

mystran = myProblem.loadAIM(aim = "mystranAIM", 
                            altName = "mystran", 
                            analysisDir = workDir + "_MYSTRAN",
                            capsIntent = "STRUCTURE")

transfers = ["Skin_Top", "Skin_Bottom", "Skin_Tip"]
for i in transfers:
    myProblem.createDataTransfer(variableName = "Cp",
                                 aimSrc = "su2",
                                 aimDest ="mystran",
                                 transferMethod = "Interpolate", #"Conserve", #
                                 capsBound = i)

# Set inputs for tetgen 
myMesh.setAnalysisVal("Tess_Params", [.05, 0.01, 20.0])
myMesh.setAnalysisVal("Preserve_Surf_Mesh", True)

# Set inputs for su2
speedofSound = 340.0 # m/s
refVelocity = 100.0 # m/s
refDensity = 1.2 # kg/m^3

su2.setAnalysisVal("Proj_Name", projectName)
su2.setAnalysisVal("Mach", refVelocity/speedofSound)
su2.setAnalysisVal("Equation_Type","compressible")
su2.setAnalysisVal("Num_Iter",10)
su2.setAnalysisVal("Output_Format", "Tecplot")
su2.setAnalysisVal("SU2_Version", "Raven")

inviscid = {"bcType" : "Inviscid", "wallTemperature" : 1.1}
su2.setAnalysisVal("Boundary_Condition", [("Skin", inviscid),
                                          ("SymmPlane", "SymmetryY"),
                                          ("Farfield","farfield")])

# Set inputs for mystran
mystran.setAnalysisVal("Proj_Name", projectName)
mystran.setAnalysisVal("Edge_Point_Max", 3)
mystran.setAnalysisVal("Edge_Point_Min", 3)

mystran.setAnalysisVal("Quad_Mesh", True)
mystran.setAnalysisVal("Tess_Params", [.5, .1, 15])
mystran.setAnalysisVal("Analysis_Type", "Static");

madeupium    = {"materialType" : "isotropic", 
                "youngModulus" : 72.0E9 , 
                "poissonRatio": 0.33, 
                "density" : 2.8E3}
mystran.setAnalysisVal("Material", ("Madeupium", madeupium))

skin  = {"propertyType" : "Shell", 
         "membraneThickness" : 0.0015, 
         "material"        : "madeupium", 
         "bendingInertiaRatio" : 1.0, # Default           
         "shearMembraneRatio"  : 5.0/6.0} # Default 

ribSpar  = {"propertyType" : "Shell", 
            "membraneThickness" : 0.05, 
            "material"        : "madeupium", 
            "bendingInertiaRatio" : 1.0, # Default           
            "shearMembraneRatio"  : 5.0/6.0} # Default 

mystran.setAnalysisVal("Property", [("Skin", skin),
                                                          ("Rib_Root", ribSpar)])

constraint = {"groupName" : "Rib_Root", 
              "dofConstraint" : 123456}
mystran.setAnalysisVal("Constraint", ("edgeConstraint", constraint))

# Run pre/post-analysis for tetgen and su2 - execute su2
aim = ["tetgen", "su2"]
for i in aim:
    print ("\nRunning PreAnalysis ......", i)
    myProblem.analysis[i].aimPreAnalysis()

    if "su2" in i:
        
        # Set scaling factor for pressure 
        myProblem.analysis[i].setAnalysisVal("Pressure_Scale_Factor",0.5*refDensity*refVelocity**2)

        ####### Run su2 ####################
        print ("\n\nRunning SU2......")
        currentDirectory = os.getcwd() # Get our current working directory 
        
        os.chdir(su2.analysisDir) # Move into test directory
        
        numberProc = 4;
        try:
            su2Run(su2.getAnalysisVal("Proj_Name") + ".cfg", numberProc) # Run SU2
        except: 
            print("SU2 FAILED!")
            myProblem.closeCAPS()
            raise SystemError
            
        os.chdir(currentDirectory) # Move back to top directory 
    
    if "tetgen" in i:    
        print ("\nRunning PostAnalysis ......", i)
        # Run AIM post-analysis
        myProblem.analysis[i].aimPostAnalysis()

# Run aim pre-analysis, excute codes, and run aim post-analysis
aim = ["mystran"]
for i in aim:
    print ("\nRunning PreAnalysis ......", i)
    myProblem.analysis[i].aimPreAnalysis()
    
    print ("\nRunning PostAnalysis ......", i)
    myProblem.analysis[i].aimPostAnalysis()

aim = ["su2", "mystran"]
for i in aim:
    
    if "su2" in i:
        print ("\nRunning PostAnalysis ......", i)
        # Run AIM post-analysis
        myProblem.analysis[i].aimPostAnalysis()

        #Execute the dataTransfer
        print ("\nExecuting dataTransfer ......")
        for j in transfers: 
            myProblem.dataBound[j].executeTransfer()
            #myProblem.dataBound[j].dataSetSrc["Pressure"].viewData()
            #myProblem.dataBound[j].dataSetDest["Pressure"].viewData()
            if j == transfers[0]:
                myProblem.dataBound[j].viewData(colormap="Blues")
            #myProblem.dataBound[j].viewData( filename = j +"Pressure", showImage=False, colormap="rainbow")
            
    
    if "mystran" in i:
        load = {"loadType" : "PressureExternal"} # Add pressure load card 
        myProblem.analysis[i].setAnalysisVal("Load", ("pressureAero", load ))   
    
        # Re-run the preAnalysis 
        print ("\nRunning PreAnalysis ......", i)
        myProblem.analysis[i].aimPreAnalysis()

        # Run mystran
        print ("\n\nRunning Mystran......")  
        currentDirectory = os.getcwd() # Get our current working directory 

        os.chdir(myProblem.analysis[i].analysisDir) # Move into test directory

        #--write_aero_loads_to_file
        os.system("mystran.exe " + projectName +  ".dat > Info.out") # Run su2 via system call

        if os.path.getsize("Info.out") == 0: # 
            print ("Mystran excution failed\n")
            myProblem.closeCAPS()
            raise SystemError
        
        os.chdir(currentDirectory) # Move back to top directory 
        
        print ("\nRunning PostAnalysis ......", i)
        # Run AIM post-analysis
        myProblem.analysis[i].aimPostAnalysis()
 
# Close CAPS 
myProblem.closeCAPS()