from __future__ import print_function

# Import pyCAPS class file
from pyCAPS import capsProblem

# Import os module   
try: 
    import os
except:
    print ("Unable to import os module")
    raise SystemExit

try: 
    import shutil
except:
    print ("Unable to import shutil module")
    raise SystemExit

try: 
    import math
except:
    print ("Unable to import math module")
    raise SystemExit

# Initialize capsProblem object
myProblem = capsProblem()

# Create working directory variable 
workDir = "Aeroelastic_Iterative"

# Create projectName vairbale
projectName = "aeroelastic_Iterative"

# Set the number of transfer iterations
numTransferIteration = 2

# Load CSM file 
myProblem.loadCAPS("./csmData/aeroelasticDataTransfer.csm")

# Load AIMs 
myProblem.loadAIM(aim = "tetgenAIM", 
                  altName= "tetgen",
                  analysisDir = workDir + "_FUN3D",
                  capsIntent = "CFD")

  
myProblem.loadAIM(aim = "fun3dAIM", 
                  altName = "fun3d", 
                  analysisDir = workDir + "_FUN3D", 
                  parents = ["tetgen"],
                  capsIntent = "CFD")

myProblem.loadAIM(aim = "mystranAIM", 
                  altName = "mystran", 
                  analysisDir = workDir + "_MYSTRAN",
                  capsIntent = "STRUCTURE")

transfers = ["Skin_Top", "Skin_Bottom", "Skin_Tip"]
for i in transfers:
    myProblem.createDataTransfer(variableName = ["Pressure", "Displacement"],
                                 aimSrc = ["fun3d", "mystran"],
                                 aimDest =["mystran", "fun3d"],
                                 transferMethod = ["Conserve", "Conserve"], #["Conserve", "Interpolate"]
                                 capsBound = i)

# Set inputs for tetgen 
myProblem.analysis["tetgen"].setAnalysisVal("Tess_Params", [.05, 0.01, 20.0])
myProblem.analysis["tetgen"].setAnalysisVal("Preserve_Surf_Mesh", True)

# Set inputs for fun3d
Mach = 0.3
Rair = 1716 #ft^2/ (s^2* R)
refTemperature = 491.4 # R
gamma = 1.4
speedofSound = math.sqrt(gamma*refTemperature*Rair) # ft/s

refVelocity = speedofSound*Mach # ft/s
refDensity = 0.0023366 # slug/ft^3
refPressure = (refDensity*speedofSound**2)/gamma

AoA = 10 # angle of attack

myProblem.analysis["fun3d"].setAnalysisVal("Proj_Name", projectName)
myProblem.analysis["fun3d"].setAnalysisVal("Mesh_ASCII_Flag", False)
myProblem.analysis["fun3d"].setAnalysisVal("Mach", Mach)
myProblem.analysis["fun3d"].setAnalysisVal("Alpha", AoA)
myProblem.analysis["fun3d"].setAnalysisVal("Equation_Type","compressible")
myProblem.analysis["fun3d"].setAnalysisVal("Viscous", "inviscid")
myProblem.analysis["fun3d"].setAnalysisVal("Num_Iter",1000)
myProblem.analysis["fun3d"].setAnalysisVal("CFL_Schedule",[1, 5.0])
myProblem.analysis["fun3d"].setAnalysisVal("CFL_Schedule_Iter", [1, 40])
myProblem.analysis["fun3d"].setAnalysisVal("Overwrite_NML", True)
myProblem.analysis["fun3d"].setAnalysisVal("Restart_Read","off")

inviscid = {"bcType" : "Inviscid", "wallTemperature" : 1.1}
myProblem.analysis["fun3d"].setAnalysisVal("Boundary_Condition", [("Skin", inviscid),
                                                                  ("SymmPlane", "SymmetryY"),
                                                                  ("Farfield","farfield")])

# Set inputs for mystran
myProblem.analysis["mystran"].setAnalysisVal("Proj_Name", projectName)
myProblem.analysis["mystran"].setAnalysisVal("Edge_Point_Max", 3)

myProblem.analysis["mystran"].setAnalysisVal("Quad_Mesh", True)
myProblem.analysis["mystran"].setAnalysisVal("Tess_Params", [.5, .1, 15])
myProblem.analysis["mystran"].setAnalysisVal("Analysis_Type", "Static");

madeupium    = {"materialType" : "isotropic", 
                "youngModulus" : 1.5E9, # lbf/ft^2 
                "poissonRatio": 0.33, 
                "density" : 5.5} #slug/ft^3

unobtainium  = {"materialType" : "isotropic", 
                "youngModulus" : 1.0E10, # lbf/ft^2  
                "poissonRatio": 0.33, 
                "density" : 6.0} #slug/ft^3 

myProblem.analysis["mystran"].setAnalysisVal("Material", [("Madeupium", madeupium), 
                                                          ("Unobtainium", unobtainium)])

skin  = {"propertyType" : "Shell", 
         "membraneThickness" : 0.0052, # ft - 1/16 in 
         "material"        : "madeupium", 
         "bendingInertiaRatio" : 1.0, # Default           
         "shearMembraneRatio"  : 5.0/6.0} # Default 

ribSpar  = {"propertyType" : "Shell", 
            "membraneThickness" : 0.0104, # ft  - 0.25 in
            "material"        : "unobtainium",
            "bendingInertiaRatio" : 1.0, # Default           
            "shearMembraneRatio"  : 5.0/6.0} # Default 

myProblem.analysis["mystran"].setAnalysisVal("Property", [("Skin", skin),
                                                          ("Ribs_and_Spars", ribSpar),
                                                          ("Rib_Root", ribSpar)])
constraint = {"groupName" : "Rib_Root", 
              "dofConstraint" : 123456}
myProblem.analysis["mystran"].setAnalysisVal("Constraint", ("edgeConstraint", constraint))


# Run pre/post-analysis for tetgen and fun3d
aim = ["tetgen", "mystran"]
for i in aim:
    print ("\nRunning PreAnalysis ......", i)
    myProblem.analysis[i].aimPreAnalysis()
    
    print ("\nRunning PostAnalysis ......",  i)
    # Run AIM post-analysis
    myProblem.analysis[i].aimPostAnalysis()


aim = ["mystran", "fun3d"]
for iter in range(numTransferIteration):
    
    i = "fun3d"
    myProblem.analysis[i].setAnalysisVal("Pressure_Scale_Factor", 0.5*refDensity*refVelocity**2) # Dirty up fun3d
    
    print ("\nRunning PreAnalysis ......", i)
    myProblem.analysis[i].aimPreAnalysis()

    ####### Run fun3d ####################
    print ("\n\nRunning FUN3D......")  
    currentDirectory = os.getcwd() # Get our current working directory 

    os.chdir(myProblem.analysis[i].analysisDir) # Move into test directory

    #--write_aero_loads_to_file --aeroelastic_external 
    cmdLineOpt = "--write_aero_loads_to_file --animation_freq -1"
    if iter != 0:
        cmdLineOpt = cmdLineOpt + " --read_surface_from_file"

    os.system("mpirun -np 10 nodet_mpi " + cmdLineOpt + " > Info.out"); # Run fun3d via system call

    if os.path.getsize("Info.out") == 0: # 
        print ("FUN3D excution failed\n")
        myProblem.closeCAPS()
        raise SystemError
    
    shutil.copy(projectName + "_tec_boundary.plt", projectName + "_tec_boundary" + "_Interation_" + str(iter) + ".plt")
    os.chdir(currentDirectory) # Move back to top directory

    print ("\nRunning PostAnalysis ......", i)
    # Run AIM post-analysis
    myProblem.analysis[i].aimPostAnalysis()

    #Execute the dataTransfer
    print ("\n\nExecuting dataTransfer \"Pressure\"......")
    for j in transfers: 
        myProblem.dataBound[j].executeTransfer("Pressure")
    
    i = "mystran"
    # External pressure load to mystran that we will inherited from fun3d
    load = {"loadType" : "PressureExternal", "loadScaleFactor" : -1.0}
    myProblem.analysis[i].setAnalysisVal("Load", ("pressureAero", load ))
    
    print ("\nRunning PreAnalysis ......", i)
    myProblem.analysis[i].aimPreAnalysis()

    ####### Run mystran ####################
    print ("\n\nRunning Mystran......")
    currentDirectory = os.getcwd() # Get our current working directory 

    os.chdir(myProblem.analysis[i].analysisDir) # Move into test directory

    os.system("mystran.exe " + projectName +  ".dat > Info.out") # Run fun3d via system call

    if os.path.getsize("Info.out") == 0: # 
        print ("Mystran excution failed\n")
        myProblem.closeCAPS()
        raise SystemError

    os.chdir(currentDirectory) # Move back to top directory 

    print ("\nRunning PostAnalysis ......", i)
    # Run AIM post-analysis
    myProblem.analysis[i].aimPostAnalysis()

    #Execute the dataTransfer
    print ("\n\nExecuting dataTransfer \"Displacement\"......")
    for j in transfers: 
        myProblem.dataBound[j].executeTransfer("Displacement")
 
# Close CAPS 
myProblem.closeCAPS()