from __future__ import print_function

# Import pyCAPS class file
from pyCAPS import capsProblem

# Import os module   
try: 
    import os
except:
    print ("Unable to import os module")
    raise SystemError

# Declare ASTROS install directory 
astrosInstallDir = "/Users/alyanaej/Projects/Solvers/ASTROS/"

# Initialize capsProblem object
myProblem = capsProblem()

# Load CSM file 
myProblem.loadCAPS("./csmData/feaSimplePlate.csm")

# Create project name 
projectName = "AstrosMultipleLoadPlate"

# Load Astros aim 
myAnalysis = myProblem.loadAIM(aim = "astrosAIM", 
                               altName = "astros", 
                               analysisDir = projectName, 
                               capsIntent = "STRUCTURE")

# Set project name so a mesh file is generated 
myAnalysis.setAnalysisVal("Proj_Name", projectName) 

# Set meshing parameters
myAnalysis.setAnalysisVal("Edge_Point_Max", 4)
myAnalysis.setAnalysisVal("Edge_Point_Min", 4)

myAnalysis.setAnalysisVal("Quad_Mesh", True)

myAnalysis.setAnalysisVal("Tess_Params", [.25,.01,15])

# Set analysis type
myAnalysis.setAnalysisVal("Analysis_Type", "Static");

# Set materials 
madeupium    = {"materialType" : "isotropic", 
                "youngModulus" : 72.0E9 , 
                "poissonRatio": 0.33, 
                "density" : 2.8E3}

myAnalysis.setAnalysisVal("Material", ("Madeupium", madeupium))

# Set properties
shell  = {"propertyType" : "Shell", 
          "membraneThickness" : 0.006, 
          "material"        : "madeupium", 
          "bendingInertiaRatio" : 1.0, # Default           
          "shearMembraneRatio"  : 5.0/6.0} # Default 

myAnalysis.setAnalysisVal("Property", ("plate", shell))

# Set constraints
constraint = {"groupName" : "plateEdge", 
              "dofConstraint" : 123456}

myAnalysis.setAnalysisVal("Constraint", ("edgeConstraint", constraint))

# Create multiple loads
numLoad = 5
loads = []
for i in range(numLoad):
    # Set load name 
    name = "load_"+ str(i)
    
    # Set load values 
    value =  {"groupName" : "plate", 
              "loadType" : "Pressure", 
              "pressureForce" : 1.e6*(i+1)}
    
    # Create temporary load tuple 
    loadElement = (name, value)
    
    # Append loadElement to laod
    loads.append( loadElement)

# Set loads    
myAnalysis.setAnalysisVal("Load", loads)

# Create multiple analysis cases 
analysisCases = []
for i in range(numLoad):
    
    # Set analysis name 
    name = loads[i][0]
    
    # set analysis value s
    value = {"analysisType" : "Static", 
             "analysisLoad" : name}
    
    # Create temporary analysis tuple 
    analysisElement = (name, value)
    
    # Append analysisElement to analysis cases 
    analysisCases.append(analysisElement)
             
# Set analysis  
myAnalysis.setAnalysisVal("Analysis", analysisCases)

# Run AIM pre-analysis
myAnalysis.aimPreAnalysis()

####### Run Astros ####################
print ("\n\nRunning Astros......")
currentDirectory = os.getcwd() # Get our current working directory 

os.chdir(myAnalysis.analysisDir) # Move into test directory

# Create symbolic links to files needed to run astros
files = ["astros.exe", # Executatble
         "ASTRO.D01",  # *.DO1 file
         "ASTRO.IDX"]  # *.IDX file
for file in files: 
    if not os.path.isfile(file):
        os.symlink(astrosInstallDir + file, file)

os.system("./astros.exe < " + projectName +  ".dat > " + projectName + ".out"); # Run Astros via system call

os.chdir(currentDirectory) # Move back to working directory 
print ("Done running Astros!")
######################################

# Run AIM post-analysis
myAnalysis.aimPostAnalysis()

# Close CAPS 
myProblem.closeCAPS()
