from __future__ import print_function

# Import pyCAPS class file
from pyCAPS import capsProblem

# Import os module   
try: 
    import os
except:
    print ("Unable to import os module")
    raise SystemError
    
# Declare ASTROS install directory 
astrosInstallDir = "/Users/alyanaej/Projects/Solvers/ASTROS/"     

# Initialize capsProblem object
myProblem = capsProblem()

# Load CSM file 
myProblem.loadCAPS("./csmData/feaSimplePlate.csm")

# Create project name 
projectName = "AstrosOptMultiLoadPlate"

# Load Astros aim 
myAnalysis = myProblem.loadAIM(aim = "astrosAIM", altName = "astros", analysisDir = projectName, capsIntent = "STRUCTURE")

# Set project name so a mesh file is generated 
myAnalysis.setAnalysisVal("Proj_Name", projectName) 

# Set meshing parameters
myAnalysis.setAnalysisVal("Edge_Point_Max", 5)
myAnalysis.setAnalysisVal("Edge_Point_Min", 5)

myAnalysis.setAnalysisVal("Quad_Mesh", True)

myAnalysis.setAnalysisVal("Tess_Params", [.25,.01,15])

myAnalysis.setAnalysisVal("File_Format", "Free")

myAnalysis.setAnalysisVal("Mesh_File_Format", "Large")

# Set analysis type
myAnalysis.setAnalysisVal("Analysis_Type", "StaticOpt");

# Set materials 
madeupium    = {"materialType" : "isotropic", 
                "youngModulus" : 72.0E9 , 
                "poissonRatio": 0.33, 
                "density" : 2.8E3,
                "yieldAllow" : 1E7}


myAnalysis.setAnalysisVal("Material", ("Madeupium", madeupium))

# Set properties
shell  = {"propertyType" : "Shell", 
          "membraneThickness" : 0.006, 
          "material"        : "madeupium", 
          "bendingInertiaRatio" : 1.0, # Default           
          "shearMembraneRatio"  : 5.0/6.0} # Default 

myAnalysis.setAnalysisVal("Property", ("plate", shell))

# Set constraints
constraint = {"groupName" : "plateEdge", 
              "dofConstraint" : 123456}

myAnalysis.setAnalysisVal("Constraint", ("edgeConstraint", constraint))

# Create multiple loads
numLoad = 2
loads = []
for i in range(numLoad):
    # Set load name 
    name = "load_"+ str(i)
    
    # Set load values 
    value =  {"groupName" : "plate", 
              "loadType" : "Pressure", 
              "pressureForce" : 1.e6*(i+1)}
    
    # Create temporary load tuple 
    loadElement = (name, value)
    
    # Append loadElement to laod
    loads.append( loadElement)

# Set loads    
myAnalysis.setAnalysisVal("Load", loads)

# Set design variables and analysis - design variable relation
propertyDV = {"groupName" : "plate",
              "initialValue" : shell["membraneThickness"],
              "lowerBound" : shell["membraneThickness"]*0.5,
              "upperBound" : shell["membraneThickness"]*1.5,
              "maxDelta"   : shell["membraneThickness"]*0.1,
              "fieldName" : "T"}

# Set design variable  
myAnalysis.setAnalysisVal("Design_Variable",  ("membranethickness", propertyDV))

# Set design constraints and responses

designConstraint1 = {"groupName" : "plate",
                    "responseType" : "STRESS",
                    "lowerBound" : -madeupium["yieldAllow"],
                    "upperBound" :  madeupium["yieldAllow"],}

designConstraint2 = {"groupName" : "plate",
                    "responseType" : "STRESS",
                    "lowerBound" : -0.5*madeupium["yieldAllow"],
                    "upperBound" :  0.5*madeupium["yieldAllow"],}    
    

# Set design constraint  
myAnalysis.setAnalysisVal("Design_Constraint",[("stress1", designConstraint1),
                                                                  ("stress2", designConstraint2)])

# Create multiple analysis cases 
analysisCases = []
for i in range(numLoad):
    
    # Set analysis name 
    name = loads[i][0]
    
    # set analysis value
    value = {"analysisType" : "Static", 
             "analysisLoad" : name,
             "analysisDesignConstraint" : "stress" + str(i+1),
             "objectiveMinMax" : "MIN",
             "objectiveResponseType" : "WEIGHT"}
    
    # Create temporary analysis tuple 
    analysisElement = (name, value)
    
    # Append analysisElement to analysis cases 
    analysisCases.append(analysisElement)
             
# Set analysis  
myAnalysis.setAnalysisVal("Analysis", analysisCases)

# Run AIM pre-analysis
myAnalysis.aimPreAnalysis()

####### Run astros####################
print ("\n\nRunning Astros......")
currentDirectory = os.getcwd() # Get our current working directory 

os.chdir(myAnalysis.analysisDir) # Move into test directory

# Create symbolic links to files needed to run astros
files = ["astros.exe", # Executatble
         "ASTRO.D01",  # *.DO1 file
         "ASTRO.IDX"]  # *.IDX file
for file in files: 
    if not os.path.isfile(file):
        os.symlink(astrosInstallDir + file, file)

os.system("./astros.exe < " + projectName +  ".dat > " + projectName + ".out"); # Run Astros via system call

os.chdir(currentDirectory) # Move back to working directory 
print ("Done running Astros!")
####### Run astros####################

# Run AIM post-analysis
myAnalysis.aimPostAnalysis()

# Close CAPS 
myProblem.closeCAPS()
