from __future__ import print_function

# Import pyCAPS class file
from pyCAPS import capsProblem

# Import os module   
try: 
    import os
except:
    print ("Unable to import os module")
    raise SystemError

# Initialize capsProblem object
myProblem = capsProblem()

# Load CSM file 
myProblem.loadCAPS("./csmData/feaWingBEM.csm")

# Load astros aim 
myProblem.loadAIM(aim = "astrosAIM", altName = "astros", analysisDir= "AstrosModalWingBEM", capsIntent = "STRUCTURE")

# Set project name so a mesh file is generated 
projectName = "pyCAPS_astros_Test"
myProblem.analysis["astros"].setAnalysisVal("Proj_Name", projectName) 

myProblem.analysis["astros"].setAnalysisVal("Edge_Point_Max", 6)

myProblem.analysis["astros"].setAnalysisVal("Quad_Mesh", True)

#myProblem.analysis["astros"].setAnalysisVal("File_Format", "Free")

myProblem.analysis["astros"].setAnalysisVal("Mesh_File_Format", "Large")

# Set analysis 
eigen = { "extractionMethod"     : "MGIV", 
          "frequencyRange"       : [0, 10000], 
          "numEstEigenvalue"     : 1,
          "numDesiredEigenvalue" : 10, 
          "eigenNormaliztion"    : "MASS"}
myProblem.analysis["astros"].setAnalysisVal("Analysis", ("EigenAnalysis", eigen))

# Set materials
unobtainium  = {"youngModulus" : 2.2E6 , 
                "poissonRatio" : .5, 
                "density"      : 7850}

madeupium    = {"materialType" : "isotropic", 
                "youngModulus" : 1.2E5 , 
                "poissonRatio" : .5, 
                "density"      : 7850}
myProblem.analysis["astros"].setAnalysisVal("Material", [("Unobtainium", unobtainium),
                                                          ("Madeupium", madeupium)])

# Set property 
shell  = {"propertyType"      : "Shell", 
          "material"          : "unobtainium", 
	      "bendingInertiaRatio" : 1.0, # Default - not necesssary           
          "shearMembraneRatio"  : 0, # Turn of shear - no materialShear
          "membraneThickness" : 0.2}
 
myProblem.analysis["astros"].setAnalysisVal("Property", ("Ribs_and_Spars", shell))

# Set constraints 
constraint = {"groupName" : ["Rib_Constraint"], 
              "dofConstraint" : 123456}

myProblem.analysis["astros"].setAnalysisVal("Constraint", ("ribConstraint", constraint))


# Run AIM pre-analysis
myProblem.analysis["astros"].aimPreAnalysis()

####### Run Astros####################
print ("\n\nRunning Astros......")  
currentDirectory = os.getcwd() # Get our current working directory 

os.chdir(myProblem.analysis["astros"].analysisDir) # Move into test directory
 
# Declare ASTROS install directory 
astrosInstallDir = "/home/durschrj/Desktop/Software_and_Libs/Astros/Linux-64bit/bin/"

# Create symbolic links to files needed to run astros
files = ["astros.exe", # Executatble
         "ASTRO.D01",  # *.DO1 file
         "ASTRO.IDX"]  # *.IDX file
for file in files: 
    if not os.path.isfile(file):
        os.symlink(astrosInstallDir + file, file)

os.system("./astros.exe < " + projectName +  ".dat > " + projectName + ".out"); # Run Astros via system call

os.chdir(currentDirectory) # Move back to working directory 
print ("Done running Astros!")

# Run AIM post-analysis
myProblem.analysis["astros"].aimPostAnalysis()

# Get Eigen-frequencies
print ("\nGetting results for natural frequencies.....")
natrualFreq = myProblem.analysis["astros"].getAnalysisOutVal("EigenFrequency")

mode = 1
for i in natrualFreq:
    print ("Natural freq (Mode {:d}) = ".format(mode) + '{:.5f} '.format(i) + "(Hz)")
    mode += 1    
 
# Close CAPS 
myProblem.closeCAPS()
