from __future__ import print_function

# Import pyCAPS class file
from pyCAPS import capsProblem 

# Import os module   
try: 
    import os
except:
    print ("Unable to import os module")
    raise SystemError

# Initialize capsProblem object
myProblem = capsProblem()

# Load CSM file
myProblem.loadCAPS("./csmData/avlWingTail.csm")

machNumber = [0.05*i for i in range(1, 20)] # Build Mach sweep with a list comprehension

sweepDesPmtr = [5, 10, 20] # Geometry design paramters

Cd = list() # Create a blank list 
Cl = list() 

# Loop through sweep design parametrs 
for ii in range(len(sweepDesPmtr)):
    
    # Set new geometry design parameter
    myProblem.geometry.setGeometryVal("sweep", sweepDesPmtr[ii])
    
    for i in range(len(machNumber)):
        analysisID = "avlSweep_" +str(sweepDesPmtr[ii]) + "_Ma_"+ str(machNumber[i])

        myProblem.loadAIM(aim = "avlAIM", altName = analysisID, analysisDir = "AVLTestDir/" + analysisID, capsIntent = "LINEARAERO")
    
        # Set Mach number
        myProblem.analysis[analysisID].setAnalysisVal("Mach", machNumber[i])

        # Set Surface information
        wing = {"groupName"  : "Wing", 
                "numChord"   : 8, 
                "spaceChord" : 1.0, 
                "numSpan"    : 12, 
                "spaceSpan"  : 1.0}

        tail = {"numChord"   : 8, 
                "spaceChord" : 1.0, 
                "numSpan"    : 12, 
                "spaceSpan"  : 1.0}

        myProblem.analysis[analysisID].setAnalysisVal("AVL_Surface", [("Wing", wing), 
                                                                      ("Vertical_Tail", tail)])

        # Run AIM pre-analysis
        myProblem.analysis[analysisID].aimPreAnalysis()

        # Run AVL 
        print (" Running AVL (sweep angle = " + str(sweepDesPmtr[ii])+ ") at Mach number = " + str(machNumber[i]) +"!")
        currentDirectory = os.getcwd() # Get our current working directory 
        
        os.chdir(myProblem.analysis[analysisID].analysisDir) # Move into test directory
        os.system("avl caps < avlInput.txt > avlOutput.txt");
        
        os.chdir(currentDirectory) # Move back to working directory 
        
        # Run AIM post-analysis
        myProblem.analysis[analysisID].aimPostAnalysis()

        # Get Drag coefficient 
        Cd.append(myProblem.analysis[analysisID].getAnalysisOutVal("CDtot"))
    
        # Get Lift coefficient 
        Cl.append(myProblem.analysis[analysisID].getAnalysisOutVal("CLtot"))
    
        # Print results
        print ("Cd = " + str(Cd[i]))
        print ("Cl = " + str(Cl[i]))

# Close CAPS 
myProblem.closeCAPS()

# Import pyplot module
try:
    from matplotlib import pyplot as plt
except:
    print ("Unable to import matplotlib.pyplot module. Drag polar will not be ploted")
    plt = None
 
# Plot data
if plt is not None:
    
    # Plot first sweep angle
    plt.plot(Cd[1:len(machNumber)], 
             Cl[1:len(machNumber)],
             'o--',
             label = "Sweep Angle = " + str(sweepDesPmtr[0]) + "$^o$")
    
    # Plot second sweep angle 
    plt.plot(Cd[len(machNumber)+1:2*len(machNumber)], 
             Cl[len(machNumber)+1:2*len(machNumber)],
             's-',
             label = "Sweep Angle = " + str(sweepDesPmtr[1]) + "$^o$")
    
    # Plot third sweep angle 
    plt.plot(Cd[2*len(machNumber)+1:3*len(machNumber)], 
             Cl[2*len(machNumber)+1:3*len(machNumber)],
             'x-.',
             label = "Sweep Angle = " + str(sweepDesPmtr[2]) + "$^o$")
    plt.legend(loc = "lower right")
    plt.title("Drag Polar for Various Sweep Angles\n (window must be closed to terminate Python script)")
    plt.xlabel("$C_D$")
    plt.ylabel("$C_L$")
    plt.show()

    
