## [importPrint]
from __future__ import print_function
## [importPrint]

## [import]
# Import pyCAPS class file
from pyCAPS import capsProblem

# Import os module   
try: 
    import os
except:
    print ("Unable to import os module")
    raise SystemExit
## [import]
	
## [initateProblem]	
# Initialize pyCAPS object
myProblem = capsProblem()
## [initateProblem]


## [geometry]
# Load CSM file 
myProblem.loadCAPS("./csmData/compositeWing.csm")
## [geometry]

## [loadAIM]
# Load nastran aim 
nastranAIM = myProblem.loadAIM(aim = "nastranAIM", 
                  	           altName = "nastran", 
                               analysisDir = "NastranCompositeWing_Freq", 
                               capsIntent = "STRUCTURE")
## [loadAIM]
				  
## [setInputs]				  
# Set project name so a mesh file is generated 
projectName = "nastran_CompositeWing"
nastranAIM.setAnalysisVal("Proj_Name", projectName) 
nastranAIM.setAnalysisVal("Edge_Point_Max", 40)
nastranAIM.setAnalysisVal("File_Format", "Small")
nastranAIM.setAnalysisVal("Mesh_File_Format", "Large")
nastranAIM.setAnalysisVal("Analysis_Type", "Modal");
## [setInputs]
		
## [defineMaterials]
Aluminum  = {"youngModulus" : 10.5E6 , 
             "poissonRatio" : 0.3, 
             "density"      : 0.1/386,
             "shearModulus" : 4.04E6}

Graphite_epoxy = {"materialType"        : "Orthotropic", 
                  "youngModulus"        : 20.8E6 ,
                  "youngModulusLateral" : 1.54E6,
                  "poissonRatio"        : 0.327,                
                  "shearModulus"        : 0.80E6,
                  "density"             : 0.059/386,
                  "tensionAllow"        : 11.2e-3,
                  "tensionAllowLateral" : 4.7e-3,
                  "compressAllow"       : 11.2e-3,
                  "compressAllowLateral": 4.7e-3,
                  "shearAllow"          : 19.0e-3,
                  "allowType"           : 1}

nastranAIM.setAnalysisVal("Material", [("Aluminum", Aluminum),
                                       ("Graphite_epoxy", Graphite_epoxy)])
## [defineMaterials]
														  
## [defineProperties]
aluminum  = {"propertyType"         : "Shell", 
             "material"             : "Aluminum",
             "bendingInertiaRatio"  : 1.0, # Default - not necesssary           
             "shearMembraneRatio"   : 0, # Turn of shear - no materialShear
             "membraneThickness"    : 0.125 }

composite  = {"propertyType"           : "Composite", 
              "shearBondAllowable"	   : 1.0e6,
              "bendingInertiaRatio"    : 1.0, # Default - not necesssary           
              "shearMembraneRatio"     : 0, # Turn off shear - no materialShear
              "compositeMaterial"      : ["Graphite_epoxy"]*8,
              "compositeThickness"     : [0.00525]*8,
              "compositeOrientation"   : [0, 0, 0, 0, -45, 45, -45, 45],
	          "symmetricLaminate"      : True, 
              "compositeFailureTheory" : "STRN" }

#nastranAIM.setAnalysisVal("Property", ("wing", aluminum))
nastranAIM.setAnalysisVal("Property", ("wing", composite))
## [defineProperties]
		
## [defineConstraints]  
constraint = {"groupName" : "root", 
              "dofConstraint" : 123456}
    
nastranAIM.setAnalysisVal("Constraint", ("root", constraint))
## [defineConstraints] 

## [defineAnalysis] 
eigen = { "extractionMethod"     : "MGIV",#"Lanczos", 
          "frequencyRange"       : [0, 10000], 
          "numEstEigenvalue"     : 1,
          "numDesiredEigenvalue" : 10, 
          "eigenNormaliztion"    : "MASS"}
    
nastranAIM.setAnalysisVal("Analysis", ("EigenAnalysis", eigen))
## [defineAnalysis]		
		
## [preAnalysis]
nastranAIM.aimPreAnalysis()
## [preAnalysis]

## [run]
####### Run Nastran####################
print ("\n\nRunning Nastran......" ) 
currentDirectory = os.getcwd() # Get our current working directory 

os.chdir(myProblem.analysis["nastran"].analysisDir) # Move into test directory
 
os.system("nastran old=no notify=no batch=no scr=yes sdirectory=./ " + projectName +  ".dat"); # Run Nastran via system call

os.chdir(currentDirectory) # Move back to working directory 
print ("Done running Nastran!")
## [run]

## [postAnalysis]
# Run AIM post-analysis
nastranAIM.aimPostAnalysis()
## [postAnalysis]

# Get Eigen-frequencies
print ("\nGetting results for natural frequencies.....")
natrualFreq = myProblem.analysis["nastran"].getAnalysisOutVal("EigenFrequency")

mode = 1
for i in natrualFreq:
    print ("Natural freq (Mode {:d}) = ".format(mode) + '{:.5f} '.format(i) + "(Hz)")
    mode += 1    
 
## [close]
# Close CAPS 
myProblem.closeCAPS()
## [close]