from __future__ import print_function

# Import pyCAPS class file
from pyCAPS import capsProblem

# Import os module   
try: 
    import os
except:
    print ("Unable to import os module")
    raise SystemError

# Import SU2 Python interface module
from parallel_computation import parallel_computation as su2Run

# Initialize capsProblem object
myProblem = capsProblem()

# Load CSM file 
myProblem.loadCAPS("./csmData/cfdMultiBody.csm")

# Set intention - could also be set during .loadAIM
myProblem.capsIntent = "CFD"

# Change a design parameter - area in the geometry 
myProblem.geometry.setGeometryVal("area", 50)

# Load Tetgen aim 
myMesh = myProblem.loadAIM(aim = "aflr3AIM", analysisDir= ".")

# Set new EGADS body tessellation parameters
myMesh.setAnalysisVal("Tess_Params", [.05, 0.01, 20.0])

# Set intial boundary layer spacing 
myMesh.setAnalysisVal("BL_Initial_Spacing", 1E-3)

# Set maximum boundary layer thickness 
myMesh.setAnalysisVal("BL_Thickness", 0.6)

# Run AIM pre-analysis
myMesh.aimPreAnalysis()

# NO analysis is needed - AFLR3 was already ran during preAnalysis

# Run AIM post-analysis
myMesh.aimPostAnalysis()

# Load SU2 aim - child of Tetgen AIM  
su2 = myProblem.loadAIM(aim = "su2AIM", 
                  altName = "su2", 
                  analysisDir = "SU2AFLR3AnalysisTest", parents = ["aflr3AIM"])

# Set SU2 Version
su2.setAnalysisVal("SU2_Version","Raven")

projectName = "pyCAPS_SU2_AFLR3"
# Set project name
su2.setAnalysisVal("Proj_Name", projectName)

# Set AoA number 
su2.setAnalysisVal("Alpha", 1.0) 

# Set Mach number 
su2.setAnalysisVal("Mach", 0.5901)

# Set equation type
su2.setAnalysisVal("Equation_Type","Compressible")

# Set number of iterations
su2.setAnalysisVal("Num_Iter",10)

# Set boundary conditions
inviscidBC1 = {"bcType" : "Inviscid"}
inviscidBC2 = {"bcType" : "Inviscid"}
su2.setAnalysisVal("Boundary_Condition", [("Wing1", inviscidBC1),
                                          ("Wing2", inviscidBC2),
                                          ("Farfield","farfield")])
 
# Run AIM pre-analysis
su2.aimPreAnalysis()

####### Run SU2 ####################
print ("\n\nRunning SU2......")
currentDirectory = os.getcwd() # Get our current working directory 

os.chdir(su2.analysisDir) # Move into test directory

numberProc = 4;
su2Run(projectName + ".cfg", numberProc) # Run SU2

os.chdir(currentDirectory) # Move back to top directory 
#######################################

# Run AIM post-analysis
su2.aimPostAnalysis()

# Get force results
print ("Total Force - Pressure + Viscous")
# Get Lift and Drag coefficients
print ("Cl = " , su2.getAnalysisOutVal("CLtot"), \
      "Cd = "  , su2.getAnalysisOutVal("CDtot"))

# Get Cmx, Cmy, and Cmz coefficients 
print ("Cmx = " , su2.getAnalysisOutVal("CMXtot"), \
      "Cmy = "  , su2.getAnalysisOutVal("CMYtot"), \
      "Cmz = "  , su2.getAnalysisOutVal("CMZtot"))
    
# Get Cx, Cy, Cz coefficients 
print ("Cx = " , su2.getAnalysisOutVal("CXtot"), \
      "Cy = "  , su2.getAnalysisOutVal("CYtot"), \
      "Cz = "  , su2.getAnalysisOutVal("CZtot"))
 
# Close CAPS 
myProblem.closeCAPS()
