import unittest

import os

from sys import version_info as pyVersion

import pyCAPS

class TestGeometry(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        cls.file = "unitGeom.csm"
        cls.analysisDir = "UnitTest"

        cls.myProblem = pyCAPS.capsProblem()

        cls.myGeometry = cls.myProblem.loadCAPS(cls.file, "basicTest")

        cls.myAnalysis = cls.myProblem.loadAIM(aim = "fun3dAIM",
                                                analysisDir = cls.analysisDir,
                                                capsIntent = "CFD")

    @classmethod
    def tearDownClass(cls):

        # Remove analysis directories
        if os.path.exists(cls.analysisDir):
            os.rmdir(cls.analysisDir)

        # Remove created files
        if os.path.isfile("myGeometry.egads"):
            os.remove("myGeometry.egads")

        if os.path.isfile("myGeometry.html"):
            os.remove("myGeometry.html")

        if os.path.isfile("geomView_0.png"):
            os.remove("geomView_0.png")

        if os.path.isfile("geomView_1.png"):
            os.remove("geomView_1.png")

        if os.path.isfile("geomView_2.png"):
            os.remove("geomView_2.png")

    # Try loading two geometries into the problem
    def test_loadCAPS(self):

        with self.assertRaises(pyCAPS.CAPSError) as e:
            myGeometry = self.myProblem.loadCAPS(self.file)

        self.assertEqual(e.exception.errorName, "CAPS_BADVALUE")

    # Retrieve geometry attribute
    def test_getAttributeVal(self):

        attributeList = self.myGeometry.getAttributeVal("capsGroup", attrLevel="Body")
        self.assertEqual(attributeList.sort(), ['Farfield', 'Wing2', 'Wing1'].sort())

        attributeList = self.myGeometry.getAttributeVal("capsGroup", attrLevel="5")
        self.assertEqual(attributeList.sort(), ['Farfield', 'Wing2', 'Wing1'].sort())

        attributeList = self.myGeometry.getAttributeVal("capsGroup", attrLevel="Node")
        self.assertEqual(attributeList.sort(), ['Farfield', 'Wing2', 'Wing1'].sort())

    # Set/get geometry value
    def test_setGeometryVal(self):

        self.myGeometry.setGeometryVal("lesweep", 20.0)
        self.assertEqual(self.myGeometry.getGeometryVal("lesweep"), 20.0)

    # Get geometry in values
    def test_getGeometryVal(self):

        # Get all input values
        valueDict = self.myGeometry.getGeometryVal()
        self.assertEqual(valueDict["aspect"], 5.0)

        # Get just the names of the input values
        valueList = self.myGeometry.getGeometryVal(namesOnly = True)
        self.assertEqual("twist" in valueList, True)

    # Get geometry out values
    def test_getGeometryOutVal(self):

        # Get all output values
        valueDict = self.myGeometry.getGeometryOutVal()
        self.assertAlmostEqual(valueDict["span"], 14.14, 2)

        # Get just the names of the output values
        valueList = self.myGeometry.getGeometryOutVal(namesOnly = True)
        self.assertEqual("cmean" in valueList, True)

    # Geometry - object consistent
    def test_geometry(self):

        self.assertEqual(self.myGeometry, self.myProblem.geometry)

    # View geometry
    def test_viewGeometry(self):

        try:

            self.myGeometry.viewGeometry(filename = "geomView", showImage =False)

            self.assertTrue(os.path.isfile("geomView_0.png"))
            self.assertTrue(os.path.isfile("geomView_1.png"))
            self.assertTrue(os.path.isfile("geomView_2.png"))

        except ImportError as e: # The machine doesn't have matplotlib
            self.assertTrue(True)

        except pyCAPS.CAPSError as e:
            raise e
        except: # Ignore other errors  TclError being thrown up by matplotlib on some machines
            pass


    # Save geometry
    def test_saveGeometry(self):

        self.myGeometry.saveGeometry()
        self.assertTrue(os.path.isfile("myGeometry.egads"))

    # Create html tree
    @unittest.skipIf(pyVersion.major > 2, "Odd behavior in Python 3")
    def test_createTree(self):

        self.myGeometry.createTree(internetAccess = True,
                                   embedJSON = True,
                                   reverseMap = True)
        self.assertTrue(os.path.isfile("myGeometry.html"))


if __name__ == '__main__':
    unittest.main()
