/*
 ************************************************************************
 *                                                                      *
 * sensCSM.c -- test configuration and tessellation sensitivities       *
 *                                                                      *
 *              Written by John Dannenhoffer @ Syracuse University      *
 *                                                                      *
 ************************************************************************
 */

/*
 * Copyright (C) 2013/2018  John F. Dannenhoffer, III (Syracuse University)
 *
 * This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *     MA  02110-1301  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include <time.h>

#include "egads.h"

#define CINT    const int
#define CDOUBLE const double
#define CCHAR   const char

#define STRNCPY(A, B, LEN) strncpy(A, B, LEN); A[LEN-1] = '\0';

#include "common.h"
#include "OpenCSM.h"
#include "udp.h"

/***********************************************************************/
/*                                                                     */
/* macros (including those that go along with common.h)                */
/*                                                                     */
/***********************************************************************/

#ifdef DEBUG
   #define DOPEN {if (dbg_fp == NULL) dbg_fp = fopen("sensCSM.dbg", "w");}
   static  FILE *dbg_fp=NULL;
#endif

                                               /* used by RALLOC macro */
//static void *realloc_temp=NULL;

#define  RED(COLOR)      (float)(COLOR / 0x10000        ) / (float)(255)
#define  GREEN(COLOR)    (float)(COLOR / 0x00100 % 0x100) / (float)(255)
#define  BLUE(COLOR)     (float)(COLOR           % 0x100) / (float)(255)

/***********************************************************************/
/*                                                                     */
/* global variables                                                    */
/*                                                                     */
/***********************************************************************/

/* global variable holding a MODL */
       char      casename[255];        /* name of case */
       char      pmtrname[255];        /* name of single design Parameter */
static void      *modl;                /* pointer to MODL */

static int       config   = 0;         /* =1 for configuration sensitivites */
static double    dtime    = 0.001;     /* nominal dtime for perturbation */
static int       outLevel = 1;         /* default output level */
static int       tessel   = 1;         /* =1 for tessellation sensitivities */

/***********************************************************************/
/*                                                                     */
/* declarations                                                        */
/*                                                                     */
/***********************************************************************/

/* declarations for high-level routines defined below */

/* declarations for support routines defined below */

/* external functions not declared in an .h file */


/***********************************************************************/
/*                                                                     */
/*   main - main program                                               */
/*                                                                     */
/***********************************************************************/

int
main(int       argc,                    /* (in)  number of arguments */
     char      *argv[])                 /* (in)  array of arguments */
{

    int       status, status2, i, nbody, ibody;
    int       imajor, iminor, builtTo, showUsage=0;
    int       ipmtr, irow, icol, ntemp, jnode, jedge, jface, itype, nlist;
    int       ntotal, nerror, inode, iedge, iface, ipnt, jpnt, periodic, iokay;
    int       npnt_tess, ntri_tess, oclass, mtype, nchild, *senses;
    CINT      *ptype, *pindx, *tris, *tric, *tempIlist;
    double    data[18], data2[18], trange[4];
    double    dist, dist2, xyz_pred[3], uv_close[2], xyz_close[3], toler;
    double    errmax, face_errmax, edge_errmax, node_errmax, EPS05=1.0e-5;
    double    distmax, face_distmax, edge_distmax, node_distmax;
    double    **face_cfg_anal=NULL, **face_cfg_fdif=NULL, face_cfg_err;
    double    **edge_cfg_anal=NULL, **edge_cfg_fdif=NULL, edge_cfg_err;
    double    **node_cfg_anal=NULL, **node_cfg_fdif=NULL, node_cfg_err;
    double    **face_tes_anal=NULL, **edge_tes_anal=NULL, **node_tes_anal=NULL;
    CDOUBLE   *xyz, *uv, *dxyz, *xyz_ptrb, *uv_ptrb;
    char      filename[255];
    CCHAR     *OCC_ver;
    ego       eref, *echilds, context;

    modl_T    *MODL;

    ROUTINE(MAIN);

    /* --------------------------------------------------------------- */

    DPRINT0("starting sensCSM");

    /* get the flags and casename from the command line */
    casename[0] = '\0';
    pmtrname[0] = '\0';

    for (i = 1; i < argc; i++) {
        if        (strcmp(argv[i], "-config") == 0) {
            config = 1;
        } else if (strcmp(argv[i], "-despmtr") == 0) {
            if (i < argc-1) {
                strcpy(pmtrname, argv[++i]);
            } else {
                showUsage = 1;
                break;
            }
        } else if (strcmp(argv[i], "-dtime") == 0) {
            if (i < argc-1) {
                sscanf(argv[++i], "%lf", &dtime);
            } else {
                showUsage = 1;
                break;
            }
        } else if (strcmp(argv[i], "-help") == 0 ||
                   strcmp(argv[i], "-h"   ) == 0   ) {
            showUsage = 1;
            break;
        } else if (strcmp(argv[i], "-outLevel") == 0) {
            if (i < argc-1) {
                sscanf(argv[++i], "%d", &outLevel);
                if (outLevel < 0) outLevel = 0;
                if (outLevel > 3) outLevel = 3;
            } else {
                showUsage = 1;
                break;
            }
        } else if (strcmp(argv[i], "--version") == 0 ||
                   strcmp(argv[i], "-version" ) == 0 ||
                   strcmp(argv[i], "-v"       ) == 0   ) {
            (void) ocsmVersion(&imajor, &iminor);
            SPRINT2(0, "OpenCSM version: %2d.%02d", imajor, iminor);
            EG_revision(&imajor, &iminor, &OCC_ver);
            SPRINT3(0, "EGADS   version: %2d.%02d (with %s)", imajor, iminor, OCC_ver);
            exit(EXIT_SUCCESS);
        } else if (strlen(casename) == 0) {
            strcpy(casename, argv[i]);
        } else {
            SPRINT0(0, "two casenames given");
            showUsage = 1;
            break;
        }
    }

    (void) ocsmVersion(&imajor, &iminor);

    if (showUsage) {
        SPRINT2(0, "sensCSM version %2d.%02d\n", imajor, iminor);
        SPRINT0(0, "proper usage: 'sensCSM [casename[.csm]] [options...]");
        SPRINT0(0, "   where [options...] = -config");
        SPRINT0(0, "                        -despmtr pmtrname");
        SPRINT0(0, "                        -dtime dtime");
        SPRINT0(0, "                        -help  -or-  -h");
        SPRINT0(0, "                        -outLevel X");
        SPRINT0(0, "STOPPING...\a");
        return EXIT_FAILURE;
    }

    /* welcome banner */
    SPRINT0(1, "**********************************************************");
    SPRINT0(1, "*                                                        *");
    SPRINT0(1, "*                    Program sensCSM                     *");
    SPRINT2(1, "*                     version %2d.%02d                      *", imajor, iminor);
    SPRINT0(1, "*                                                        *");
    SPRINT0(1, "*        written by John Dannenhoffer, 2010/2018         *");
    SPRINT0(1, "*                                                        *");
    SPRINT0(1, "**********************************************************\n");

    SPRINT1(1, "    casename   = %s", casename  );
    SPRINT1(1, "    config     = %d", config    );
    SPRINT1(1, "    despmtr    = %s", pmtrname  );
    SPRINT1(1, "    dtime      = %f", dtime     );
    SPRINT1(1, "    outLevel   = %d", outLevel  );
    SPRINT1(1, "    tessel     = %d", tessel    );
    SPRINT0(1, " ");

    /* set OCSMs output level */
    (void) ocsmSetOutLevel(outLevel);

    /* strip off .csm (which is assumed to be at the end) if present */
    if (strlen(casename) > 0) {
        strcpy(filename, casename);
        if (strstr(casename, ".csm") == NULL) {
            strcat(filename, ".csm");
        }
    } else {
        filename[0] = '\0';
    }

    /* read the .csm file and create the MODL */
    status   = ocsmLoad(filename, &modl);
    MODL = (modl_T*)modl;

    SPRINT3(1, "--> ocsmLoad(%s) -> status=%d (%s)",
            filename, status, ocsmGetText(status));
    CHECK_STATUS(ocsmLoad);

    /* check that Branches are properly ordered */
    status   = ocsmCheck(modl);
    SPRINT2(0, "--> ocsmCheck() -> status=%d (%s)",
            status, ocsmGetText(status));
    CHECK_STATUS(ocsmCheck);

    /* print out the global Attributes, Parameters, and Branches */
    SPRINT0(1, "External Parameter(s):");
    if (outLevel > 0) {
        status = ocsmPrintPmtrs(modl, stdout);
        CHECK_STATUS(ocsmPrintPmtrs);
    }

    SPRINT0(1, "Branch(es):");
    if (outLevel > 0) {
        status = ocsmPrintBrchs(modl, stdout);
        CHECK_STATUS(ocsmPrintBrchs);
    }

    SPRINT0(1, "Global Attribute(s):");
    if (outLevel > 0) {
        status = ocsmPrintAttrs(modl, stdout);
        CHECK_STATUS(ocsmPrintAttrs);
    }

    /* build the Bodys from the MODL */
    nbody    = 0;
    status   = ocsmBuild(modl, 0, &builtTo, &nbody, NULL);
    SPRINT4(1, "--> ocsmBuild -> status=%d (%s), builtTo=%d, nbody=%d",
            status, ocsmGetText(status), builtTo, nbody);
    CHECK_STATUS(ocsmBuild);

    /* print out the Bodys */
    SPRINT0(1, "Body(s):");
    if (outLevel > 0) {
        status = ocsmPrintBodys(modl, stdout);
        CHECK_STATUS(ocsmPrintBodys);
    }

    /* check configuration and tessellation sensitivities */
    ntotal  = 0;              // total number of errors exceeding tolerance
    errmax  = 0;              // maximmum error   (configuration)
    distmax = 0;              // maximum distance (tessellation)

    /* loop through all design Parameters */
    for (ipmtr = 1; ipmtr <= MODL->npmtr; ipmtr++) {
        if (MODL->pmtr[ipmtr].type != OCSM_EXTERNAL) continue;

        if (strlen(pmtrname) > 0 &&
            strcmp(pmtrname, MODL->pmtr[ipmtr].name) != 0) continue;

        for (irow = 1; irow <= MODL->pmtr[ipmtr].nrow; irow++) {
            for (icol = 1; icol <= MODL->pmtr[ipmtr].ncol; icol++) {

                SPRINT0(0, "=========================================================\n");

                /* set the velocity */
                status = ocsmSetVelD(MODL, 0, 0, 0, 0.0);
                CHECK_STATUS(ocsmSetVelD);
                status = ocsmSetVelD(MODL, ipmtr, irow, icol, 1.0);
                CHECK_STATUS(ocsmSetVelD);

                /* build needed to propagate velocities to the Branch arguments */
                ntemp    = 0;
                status   = ocsmBuild(MODL, 0, &builtTo, &ntemp, NULL);
                CHECK_STATUS(ocsmBuild);

//$$$                if (config) {
                    SPRINT0(0, "\n*********************************************************");
                    if (MODL->pmtr[ipmtr].nrow == 1 &&
                        MODL->pmtr[ipmtr].ncol == 1   ) {
                        SPRINT1(0, "Starting configuration sensitivity wrt \"%s\"",
                                MODL->pmtr[ipmtr].name);
                    } else {
                        SPRINT3(0, "Starting configuration sensitivity wrt \"%s[%d,%d]\"",
                                MODL->pmtr[ipmtr].name, irow, icol);
                    }
                    SPRINT0(0, "*********************************************************\n");

                    /* perform configuration sensitivity checks for each Body on the stack */
                    for (ibody = 1; ibody <= MODL->nbody; ibody++) {
                        if (MODL->body[ibody].onstack != 1) continue;

                        /* analytic configuration sensitivities (if possible) */
                        SPRINT1(0, "Computing analytic sensitivities (if possible) for ibody=%d",
                                ibody);
//$$$                        status = removePerturbation(MODL);
//$$$                        CHECK_STATUS(removePerturbation);
                        status = ocsmSetDtime(MODL, 0);
                        CHECK_STATUS(ocsmSetDtime);

                        /* save analytic configuration sensitivity of each Face */
                        MALLOC(face_cfg_anal, double*, MODL->body[ibody].nface+1);
                        for (iface = 0; iface <= MODL->body[ibody].nface; iface++) {
                            face_cfg_anal[iface] = NULL;
                        }

                        for (iface = 1; iface <= MODL->body[ibody].nface; iface++) {
                            status = EG_getTessFace(MODL->body[ibody].etess, iface,
                                                    &npnt_tess, &xyz, &uv, &ptype, &pindx,
                                                    &ntri_tess, &tris, &tric);
                            CHECK_STATUS(EG_getTessFace);
                            if (npnt_tess <= 0) {
                                SPRINT3(0, "ERROR:: EG_getTessFace -> status=%d (%s), npnt_tess=%d",
                                        status, ocsmGetText(status), npnt_tess);
                                status = EXIT_FAILURE;
                                goto cleanup;
                            }

                            MALLOC(face_cfg_anal[iface], double, 3*npnt_tess);

                            status = ocsmGetVel(MODL, ibody, OCSM_FACE, iface, npnt_tess, NULL, face_cfg_anal[iface]);
                            CHECK_STATUS(ocsmGetVel);
                        }

                        /* save analytic configuration sensitivity of each Edge */
                        MALLOC(edge_cfg_anal, double*, MODL->body[ibody].nedge+1);
                        for (iedge = 0; iedge <= MODL->body[ibody].nedge; iedge++) {
                            edge_cfg_anal[iedge] = NULL;
                        }

                        for (iedge = 1; iedge <= MODL->body[ibody].nedge; iedge++) {
                            status = EG_getTessEdge(MODL->body[ibody].etess, iedge,
                                                    &npnt_tess, &xyz, &uv);
                            CHECK_STATUS(EG_getTessEdge);
                            if (npnt_tess <= 0) {
                                SPRINT3(0, "ERROR:: EG_getTessEdge -> status=%d (%s), npnt_tess=%d",
                                        status, ocsmGetText(status), npnt_tess);
                                status = EXIT_FAILURE;
                                goto cleanup;
                            }

                            MALLOC(edge_cfg_anal[iedge], double, 3*npnt_tess);

                            status = ocsmGetVel(MODL, ibody, OCSM_EDGE, iedge, npnt_tess, NULL, edge_cfg_anal[iedge]);
                            CHECK_STATUS(ocsmGetVel);
                        }

                        /* save analytic configuration sensitivity of each Node */
                        MALLOC(node_cfg_anal, double*, MODL->body[ibody].nnode+1);
                        for (inode = 0; inode <= MODL->body[ibody].nnode; inode++) {
                            node_cfg_anal[inode] = NULL;
                        }

                        for (inode = 1; inode <= MODL->body[ibody].nnode; inode++) {
                            npnt_tess = 1;

                            MALLOC(node_cfg_anal[inode], double, 3*npnt_tess);

                            status = ocsmGetVel(MODL, ibody, OCSM_NODE, inode, npnt_tess, NULL, node_cfg_anal[inode]);
                            CHECK_STATUS(ocsmGetVel);
                        }

                        /* if there is a perturbation, return that finite differences were used */
                        if (MODL->perturb != NULL) {
                            SPRINT1(0, "\nSensitivity checks complete with %8d total errors (    finite diffs   )",
                                    ntotal);
                            status = EXIT_SUCCESS;
                            goto cleanup;
                        }

                        /* finite difference configuration sensitivities */
                        SPRINT1(0, "Computing finite difference sensitivities for ibody=%d",
                                ibody);
//$$$                        status = createPerturbation(MODL);
//$$$                        CHECK_STATUS(createPerturbation);
                        status = ocsmSetDtime(MODL, dtime);
                        CHECK_STATUS(ocsmSetDtime);

                        /* save finite difference configuration sensitivity of each Face */
                        MALLOC(face_cfg_fdif, double*, MODL->body[ibody].nface+1);
                        for (iface = 0; iface <= MODL->body[ibody].nface; iface++) {
                            face_cfg_fdif[iface] = NULL;
                        }

                        for (iface = 1; iface <= MODL->body[ibody].nface; iface++) {
                            status = EG_getTessFace(MODL->body[ibody].etess, iface,
                                                    &npnt_tess, &xyz, &uv, &ptype, &pindx,
                                                    &ntri_tess, &tris, &tric);
                            CHECK_STATUS(EG_getTessFace);

                            MALLOC(face_cfg_fdif[iface], double, 3*npnt_tess);

                            status = ocsmGetVel(MODL, ibody, OCSM_FACE, iface, npnt_tess, NULL, face_cfg_fdif[iface]);
                            CHECK_STATUS(ocsmGetVel);
                        }

                        /* save finite difference configuration sensitivity of each Edge */
                        MALLOC(edge_cfg_fdif, double*, MODL->body[ibody].nedge+1);
                        for (iedge = 0; iedge <= MODL->body[ibody].nedge; iedge++) {
                            edge_cfg_fdif[iedge] = NULL;
                        }

                        for (iedge = 1; iedge <= MODL->body[ibody].nedge; iedge++) {
                            status = EG_getTessEdge(MODL->body[ibody].etess, iedge,
                                                    &npnt_tess, &xyz, &uv);
                            CHECK_STATUS(EG_getTessEdge);

                            MALLOC(edge_cfg_fdif[iedge], double, 3*npnt_tess);

                            status = ocsmGetVel(MODL, ibody, OCSM_EDGE, iedge, npnt_tess, NULL, edge_cfg_fdif[iedge]);
                            CHECK_STATUS(ocsmGetVel);
                        }

                        /* save finite difference configuration sensitivity of each Node */
                        MALLOC(node_cfg_fdif, double*, MODL->body[ibody].nnode+1);
                        for (inode = 0; inode <= MODL->body[ibody].nnode; inode++) {
                            node_cfg_fdif[inode] = NULL;
                        }

                        for (inode = 1; inode <= MODL->body[ibody].nnode; inode++) {
                            npnt_tess = 1;

                            MALLOC(node_cfg_fdif[inode], double, 3*npnt_tess);

                            status = ocsmGetVel(MODL, ibody, OCSM_NODE, inode, npnt_tess, NULL, node_cfg_fdif[inode]);
                            CHECK_STATUS(ocsmGetVel);
                        }

//$$$                        status = removePerturbation(MODL);
//$$$                        CHECK_STATUS(removePerturbation);
                        status = ocsmSetDtime(MODL, 0);
                        CHECK_STATUS(ocsmSetDtime);

                        /* compare configuration sensitivities */
                        node_errmax = 0;
                        edge_errmax = 0;
                        face_errmax = 0;

                        if (MODL->pmtr[ipmtr].nrow == 1 &&
                            MODL->pmtr[ipmtr].ncol == 1   ) {
                            SPRINT2(0, "\nComparing configuration sensitivities wrt \"%s\" for ibody=%d",
                                    MODL->pmtr[ipmtr].name, ibody);
                        } else {
                            SPRINT4(0, "\nComparing configuration sensitivities wrt \"%s[%d,%d]\" for ibody=%d",
                                    MODL->pmtr[ipmtr].name, irow, icol, ibody);
                        }

                        /* compare configuration sensitivities for each Face */
                        nerror = 0;
                        for (iface = 1; iface <= MODL->body[ibody].nface; iface++) {
                            status = EG_getTessFace(MODL->body[ibody].etess, iface,
                                                    &npnt_tess, &xyz, &uv, &ptype, &pindx,
                                                    &ntri_tess, &tris, &tric);
                            CHECK_STATUS(EG_getTessFace);

#ifndef __clang_analyzer__
                            for (ipnt = 0; ipnt < npnt_tess; ipnt++) {
                                for (i = 0; i < 3; i++) {
                                    face_cfg_err = face_cfg_anal[iface][3*ipnt+i] - face_cfg_fdif[iface][3*ipnt+i];
                                    if (fabs(face_cfg_err) > face_errmax) face_errmax = fabs(face_cfg_err);
                                    if (fabs(face_cfg_err) > EPS05) {
                                        if (nerror < 20 || fabs(face_cfg_err) >= face_errmax) {
                                            SPRINT9(0, "iface=%4d,  ipnt=%4d,  %1d anal=%16.8f,  fd=%16.8f,  err=%16.8f (at %10.4f %10.4f %10.4f)",
                                                    iface, ipnt, i, face_cfg_anal[iface][3*ipnt+i], face_cfg_fdif[iface][3*ipnt+i], face_cfg_err,
                                                    xyz[3*ipnt], xyz[3*ipnt+1], xyz[3*ipnt+2]);
                                        }
                                        nerror++;
                                        ntotal++;
                                    }
                                }
                            }

                            FREE(face_cfg_anal[iface]);
                            FREE(face_cfg_fdif[iface]);
#endif
                        }
                        errmax = MAX(errmax, face_errmax);
                        SPRINT3(0, "    d(Face)/d(%s) check complete with %8d total errors (errmax=%12.4e)",
                                MODL->pmtr[ipmtr].name, nerror, face_errmax);

                        /* compare configuration sensitivities for each Edge */
                        nerror = 0;
                        for (iedge = 1; iedge <= MODL->body[ibody].nedge; iedge++) {
                            status = EG_getTessEdge(MODL->body[ibody].etess, iedge,
                                                    &npnt_tess, &xyz, &uv);
                            CHECK_STATUS(EG_getTessEdge);

#ifndef __clang_analyzer__
                            for (ipnt = 0; ipnt < npnt_tess; ipnt++) {
                                for (i = 0; i < 3; i++) {
                                    edge_cfg_err = edge_cfg_anal[iedge][3*ipnt+i] - edge_cfg_fdif[iedge][3*ipnt+i];
                                    if (fabs(edge_cfg_err) > edge_errmax) edge_errmax = fabs(edge_cfg_err);
                                    if (fabs(edge_cfg_err) > EPS05) {
                                        if (nerror < 20 || fabs(edge_cfg_err) >= edge_errmax) {
                                            SPRINT9(0, "iedge=%4d,  ipnt=%4d,  %1d anal=%16.8f,  fd=%16.8f,  err=%16.8f (at %10.4f %10.4f %10.4f)",
                                                    iedge, ipnt, i, edge_cfg_anal[iedge][3*ipnt+i],
                                                    edge_cfg_fdif[iedge][3*ipnt+i], edge_cfg_err,
                                                    xyz[3*ipnt], xyz[3*ipnt+1], xyz[3*ipnt+2]);
                                        }
                                        nerror++;
                                        ntotal++;
                                    }
                                }
                            }

                            FREE(edge_cfg_anal[iedge]);
                            FREE(edge_cfg_fdif[iedge]);
#endif
                        }
                        errmax = MAX(errmax, edge_errmax);
                        SPRINT3(0, "    d(Edge)/d(%s) check complete with %8d total errors (errmax=%12.4e)",
                                MODL->pmtr[ipmtr].name, nerror, edge_errmax);

                        /* compare configuration sensitivities for each Node */
                        nerror = 0;
                        for (inode = 1; inode <= MODL->body[ibody].nnode; inode++) {
                            npnt_tess = 1;

#ifndef __clang_analyzer__
                            for (ipnt = 0; ipnt < npnt_tess; ipnt++) {
                                for (i = 0; i < 3; i++) {
                                    node_cfg_err = node_cfg_anal[inode][3*ipnt+i] - node_cfg_fdif[inode][3*ipnt+i];
                                    if (fabs(node_cfg_err) > node_errmax) node_errmax = fabs(node_cfg_err);
                                    if (fabs(node_cfg_err) > EPS05) {
                                        if (nerror < 20 || fabs(node_cfg_err) >= node_errmax) {
                                            SPRINT9(0, "inode=%4d,  ipnt=%4d,  %1d anal=%16.8f,  fd=%16.8f,  err=%16.8f (at %10.4f %10.4f %10.4f)",
                                                    inode, ipnt, i, node_cfg_anal[inode][3*ipnt+i], node_cfg_fdif[inode][3*ipnt+i], node_cfg_err,
                                                    MODL->body[ibody].node[inode].x,
                                                    MODL->body[ibody].node[inode].y,
                                                    MODL->body[ibody].node[inode].z);
                                        }
                                        nerror++;
                                        ntotal++;
                                    }
                                }
                            }

                            FREE(node_cfg_anal[inode]);
                            FREE(node_cfg_fdif[inode]);
#endif
                        }
                        errmax = MAX(errmax, node_errmax);
                        SPRINT3(0, "    d(Node)/d(%s) check complete with %8d total errors (errmax=%12.4e)",
                                MODL->pmtr[ipmtr].name, nerror, node_errmax);

                        FREE(face_cfg_anal);
                        FREE(face_cfg_fdif);
                        FREE(edge_cfg_anal);
                        FREE(edge_cfg_fdif);
                        FREE(node_cfg_anal);
                        FREE(node_cfg_fdif);
                    }
//$$$                }

//$$$                if (tessel) {
                    SPRINT0(0, "\n*********************************************************");
                    if (MODL->pmtr[ipmtr].nrow == 1 &&
                        MODL->pmtr[ipmtr].ncol == 1   ) {
                        SPRINT1(0, "Starting tessellation sensitivity wrt \"%s\"",
                                MODL->pmtr[ipmtr].name);
                    } else {
                        SPRINT3(0, "Starting tessellation sensitivity wrt \"%s[%d,%d]\"",
                                MODL->pmtr[ipmtr].name, irow, icol);
                    }
                    SPRINT0(0, "*********************************************************\n");

                    /* perform tessellation sensitivity checks for each Body on the stack */
                    for (ibody = 1; ibody <= MODL->nbody; ibody++) {
                        if (MODL->body[ibody].onstack != 1) continue;

                        /* analytic tessellation sensitivities (if possible) */
                        SPRINT1(0, "Computing analytic sensitivities (if possible) for ibody=%d",
                                ibody);
//$$$                        status = removePerturbation(MODL);
//$$$                        CHECK_STATUS(removePerturbation);
                        status = ocsmSetDtime(MODL, 0);
                        CHECK_STATUS(ocsmSetDtime);

                        /* save analytic tessellation sensitivity of each Face */
                        MALLOC(face_tes_anal, double*, MODL->body[ibody].nface+1);
                        for (iface = 0; iface <= MODL->body[ibody].nface; iface++) {
                            face_tes_anal[iface] = NULL;
                        }

                        for (iface = 1; iface <= MODL->body[ibody].nface; iface++) {
                            status = EG_getTessFace(MODL->body[ibody].etess, iface,
                                                    &npnt_tess, &xyz, &uv, &ptype, &pindx,
                                                    &ntri_tess, &tris, &tric);
                            CHECK_STATUS(EG_getTessFace);
                            if (npnt_tess <= 0) {
                                SPRINT3(0, "ERROR:: EG_getTessFace -> status=%d (%s), npnt_tess=%d",
                                        status, ocsmGetText(status), npnt_tess);
                                status = EXIT_FAILURE;
                                goto cleanup;
                            }

                            status = ocsmGetTessVel(MODL, ibody, OCSM_FACE, iface, &dxyz);
                            CHECK_STATUS(ocsmGetTessVel);

                            MALLOC(face_tes_anal[iface], double, 3*npnt_tess);

                            for (ipnt = 0; ipnt < npnt_tess; ipnt++) {
                                face_tes_anal[iface][3*ipnt  ] = dxyz[3*ipnt  ];
                                face_tes_anal[iface][3*ipnt+1] = dxyz[3*ipnt+1];
                                face_tes_anal[iface][3*ipnt+2] = dxyz[3*ipnt+2];
                            }
                        }

                        /* save analytic tessellation sensitivity of each Edge */
                        MALLOC(edge_tes_anal, double*, MODL->body[ibody].nedge+1);
                        for (iedge = 0; iedge <= MODL->body[ibody].nedge; iedge++) {
                            edge_tes_anal[iedge] = NULL;
                        }

                        for (iedge = 1; iedge <= MODL->body[ibody].nedge; iedge++) {
                            status = EG_getTessEdge(MODL->body[ibody].etess, iedge,
                                                    &npnt_tess, &xyz, &uv);
                            CHECK_STATUS(EG_getTessEdge);
                            if (npnt_tess <= 0) {
                                SPRINT3(0, "ERROR:: EG_getTessEdge -> status=%d (%s), npnt_tess=%d",
                                        status, ocsmGetText(status), npnt_tess);
                                status = EXIT_FAILURE;
                                goto cleanup;
                            }

                            status = ocsmGetTessVel(MODL, ibody, OCSM_EDGE, iedge, &dxyz);
                            CHECK_STATUS(ocsmGetTessVel);

                            MALLOC(edge_tes_anal[iedge], double, 3*npnt_tess);

                            for (ipnt = 0; ipnt < npnt_tess; ipnt++) {
                                edge_tes_anal[iedge][3*ipnt  ] = dxyz[3*ipnt  ];
                                edge_tes_anal[iedge][3*ipnt+1] = dxyz[3*ipnt+1];
                                edge_tes_anal[iedge][3*ipnt+2] = dxyz[3*ipnt+2];
                            }
                        }

                        /* save analytic tessellation sensitivity of each Node */
                        MALLOC(node_tes_anal, double*, MODL->body[ibody].nnode+1);
                        for (inode = 0; inode <= MODL->body[ibody].nnode; inode++) {
                            node_tes_anal[inode] = NULL;
                        }

                        for (inode = 1; inode <= MODL->body[ibody].nnode; inode++) {
                            npnt_tess = 1;

                            status = ocsmGetTessVel(MODL, ibody, OCSM_NODE, inode, &dxyz);
                            CHECK_STATUS(ocsmGetTessVel);

                            MALLOC(node_tes_anal[inode], double, 3*npnt_tess);

                            for (ipnt = 0; ipnt < npnt_tess; ipnt++) {
                                node_tes_anal[inode][3*ipnt  ] = dxyz[3*ipnt  ];
                                node_tes_anal[inode][3*ipnt+1] = dxyz[3*ipnt+1];
                                node_tes_anal[inode][3*ipnt+2] = dxyz[3*ipnt+2];
                            }
                        }

                        /* if there is a perturbation, return that finite differences were used */
                        if (MODL->perturb != NULL) {
                            SPRINT1(0, "\nSensitivity checks complete with %8d total errors (    finite diffs   )",
                                    ntotal);
                            status = EXIT_SUCCESS;
                            goto cleanup;
                        }

                        /* finite difference tessellation sensitivities */
                        SPRINT1(0, "Computing finite difference sensitivities for ibody=%d",
                                ibody);
//$$$                        status = createPerturbation(MODL);
//$$$                        CHECK_STATUS(createPerturbation);
                        status = ocsmSetDtime(MODL, dtime);
                        CHECK_STATUS(ocsmSetDtime);

                        /* compare tessellation sensitivities */
                        node_distmax = 0;
                        edge_distmax = 0;
                        face_distmax = 0;

                        if (MODL->pmtr[ipmtr].nrow == 1 &&
                            MODL->pmtr[ipmtr].ncol == 1   ) {
                            SPRINT2(0, "\nComparing tessellation sensitivities wrt \"%s\" for ibody=%d",
                                    MODL->pmtr[ipmtr].name, ibody);
                        } else {
                            SPRINT4(0, "\nComparing tessellation sensitivities wrt \"%s[%d,%d]\" for ibody=%d",
                                    MODL->pmtr[ipmtr].name, irow, icol, ibody);
                        }

                        /* find the distance between the analytical tessellation sensitivity with
                           the perturbed Faces (made by finite differences) */
                        nerror = 0;
                        for (iface = 1; iface <= MODL->body[ibody].nface; iface++) {
                            status = EG_getTessFace(MODL->body[ibody].etess, iface,
                                                    &npnt_tess, &xyz, &uv, &ptype, &pindx,
                                                    &ntri_tess, &tric, &tric);
                            CHECK_STATUS(EG_getTessFace);

                            status = EG_attributeRet(MODL->perturb->body[ibody].ebody, ".fMap",
                                                     &itype, &nlist, &tempIlist, NULL, NULL);
                            if (status == SUCCESS) {
                                jface = tempIlist[iface-1];
                            } else {
                                jface = iface;
                            }

                            status = EG_getTolerance(MODL->perturb->body[ibody].face[iface].eface, &toler);
                            CHECK_STATUS(EG_getTolerance);

                            for (ipnt = 0; ipnt < npnt_tess; ipnt++) {
                                xyz_pred[0] = xyz[3*ipnt  ] + MODL->dtime * face_tes_anal[iface][3*ipnt  ];
                                xyz_pred[1] = xyz[3*ipnt+1] + MODL->dtime * face_tes_anal[iface][3*ipnt+1];
                                xyz_pred[2] = xyz[3*ipnt+2] + MODL->dtime * face_tes_anal[iface][3*ipnt+2];

                                status = EG_invEvaluate(MODL->perturb->body[ibody].face[jface].eface,
                                                        xyz_pred, uv_close, xyz_close);
                                CHECK_STATUS(EG_getTessFace);

                                dist = sqrt(SQR(xyz_pred[0]-xyz_close[0])
                                           +SQR(xyz_pred[1]-xyz_close[1])
                                           +SQR(xyz_pred[2]-xyz_close[2]));

                                /* check to see if any tessellation point is closer than the
                                   values returned by EG_invEvaluate */
                                if (dist > toler) {
                                    status = EG_getTessFace(MODL->perturb->body[ibody].etess, jface,
                                                            &npnt_tess, &xyz_ptrb, &uv_ptrb, &ptype, &pindx,
                                                            &ntri_tess, &tric, &tric);
                                    CHECK_STATUS(EG_getTessFace);

                                    iokay = 1;         /* means that EG_invEvaluate worked */
                                    for (jpnt = 0; jpnt < npnt_tess; jpnt++) {
                                        dist2 = sqrt(SQR(xyz_pred[0]-xyz_ptrb[3*jpnt  ])
                                                    +SQR(xyz_pred[1]-xyz_ptrb[3*jpnt+1])
                                                    +SQR(xyz_pred[2]-xyz_ptrb[3*jpnt+2]));
                                        if (dist2 < dist) {
                                            dist  = dist2;
                                            iokay = 0;
                                        }
                                    }
                                    if (iokay == 0) {
                                        SPRINT5(0, "WARNING:: EG_invEvaluate failed for ibody=%5d, jface=%5d, ipnt=%5d (%5d,%5d)",
                                                ibody, jface, ipnt, ptype[ipnt], pindx[ipnt]);
                                    }
                                }

                                if (dist > face_distmax) face_distmax = dist;
                                if (dist > EPS05) {
                                    if (nerror < 20 || dist >= face_distmax) {
                                        SPRINT6(0, "iface=%4d,  ipnt=%4d,  dist=%16.8f (at %10.4f %10.4f %10.4f)",
                                                iface, ipnt, dist, xyz[3*ipnt], xyz[3*ipnt+1], xyz[3*ipnt+2]);
                                    }
                                    nerror++;
                                    ntotal++;
                                }
                            }
                            FREE(face_tes_anal[iface]);
                        }
                        distmax = MAX(distmax, face_distmax);
                        SPRINT3(0, "    d(Face)/d(%s) check complete with %8d total errors (distmax=%12.4e)",
                                MODL->pmtr[ipmtr].name, nerror, face_distmax);

                        /* find the distance between the analytical tessellation sensitivity with
                           the perturbed Edges (made by finite differences) */

                        nerror = 0;
                        for (iedge = 1; iedge <= MODL->body[ibody].nedge; iedge++) {
                            status = EG_getTopology(MODL->body[ibody].edge[iedge].eedge,
                                                    &eref, &oclass, &mtype,
                                                    data, &nchild, &echilds, &senses);
                            CHECK_STATUS(EG_getTopology);

                            if (mtype == DEGENERATE) continue;

                            status = EG_getTessEdge(MODL->body[ibody].etess, iedge,
                                                    &npnt_tess, &xyz, &uv);
                            CHECK_STATUS(EG_getTessEdge);

                            status = EG_attributeRet(MODL->perturb->body[ibody].ebody, ".eMap",
                                                     &itype, &nlist, &tempIlist, NULL, NULL);
                            if (status == SUCCESS) {
                                jedge = tempIlist[iedge-1];
                            } else {
                                jedge = iedge;
                            }

                            status = EG_getTolerance(MODL->perturb->body[ibody].edge[iedge].eedge, &toler);
                            CHECK_STATUS(EG_getTolerance);

                            for (ipnt = 0; ipnt < npnt_tess; ipnt++) {
                                xyz_pred[0] = xyz[3*ipnt  ] + MODL->dtime * edge_tes_anal[iedge][3*ipnt  ];
                                xyz_pred[1] = xyz[3*ipnt+1] + MODL->dtime * edge_tes_anal[iedge][3*ipnt+1];
                                xyz_pred[2] = xyz[3*ipnt+2] + MODL->dtime * edge_tes_anal[iedge][3*ipnt+2];

                                status = EG_invEvaluate(MODL->perturb->body[ibody].edge[jedge].eedge,
                                                        xyz_pred, uv_close, xyz_close);
                                CHECK_STATUS(EG_invEvaluate);

                                dist = sqrt(SQR(xyz_pred[0]-xyz_close[0])
                                           +SQR(xyz_pred[1]-xyz_close[1])
                                           +SQR(xyz_pred[2]-xyz_close[2]));

                                /* check to see if either of the endpoints is closer than the
                                   values returned by EG_invEvaluate */
                                if (dist > toler) {
                                    status = EG_getRange(MODL->perturb->body[ibody].edge[jedge].eedge, trange, &periodic);
                                    CHECK_STATUS(EG_getRange);

                                    status = EG_evaluate(MODL->perturb->body[ibody].edge[jedge].eedge, &(trange[0]), data2);
                                    CHECK_STATUS(EG_evaluate);

                                    dist2 = sqrt(SQR(xyz_pred[0]-data2[0])
                                                +SQR(xyz_pred[1]-data2[1])
                                                +SQR(xyz_pred[2]-data2[2]));
                                    if (dist2 < dist) {
                                        SPRINT3(0, "WARNING:: EG_invEvaluate failed for ibody=%5d, jedge=%5d, ipnt=%5d",
                                                ibody, jedge, ipnt);

                                        dist = dist2;
                                    }

                                    status = EG_evaluate(MODL->perturb->body[ibody].edge[jedge].eedge, &(trange[1]), data2);
                                    CHECK_STATUS(EG_evaluate);

                                    dist2 = sqrt(SQR(xyz_pred[0]-data2[0])
                                                +SQR(xyz_pred[1]-data2[1])
                                                +SQR(xyz_pred[2]-data2[2]));
                                    if (dist2 < dist) {
                                        SPRINT3(0, "WARNING:: EG_invEvaluate failed for ibody=%5d, jedge=%5d, ipnt=%5d",
                                                ibody, jedge, ipnt);

                                        dist = dist2;
                                    }
                                }

                                if (dist > edge_distmax) edge_distmax = dist;
                                if (dist > EPS05) {
                                    if (nerror < 20 || dist >= edge_distmax) {
                                        SPRINT6(0, "iedge=%4d,  ipnt=%4d,  dist=%16.8f (at %10.4f %10.4f %10.4f)",
                                                iedge, ipnt, dist, xyz[3*ipnt], xyz[3*ipnt+1], xyz[3*ipnt+2]);
                                    }
                                    nerror++;
                                    ntotal++;
                                }
                            }
                            FREE(edge_tes_anal[iedge]);
                        }
                        distmax = MAX(distmax, edge_distmax);
                        SPRINT3(0, "    d(Edge)/d(%s) check complete with %8d total errors (distmax=%12.4e)",
                                MODL->pmtr[ipmtr].name, nerror, edge_distmax);

                        /* find the distance between the analytical tessellation sensitivity with
                           the perturbed Nodes (made by finite differences) */
                        nerror = 0;
                        for (inode = 1; inode <= MODL->body[ibody].nnode; inode++) {
                            npnt_tess = 1;

                            /* if .nMap exists, use mapped Node location in the perturbed Body */
                            status = EG_attributeRet(MODL->perturb->body[ibody].ebody, ".nMap",
                                                     &itype, &nlist, &tempIlist, NULL, NULL);
                            if (status == SUCCESS) {
                                jnode = tempIlist[inode-1];
                            } else {
                                jnode = inode;
                            }

                            for (ipnt = 0; ipnt < npnt_tess; ipnt++) {
                                xyz_pred[0] = MODL->body[ibody].node[inode].x + MODL->dtime * node_tes_anal[inode][0];
                                xyz_pred[1] = MODL->body[ibody].node[inode].y + MODL->dtime * node_tes_anal[inode][1];
                                xyz_pred[2] = MODL->body[ibody].node[inode].z + MODL->dtime * node_tes_anal[inode][2];

                                dist = sqrt(SQR(xyz_pred[0]-MODL->perturb->body[ibody].node[jnode].x)
                                           +SQR(xyz_pred[1]-MODL->perturb->body[ibody].node[jnode].y)
                                           +SQR(xyz_pred[2]-MODL->perturb->body[ibody].node[jnode].z));
                                if (dist > node_distmax) node_distmax = dist;
                                if (dist > EPS05) {
                                    if (nerror < 20 || dist >= node_distmax) {
                                        SPRINT6(0, "inode=%4d,  ipnt=%4d,  dist=%16.8f (at %10.4f %10.4f %10.4f)",
                                                inode, ipnt, dist,
                                                MODL->body[ibody].node[inode].x,
                                                MODL->body[ibody].node[inode].y,
                                                MODL->body[ibody].node[inode].z);
                                    }
                                    nerror++;
                                    ntotal++;
                                }
                            }

                            FREE(node_tes_anal[inode]);
                        }
                        distmax = MAX(distmax, node_distmax);
                        SPRINT3(0, "    d(Node)/d(%s) check complete with %8d total errors (distmax=%12.4e)",
                                MODL->pmtr[ipmtr].name, nerror, node_distmax);

                        FREE(face_tes_anal);
                        FREE(edge_tes_anal);
                        FREE(node_tes_anal);

//$$$                        status = removePerturbation(MODL);
//$$$                        CHECK_STATUS(removePerturbation);
                        status = ocsmSetDtime(MODL, 0);
                        CHECK_STATUS(ocsmSetDtime);
                    }
//$$$                }
            }
        }
        SPRINT0(0, " ");
    }

    SPRINT0(0, "==> sensCSM completed successfully");
    status = EXIT_SUCCESS;

    /* cleanup and exit */
cleanup:
    context = MODL->context;

    /* remove all Bodys and etess objects */
    for (ibody = 1; ibody <= MODL->nbody; ibody++) {
        if (MODL->body[ibody].etess != NULL) {
            status2 = EG_deleteObject(MODL->body[ibody].etess);
            SPRINT3(2, "--> EG_deleteObject(etess[%d]) -> status=%d (%s)",
                    ibody, status2, ocsmGetText(status2));

            MODL->body[ibody].etess = NULL;
        }

        if (MODL->body[ibody].ebody != NULL) {
            status2 = EG_deleteObject(MODL->body[ibody].ebody);
            SPRINT3(2, "--> EG_deleteObject(ebody[%d]) => status=%d (%s)",
                    ibody, status2, ocsmGetText(status2));

            MODL->body[ibody].ebody = NULL;
        }
    }

    /* free up the modl */
    status2 = ocsmFree(modl);
    SPRINT2(1, "--> ocsmFree() -> status=%d (%s)",
            status2, ocsmGetText(status2));

    /* remove tmp files (if they exist) and clean up the udp storage */
    status2 = ocsmFree(NULL);
    SPRINT2(1, "--> ocsmFree(NULL) -> status=%d (%s)",
            status2, ocsmGetText(status2));

    /* remove hte context */
    if (context != NULL) {
        status2 = EG_setOutLevel(context, 0);
        if (status2 < 0) {
            SPRINT2(0, "EG_setOutLevel -> status=%d (%s)",
                    status2, ocsmGetText(status2));
        }

        status2 = EG_close(context);
        SPRINT2(1, "--> EG_close() -> status=%d (%s)",
                status2, ocsmGetText(status2));
    }

    /* report final statistics */
    if (status == SUCCESS) {
        SPRINT3(0, "\nSensitivity checks complete with %8d total errors (errmax=%12.4e, distmax=%12.4e)",
                ntotal, errmax+1.0e-20, distmax+1.0e-20);
    } else {
        SPRINT1(0, "\nSensitivity checks not complete because error \"%s\" was detected",
                ocsmGetText(status));
    }

    /* these statements are in case we used an error return to go to cleanup */
    if (face_cfg_anal != NULL) free(face_cfg_anal);
    if (face_cfg_fdif != NULL) free(face_cfg_fdif);
    if (edge_cfg_anal != NULL) free(edge_cfg_anal);
    if (edge_cfg_fdif != NULL) free(edge_cfg_fdif);
    if (node_cfg_anal != NULL) free(node_cfg_anal);
    if (node_cfg_fdif != NULL) free(node_cfg_fdif);

    if (status < 0) status = EXIT_FAILURE;

    return status;
}
