/*
 ************************************************************************
 *                                                                      *
 * udfEditAttr -- edit Attributes in a Body                             *
 *                                                                      *
 *            Written by John Dannenhoffer @ Syracuse University        *
 *            Patterned after code written by Bob Haimes  @ MIT         *
 *                                                                      *
 ************************************************************************
 */

/*
 * Copyright (C) 2011/2018  John F. Dannenhoffer, III (Syracuse University)
 *
 * This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *     MA  02110-1301  USA
 */

#define NUMUDPARGS 6
#define NUMUDPINPUTBODYS 1
#include "udpUtilities.h"

/* shorthands for accessing argument values and velocities */
#define ATTRNAME( IUDP)  ((char   *) (udps[IUDP].arg[0].val))
#define INPUT(    IUDP)  ((char   *) (udps[IUDP].arg[1].val))
#define OUTPUT(   IUDP)  ((char   *) (udps[IUDP].arg[2].val))
#define OVERWRITE(IUDP)  ((int    *) (udps[IUDP].arg[3].val))[0]
#define FILENAME( IUDP)  ((char   *) (udps[IUDP].arg[4].val))
#define NCHANGE(  IUDP)  ((double *) (udps[IUDP].arg[5].val))[0]

/* data about possible arguments */
static char  *argNames[NUMUDPARGS] = {"attrname",  "input",     "output",   "overwrite", "filename", "nchange", };
static int    argTypes[NUMUDPARGS] = {ATTRSTRING,  ATTRSTRING,  ATTRSTRING, ATTRINT,     ATTRSTRING, -ATTRREAL  };
static int    argIdefs[NUMUDPARGS] = {0,           0,           0,          0,           0,          0,         };
static double argDdefs[NUMUDPARGS] = {0.,          0.,          0.,         0.,          0.,         0.,        };

/* get utility routines: udpErrorStr, udpInitialize, udpReset, udpSet,
                         udpGet, udpVel, udpClean, udpMesh */
#include "udpUtilities.c"

#define           MIN(A,B)        (((A) < (B)) ? (A) : (B))
#define           MAX(A,B)        (((A) < (B)) ? (B) : (A))

/* prototype for function defined below */
static int editAttrs(ego ebody, char attrname[], char input[], char output[],
                     int overwrite, int *nchange);
static int processFile(ego ebody, char filename[], int *nchange);
static int matches(char pattern[], const char string[]);
static int editEgo(const char attrname[], int atype, int alen,
                     const int *tempIlist, const double *tempRlist, const char *tempClist,
                     ego eobj, int overwrite, int *nchange);
static int getToken(char *text, int nskip, char sep, int maxtok, char *token);
static int getBodyTopos(ego ebody, ego esrc, int oclass, int *nlist, ego **elist);


/*
 ************************************************************************
 *                                                                      *
 *   udpExecute - execute the primitive                                 *
 *                                                                      *
 ************************************************************************
 */

int
udpExecute(ego  emodel,                 /* (in)  Model containing Body */
           ego  *ebody,                 /* (out) Body pointer */
           int  *nMesh,                 /* (out) number of associated meshes */
           char *string[])              /* (out) error message */
{
    int     status = EGADS_SUCCESS;

    int     oclass, mtype, nchild, *senses, nchange;
    double  data[4];
    ego     context, eref, *ebodys;

    ROUTINE(udpExecute);

    /* --------------------------------------------------------------- */

#ifdef DEBUG
    printf("udpExecute(emodel=%llx)\n", (long long)emodel);
    printf("attrname( 0) = %s\n", ATTRNAME( 0));
    printf("input(    0) = %s\n", INPUT(    0));
    printf("output(   0) = %s\n", OUTPUT(   0));
    printf("overwrite(0) = %d\n", OVERWRITE(0));
    printf("filename( 0) = %s\n", FILENAME( 0));
#endif

    /* default return values */
    *ebody  = NULL;
    *nMesh  = 0;
    *string = NULL;

    /* check that Model was input that contains on Body */
    status = EG_getTopology(emodel, &eref, &oclass, &mtype,
                            data, &nchild, &ebodys, &senses);
    if (status < EGADS_SUCCESS) goto cleanup;

    if (oclass != MODEL) {
        printf(" udpExecute: expecting a Model\n");
        status = EGADS_NOTMODEL;
        goto cleanup;
    } else if (nchild != 1) {
        printf(" udpExecute: expecting Model to contain one Body (not %d)\n", nchild);
        status = EGADS_NOTBODY;
        goto cleanup;
    }

    status = EG_getContext(emodel, &context);
    if (status < EGADS_SUCCESS) goto cleanup;

    /* if filename is specified, then process it and skip everything else */
    if (STRLEN(FILENAME(0)) > 0) {

    /* check arguments */
    } else {
        if        (STRLEN(INPUT( 0)) == 1) {
        } else if (STRLEN(OUTPUT(0)) == 1) {
        } else if (STRLEN(INPUT( 0)) != STRLEN(OUTPUT(0))) {
            printf(" udpExecute: input and output should be same length\n");
            status  = EGADS_RANGERR;
            goto cleanup;
        }

        if        (udps[0].arg[3].size >= 2) {
            printf(" udpExecute: overwrite should be a scalar\n");
            status  = EGADS_RANGERR;
            goto cleanup;

        } else if (NINT(OVERWRITE(0)) < 0 && NINT(OVERWRITE(1)) > 4) {
            printf(" udpExecute: overwrite = %d should be between 0 and 4\n", OVERWRITE(0));
            status  = EGADS_RANGERR;
            goto cleanup;

        }
    }

    /* cache copy of arguments for future use */
    status = cacheUdp();
    if (status < 0) {
        printf(" udpExecute: problem caching arguments\n");
        goto cleanup;
    }

#ifdef DEBUG
    printf("attrname( %d) = %s\n", numUdp, ATTRNAME( numUdp));
    printf("input(    %d) = %s\n", numUdp, INPUT(    numUdp));
    printf("output(   %d) = %s\n", numUdp, OUTPUT(   numUdp));
    printf("overwrite(%d) = %d\n", numUdp, OVERWRITE(numUdp));
    printf("filename( %d) = %s\n", numUdp, FILENAME( numUdp));
#endif

    /* make a copy of the Body (so that it does not get removed
     when OpenCSM deletes emodel) */
    status = EG_copyObject(ebodys[0], NULL, ebody);
    if (status < EGADS_SUCCESS) goto cleanup;

    /* edit the attributes */
    if (STRLEN(FILENAME(numUdp)) > 0) {
        status = processFile(*ebody, FILENAME(numUdp), &nchange);
        if (status < 0) {
            printf(" udpExecute: problem in processFile\n");
            goto cleanup;
        }
    } else {
        status = editAttrs(*ebody, ATTRNAME(numUdp), INPUT(numUdp), OUTPUT(numUdp), OVERWRITE(numUdp), &nchange);
        if (status < 0) {
            printf(" udpExecute: problem in editAttrs\n");
            goto cleanup;
        }
    }

    /* add a special Attribute to the Body to tell OpenCSM that there
       is no topological change and hence it should not adjust the
       Attributes on the Body in finishBody() */
    status = EG_attributeAdd(*ebody, "__noTopoChange__", ATTRSTRING,
                             0, NULL, NULL, "udfEditAttr");
    if (status < 0) {
        printf(" udpExecute: problem setting __noTopoChange__\n");
        goto cleanup;
    }

    /* set the output value */
    NCHANGE(0) = (double)nchange;

    /* the copy of the Body that was annotated is returned */
    udps[numUdp].ebody = *ebody;

cleanup:
    if (status != EGADS_SUCCESS) {
        *string = udpErrorStr(status);
    }

    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpSensitivity - return sensitivity derivatives for the "real" argument *
 *                                                                      *
 ************************************************************************
 */

int
udpSensitivity(ego    ebody,            /* (in)  Body pointer */
   /*@unused@*/int    npnt,             /* (in)  number of points */
   /*@unused@*/int    entType,          /* (in)  OCSM entity type */
   /*@unused@*/int    entIndex,         /* (in)  OCSM entity index (bias-1) */
   /*@unused@*/double uvs[],            /* (in)  parametric coordinates for evaluation */
   /*@unused@*/double vels[])           /* (out) velocities */
{
    int iudp, judp;

    ROUTINE(udpSensitivity);

    /* --------------------------------------------------------------- */

    /* check that ebody matches one of the ebodys */
    iudp = 0;
    for (judp = 1; judp <= numUdp; judp++) {
        if (ebody == udps[judp].ebody) {
            iudp = judp;
            break;
        }
    }
    if (iudp <= 0) {
        return EGADS_NOTMODEL;
    }

    /* this routine is not written yet */
    return EGADS_NOLOAD;
}


/*
 ************************************************************************
 *                                                                      *
 *   editAttrs - edit the Attributes                                    *
 *                                                                      *
 ************************************************************************
 */

static int
editAttrs(ego    ebody,                 /* (in)  EGADS Body */
          char   attrname[],            /* (in)  name of attribute to edit */
          char   input[],               /* (in)  types of entity to edit from */
          char   output[],              /* (in)  types of entity to edit to */
          int    overwrite,             /* (in)  =0 do not overwrite */
                                        /*       =1        overwrite */
                                        /*       =2        use smaller value */
                                        /*       =3        use larger  value */
                                        /*       =4        use sum or concatenation */
          int    *nchange)              /* (out) number of changes */
{
    int       status = 0;               /* (out) return status */

    int       nnode, inode, nedge, iedge, nface, iface, nlist, ilist;
    int       nedit, iedit, nattr, iattr, atype, alen;
    CINT      *tempIlist;
    CDOUBLE   *tempRlist;
    char      intype, outtype;
    CCHAR     *tempClist, *aname;
    ego       *enodes=NULL, *eedges=NULL, *efaces=NULL, *elist=NULL;

    ROUTINE(editAttrs);

    /* --------------------------------------------------------------- */

    /* initialize the output */
    *nchange = 0;

    /* get the list of Nodes, Edges, and Faces associated with ebody */
    status = EG_getBodyTopos(ebody, NULL, NODE, &nnode, &enodes);
    if (status < EGADS_SUCCESS) goto cleanup;

    status = EG_getBodyTopos(ebody, NULL, EDGE, &nedge, &eedges);
    if (status < EGADS_SUCCESS) goto cleanup;

    status = EG_getBodyTopos(ebody, NULL, FACE, &nface, &efaces);
    if (status < EGADS_SUCCESS) goto cleanup;

    /* determone number of edits */
    nedit = MAX(STRLEN(input), STRLEN(output));

    /* loop through all the edits */
    for (iedit = 0; iedit < nedit; iedit++) {
        if (STRLEN(input) > 1) {
            intype = input[iedit];
        } else {
            intype = input[0];
        }

        if (STRLEN(output) > 1) {
            outtype = output[iedit];
        } else {
            outtype = output[0];
        }

        /* delete an attribute */
        if (intype == 'd' || intype == 'D') {

            /* delete Body attribute */
            if (outtype == 'b' || outtype == 'B') {
                status = EG_attributeNum(ebody, &nattr);
                CHECK_STATUS(EG_attributeNum);

                for (iattr = nattr; iattr > 0; iattr--) {
                    status = EG_attributeGet(ebody, iattr, &aname, &atype, &alen,
                                             &tempIlist, &tempRlist, &tempClist);
                    CHECK_STATUS(EG_attributeGet);

                    if (matches(attrname, aname)) {
                        status = EG_attributeDel(ebody, aname);
                        CHECK_STATUS(EG_attributeDel);

                        (*nchange)++;
                    }
                }

            /* delete Node attribute */
            } else if (outtype == 'n' || outtype == 'N') {
                for (inode = 0; inode < nnode; inode++) {
                    status = EG_attributeNum(enodes[inode], &nattr);
                    CHECK_STATUS(EG_attributeNum);

                    for (iattr = nattr; iattr > 0; iattr--) {
                        status = EG_attributeGet(enodes[inode], iattr, &aname, &atype, &alen,
                                                 &tempIlist, &tempRlist, &tempClist);
                        CHECK_STATUS(EG_attributeGet);

                        if (matches(attrname, aname)) {
                            status = EG_attributeDel(enodes[inode], aname);
                            CHECK_STATUS(EG_attributeDel);

                            (*nchange)++;
                        }
                    }
                }

            /* delete Edge attribute */
            } else if (outtype == 'e' || outtype == 'E') {
                for (iedge = 0; iedge < nedge; iedge++) {
                    status = EG_attributeNum(eedges[iedge], &nattr);
                    CHECK_STATUS(EG_attributeNum);

                    for (iattr = nattr; iattr > 0; iattr--) {
                        status = EG_attributeGet(eedges[iedge], iattr, &aname, &atype, &alen,
                                                 &tempIlist, &tempRlist, &tempClist);
                        CHECK_STATUS(EG_attributeGet);

                        if (matches(attrname, aname)) {
                            status = EG_attributeDel(eedges[iedge], aname);
                            CHECK_STATUS(EG_attributeDel);

                            (*nchange)++;
                        }
                    }
                }

            /* delete Face attribute */
            } else if (outtype == 'f' || outtype == 'F') {
                for (iface = 0; iface < nface; iface++) {
                    status = EG_attributeNum(efaces[iface], &nattr);
                    CHECK_STATUS(EG_attributeNum);

                    for (iattr = nattr; iattr > 0; iattr--) {
                        status = EG_attributeGet(efaces[iface], iattr, &aname, &atype, &alen,
                                                 &tempIlist, &tempRlist, &tempClist);
                        CHECK_STATUS(EG_attributeGet);

                        if (matches(attrname, aname)) {
                            status = EG_attributeDel(efaces[iface], aname);
                            CHECK_STATUS(EG_attributeDel);

                            (*nchange)++;
                        }
                    }
                }
            } else {
                printf("outype=%c is not valid for intype=D\n", outtype);
                status = -999;
                goto cleanup;
            }

        /* propagate attribute from the Body */
        } else if (intype == 'b' || intype == 'B') {

            /* look through all Body Attributes, looking for a match */
            status = EG_attributeNum(ebody, &nattr);
            CHECK_STATUS(EG_attributeNum);

            for (iattr = 1; iattr <= nattr; iattr++) {
                status = EG_attributeGet(ebody, iattr, &aname, &atype, &alen,
                                         &tempIlist, &tempRlist, &tempClist);
                CHECK_STATUS(EG_attributeGet);

                if (matches(attrname, aname)) {

                    /* propagate from Body to Body */
                    if (outtype == 'b' || outtype == 'B') {

                    /* propagate from Body to Node */
                    } else if (outtype == 'n' || outtype == 'N') {
                        for (inode = 0; inode < nnode; inode++) {
                            status = editEgo(aname, atype, alen, tempIlist, tempRlist, tempClist,
                                             enodes[inode], overwrite, nchange);
                            CHECK_STATUS(editEgo);
                        }

                    /* propagate from Body to Edge */
                    } else if (outtype == 'e' || outtype == 'E') {
                        for (iedge = 0; iedge < nedge; iedge++) {
                            status = editEgo(aname, atype, alen, tempIlist, tempRlist, tempClist,
                                             eedges[iedge], overwrite, nchange);
                            CHECK_STATUS(editEgo);
                        }

                    /* propagate from Body to Face */
                    } else if (outtype == 'f' || outtype == 'F') {
                        for (iface = 0; iface < nface; iface++) {
                            status = editEgo(aname, atype, alen, tempIlist, tempRlist, tempClist,
                                             efaces[iface], overwrite, nchange);
                            CHECK_STATUS(editEgo);
                        }

                    } else {
                        printf("outtype=%c is not valid for intype=B\n", outtype);
                        status = -999;
                        goto cleanup;
                    }
                }
            }

        /* edit propagate attribute from the Nodes */
        } else if (intype == 'n' || intype == 'N') {
            for (inode = 0; inode < nnode; inode++) {

                /* look through all Node Attributes, looking for a match */
                status = EG_attributeNum(enodes[inode], &nattr);
                CHECK_STATUS(EG_attributeNum);

                for (iattr = 1; iattr <= nattr; iattr++) {
                    status = EG_attributeGet(enodes[inode], iattr, &aname, &atype, &alen,
                                             &tempIlist, &tempRlist, &tempClist);
                    CHECK_STATUS(EG_attributeGet);

                    if (matches(attrname, aname)) {

                        /* propagate from Node to Body */
                        if (outtype == 'b' || outtype == 'B') {
                            status = editEgo(aname, atype, alen, tempIlist, tempRlist, tempClist,
                                             ebody, overwrite, nchange);
                            CHECK_STATUS(editEgo);

                        /* propagate from Node to Node */
                        } else if (outtype == 'n' || outtype == 'N') {

                        /* propagate from Node to Edge */
                        } else if (outtype == 'e' || outtype == 'E') {
                            status = EG_getBodyTopos(ebody, enodes[inode], EDGE, &nlist, &elist);
                            CHECK_STATUS(EG_getBodyTopos);

                            for (ilist = 0; ilist < nlist; ilist++) {
                                status = editEgo(aname, atype, alen, tempIlist, tempRlist, tempClist,
                                                 elist[ilist], overwrite, nchange);
                                CHECK_STATUS(editEgo);
                            }

                            EG_free(elist);
                            elist = NULL;

                        /* propagate from Node to Face */
                        } else if (outtype == 'f' || outtype == 'F') {
                            status = EG_getBodyTopos(ebody, enodes[inode], FACE, &nlist, &elist);
                            CHECK_STATUS(EG_getBodyTopos);

                            for (ilist = 0; ilist < nlist; ilist++) {
                                status = editEgo(aname, atype, alen, tempIlist, tempRlist, tempClist,
                                                 elist[ilist], overwrite, nchange);
                                CHECK_STATUS(editEgo);
                            }

                            EG_free(elist);
                            elist = NULL;

                        } else {
                            printf("outtype=%c is not valid for intype=N\n", outtype);
                            status = -999;
                            goto cleanup;
                        }
                    }
                }
            }

        /* propagate attribute from the Edges */
        } else if (intype == 'e' || intype == 'E') {
            for (iedge = 0; iedge < nedge; iedge++) {

                /* look through all Edge Attributes, looking for a match */
                status = EG_attributeNum(eedges[iedge], &nattr);
                CHECK_STATUS(EG_attributeNum);

                for (iattr = 1; iattr <= nattr; iattr++) {
                    status = EG_attributeGet(eedges[iedge], iattr, &aname, &atype, &alen,
                                             &tempIlist, &tempRlist, &tempClist);
                    CHECK_STATUS(EG_attributeGet);

                    if (matches(attrname, aname)) {

                        /* propagate from Edge to Body */
                        if (outtype == 'b' || outtype== 'B') {
                            status = editEgo(aname, atype, alen, tempIlist, tempRlist, tempClist,
                                             ebody, overwrite, nchange);
                            CHECK_STATUS(editEgo);

                        /* propagate from Edge to Node */
                        } else if (outtype == 'n' || outtype == 'N') {
                            status = EG_getBodyTopos(ebody, eedges[iedge], NODE, &nlist, &elist);
                            CHECK_STATUS(EG_getBodyTopos);

                            for (ilist = 0; ilist < nlist; ilist++) {
                                status = editEgo(aname, atype, alen, tempIlist, tempRlist, tempClist,
                                                 elist[ilist], overwrite, nchange);
                                CHECK_STATUS(editEgo);
                            }

                            EG_free(elist);
                            elist = NULL;

                        /* propagate from Edge to Edge */
                        } else if (outtype == 'e' || outtype == 'E') {

                        /* propagate from Edge to Face */
                        } else if (outtype == 'f' || outtype == 'F') {
                            status = EG_getBodyTopos(ebody, eedges[iedge], FACE, &nlist, &elist);
                            CHECK_STATUS(EG_getBodyTopos);

                            for (ilist = 0; ilist < nlist; ilist++) {
                                status = editEgo(aname, atype, alen, tempIlist, tempRlist, tempClist,
                                                 elist[ilist], overwrite, nchange);
                                CHECK_STATUS(editEgo);
                            }

                            EG_free(elist);
                            elist = NULL;

                        } else {
                            printf("outtype=%c is not valid for intype=E\n", outtype);
                            status = -999;
                            goto cleanup;
                        }
                    }
                }
            }

        /* propagate attribute from the Faces */
        } else if (intype == 'f' || intype == 'F') {
            for (iface = 0; iface < nface; iface++) {

                /* look through all Face Attributes, looking for a match */
                status = EG_attributeNum(efaces[iface], &nattr);
                CHECK_STATUS(EG_attributeNum);

                for (iattr = 1; iattr <= nattr; iattr++) {
                    status = EG_attributeGet(efaces[iface], iattr, &aname, &atype, &alen,
                                             &tempIlist, &tempRlist, &tempClist);
                    CHECK_STATUS(EG_attributeGet);

                    if (matches(attrname, aname)) {

                        /* propagate from Face to to Body */
                        if (outtype == 'b' || outtype == 'B') {
                            status = editEgo(aname, atype, alen, tempIlist, tempRlist, tempClist,
                                             ebody, overwrite, nchange);
                            CHECK_STATUS(editEgo);

                        /* propagate from Face edit to Node */
                        } else if (outtype == 'n' || outtype == 'N') {
                            status = EG_getBodyTopos(ebody, efaces[iface], NODE, &nlist, &elist);
                            CHECK_STATUS(EG_getBodyTopos);

                            for (ilist = 0; ilist < nlist; ilist++) {
                                status = editEgo(aname, atype, alen, tempIlist, tempRlist, tempClist,
                                                 elist[ilist], overwrite, nchange);
                                CHECK_STATUS(editEgo);
                            }

                            EG_free(elist);
                            elist = NULL;

                        /* propagate from Face to Edge */
                        } else if (outtype == 'e' || outtype == 'E') {
                            status = EG_getBodyTopos(ebody, efaces[iface], EDGE, &nlist, &elist);
                            CHECK_STATUS(EG_getBodyTopos);

                            for (ilist = 0; ilist < nlist; ilist++) {
                                status = editEgo(aname, atype, alen, tempIlist, tempRlist, tempClist,
                                                 elist[ilist], overwrite, nchange);
                                CHECK_STATUS(editEgo);
                            }

                            EG_free(elist);
                            elist = NULL;

                        /* propagate from Face to Face */
                        } else if (outtype == 'f' || outtype == 'F') {

                        } else {
                            printf("outtype=%c is not valid for intype=F\n", outtype);
                            status = -999;
                            goto cleanup;
                        }
                    }
                }
            }

        } else {
            printf("intype=%c is not valid\n", intype);
            status = -999;
            goto cleanup;
        }
    }

cleanup:
    if (enodes != NULL) EG_free(enodes);
    if (eedges != NULL) EG_free(eedges);
    if (efaces != NULL) EG_free(efaces);
    if (elist  != NULL) EG_free(elist );

    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   processFile - special file processing mode                         *
 *                                                                      *
 ************************************************************************
 */

static int
processFile(ego    ebody,               /* (in)  EGADS Body */
            char   filename[],          /* (in)  file that contains directives */
            int    *nchange)            /* (out) number of changes */
{
    int       status = 0;               /* (out) return status */

    int       nnode, nedge, nface, isel, nsel, istype, itoken;
    int       nlist, ilist, nnbor, inbor, iremove, atype, alen;
    int       status1, status2;
    CINT      *tempIlist;
    CDOUBLE   *tempRlist;
    char      templine[256], token[256], attrName[256], attrValu[256], newValu[256];
    CCHAR     *tempClist;
    ego       *esel=NULL, *elist=NULL, *enbor=NULL;
    FILE      *fp;

    ROUTINE(processFile);

    /* --------------------------------------------------------------- */

    /* default return */
    *nchange = 0;

    /* find number of Nodes, Edges, and Faces in ebody */
    status = EG_getBodyTopos(ebody, NULL, NODE, &nnode, NULL);
    if (status < EGADS_SUCCESS) goto cleanup;

    status = EG_getBodyTopos(ebody, NULL, EDGE, &nedge, NULL);
    if (status < EGADS_SUCCESS) goto cleanup;

    status = EG_getBodyTopos(ebody, NULL, FACE, &nface, NULL);
    if (status < EGADS_SUCCESS) goto cleanup;

    /* allocate a list of EGOs that is big enough for all Nodes, Edges, or Faces */
    if (nnode > nedge && nnode > nface) {
        nsel = nnode;
    } else if (nedge > nface) {
        nsel = nedge;
    } else {
        nsel = nface;
    }

    esel = (ego*) malloc(nsel*sizeof(ego));
    nsel = 0;

    istype = 0;

    /* open the file */
    fp = fopen(filename, "r");
    if (fp == NULL) {
        printf(" processFile: could not open file \"%s\"\n", filename);
        status = EGADS_NOTFOUND;
        goto cleanup;
    }

    /* read until end of file */
    while (!feof(fp)) {

        /* read the next line */
        (void) fgets(templine, 255, fp);

        if (feof(fp)) break;

        printf("    processing: %s", templine);

        if (templine[0] == '#') continue;

        /* overwrite the \n */
        if (templine[STRLEN(templine)-1] == '\n') {
            templine[STRLEN(templine)-1] = '\0';
        }

        /* get and process the first token */
        itoken = 0;
        status = getToken(templine, itoken, ' ', 255, token);

        if (status == 0) {              /* null token */
            continue;

        } else if (strcmp(token, "END") == 0 ||
                   strcmp(token, "end") == 0   ) {
            break;

        } else if (strcmp(token, "FACE") == 0 ||
                   strcmp(token, "face") == 0   ) {
            istype = OCSM_FACE;
            status = EG_getBodyTopos(ebody, NULL, FACE, &nlist, &elist);
            if (status < EGADS_SUCCESS) goto cleanup;

            nsel = nlist;
            for (ilist = 0; ilist < nlist; ilist++) {
                esel[ilist] = elist[ilist];
            }

            EG_free(elist);
            elist = NULL;

        } else if (strcmp(token, "EDGE") == 0 ||
                   strcmp(token, "edge") == 0   ) {
            istype = OCSM_EDGE;
            status = EG_getBodyTopos(ebody, NULL, EDGE, &nlist, &elist);
            if (status < EGADS_SUCCESS) goto cleanup;

            nsel = nlist;
            for (ilist = 0; ilist < nlist; ilist++) {
                esel[ilist] = elist[ilist];
            }

            EG_free(elist);
            elist = NULL;

        } else if (strcmp(token, "NODE") == 0 ||
                   strcmp(token, "node") == 0   ) {
            istype = OCSM_NODE;
            status = EG_getBodyTopos(ebody, NULL, NODE, &nlist, &elist);
            if (status < EGADS_SUCCESS) goto cleanup;

            nsel = nlist;
            for (ilist = 0; ilist < nlist; ilist++) {
                esel[ilist] = elist[ilist];
            }

            EG_free(elist);
            elist = NULL;

        } else if (strcmp(token, "AND") == 0 ||
                   strcmp(token, "and") == 0   ) {
            if (istype == 0) {
                printf(" processFile: AND has to follow NODE, EDGE, FACE, or AND\n");
                status = -999;
                goto cleanup;
            }

        } else if (strcmp(token, "SET") == 0 ||
                   strcmp(token, "set") == 0   ) {

            /* apply the Attribute name/value to everything in esel */
            itoken = 0;
            while (1) {
                status = getToken(templine, ++itoken, ' ', 255, token);
                if (status <= 0) break;

                status1 = getToken(token, 0, '=', 255, attrName);
                status2 = getToken(token, 1, '=', 255, attrValu);

                if (status1 <= 0 || status2 <= 0) {
                    printf(" processFile: token is not name=value\n");
                    status = -999;
                    goto cleanup;
                } else if (istype == 0) {
                    printf(" processFile: nothing selected, so not setting %s=%s\n",
                           attrName, attrValu);
                    break;
                }

                for (isel = 0; isel < nsel; isel++) {
                    status = EG_attributeAdd(esel[isel], attrName, ATTRSTRING, 0,
                                             NULL, NULL, attrValu);
                    if (status < EGADS_SUCCESS) goto cleanup;
                    (*nchange)++;
                }
            }

            continue;

        } else if (strcmp(token, "ADD") == 0 ||
                   strcmp(token, "add") == 0   ) {

            /* append the Attribute name/value to everything in esel */
            itoken = 0;
            while (1) {
                status = getToken(templine, ++itoken, ' ', 255, token);
                if (status <= 0) break;

                status1 = getToken(token, 0, '=', 255, attrName);
                status2 = getToken(token, 1, '=', 255, attrValu);

                if (status1 <= 0 || status2 <= 0) {
                    printf(" processFile: token is not name=value\n");
                    status = -999;
                    goto cleanup;
                }

                for (isel = 0; isel < nsel; isel++) {
                    status = EG_attributeRet(esel[isel], attrName, &atype, &alen,
                                             &tempIlist, &tempRlist, &tempClist);

                    if (status == EGADS_SUCCESS && atype == ATTRSTRING) {
                        snprintf(newValu, 255, "%s;%s", tempClist, attrValu);
                    } else {
                        snprintf(newValu, 255, "%s",               attrValu);
                    }

                    status = EG_attributeAdd(esel[isel], attrName, ATTRSTRING, 0,
                                             NULL, NULL, newValu);
                    if (status < EGADS_SUCCESS) goto cleanup;
                    (*nchange)++;
                }
            }

            continue;

        } else {
            printf(" processFile: unexpected token \"%s\"\n", token);
            status =  EGADS_NONAME;
            goto cleanup;
        }

        /* do not process rest of tokens if nothing is selected */
        if (nsel <= 0) {
            istype = 0;
            continue;
        }

        /* process the rest of the tokens */
        status = getToken(templine, 1, ' ', 255, token);

        if        (strcmp(token, "HAS") == 0 ||
                   strcmp(token, "has") == 0   ) {

            /* remove entries from esel if it does not match name/value pairs */
            for (isel = nsel-1; isel >= 0; isel--) {
                for (itoken = 2; itoken < 10; itoken++) {

                    /* get the name/value */
                    status = getToken(templine, itoken, ' ', 255, token);
                    if (status <= 0) break;

                    status1 = getToken(token, 0, '=', 255, attrName);
                    status2 = getToken(token, 1, '=', 255, attrValu);

                    if (status1 <= 0 || status2 <= 0) {
                        printf(" processFile: token is not name=value\n");
                        status = -999;
                        goto cleanup;
                    }
                    /* remove entry from esel if it does not match name/value */
                    status = EG_attributeRet(esel[isel], attrName,
                                             &atype, &alen, &tempIlist, &tempRlist, &tempClist);
                    if (status != EGADS_SUCCESS) {
                        esel[isel] = esel[nsel-1];
                        nsel--;
                        break;
                    } else if (atype != ATTRSTRING) {
                        esel[isel] = esel[nsel-1];
                        nsel--;
                        break;
                    } else if (matches(attrValu, tempClist) == 0) {
                        esel[isel] = esel[nsel-1];
                        nsel--;
                        break;
                    }
                }
            }

        } else if (strcmp(token, "ADJ2FACE") == 0 ||
                   strcmp(token, "adj2face") == 0   ) {

            /* make elist of all Faces */
            status = EG_getBodyTopos(ebody, NULL, FACE, &nlist, &elist);
            if (status < EGADS_SUCCESS) goto cleanup;

            /* remove entries from elist if it does not match name/value pairs */
            for (ilist = nlist-1; ilist >= 0; ilist--) {
                for (itoken = 2; itoken < 10; itoken++) {

                    /* get the name/value */
                    status = getToken(templine, itoken, ' ', 255, token);
                    if (status <= 0) break;

                    status1 = getToken(token, 0, '=', 255, attrName);
                    status2 = getToken(token, 1, '=', 255, attrValu);

                    if (status1 <= 0 || status2 <= 0) {
                        printf(" processFile: token is not name=value\n");
                        status = -999;
                        goto cleanup;
                    }
                    /* remove entry from elist if it does not match name/value */
                    status = EG_attributeRet(elist[ilist], attrName,
                                             &atype, &alen, &tempIlist, &tempRlist, &tempClist);
                    if (status != EGADS_SUCCESS) {
                        elist[ilist] = elist[nlist-1];
                        nlist--;
                        break;
                    } else if (atype != ATTRSTRING) {
                        elist[ilist] = elist[nlist-1];
                        nlist--;
                        break;
                    } else if (matches(attrValu, tempClist) == 0) {
                        elist[ilist] = elist[nlist-1];
                        nlist--;
                        break;
                    }
                }
            }

            /* for each entry in esel, remove it if it has no neighbors on elist */
            for (isel = nsel-1; isel >= 0; isel--) {
                status = getBodyTopos(ebody, esel[isel], FACE, &nnbor, &enbor);
                if (status < EGADS_SUCCESS) goto cleanup;

                iremove = 1;

                for (ilist = 0;  ilist < nlist; ilist++) {
                    for (inbor = 0; inbor < nnbor; inbor++) {
                        if (elist[ilist] == enbor[inbor]) {
                            iremove = 0;
                        }
                    }
                }

                if (iremove > 0) {
                    esel[isel] = esel[nsel-1];
                    nsel--;
                }

                EG_free(enbor);
                enbor = NULL;
            }

            EG_free(elist);
            elist = NULL;

        } else if (strcmp(token, "ADJ2EDGE") == 0 ||
                   strcmp(token, "adj2edge") == 0   ) {

            /* make elist of all Edges */
            status = EG_getBodyTopos(ebody, NULL, EDGE, &nlist, &elist);
            if (status < EGADS_SUCCESS) goto cleanup;

            /* remove entries from elist if it does not match name/value pairs */
            for (ilist = nlist-1; ilist >= 0; ilist--) {
                for (itoken = 2; itoken < 10; itoken++) {

                    /* get the name/value */
                    status = getToken(templine, itoken, ' ', 255, token);
                    if (status <= 0) break;

                    status1 = getToken(token, 0, '=', 255, attrName);
                    status2 = getToken(token, 1, '=', 255, attrValu);

                    if (status1 <= 0 || status2 <= 0) {
                        printf(" processFile: token is not name=value\n");
                        status = -999;
                        goto cleanup;
                    }
                    /* remove entry from elist if it does not match name/value */
                    status = EG_attributeRet(elist[ilist], attrName,
                                             &atype, &alen, &tempIlist, &tempRlist, &tempClist);
                    if (status != EGADS_SUCCESS) {
                        elist[ilist] = elist[nlist-1];
                        nlist--;
                        break;
                    } else if (atype != ATTRSTRING) {
                        elist[ilist] = elist[nlist-1];
                        nlist--;
                        break;
                    } else if (matches(attrValu, tempClist) == 0) {
                        elist[ilist] = elist[nlist-1];
                        nlist--;
                        break;
                    }
                }
            }

            /* for each entry in esel, remove it if it has no neighbors on elist */
            for (isel = nsel-1; isel >= 0; isel--) {
                status = getBodyTopos(ebody, esel[isel], EDGE, &nnbor, &enbor);
                if (status < EGADS_SUCCESS) goto cleanup;

                iremove = 1;

                for (ilist = 0;  ilist < nlist; ilist++) {
                    for (inbor = 0; inbor < nnbor; inbor++) {
                        if (elist[ilist] == enbor[inbor]) {
                            iremove = 0;
                        }
                    }
                }

                if (iremove > 0) {
                    esel[isel] = esel[nsel-1];
                    nsel--;
                }

                EG_free(enbor);
                enbor = NULL;
            }

            EG_free(elist);
            elist = NULL;

        } else if (strcmp(token, "ADJ2NODE") == 0 ||
                   strcmp(token, "adj2node") == 0   ) {

            /* make elist of all Faces */
            status = EG_getBodyTopos(ebody, NULL, NODE, &nlist, &elist);
            if (status < EGADS_SUCCESS) goto cleanup;

            /* remove entries from elist if it does not match name/value pairs */
            for (ilist = nlist-1; ilist >= 0; ilist--) {
                for (itoken = 2; itoken < 10; itoken++) {

                    /* get the name/value */
                    status = getToken(templine, itoken, ' ', 255, token);
                    if (status <= 0) break;

                    status1 = getToken(token, 0, '=', 255, attrName);
                    status2 = getToken(token, 1, '=', 255, attrValu);

                    if (status1 <= 0 || status2 <= 0) {
                        printf(" processFile: token is not name=value\n");
                        status = -999;
                        goto cleanup;
                    }
                    /* remove entry from elist if it does not match name/value */
                    status = EG_attributeRet(elist[ilist], attrName,
                                             &atype, &alen, &tempIlist, &tempRlist, &tempClist);
                    if (status != EGADS_SUCCESS) {
                        elist[ilist] = elist[nlist-1];
                        nlist--;
                        break;
                    } else if (atype != ATTRSTRING) {
                        elist[ilist] = elist[nlist-1];
                        nlist--;
                        break;
                    } else if (matches(attrValu, tempClist) == 0) {
                        elist[ilist] = elist[nlist-1];
                        nlist--;
                        break;
                    }
                }
            }

            /* for each entry in esel, remove it if it has no neighbors on elist */
            for (isel = nsel-1; isel >= 0; isel--) {
                status = getBodyTopos(ebody, esel[isel], NODE, &nnbor, &enbor);
                if (status < EGADS_SUCCESS) goto cleanup;

                iremove = 1;

                for (ilist = 0;  ilist < nlist; ilist++) {
                    for (inbor = 0; inbor < nnbor; inbor++) {
                        if (elist[ilist] == enbor[inbor]) {
                            iremove = 0;
                        }
                    }
                }

                if (iremove > 0) {
                    esel[isel] = esel[nsel-1];
                    nsel--;
                }

                EG_free(enbor);
                enbor = NULL;
            }

            EG_free(elist);
            elist = NULL;

        } else {
            printf(" processFile: illegal specifier (not HAS, ADJ2NODE, ADJ2EDGE, or ADJ2FACE");
            status = -999;
            goto cleanup;
        }

        EG_free(elist);
        elist = NULL;
    }

    fclose(fp);

cleanup:
    EG_free(elist);
    EG_free(enbor);

    if (esel != NULL) free(esel);

    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   matches - check if two strings match (with * + ? wildcards)        *
 *                                                                      *
 ************************************************************************
 */

static int
matches(char pattern[],                /* (in)  pattern */
  const char string[])                 /* (in)  string to check */
{

    /* --------------------------------------------------------------- */

    /* matching rules:
       ? matches any one character
       + matches one or more characters
       * matches zero or more characters

       use an adaptation of the recursive code found at:
       http://codegolf.stackexchange.com/questions/467/implement-glob-matcher
    */

    if        (*pattern == '*') {
        return matches(pattern+1, string) || (*string && matches(pattern, string+1));

    } else if (*pattern == '+') {
        return *string && (matches(pattern+1, ++string) || matches(pattern, string));

    } else if (*pattern == '?') {
        return *string && matches(pattern+1, string+1);

    } else {
        return *string == *pattern++ && (!*string++ || matches(pattern, string));

    }
}


/*
 ************************************************************************
 *                                                                      *
 *   editEgo - edit an Attribute on an ego                              *
 *                                                                      *
 ************************************************************************
 */

static int
editEgo(
    const char   attrname[],            /* (in)  Attribute name */
          int    atype,                 /* (in)  Attribute type */
          int    alen,                  /* (in)  Attribute length */
    const int    *tempIlist,            /* (in)  Attribute int  values */
    const double *tempRlist,            /* (in)  Attribute real values */
    const char   *tempClist,            /* (in)  Attribute char values */
          ego    eobj,                  /* (in)  ego to receive Attribute */
          int    overwrite,             /* (in)  =0 do not overwrite */
                                        /*       =1        overwrite */
                                        /*       =2        use smaller value */
                                        /*       =3        use larger  value */
                                        /*       =4        use sum or concatenation */
          int    *nchange)              /* (both) number of changes */
{
    int     status = EGADS_SUCCESS;     /* (out) return status */

    int     btype, blen, ivalue;
    CINT    *tmpIlist;
    double  value;
    CDOUBLE *tmpRlist;
    char    *strnew=NULL;
    CCHAR   *tmpClist;

    /* --------------------------------------------------------------- */

    /* check to see if Attribute already exists */
    status = EG_attributeRet(eobj, attrname, &btype, &blen,
                             &tmpIlist, &tmpRlist, &tmpClist);

    /* if overwrite==0 (do not overwrite) only save Attribute if it
                        does not already exist */
    if (overwrite == 0) {
        if (status != EGADS_SUCCESS) {
            status = EG_attributeAdd(eobj, attrname, atype, alen,
                                     tempIlist, tempRlist, tempClist);
            (*nchange)++;
        }

    /* if overwrite==1 (overwrite) save the Attribute */
    } else if (overwrite == 1) {
        status = EG_attributeAdd(eobj, attrname, atype, alen,
                                 tempIlist, tempRlist, tempClist);
        (*nchange)++;

    /* if overwrite==2 (use smaller value) save the smaller of the values */
    } else if (overwrite == 2) {
        if (status != EGADS_SUCCESS) {
            status = EG_attributeAdd(eobj, attrname, atype, alen,
                                     tempIlist, tempRlist, tempClist);
            (*nchange)++;
        } else if (atype != btype) {
            printf("existing (%d) and new (%d) Attributes have to be same type\n", atype, btype);
            status = EGADS_ATTRERR;
        } else if (atype == ATTRINT) {
            if (alen != 1 || blen != 1) {
                printf("can only take smaller value for scalar Attributes (alen=%d, blen=%d)\n", alen, blen);
                status = EGADS_ATTRERR;
            } else {
                ivalue = MIN(tempIlist[0], tmpIlist[0]);

                status = EG_attributeAdd(eobj, attrname, ATTRINT, 1,
                                         &ivalue, NULL, NULL);
                (*nchange)++;
            }
        } else if (atype == ATTRREAL) {
            if (alen != 1 || blen != 1) {
                printf("can only take smaller value for scalar Attributes (alen=%d, blen=%d)\n", alen, blen);
                status = EGADS_ATTRERR;
            } else {
                value = MIN(tempRlist[0], tmpRlist[0]);

                status = EG_attributeAdd(eobj, attrname, ATTRREAL, 1,
                                         NULL, &value, NULL);
                (*nchange)++;
            }
        } else if (atype == ATTRSTRING) {
            if (strcmp(tempClist, tmpClist) < 0) {
                status = EG_attributeAdd(eobj, attrname, ATTRSTRING, 0,
                                         NULL, NULL, tempClist);
                (*nchange)++;
            }
        }

    /* if overwrite==3 (use larger value) save the larger of the values */
    } else if (overwrite == 3) {
        if (status != EGADS_SUCCESS) {
            status = EG_attributeAdd(eobj, attrname, atype, alen,
                                     tempIlist, tempRlist, tempClist);
            (*nchange)++;
        } else if (atype != btype) {
            printf("existing (%d) and new (%d) Attributes have to be same type\n", atype, btype);
            status = EGADS_ATTRERR;
        } else if (atype == ATTRINT) {
            if (alen != 1 || blen != 1) {
                printf("can only take larger value for scalar Attributes (alen=%d, blen=%d)\n", alen, blen);
                status = EGADS_ATTRERR;
            } else {
                ivalue = MAX(tempIlist[0], tmpIlist[0]);

                status = EG_attributeAdd(eobj, attrname, ATTRINT, 1,
                                         &ivalue, NULL, NULL);
                (*nchange)++;
            }
        } else if (atype == ATTRREAL) {
            if (alen != 1 || blen != 1) {
                printf("can only take larger value for scalar Attributes (alen=%d, blen=%d)\n", alen, blen);
                status = EGADS_ATTRERR;
            } else {
                value = MAX(tempRlist[0], tmpRlist[0]);

                status = EG_attributeAdd(eobj, attrname, ATTRREAL, 1,
                                         NULL, &value, NULL);
                (*nchange)++;
            }
        } else if (atype == ATTRSTRING) {
            if (strcmp(tempClist, tmpClist) > 0) {
                status = EG_attributeAdd(eobj, attrname, ATTRSTRING, 0,
                                         NULL, NULL, tempClist);
                (*nchange)++;
            }
        }

    /* if overwrite==4 (use sum) save the sum of the values (or concatenate strings) */
    } else if (overwrite == 4) {
        if (status != EGADS_SUCCESS) {
            status = EG_attributeAdd(eobj, attrname, atype, alen,
                                     tempIlist, tempRlist, tempClist);
            (*nchange)++;
        } else if (atype != btype) {
            printf("existing (%d) and new (%d) Attributes have to be same type\n", atype, btype);
            status = EGADS_ATTRERR;
        } else if (atype == ATTRINT) {
            if (alen != 1 || blen != 1) {
                printf("can only take sum for scalar Attributes (alen=%d, blen=%d)\n", alen, blen);
                status = EGADS_ATTRERR;
            } else {
                ivalue = tempIlist[0] + tmpIlist[0];

                status = EG_attributeAdd(eobj, attrname, ATTRINT, 1,
                                         &ivalue, NULL, NULL);
                (*nchange)++;
            }
        } else if (atype == ATTRREAL) {
            if (alen != 1 || blen != 1) {
                printf("can only take sum for scalar Attributes (alen=%d, blen=%d)\n", alen, blen);
                status = EGADS_ATTRERR;
            } else {
                value = tempRlist[0] + tmpRlist[0];

                status = EG_attributeAdd(eobj, attrname, ATTRREAL, 1,
                                         NULL, &value, NULL);
                (*nchange)++;
            }
        } else if (atype == ATTRSTRING) {
            strnew = (char*) malloc((STRLEN(tempClist)+STRLEN(tmpClist)+2)*sizeof(char));
            if (strnew == NULL) {
                status = EGADS_MALLOC;
            } else {
                strcpy(strnew, tmpClist );
                strcat(strnew, tempClist);

                status = EG_attributeAdd(eobj, attrname, ATTRSTRING, 0,
                                         NULL, NULL, strnew);
                (*nchange)++;

                free(strnew);
            }
        }
    }

    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   getToken - get a token from a string                               *
 *                                                                      *
 ************************************************************************
 */

static int
getToken(char   *text,                  /* (in)  full text */
         int    nskip,                  /* (in)  tokens to skip */
         char   sep,                    /* (in)  seperator character */
         int    maxtok,                 /* (in)  size of token */
         char   *token)                 /* (out) token */
{
    int     lentok, ibeg, i, j, count, iskip;

    /* --------------------------------------------------------------- */

    token[0] = '\0';
    lentok   = 0;

    /* skip past white spaces at beginning of text */
    ibeg = 0;
    while (ibeg < STRLEN(text)) {
        if (text[ibeg] != ' ' && text[ibeg] != '\t') break;
        ibeg++;
    }

    /* count the number of separators */
    count = 0;
    for (i = ibeg; i < STRLEN(text); i++) {
        if (text[i] == sep) {
            count++;
            for (j = i+1; j < STRLEN(text); j++) {
                if (text[j] != sep) break;
                i++;
            }
        }
    }

    if (count < nskip) return 0;

    /* skip over nskip tokens */
    i = ibeg;
    for (iskip = 0; iskip < nskip; iskip++) {
        while (text[i] != sep) {
            i++;
        }
        while (text[i] == sep) {
            i++;
        }
    }

    /* extract the token we are looking for */
    while (text[i] != sep && i < STRLEN(text)) {
        token[lentok++] = text[i++];
        token[lentok  ] = '\0';

        if (lentok >= maxtok-1) {
            printf("ERROR:: token exceeds maxtok=%d\n", maxtok);
            break;
        }
    }

    return STRLEN(token);
}


/*
 ************************************************************************
 *                                                                      *
 *   getBodyTopos - extension to get neighbors f same oclass            *
 *                                                                      *
 ************************************************************************
 */

static int
getBodyTopos(ego    ebody,
             ego    esrc,
             int    oclass,
             int    *nlist,
             ego    **elist)
{

    int    status = EGADS_SUCCESS;

    int    inode, nnode, iedge, nedge, iface, nface, ilist;
    ego    *enodes, *eedges, *efaces;

    /* --------------------------------------------------------------- */

    *nlist = 0;
    *elist = NULL;

    if (ebody == NULL) {
        printf("ebody is NULL\n");
        status = -999;
        goto cleanup;
    } else if (ebody->magicnumber != MAGIC) {
        printf("ebody has wrong magic number\n");
        status = -999;
        goto cleanup;
    } else if (ebody->oclass != BODY) {
        printf("ebody is not a Body\n");
        status = -999;
        goto cleanup;
    } else if (esrc == NULL) {
        printf("ersc is NULL\n");
        status = -999;
        goto cleanup;
    } else if (esrc->magicnumber != MAGIC) {
        printf("esrc has wrong magic number\n");
        status = -999;
        goto cleanup;
    }

    if        (oclass == NODE) {
        if (esrc->oclass == EDGE || esrc->oclass == FACE) {
            status = EG_getBodyTopos(ebody, esrc, oclass, nlist, elist);
        } else {
            printf("cannot proocess NODE/?\n");
            status = -999;
        }
    } else if (oclass == EDGE) {
        if (esrc->oclass == NODE || esrc->oclass == FACE) {
            status = EG_getBodyTopos(ebody, esrc, oclass, nlist, elist);
        } else if (esrc->oclass != EDGE) {
            printf("cannot process EDGE/?\n");
            status = -999;
        } else {

            /* allocate the returned array (big enough for all Edges,
               which is probably a bit wasteful) */
            status = EG_getBodyTopos(ebody, NULL, FACE, &nedge, NULL);
            if (status < EGADS_SUCCESS) goto cleanup;

            *elist = EG_alloc(nedge*sizeof(ego));
            if (*elist == NULL) {
                status = EGADS_MALLOC;
                goto cleanup;
            }

            /* make a list of all Nodes that are adjacent to Edge */
            status = EG_getBodyTopos(ebody, esrc, NODE, &nnode, &enodes);
            if (status < EGADS_SUCCESS) goto cleanup;

            /* loop through all Nodes (that are adjacent to esrc) and add
               its Edges to elist (if not esrc and if not already in list) */
            for (inode = 0; inode < nnode; inode++) {
                status = EG_getBodyTopos(ebody, enodes[inode], EDGE, &nedge, &eedges);
                if (status < EGADS_SUCCESS) goto cleanup;

                for (iedge = 0; iedge < nedge; iedge++) {
                    if (eedges[iedge] == esrc) continue;

                    /* add the Edge to the list (possibly removed below) */
                    (*elist)[(*nlist)++] = eedges[iedge];

                    /* remove the Edge if it already was in list */
                    for (ilist = 0; ilist < (*nlist)-1; ilist++) {
                        if (eedges[iedge] == (*elist)[ilist]) {
                            (*nlist)--;
                            break;
                        }
                    }
                }

                EG_free(eedges);
            }

            EG_free(enodes);
        }
    } else if (oclass == FACE) {
        if (esrc->oclass == NODE || esrc->oclass == EDGE) {
            status = EG_getBodyTopos(ebody, esrc, oclass, nlist, elist);
        } else if (esrc->oclass != FACE) {
            printf("cannot process FACE/?\n");
            status = -999;
        } else {

            /* allocate the returned array (big enough for all Faces,
               which is probably a bit wasteful) */
            status = EG_getBodyTopos(ebody, NULL, FACE, &nface, NULL);
            if (status < EGADS_SUCCESS) goto cleanup;

            *elist = EG_alloc(nface*sizeof(ego));
            if (*elist == NULL) {
                status = EGADS_MALLOC;
                goto cleanup;
            }

            /* make a list of all Edges that are adjacent to Face */
            status = EG_getBodyTopos(ebody, esrc, EDGE, &nedge, &eedges);
            if (status < EGADS_SUCCESS) goto cleanup;

            /* loop through all Edges (that are adjacent to esrc) and add
               its Faces to elist (if not esrc and if not already in list) */
            for (iedge = 0; iedge < nedge; iedge++) {
                status = EG_getBodyTopos(ebody, eedges[iedge], FACE, &nface, &efaces);
                if (status < EGADS_SUCCESS) goto cleanup;

                for (iface = 0; iface < nface; iface++) {
                    if (efaces[iface] == esrc) continue;

                    /* add the Face to the list (possibly removed below) */
                    (*elist)[(*nlist)++] = efaces[iface];

                    /* remove the Face if it already was in list */
                    for (ilist = 0; ilist < (*nlist)-1; ilist++) {
                        if (efaces[iface] == (*elist)[ilist]) {
                            (*nlist)--;
                            break;
                        }
                    }
                }

                EG_free(efaces);
            }

            EG_free(eedges);
        }
    }

cleanup:
    return status;
}
