/*
 * OpenCSM/EGADS -- Extended User Defined Primitive Interface
 *
 */

/*
 * Copyright (C) 2010/2018  John F. Dannenhoffer, III (Syracuse University)
 *
 * This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *     MA  02110-1301  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef WIN32
   #include <windows.h>
#else
   #include <dlfcn.h>
#endif
#include "egads.h"

#ifdef WIN32
   #define DLL HINSTANCE
#else
   #define DLL void *
#endif

#define MAXPRIM 32

typedef int (*DLLFunc) (void);
typedef int (*udpI)    (int *, char ***, int **, int **, double **);
typedef int (*udpN)    (int *);
typedef int (*udpR)    (int);
typedef int (*udpC)    (ego);
typedef int (*udpS)    (char *, void *, int);
typedef int (*udpE)    (ego, ego *, int *, char **);
typedef int (*udpG)    (ego, char *, char **);
typedef int (*udpM)    (ego, int, int *, int *, int *, double **);
typedef int (*udpV)    (ego, char *, double *, int);
typedef int (*udpP)    (ego, int, int, int, double *, double *);

static char *udpName[ MAXPRIM];
static DLL   udpDLL[  MAXPRIM];
static udpI  udpInit[ MAXPRIM];
static udpN  udpNumB[ MAXPRIM];
static udpR  udpReset[MAXPRIM];
static udpC  udpClean[MAXPRIM];
static udpS  udpSet[  MAXPRIM];
static udpE  udpExec[ MAXPRIM];
static udpG  udpGet[  MAXPRIM];
static udpM  udpMesh[ MAXPRIM];
static udpV  udpVel[  MAXPRIM];
static udpP  udpSens[ MAXPRIM];

static int udp_nPrim = 0;

/* ************************* Utility Functions ***************************** */

static /*@null@*/ DLL udpDLopen(const char *name)
{
    int  i, len;
    DLL  dll;
    char *full;

    if (name == NULL) {
        printf(" Information: Dynamic Loader invoked with NULL name!\n");
        return NULL;
    }
    len  = strlen(name);
    full = (char *) malloc((len+5)*sizeof(char));
    if (full == NULL) {
        printf(" Information: Dynamic Loader MALLOC Error!\n");
        return NULL;
    }
    for (i = 0; i < len; i++) full[i] = name[i];
    full[len  ] = '.';
#ifdef WIN32
    full[len+1] = 'D';
    full[len+2] = 'L';
    full[len+3] = 'L';
    full[len+4] =  0;
    dll = LoadLibrary(full);
#else
    full[len+1] = 's';
    full[len+2] = 'o';
    full[len+3] =  0;
    dll = dlopen(full, RTLD_NOW | RTLD_NODELETE);
#endif

    if (!dll) {
        printf(" Information: Dynamic Loader Error for %s\n", full);
#ifndef WIN32
        printf("              %s\n", dlerror());
#endif
        free(full);
        return NULL;
    }

    free(full);
    return dll;
}


static void udpDLclose(/*@unused@*/ /*@only@*/ DLL dll)
{
#ifdef WIN32
    FreeLibrary(dll);
#else
    dlclose(dll);
#endif
}


static DLLFunc udpDLget(DLL dll,   const char *symname,
                        /*@null@*/ const char *name)
{
    DLLFunc data;

#ifdef WIN32
    data = (DLLFunc) GetProcAddress(dll, symname);
#else
    data = (DLLFunc) dlsym(dll, symname);
#endif
    if ((data == NULL) && (name != NULL))
        printf(" Information: Couldn't get symbol %s in %s\n", symname, name);
    return data;
}


static int udpDLoaded(const char *name)
{
    int i;

    for (i = 0; i < udp_nPrim; i++)
        if (strcmp(name, udpName[i]) == 0) return i;

    return -1;
}


static int udpDYNload(const char *name)
{
    int i, len, ret;
    DLL dll;

    if (udp_nPrim >= MAXPRIM) {
        printf(" Information: Number of Primitives > %d!\n", MAXPRIM);
        return EGADS_INDEXERR;
    }
    dll = udpDLopen(name);
    if (dll == NULL) return EGADS_NULLOBJ;

    ret = udp_nPrim;
    udpInit[ ret] = (udpI) udpDLget(dll, "udpInitialize",  name);
    udpNumB[ ret] = (udpN) udpDLget(dll, "udpNumBodys",    name);
    udpReset[ret] = (udpR) udpDLget(dll, "udpReset",       name);
    udpClean[ret] = (udpC) udpDLget(dll, "udpClean",       name);
    udpSet[  ret] = (udpS) udpDLget(dll, "udpSet",         name);
    udpExec[ ret] = (udpE) udpDLget(dll, "udpExecute",     name);
    udpGet[  ret] = (udpG) udpDLget(dll, "udpGet",         NULL);
    udpMesh[ ret] = (udpM) udpDLget(dll, "udpMesh",        NULL);
    udpVel[  ret] = (udpV) udpDLget(dll, "udpVel",         name);
    udpSens[ ret] = (udpP) udpDLget(dll, "udpSensitivity", name);
    if ((udpInit[ ret] == NULL) ||
        (udpNumB[ ret] == NULL) ||
        (udpReset[ret] == NULL) ||
        (udpClean[ret] == NULL) ||
        (udpSet[  ret] == NULL) ||
        (udpExec[ ret] == NULL) ||
        (udpVel[  ret] == NULL) ||
        (udpSens[ ret] == NULL)   ) {
        udpDLclose(dll);
        return EGADS_EMPTY;
    }

    len  = strlen(name) + 1;
    udpName[ret] = (char *) malloc(len*sizeof(char));
    if (udpName[ret] == NULL) {
        udpDLclose(dll);
        return EGADS_MALLOC;
    }
    for (i = 0; i < len; i++) udpName[ret][i] = name[i];
    udpDLL[ret] = dll;
    udp_nPrim++;

    return ret;
}


/* ************************* Exposed Functions ***************************** */


int udp_initialize(const char primName[],
                   int        *nArgs,
                   char       **name[],
                   int        *type[],
                   int        *idefault[],
                   double     *ddefault[])
{
    int i;

    i = udpDLoaded(primName);

    if (i == -1) {
        i = udpDYNload(primName);
        if (i < 0) return i;
    }

    return udpInit[i](nArgs, name, type, idefault, ddefault);
}


int udp_numBodys(const char primName[])
{
    int i;

    i = udpDLoaded(primName);
    if (i == -1) return EGADS_NOTFOUND;

    return udpNumB[i](0);
}


int udp_clrArguments(const char primName[])
{
    int i;

    i = udpDLoaded(primName);
    if (i == -1) return EGADS_NOTFOUND;

    return udpReset[i](0);
}


int udp_clean(const char primName[],
             ego         body)
{
    int i;

    i = udpDLoaded(primName);
    if (i == -1) return EGADS_NOTFOUND;

    return udpClean[i](body);
}


int udp_setArgument(const char primName[],
                    char       name[],
                    void       *value,
                    int        nvalue)
{
    int i;

    i = udpDLoaded(primName);
    if (i == -1) return EGADS_NOTFOUND;

    return udpSet[i](name, value, nvalue);
}


int udp_executePrim(const char primName[],
                    ego        context,
                    ego        *body,
                    int        *nMesh,
                    char       *string[])
{
    int i;

    i = udpDLoaded(primName);
    if (i == -1) return EGADS_NOTFOUND;

    return udpExec[i](context, body, nMesh, string);
}


int udp_getOutput(const char primName[],
                  ego        body,
                  char       name[],
                  void       *value)
{
  int i;

  i = udpDLoaded(primName);
  if (i == -1) return EGADS_NOTFOUND;
  if (udpGet[i] == NULL) return EGADS_EMPTY;

  return udpGet[i](body, name, value);
}


int udp_getMesh(const char primName[],
                ego        body,
                int        imesh,
                int        *imax,
                int        *jmax,
                int        *kmax,
                double     *mesh[])
{
    int i;

    i = udpDLoaded(primName);
    if (i == -1) return EGADS_NOTFOUND;
    if (udpMesh[i] == NULL) return EGADS_EMPTY;

    return udpMesh[i](body, imesh, imax, jmax, kmax, mesh);
}


int udp_setVelocity(const char primName[],
                    ego        body,
                    char       name[],
                    double     value[],
                    int        nvalue)
{
    int i;

    i = udpDLoaded(primName);
    if (i == -1) return EGADS_NOTFOUND;

    return udpVel[i](body, name, value, nvalue);
}


int udp_sensitivity(const char primName[],
                    ego        body,
                    int        npts,
                    int        entType,
                    int        entIndex,
                    double     uvs[],
                    double     vels[])
{
    int i;

    i = udpDLoaded(primName);
    if (i == -1) return EGADS_NOTFOUND;

    return udpSens[i](body, npts, entType, entIndex, uvs, vels);
}


void udp_cleanupAll()
{
    int i;

    if (udp_nPrim == 0) return;

    for (i = 0; i < udp_nPrim; i++) {
        udpReset[i](1);
        free(udpName[i]);
        udpDLclose(udpDLL[i]);
    }

    udp_nPrim = 0;
}
